<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/kw/clock.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/kw/clock.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * kirkwood clocks
 *
 * timers count down to zero.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"

#include "ureg.h"

enum {
	Tcycles		= CLOCKFREQ / HZ,	/* cycles per clock tick */
	Dogperiod	= 15 * CLOCKFREQ, /* at most 21 s.; must fit in ulong */
	MaxPeriod	= Tcycles,
	MinPeriod	= MaxPeriod / 100,

	/* timer ctl bits */
	Tmr0enable	= 1&lt;&lt;0,
	Tmr0reload	= 1&lt;&lt;1,	/* at 0 count, load timer0 from reload0 */
	Tmr1enable	= 1&lt;&lt;2,
	Tmr1reload	= 1&lt;&lt;3,	/* at 0 count, load timer1 from reload1 */
	TmrWDenable	= 1&lt;&lt;4,
	TmrWDreload	= 1&lt;&lt;5,
};

typedef struct TimerReg TimerReg;
struct TimerReg
{
	ulong	ctl;
	ulong	pad[3];
	ulong	reload0;
	ulong	timer0;			/* cycles until zero */
	ulong	reload1;
	ulong	timer1;			/* cycles until zero */
	ulong	reloadwd;
	ulong	timerwd;
};

static int ticks; /* for sanity checking; m-&gt;ticks doesn't always get updated */

static void
clockintr(Ureg *ureg, void *arg)
{
	TimerReg *tmr = arg;
	static int nesting;

	tmr-&gt;timerwd = Dogperiod;		/* reassure the watchdog */
	ticks++;
	coherence();

	if (nesting == 0) {	/* if the clock interrupted itself, bail out */
		++nesting;
		timerintr(ureg, 0);
		--nesting;
	}

	intrclear(Irqbridge, IRQcputimer0);
}

/* stop clock interrupts and disable the watchdog timer */
void
clockshutdown(void)
{
	TimerReg *tmr = (TimerReg *)soc.clock;

	tmr-&gt;ctl = 0;
	coherence();
}

void
clockinit(void)
{
	int i, s;
	CpucsReg *cpu = (CpucsReg *)soc.cpu;
	TimerReg *tmr = (TimerReg *)soc.clock;

	clockshutdown();

	/*
	 * verify sanity of timer0
	 */

	intrenable(Irqbridge, IRQcputimer0, clockintr, tmr, "clock0");
	s = spllo();			/* risky */
	/* take any deferred clock (&amp; other) interrupts here */
	splx(s);

	/* adjust m-&gt;bootdelay, used by delay()? */
	m-&gt;ticks = ticks = 0;
	m-&gt;fastclock = 0;

	tmr-&gt;timer0 = 1;
	tmr-&gt;ctl = Tmr0enable;		/* just once */
	coherence();

	s = spllo();			/* risky */
	for (i = 0; i &lt; 10 &amp;&amp; ticks == 0; i++) {
		delay(1);
		coherence();
	}
	splx(s);
	if (ticks == 0) {
		serialputc('?');
		if (tmr-&gt;timer0 == 0)
			panic("clock not interrupting");
		else if (tmr-&gt;timer0 == tmr-&gt;reload0)
			panic("clock not ticking");
		else
			panic("clock running very slowly");
	}

	/*
	 * configure all timers
	 */
	clockshutdown();
	tmr-&gt;reload0 = tmr-&gt;timer0 = Tcycles;	/* tick clock */
	tmr-&gt;reload1 = tmr-&gt;timer1 = ~0;	/* cycle clock */
	tmr-&gt;timerwd = Dogperiod;		/* watch dog timer */
	coherence();
	tmr-&gt;ctl = Tmr0enable | Tmr0reload | Tmr1enable | Tmr1reload |
		TmrWDenable;
	cpu-&gt;rstout |= RstoutWatchdog;
	coherence();
}

void
timerset(Tval next)
{
	int offset;
	TimerReg *tmr = (TimerReg *)soc.clock;

	offset = next - fastticks(nil);
	if(offset &lt; MinPeriod)
		offset = MinPeriod;
	else if(offset &gt; MaxPeriod)
		offset = MaxPeriod;
	tmr-&gt;timer0 = offset;
	coherence();
}

uvlong
fastticks(uvlong *hz)
{
	uvlong now;
	int s;

	if(hz)
		*hz = CLOCKFREQ;
	s = splhi();
	/* zero low ulong of fastclock */
	now = (m-&gt;fastclock &amp; ~(uvlong)~0ul) | perfticks();
	if(now &lt; m-&gt;fastclock)		/* low bits must have wrapped */
		now += 1ll &lt;&lt; 32;
	m-&gt;fastclock = now;
	splx(s);
	return now;
}

ulong
perfticks(void)
{
	TimerReg *tmr = (TimerReg *)soc.clock;

	return ~tmr-&gt;timer1;
}

long
lcycles(void)
{
	return perfticks();
}

ulong
µs(void)
{
	return fastticks2us(fastticks(nil));
}

void
microdelay(int l)
{
	int i;

	l *= m-&gt;delayloop;
	l /= 1000;
	if(l &lt;= 0)
		l = 1;
	for(i = 0; i &lt; l; i++)
		;
}

void
delay(int l)
{
	ulong i, j;

	j = m-&gt;delayloop;
	while(l-- &gt; 0)
		for(i=0; i &lt; j; i++)
			;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

