<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/omap/dma.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/omap/dma.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * omap3530 system dma controller
 *
 * terminology: a block consist of frame(s), a frame consist of elements
 * (uchar, ushort, or ulong sized).
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"
#include "../port/netif.h"

enum {
	Nirq	= 4,
	Baseirq	= 12,

	Nchan	= 32,
};

/*
 * has a sw reset bit
 * dma req lines 1, 2, 6, 63 are available for `system expansion'
 */

typedef struct Regs Regs;
typedef struct Dchan Dchan;
struct Regs {
	uchar	_pad0[8];
	/* bitfield of intrs pending, by Dchan; write 1s to clear */
	ulong	irqsts[Nirq];
	ulong	irqen[Nirq];	/* bitfield of intrs enabled, by Dchan */
	ulong	syssts;		/* 1&lt;&lt;0 is Resetdone */
	ulong	syscfg;		/* 1&lt;&lt;1 is Softreset */
	uchar	_pad1[0x64 - 0x30];

	ulong	caps[5];	/* caps[1] not defined */
	ulong	gcr;		/* knobs */
	ulong	_pad2;

	struct Dchan {
		ulong	ccr;	/* chan ctrl: incr, etc. */
		ulong	clnkctrl; /* link ctrl */
		ulong	cicr;	/* intr ctrl */
		ulong	csr;	/* status */
		ulong	csdp;	/* src &amp; dest params */
		ulong	cen;	/* element # */
		ulong	cfn;	/* frame # */
		ulong	cssa;	/* src start addr */
		ulong	cdsa;	/* dest start addr */
		ulong	csei;	/* src element index */
		ulong	csfi;	/* src frame index | pkt size */
		ulong	cdei;	/* dest element index */
		ulong	cdfi;	/* dest frame index | pkt size */
		ulong	csac;	/* src addr value (read-only?) */
		ulong	cdac;	/* dest addr value */
		ulong	ccen;	/* curr transferred element # (in frame) */
		ulong	ccfn;	/* curr transferred frame # (in xfer) */
		ulong	color;
		uchar	_pad3[24];
	} chan[Nchan];
};

enum {
	/* cicr/csr bits */
	Blocki	= 1 &lt;&lt; 5,

	/* ccr bits */
	Enable	= 1 &lt;&lt; 7,
};

typedef struct Xfer Xfer;
static struct Xfer {
	Rendez	*rend;
	int	*done;		/* flag to set on intr */
} xfer[Nirq];

int
isdmadone(int irq)
{
	Dchan *cp;
	Regs *regs = (Regs *)PHYSSDMA;

	cp = regs-&gt;chan + irq;
	return cp-&gt;csr &amp; Blocki;
}

static void
dmaintr(Ureg *, void *a)
{
	int i = (int)a;			/* dma request &amp; chan # */
	Dchan *cp;
	Regs *regs = (Regs *)PHYSSDMA;

	assert(i &gt;= 0 &amp;&amp; i &lt; Nirq);

	*xfer[i].done = 1;
	assert(xfer[i].rend != nil);
	wakeup(xfer[i].rend);

	cp = regs-&gt;chan + i;
	if(!(cp-&gt;csr &amp; Blocki))
		iprint("dmaintr: req %d: Blocki not set; csr %#lux\n",
			i, cp-&gt;csr);
	cp-&gt;csr |= cp-&gt;csr;			/* extinguish intr source */
	coherence();
	regs-&gt;irqsts[i] = regs-&gt;irqsts[i];	/* extinguish intr source */
	coherence();
	regs-&gt;irqen[i] &amp;= ~(1 &lt;&lt; i);
	coherence();

	xfer[i].rend = nil;
	coherence();
}

void
zerowds(ulong *wdp, int cnt)
{
	while (cnt-- &gt; 0)
		*wdp++ = 0;
}

static int
istestdmadone(void *arg)
{
	return *(int *)arg;
}

void
dmainit(void)
{
	int n;
	char name[16];
	Dchan *cp;
	Regs *regs = (Regs *)PHYSSDMA;

	if (probeaddr((uintptr)&amp;regs-&gt;syssts) &lt; 0)
		panic("dmainit: no syssts reg");
	regs-&gt;syssts = 0;
	coherence();
	regs-&gt;syscfg |= 1&lt;&lt;1;		/* Softreset */
	coherence();
	while(!(regs-&gt;syssts &amp; (1&lt;&lt;0)))	/* Resetdone? */
		;

	for (n = 0; n &lt; Nchan; n++) {
		cp = regs-&gt;chan + n;
		cp-&gt;ccr = 0;
		cp-&gt;clnkctrl = 0;
		cp-&gt;cicr = 0;
		cp-&gt;csr = 0;
		cp-&gt;csdp = 0;
		cp-&gt;cen = cp-&gt;cfn = 0;
		cp-&gt;cssa = cp-&gt;cdsa = 0;
		cp-&gt;csei = cp-&gt;csfi = 0;
		cp-&gt;cdei = cp-&gt;cdfi = 0;
//		cp-&gt;csac = cp-&gt;cdac = 0;		// ro
		cp-&gt;ccen = cp-&gt;ccfn = 0;
		cp-&gt;color = 0;
	}
	zerowds((void *)regs-&gt;irqsts, sizeof regs-&gt;irqsts / sizeof(ulong));
	zerowds((void *)regs-&gt;irqen,  sizeof regs-&gt;irqen / sizeof(ulong));
	coherence();

	regs-&gt;gcr = 65;			/* burst size + 1 */
	coherence();

	for (n = 0; n &lt; Nirq; n++) {
		snprint(name, sizeof name, "dma%d", n);
		intrenable(Baseirq + n, dmaintr, (void *)n, nil, name);
	}
}

enum {
	Testbyte	= 0252,
	Testsize	= 256,
	Scratch		= MB,
};

/*
 * try to confirm sane operation
 */
void
dmatest(void)
{
	int n, done;
	uchar *bp;
	static ulong pat = 0x87654321;
	static Rendez trendez;

	if (up == nil)
		panic("dmatest: up not set yet");
	bp = (uchar *)KADDR(PHYSDRAM + 128*MB);
	memset(bp, Testbyte, Scratch);
	done = 0;
	dmastart((void *)PADDR(bp), Postincr, (void *)PADDR(&amp;pat), Const,
		Testsize, &amp;trendez, &amp;done);
	sleep(&amp;trendez, istestdmadone, &amp;done);
	cachedinvse(bp, Scratch);

	if (((ulong *)bp)[0] != pat)
		panic("dmainit: copied incorrect data %#lux != %#lux",
			((ulong *)bp)[0], pat);
	for (n = Testsize; n &lt; Scratch &amp;&amp; bp[n] != Testbyte; n++)
		;
	if (n &gt;= Scratch)
		panic("dmainit: ran wild over memory, clobbered ≥%,d bytes", n);
	if (bp[n] == Testbyte &amp;&amp; n != Testsize)
		iprint("dma: %d-byte dma stopped after %d bytes!\n",
			Testsize, n);
}

/* addresses are physical */
int
dmastart(void *to, int tmode, void *from, int fmode, uint len, Rendez *rend,
	int *done)
{
	int irq, chan;
	uint ruplen;
	Dchan *cp;
	Regs *regs = (Regs *)PHYSSDMA;
	static Lock alloclck;

	/* allocate free irq (and chan) */
	ilock(&amp;alloclck);
	for (irq = 0; irq &lt; Nirq &amp;&amp; xfer[irq].rend != nil; irq++)
		;
	if (irq &gt;= Nirq)
		panic("dmastart: no available irqs; too many concurrent dmas");
	chan = irq;
	xfer[irq].rend = rend;			/* for wakeup at intr time */
	xfer[irq].done = done;
	*done = 0;
	iunlock(&amp;alloclck);

	ruplen = ROUNDUP(len, sizeof(ulong));
	assert(to != from);

	cp = regs-&gt;chan + chan;
	cp-&gt;ccr &amp;= ~Enable;			/* paranoia */
	cp-&gt;cicr = 0;
	regs-&gt;irqen[irq] &amp;= ~(1 &lt;&lt; chan);
	coherence();

	cp-&gt;csdp = 2;				/* 2 = log2(sizeof(ulong)) */
	cp-&gt;cssa = (uintptr)from;
	cp-&gt;cdsa = (uintptr)to;
	cp-&gt;ccr = tmode &lt;&lt; 14 | fmode &lt;&lt; 12;
	cp-&gt;csei = cp-&gt;csfi = cp-&gt;cdei = cp-&gt;cdfi = 1;
	cp-&gt;cen = ruplen / sizeof(ulong);	/* ulongs / frame */
	cp-&gt;cfn = 1;				/* 1 frame / xfer */
	cp-&gt;cicr = Blocki;			/* intr at end of block */

	regs-&gt;irqen[irq] |= 1 &lt;&lt; chan;
	coherence();

	cp-&gt;ccr |= Enable;			/* fire! */
	coherence();

	return irq;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

