<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/9/pc/usbohci.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/9/pc/usbohci.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * USB Open Host Controller Interface (Ohci) driver
 *
 * BUGS:
 * - Missing isochronous input streams.
 * - Too many delays and ilocks.
 * - bandwidth admission control must be done per-frame.
 * - Buffering could be handled like in uhci, to avoid
 * needed block allocation and avoid allocs for small Tds.
 * - must warn of power overruns.
 */

#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"../port/error.h"

#include	"../port/usb.h"

typedef struct Ctlio Ctlio;
typedef struct Ctlr Ctlr;
typedef struct Ed Ed;
typedef struct Edpool Edpool;
typedef struct Epx Epx;
typedef struct Hcca Hcca;
typedef struct Isoio Isoio;
typedef struct Ohci Ohci;
typedef struct Qio Qio;
typedef struct Qtree Qtree;
typedef struct Td Td;
typedef struct Tdpool Tdpool;

enum
{
	Incr		= 64,		/* for Td and Ed pools */

	Align		= 0x20,		/* OHCI only requires 0x10 */
					/* use always a power of 2 */

	Abortdelay	= 1,		/* delay after cancelling Tds (ms) */
	Tdatomic		= 8,		/* max nb. of Tds per bulk I/O op. */
	Enabledelay	= 100,		/* waiting for a port to enable */


	/* Queue states (software) */
	Qidle		= 0,
	Qinstall,
	Qrun,
	Qdone,
	Qclose,
	Qfree,

	/* Ed control bits */
	Edmpsmask	= 0x7ff,	/* max packet size */
	Edmpsshift	= 16,
	Edlow		= 1 &lt;&lt; 13,	/* low speed */
	Edskip		= 1 &lt;&lt; 14,	/* skip this ed */
	Ediso		= 1 &lt;&lt; 15,	/* iso Tds used */
	Edtddir		= 0,		/* get dir from td */
	Edin		= 2 &lt;&lt; 11,	/* direction in */
	Edout		= 1 &lt;&lt; 11,	/* direction out */
	Eddirmask	= 3 &lt;&lt; 11,	/* direction bits */
	Edhalt		= 1,		/* halted (in head ptr) */
	Edtoggle	= 2,		/* toggle (in head ptr) 1 == data1 */

	/* Td control bits */
	Tdround		= 1&lt;&lt;18,	/* (rounding) short packets ok */
	Tdtoksetup	= 0&lt;&lt;19,	/* setup packet */
	Tdtokin		= 2&lt;&lt;19,	/* in packet */
	Tdtokout	= 1&lt;&lt;19,	/* out packet */
	Tdtokmask	= 3&lt;&lt;19,	/* in/out/setup bits */
	Tdnoioc		= 7&lt;&lt;21,	/* intr. cnt. value for no interrupt */
	Tdusetog	= 1&lt;&lt;25,	/* use toggle from Td (1) or Ed (0) */
	Tddata1		= 1&lt;&lt;24,	/* data toggle (1 == data1) */
	Tddata0		= 0&lt;&lt;24,
	Tdfcmask	= 7,		/* frame count (iso) */
	Tdfcshift	= 24,
	Tdsfmask	= 0xFFFF,	/* starting frame (iso) */
	Tderrmask	= 3,		/* error counter */
	Tderrshift	= 26,
	Tdccmask	= 0xf,		/* condition code (status) */
	Tdccshift	= 28,
	Tdiccmask	= 0xf,		/* condition code (iso, offsets) */
	Tdiccshift	= 12,

	Ntdframes	= 0x10000,	/* # of different iso frame numbers */

	/* Td errors (condition code) */
	Tdok		= 0,
	Tdcrc		= 1,
	Tdbitstuff	= 2,
	Tdbadtog	= 3,
	Tdstalled	= 4,
	Tdtmout		= 5,
	Tdpidchk	= 6,
	Tdbadpid	= 7,
	Tddataovr	= 8,
	Tddataund	= 9,
	Tdbufovr	= 0xC,
	Tdbufund	= 0xD,
	Tdnotacc	= 0xE,

	/* control register */
	Cple		= 0x04,		/* periodic list enable */
	Cie		= 0x08,		/* iso. list enable */
	Ccle		= 0x10,		/* ctl list enable */
	Cble		= 0x20,		/* bulk list enable */
	Cfsmask		= 3 &lt;&lt; 6,	/* functional state... */
	Cfsreset	= 0 &lt;&lt; 6,
	Cfsresume	= 1 &lt;&lt; 6,
	Cfsoper		= 2 &lt;&lt; 6,
	Cfssuspend	= 3 &lt;&lt; 6,

	/* command status */
	Sblf =	1 &lt;&lt; 2,			/* bulk list (load) flag */
	Sclf =	1 &lt;&lt; 1,			/* control list (load) flag */
	Shcr =	1 &lt;&lt; 0,			/* host controller reset */

	/* intr enable */
	Mie =	1 &lt;&lt; 31,
	Oc =	1 &lt;&lt; 30,
	Rhsc =	1 &lt;&lt; 6,
	Fno =	1 &lt;&lt; 5,
	Ue =	1 &lt;&lt; 4,
	Rd =	1 &lt;&lt; 3,
	Sf =	1 &lt;&lt; 2,
	Wdh =	1 &lt;&lt; 1,
	So =	1 &lt;&lt; 0,

	Fmaxpktmask = 0x7fff,
	Fmaxpktshift = 16,
	HcRhDescA_POTPGT_MASK =	0xff &lt;&lt; 24,
	HcRhDescA_POTPGT_SHIFT =	24,

	/* Rh status */
	Lps =	1 &lt;&lt; 0,
	Cgp =	1 &lt;&lt; 0,
	Oci =	1 &lt;&lt; 1,
	Psm =	1 &lt;&lt; 8,
	Nps =	1 &lt;&lt; 9,
	Drwe =	1 &lt;&lt; 15,
	Srwe =	1 &lt;&lt; 15,
	Lpsc =	1 &lt;&lt; 16,
	Ccic =	1 &lt;&lt; 17,
	Crwe =	1 &lt;&lt; 31,

	/* port status */
	Ccs =	0x00001,	/* current connect status */
	Pes =	0x00002,	/* port enable status */
	Pss =	0x00004,	/* port suspend status */
	Poci =	0x00008,	/* over current indicator */
	Prs =	0x00010,	/* port reset status */
	Pps =	0x00100,	/* port power status */
	Lsda =	0x00200,	/* low speed device attached */
	Csc =	0x10000,	/* connect status change */
	Pesc =	0x20000,	/* enable status change */
	Pssc =	0x40000,	/* suspend status change */
	Ocic =	0x80000,	/* over current ind. change */
	Prsc =	0x100000,	/* reset status change */

	/* port status write bits */
	Cpe =	0x001,		/* clear port enable */
	Spe =	0x002,		/* set port enable */
	Spr =	0x010,		/* set port reset */
	Spp =	0x100,		/* set port power */
	Cpp =	0x200,		/* clear port power */

};

/*
 * Endpoint descriptor. (first 4 words used by hardware)
 */
struct Ed {
	ulong	ctrl;
	ulong	tail;		/* transfer descriptor */
	ulong	head;
	ulong	nexted;

	Ed*	next;		/* sw; in free list or next in list */
	Td*	tds;		/* in use by current xfer; all for iso */
	Ep*	ep;		/* debug/align */
	Ed*	inext;		/* debug/align (dump interrupt eds). */
};

/*
 * Endpoint I/O state (software), per direction.
 */
struct Qio
{
	QLock;			/* for the entire I/O process */
	Rendez;			/* wait for completion */
	Ed*	ed;		/* to place Tds on it */
	int	sched;		/* queue number (intr/iso) */
	int	toggle;		/* Tddata0/Tddata1 */
	ulong	usbid;		/* device/endpoint address */
	int	tok;		/* Tdsetup, Tdtokin, Tdtokout */
	long	iotime;		/* last I/O time; to hold interrupt polls */
	int	debug;		/* for the endpoint */
	char*	err;		/* error status */
	int	state;		/* Qidle -&gt; Qinstall -&gt; Qrun -&gt; Qdone | Qclose */
	long	bw;		/* load (intr/iso) */
};

struct Ctlio
{
	Qio;			/* single Ed for all transfers */
	uchar*	data;		/* read from last ctl req. */
	int	ndata;		/* number of bytes read */
};

struct Isoio
{
	Qio;
	int	nframes;	/* number of frames for a full second */
	Td*	atds;		/* Tds avail for further I/O */
	int	navail;		/* number of avail Tds */
	ulong	frno;		/* next frame number avail for I/O */
	ulong	left;		/* remainder after rounding Hz to samples/ms */
	int	nerrs;		/* consecutive errors on iso I/O */
};

/*
 * Transfer descriptor. Size must be multiple of 32
 * First block is used by hardware (aligned to 32).
 */
struct Td
{
	ulong	ctrl;
	ulong	cbp;		/* current buffer pointer */
	ulong	nexttd;
	ulong	be;
	ushort	offsets[8];	/* used by Iso Tds only */

	Td*	next;		/* in free or Ed tds list */
	Td*	anext;		/* in avail td list (iso) */
	Ep*	ep;		/* using this Td for I/O */
	Qio*	io;		/* using this Td for I/O */
	Block*	bp;		/* data for this Td */
	ulong	nbytes;		/* bytes in this Td */
	ulong	cbp0;		/* initial value for cbp */
	ulong	last;		/* true for last Td in Qio */
};

/*
 * Host controller communication area (hardware)
 */
struct Hcca
{
	ulong	intrtable[32];
	ushort	framenumber;
	ushort	pad1;
	ulong	donehead;
	uchar	reserved[116];
};

/*
 * I/O registers
 */
struct Ohci
{
	/* control and status group */
	ulong	revision;		/*00*/
	ulong	control;		/*04*/
	ulong	cmdsts;			/*08*/
	ulong	intrsts;			/*0c*/
	ulong	intrenable;		/*10*/
	ulong	intrdisable;		/*14*/

	/* memory pointer group */
	ulong	hcca;			/*18*/
	ulong	periodcurred;		/*1c*/
	ulong	ctlheaded;		/*20*/
	ulong	ctlcurred;		/*24*/
	ulong	bulkheaded;		/*28*/
	ulong	bulkcurred;		/*2c*/
	ulong	donehead;		/*30*/

	/* frame counter group */
	ulong	fminterval;		/*34*/
	ulong	fmremaining;		/*38*/
	ulong	fmnumber;		/*3c*/
	ulong	periodicstart;		/*40*/
	ulong	lsthreshold;		/*44*/

	/* root hub group */
	ulong	rhdesca;		/*48*/
	ulong	rhdescb;		/*4c*/
	ulong	rhsts;			/*50*/
	ulong	rhportsts[15];		/*54*/
	ulong	pad25[20];		/*90*/

	/* unknown */
	ulong	hostueaddr;		/*e0*/
	ulong	hostuests;		/*e4*/
	ulong	hosttimeoutctrl;		/*e8*/
	ulong	pad59;			/*ec*/
	ulong	pad60;			/*f0*/
	ulong	hostrevision;		/*f4*/
	ulong	pad62[2];
					/*100*/
};

/*
 * Endpoint tree (software)
 */
struct Qtree
{
	int	nel;
	int	depth;
	ulong*	bw;
	Ed**	root;
};

struct Tdpool
{
	Lock;
	Td*	free;
	int	nalloc;
	int	ninuse;
	int	nfree;
};

struct Edpool
{
	Lock;
	Ed*	free;
	int	nalloc;
	int	ninuse;
	int	nfree;
};

struct Ctlr
{
	Lock;			/* for ilock; lists and basic ctlr I/O */
	QLock	resetl;		/* lock controller during USB reset */
	int	active;
	Ctlr*	next;
	int	nports;

	Ohci*	ohci;		/* base I/O address */
	Hcca*	hcca;		/* intr/done Td lists (used by hardware) */
	int	overrun;	/* sched. overrun */
	Ed*	intrhd;		/* list of intr. eds in tree */
	Qtree*	tree;		/* tree for t Ep i/o */
	int	ntree;		/* number of dummy Eds in tree */
	Pcidev*	pcidev;
};

#define dqprint		if(debug || io &amp;&amp; io-&gt;debug)print
#define ddqprint		if(debug&gt;1 || (io &amp;&amp; io-&gt;debug&gt;1))print
#define diprint		if(debug || iso &amp;&amp; iso-&gt;debug)print
#define ddiprint		if(debug&gt;1 || (iso &amp;&amp; iso-&gt;debug&gt;1))print
#define TRUNC(x, sz)	((x) &amp; ((sz)-1))

static int ohciinterrupts[Nttypes];
static char* iosname[] = { "idle", "install", "run", "done", "close", "FREE" };

static int debug;
static Edpool edpool;
static Tdpool tdpool;
static Ctlr* ctlrs[Nhcis];

static	QLock	usbhstate;	/* protects name space state */

static int	schedendpt(Ctlr *ub, Ep *ep);
static void	unschedendpt(Ctlr *ub, Ep *ep);
static long	qtd(Ctlr*, Ep*, int, Block*, uchar*, uchar*, int, ulong);

static char* errmsgs[] =
{
[Tdcrc]		"crc error",
[Tdbitstuff]	"bit stuffing error",
[Tdbadtog]	"bad toggle",
[Tdstalled]	Estalled,
[Tdtmout]	"timeout error",
[Tdpidchk]	"pid check error",
[Tdbadpid]	"bad pid",
[Tddataovr]	"data overrun",
[Tddataund]	"data underrun",
[Tdbufovr]	"buffer overrun",
[Tdbufund]	"buffer underrun",
[Tdnotacc]	"not accessed"
};

static void*
pa2ptr(ulong pa)
{
	if(pa == 0)
		return nil;
	else
		return KADDR(pa);
}

static ulong
ptr2pa(void *p)
{
	if(p == nil)
		return 0;
	else
		return PADDR(p);
}

static void
waitSOF(Ctlr *ub)
{
	int frame = ub-&gt;hcca-&gt;framenumber &amp; 0x3f;

	do {
		delay(2);
	} while(frame == (ub-&gt;hcca-&gt;framenumber &amp; 0x3f));
}

static char*
errmsg(int err)
{

	if(err &lt; nelem(errmsgs))
		return errmsgs[err];
	return nil;
}

static Ed*
ctlhd(Ctlr *ctlr)
{
	return pa2ptr(ctlr-&gt;ohci-&gt;ctlheaded);
}

static Ed*
bulkhd(Ctlr *ctlr)
{
	return pa2ptr(ctlr-&gt;ohci-&gt;bulkheaded);
}

static void
edlinked(Ed *ed, Ed *next)
{
	if(ed == nil)
		print("edlinked: nil ed: pc %#p\n", getcallerpc(&amp;ed));
	ed-&gt;nexted = ptr2pa(next);
	ed-&gt;next = next;
}

static void
setctlhd(Ctlr *ctlr, Ed *ed)
{
	ctlr-&gt;ohci-&gt;ctlheaded = ptr2pa(ed);
	if(ed != nil)
		ctlr-&gt;ohci-&gt;cmdsts |= Sclf;	/* reload it on next pass */
}

static void
setbulkhd(Ctlr *ctlr, Ed *ed)
{
	ctlr-&gt;ohci-&gt;bulkheaded = ptr2pa(ed);
	if(ed != nil)
		ctlr-&gt;ohci-&gt;cmdsts |= Sblf;	/* reload it on next pass */
}

static void
unlinkctl(Ctlr *ctlr, Ed *ed)
{
	Ed *this, *prev, *next;

	ctlr-&gt;ohci-&gt;control &amp;= ~Ccle;
	waitSOF(ctlr);
	this = ctlhd(ctlr);
	ctlr-&gt;ohci-&gt;ctlcurred = 0;
	prev = nil;
	while(this != nil &amp;&amp; this != ed){
		prev = this;
		this = this-&gt;next;
	}
	if(this == nil){
		print("unlinkctl: not found\n");
		return;
	}
	next = this-&gt;next;
	if(prev == nil)
		setctlhd(ctlr, next);
	else
		edlinked(prev, next);
	ctlr-&gt;ohci-&gt;control |= Ccle;
	edlinked(ed, nil);		/* wipe out next field */
}

static void
unlinkbulk(Ctlr *ctlr, Ed *ed)
{
	Ed *this, *prev, *next;

	ctlr-&gt;ohci-&gt;control &amp;= ~Cble;
	waitSOF(ctlr);
	this = bulkhd(ctlr);
	ctlr-&gt;ohci-&gt;bulkcurred = 0;
	prev = nil;
	while(this != nil &amp;&amp; this != ed){
		prev = this;
		this = this-&gt;next;
	}
	if(this == nil){
		print("unlinkbulk: not found\n");
		return;
	}
	next = this-&gt;next;
	if(prev == nil)
		setbulkhd(ctlr, next);
	else
		edlinked(prev, next);
	ctlr-&gt;ohci-&gt;control |= Cble;
	edlinked(ed, nil);		/* wipe out next field */
}

static void
edsetaddr(Ed *ed, ulong addr)
{
	ulong ctrl;

	ctrl = ed-&gt;ctrl &amp; ~((Epmax&lt;&lt;7)|Devmax);
	ctrl |= (addr &amp; ((Epmax&lt;&lt;7)|Devmax));
	ed-&gt;ctrl = ctrl;
}

static void
edsettog(Ed *ed, int c)
{
	if(c != 0)
		ed-&gt;head |= Edtoggle;
	else
		ed-&gt;head &amp;= ~Edtoggle;
}

static int
edtoggle(Ed *ed)
{
	return ed-&gt;head &amp; Edtoggle;
}

static int
edhalted(Ed *ed)
{
	return ed-&gt;head &amp; Edhalt;
}

static int
edmaxpkt(Ed *ed)
{
	return (ed-&gt;ctrl &gt;&gt; Edmpsshift) &amp; Edmpsmask;
}

static void
edsetmaxpkt(Ed *ed, int m)
{
	ulong c;

	c = ed-&gt;ctrl &amp; ~(Edmpsmask &lt;&lt; Edmpsshift);
	ed-&gt;ctrl = c | ((m&amp;Edmpsmask) &lt;&lt; Edmpsshift);
}

static int
tderrs(Td *td)
{
	return (td-&gt;ctrl &gt;&gt; Tdccshift) &amp; Tdccmask;
}

static int
tdtok(Td *td)
{
	return (td-&gt;ctrl &amp; Tdtokmask);
}

static Td*
tdalloc(void)
{
	Td *td;
	Td *pool;
	int i;

	lock(&amp;tdpool);
	if(tdpool.free == nil){
		ddprint("ohci: tdalloc %d Tds\n", Incr);
		pool = xspanalloc(Incr*sizeof(Td), Align, 0);
		if(pool == nil)
			panic("tdalloc");
		for(i=Incr; --i&gt;=0;){
			pool[i].next = tdpool.free;
			tdpool.free = &amp;pool[i];
		}
		tdpool.nalloc += Incr;
		tdpool.nfree += Incr;
	}
	tdpool.ninuse++;
	tdpool.nfree--;
	td = tdpool.free;
	tdpool.free = td-&gt;next;
	memset(td, 0, sizeof(Td));
	unlock(&amp;tdpool);

	assert(((uintptr)td &amp; 0xF) == 0);
	return td;
}

static void
tdfree(Td *td)
{
	if(td == 0)
		return;
	freeb(td-&gt;bp);
	td-&gt;bp = nil;
	lock(&amp;tdpool);
	if(td-&gt;nexttd == 0x77777777)
		panic("ohci: tdfree: double free");
	memset(td, 7, sizeof(Td));	/* poison */
	td-&gt;next = tdpool.free;
	tdpool.free = td;
	tdpool.ninuse--;
	tdpool.nfree++;
	unlock(&amp;tdpool);
}

static Ed*
edalloc(void)
{
	Ed *ed, *pool;
	int i;

	lock(&amp;edpool);
	if(edpool.free == nil){
		ddprint("ohci: edalloc %d Eds\n", Incr);
		pool = xspanalloc(Incr*sizeof(Ed), Align, 0);
		if(pool == nil)
			panic("edalloc");
		for(i=Incr; --i&gt;=0;){
			pool[i].next = edpool.free;
			edpool.free = &amp;pool[i];
		}
		edpool.nalloc += Incr;
		edpool.nfree += Incr;
	}
	edpool.ninuse++;
	edpool.nfree--;
	ed = edpool.free;
	edpool.free = ed-&gt;next;
	memset(ed, 0, sizeof(Ed));
	unlock(&amp;edpool);

	return ed;
}

static void
edfree(Ed *ed)
{
	Td *td, *next;
	int i;

	if(ed == 0)
		return;
	i = 0;
	for(td = ed-&gt;tds; td != nil; td = next){
		next = td-&gt;next;
		tdfree(td);
		if(i++ &gt; 2000){
			print("ohci: bug: ed with more than 2000 tds\n");
			break;
		}
	}
	lock(&amp;edpool);
	if(ed-&gt;nexted == 0x99999999)
		panic("ohci: edfree: double free");
	memset(ed, 9, sizeof(Ed));	/* poison */
	ed-&gt;next = edpool.free;
	edpool.free = ed;
	edpool.ninuse--;
	edpool.nfree++;
	unlock(&amp;edpool);
	ddprint("edfree: ed %#p\n", ed);
}

/*
 * return smallest power of 2 &gt;= n
 */
static int
flog2(int n)
{
	int i;

	for(i = 0; (1 &lt;&lt; i) &lt; n; i++)
		;
	return i;
}

/*
 * return smallest power of 2 &lt;= n
 */
static int
flog2lower(int n)
{
	int i;

	for(i = 0; (1 &lt;&lt; (i + 1)) &lt;= n; i++)
		;
	return i;
}

static int
pickschedq(Qtree *qt, int pollival, ulong bw, ulong limit)
{
	int i, j, d, upperb, q;
	ulong best, worst, total;

	d = flog2lower(pollival);
	if(d &gt; qt-&gt;depth)
		d = qt-&gt;depth;
	q = -1;
	worst = 0;
	best = ~0;
	upperb = (1 &lt;&lt; (d+1)) - 1;
	for(i = (1 &lt;&lt; d) - 1; i &lt; upperb; i++){
		total = qt-&gt;bw[0];
		for(j = i; j &gt; 0; j = (j - 1) / 2)
			total += qt-&gt;bw[j];
		if(total &lt; best){
			best = total;
			q = i;
		}
		if(total &gt; worst)
			worst = total;
	}
	if(worst + bw &gt;= limit)
		return -1;
	return q;
}

static int
schedq(Ctlr *ctlr, Qio *io, int pollival)
{
	int q;
	Ed *ted;

	q = pickschedq(ctlr-&gt;tree, pollival, io-&gt;bw, ~0);
	ddqprint("ohci: sched %#p q %d, ival %d, bw %ld\n", io, q, pollival, io-&gt;bw);
	if(q &lt; 0){
		print("ohci: no room for ed\n");
		return -1;
	}
	ctlr-&gt;tree-&gt;bw[q] += io-&gt;bw;
	ted = ctlr-&gt;tree-&gt;root[q];
	io-&gt;sched = q;
	edlinked(io-&gt;ed, ted-&gt;next);
	edlinked(ted, io-&gt;ed);
	io-&gt;ed-&gt;inext = ctlr-&gt;intrhd;
	ctlr-&gt;intrhd = io-&gt;ed;
	return 0;
}

static void
unschedq(Ctlr *ctlr, Qio *qio)
{
	int q;
	Ed *prev, *this, *next;
	Ed **l;

	q = qio-&gt;sched;
	if(q &lt; 0)
		return;
	ctlr-&gt;tree-&gt;bw[q] -= qio-&gt;bw;

	prev = ctlr-&gt;tree-&gt;root[q];
	this = prev-&gt;next;
	while(this != nil &amp;&amp; this != qio-&gt;ed){
		prev = this;
		this = this-&gt;next;
	}
	if(this == nil)
		print("ohci: unschedq %d: not found\n", q);
	else{
		next = this-&gt;next;
		edlinked(prev, next);
	}
	waitSOF(ctlr);
	for(l = &amp;ctlr-&gt;intrhd; *l != nil; l = &amp;(*l)-&gt;inext)
		if(*l == qio-&gt;ed){
			*l = (*l)-&gt;inext;
			return;
		}
	print("ohci: unschedq: ed %#p not found\n", qio-&gt;ed);
}

static char*
seprinttdtok(char *s, char *e, int tok)
{
	switch(tok){
	case Tdtoksetup:
		s = seprint(s, e, " setup");
		break;
	case Tdtokin:
		s = seprint(s, e, " in");
		break;
	case Tdtokout:
		s = seprint(s, e, " out");
		break;
	}
	return s;
}


static char*
seprinttd(char *s, char *e, Td *td, int iso)
{
	int i;
	Block *bp;

	if(td == nil)
		return seprint(s, e, "&lt;nil td&gt;\n");
	s = seprint(s, e, "%#p ep %#p ctrl %#p", td, td-&gt;ep, td-&gt;ctrl);
	s = seprint(s, e, " cc=%#ulx", (td-&gt;ctrl &gt;&gt; Tdccshift) &amp; Tdccmask);
	if(iso == 0){
		if((td-&gt;ctrl &amp; Tdround) != 0)
			s = seprint(s, e, " rnd");
		s = seprinttdtok(s, e, td-&gt;ctrl &amp; Tdtokmask);
		if((td-&gt;ctrl &amp; Tdusetog) != 0)
			s = seprint(s, e, " d%d", (td-&gt;ctrl &amp; Tddata1) ? 1 : 0);
		else
			s = seprint(s, e, " d-");
		s = seprint(s, e, " ec=%uld", (td-&gt;ctrl &gt;&gt; Tderrshift) &amp; Tderrmask);
	}else{
		s = seprint(s, e, " fc=%uld", (td-&gt;ctrl &gt;&gt; Tdfcshift) &amp; Tdfcmask);
		s = seprint(s, e, " sf=%uld", td-&gt;ctrl &amp; Tdsfmask);
	}
	s = seprint(s, e, " cbp0 %#p cbp %#p next %#p be %#p %s",
		td-&gt;cbp0, td-&gt;cbp, td-&gt;nexttd, td-&gt;be, td-&gt;last ? "last" : "");
	s = seprint(s, e, "\n\t\t%ld bytes", td-&gt;nbytes);
	if((bp = td-&gt;bp) != nil){
		s = seprint(s, e, " rp %#p wp %#p ", bp-&gt;rp, bp-&gt;wp);
		if(BLEN(bp) &gt; 0)
			s = seprintdata(s, e, bp-&gt;rp, bp-&gt;wp - bp-&gt;rp);
	}
	if(iso == 0)
		return seprint(s, e, "\n");
	s = seprint(s, e, "\n\t\t");
	/* we use only offsets[0] */
	i = 0;
	s = seprint(s, e, "[%d] %#ux cc=%#ux sz=%ud\n", i, td-&gt;offsets[i],
		(td-&gt;offsets[i] &gt;&gt; Tdiccshift) &amp; Tdiccmask,
		td-&gt;offsets[i] &amp; 0x7FF);
	return s;
}

static void
dumptd(Td *td, char *p, int iso)
{
	static char buf[512];	/* Too much */
	char *s;

	s = seprint(buf, buf+sizeof(buf), "%s: ", p);
	s = seprinttd(s, buf+sizeof(buf), td, iso);
	if(s &gt; buf &amp;&amp; s[-1] != '\n')
		s[-1] = '\n';
	print("\t%s", buf);
}

static void
dumptds(Td *td, char *p, int iso)
{
	int i;

	for(i = 0; td != nil; td = td-&gt;next){
		dumptd(td, p, iso);
		if(td-&gt;last)
			break;
		if(tdtok(td) == Tdtokin &amp;&amp; ++i &gt; 2){
			print("\t\t...\n");
			break;
		}
	}
}

static void
dumped(Ed *ed)
{
	char *buf, *s, *e;

	if(ed == nil){
		print("&lt;null ed&gt;\n");
		return;
	}
	buf = malloc(512);
	/* no waserror; may want to use from interrupt context */
	if(buf == nil)
		return;
	e = buf+512;
	s = seprint(buf, e, "\ted %#p: ctrl %#p", ed, ed-&gt;ctrl);
	if((ed-&gt;ctrl &amp; Edskip) != 0)
		s = seprint(s, e, " skip");
	if((ed-&gt;ctrl &amp; Ediso) != 0)
		s = seprint(s, e, " iso");
	if((ed-&gt;ctrl &amp; Edlow) != 0)
		s = seprint(s, e, " low");
	s = seprint(s, e, " d%d", (ed-&gt;head &amp; Edtoggle) ? 1 : 0);
	if((ed-&gt;ctrl &amp; Eddirmask) == Edin)
		s = seprint(s, e, " in");
	if((ed-&gt;ctrl &amp; Eddirmask) == Edout)
		s = seprint(s, e, " out");
	if(edhalted(ed))
		s = seprint(s, e, " hlt");
	s = seprint(s, e, " ep%uld.%uld", (ed-&gt;ctrl&gt;&gt;7)&amp;Epmax, ed-&gt;ctrl&amp;0x7f);
	s = seprint(s, e, " maxpkt %uld", (ed-&gt;ctrl&gt;&gt;Edmpsshift)&amp;Edmpsmask);
	seprint(s, e, " tail %#p head %#p next %#p\n",ed-&gt;tail,ed-&gt;head,ed-&gt;nexted);
	print("%s", buf);
	free(buf);
	if(ed-&gt;tds != nil &amp;&amp; (ed-&gt;ctrl &amp; Ediso) == 0)
		dumptds(ed-&gt;tds, "td", 0);
}

static char*
seprintio(char *s, char *e, Qio *io, char *pref)
{
	s = seprint(s, e, "%s qio %#p ed %#p", pref, io, io-&gt;ed);
	s = seprint(s, e, " tog %d iot %ld err %s id %#ulx",
		io-&gt;toggle, io-&gt;iotime, io-&gt;err, io-&gt;usbid);
	s = seprinttdtok(s, e, io-&gt;tok);
	s = seprint(s, e, " %s\n", iosname[io-&gt;state]);
	return s;
}

static char*
seprintep(char* s, char* e, Ep *ep)
{
	Isoio *iso;
	Qio *io;
	Ctlio *cio;

	if(ep == nil)
		return seprint(s, e, "&lt;nil ep&gt;\n");
	if(ep-&gt;aux == nil)
		return seprint(s, e, "no mdep\n");
	switch(ep-&gt;ttype){
	case Tctl:
		cio = ep-&gt;aux;
		s = seprintio(s, e, cio, "c");
		s = seprint(s, e, "\trepl %d ndata %d\n", ep-&gt;rhrepl, cio-&gt;ndata);
		break;
	case Tbulk:
	case Tintr:
		io = ep-&gt;aux;
		if(ep-&gt;mode != OWRITE)
			s = seprintio(s, e, &amp;io[OREAD], "r");
		if(ep-&gt;mode != OREAD)
			s = seprintio(s, e, &amp;io[OWRITE], "w");
		break;
	case Tiso:
		iso = ep-&gt;aux;
		s = seprintio(s, e, iso, "w");
		s = seprint(s, e, "\tntds %d avail %d frno %uld left %uld next avail %#p\n",
			iso-&gt;nframes, iso-&gt;navail, iso-&gt;frno, iso-&gt;left, iso-&gt;atds);
		break;
	}
	return s;
}

static char*
seprintctl(char *s, char *se, ulong ctl)
{
	s = seprint(s, se, "en=");
	if((ctl&amp;Cple) != 0)
		s = seprint(s, se, "p");
	if((ctl&amp;Cie) != 0)
		s = seprint(s, se, "i");
	if((ctl&amp;Ccle) != 0)
		s = seprint(s, se, "c");
	if((ctl&amp;Cble) != 0)
		s = seprint(s, se, "b");
	switch(ctl &amp; Cfsmask){
	case Cfsreset:
		return seprint(s, se, " reset");
	case Cfsresume:
		return seprint(s, se, " resume");
	case Cfsoper:
		return seprint(s, se, " run");
	case Cfssuspend:
		return seprint(s, se, " suspend");
	default:
		return seprint(s, se, " ???");
	}
}

static void
dump(Hci *hp)
{
	Ctlr *ctlr;
	Ed *ed;
	char cs[20];

	ctlr = hp-&gt;aux;
	ilock(ctlr);
	seprintctl(cs, cs+sizeof(cs), ctlr-&gt;ohci-&gt;control);
	print("ohci ctlr %#p: frno %#ux ctl %#lux %s sts %#lux intr %#lux\n",
		ctlr, ctlr-&gt;hcca-&gt;framenumber, ctlr-&gt;ohci-&gt;control, cs,
		ctlr-&gt;ohci-&gt;cmdsts, ctlr-&gt;ohci-&gt;intrsts);
	print("ctlhd %#ulx cur %#ulx bulkhd %#ulx cur %#ulx done %#ulx\n",
		ctlr-&gt;ohci-&gt;ctlheaded, ctlr-&gt;ohci-&gt;ctlcurred,
		ctlr-&gt;ohci-&gt;bulkheaded, ctlr-&gt;ohci-&gt;bulkcurred,
		ctlr-&gt;ohci-&gt;donehead);
	if(ctlhd(ctlr) != nil)
		print("[ctl]\n");
	for(ed = ctlhd(ctlr); ed != nil; ed = ed-&gt;next)
		dumped(ed);
	if(bulkhd(ctlr) != nil)
		print("[bulk]\n");
	for(ed = bulkhd(ctlr); ed != nil; ed = ed-&gt;next)
		dumped(ed);
	if(ctlr-&gt;intrhd != nil)
		print("[intr]\n");
	for(ed = ctlr-&gt;intrhd; ed != nil; ed = ed-&gt;inext)
		dumped(ed);
	if(ctlr-&gt;tree-&gt;root[0]-&gt;next != nil)
		print("[iso]");
	for(ed = ctlr-&gt;tree-&gt;root[0]-&gt;next; ed != nil; ed = ed-&gt;next)
		dumped(ed);
	print("%d eds in tree\n", ctlr-&gt;ntree);
	iunlock(ctlr);
	lock(&amp;tdpool);
	print("%d tds allocated = %d in use + %d free\n",
		tdpool.nalloc, tdpool.ninuse, tdpool.nfree);
	unlock(&amp;tdpool);
	lock(&amp;edpool);
	print("%d eds allocated = %d in use + %d free\n",
		edpool.nalloc, edpool.ninuse, edpool.nfree);
	unlock(&amp;edpool);
}

/*
 * Compute size for the next iso Td and setup its
 * descriptor for I/O according to the buffer size.
 */
static void
isodtdinit(Ep *ep, Isoio *iso, Td *td)
{
	Block *bp;
	long size;
	int i;

	bp = td-&gt;bp;
	assert(bp != nil &amp;&amp; BLEN(bp) == 0);
	size = (ep-&gt;hz+iso-&gt;left) * ep-&gt;pollival / 1000;
	iso-&gt;left = (ep-&gt;hz+iso-&gt;left) * ep-&gt;pollival % 1000;
	size *= ep-&gt;samplesz;
	if(size &gt; ep-&gt;maxpkt){
		print("ohci: ep%d.%d: size &gt; maxpkt\n",
			ep-&gt;dev-&gt;nb, ep-&gt;nb);
		print("size = %uld max = %ld\n", size, ep-&gt;maxpkt);
		size = ep-&gt;maxpkt;
	}
	td-&gt;nbytes = size;
	memset(bp-&gt;wp, 0, size);	/* in case we don't fill it on time */
	td-&gt;cbp0 = td-&gt;cbp = ptr2pa(bp-&gt;rp) &amp; ~0xFFF;
	td-&gt;ctrl = TRUNC(iso-&gt;frno, Ntdframes);
	td-&gt;offsets[0] = (ptr2pa(bp-&gt;rp) &amp; 0xFFF);
	td-&gt;offsets[0] |= (Tdnotacc &lt;&lt; Tdiccshift);
	/* in case the controller checks out the offests... */
	for(i = 1; i &lt; nelem(td-&gt;offsets); i++)
		td-&gt;offsets[i] = td-&gt;offsets[0];
	td-&gt;be = ptr2pa(bp-&gt;rp + size - 1);
	td-&gt;ctrl |= (0 &lt;&lt; Tdfcshift);	/* frame count is 1 */

	iso-&gt;frno = TRUNC(iso-&gt;frno + ep-&gt;pollival, Ntdframes);
}

/*
 * start I/O on the dummy td and setup a new dummy to fill up.
 */
static void
isoadvance(Ep *ep, Isoio *iso, Td *td)
{
	Td *dtd;

	dtd = iso-&gt;atds;
	iso-&gt;atds = dtd-&gt;anext;
	iso-&gt;navail--;
	dtd-&gt;anext = nil;
	dtd-&gt;bp-&gt;wp = dtd-&gt;bp-&gt;rp;
	dtd-&gt;nexttd = 0;
	td-&gt;nexttd = ptr2pa(dtd);
	isodtdinit(ep, iso, dtd);
	iso-&gt;ed-&gt;tail = ptr2pa(dtd);
}

static int
isocanwrite(void *a)
{
	Isoio *iso;

	iso = a;
	return iso-&gt;state == Qclose || iso-&gt;err != nil ||
		iso-&gt;navail &gt; iso-&gt;nframes / 2;
}

/*
 * Service a completed/failed Td from the done queue.
 * It may be of any transfer type.
 * The queue is not in completion order.
 * (It's actually in reverse completion order).
 *
 * When an error, a short packet, or a last Td is found
 * we awake the process waiting for the transfer.
 * Although later we will process other Tds completed
 * before, epio won't be able to touch the current Td
 * until interrupt returns and releases the lock on the
 * controller.
 */
static void
qhinterrupt(Ctlr *, Ep *ep, Qio *io, Td *td, int)
{
	Block *bp;
	int mode, err;
	Ed *ed;

	ed = io-&gt;ed;
	if(io-&gt;state != Qrun)
		return;
	if(tdtok(td) == Tdtokin)
		mode = OREAD;
	else
		mode = OWRITE;
	bp = td-&gt;bp;
	err = tderrs(td);

	switch(err){
	case Tddataovr:			/* Overrun is not an error */
		break;
	case Tdok:
		/* virtualbox doesn't always report underflow on short packets */
		if(td-&gt;cbp == 0)
			break;
		/* fall through */
	case Tddataund:
		/* short input packets are ok */
		if(mode == OREAD){
			if(td-&gt;cbp == 0)
				panic("ohci: short packet but cbp == 0");
			/*
			 * td-&gt;cbp and td-&gt;cbp0 are the real addresses
			 * corresponding to virtual addresses bp-&gt;wp and
			 * bp-&gt;rp respectively.
			 */
			bp-&gt;wp = bp-&gt;rp + (td-&gt;cbp - td-&gt;cbp0);
			if(bp-&gt;wp &lt; bp-&gt;rp)
				panic("ohci: wp &lt; rp");
			/*
			 * It's ok. clear error and flag as last in xfer.
			 * epio must ignore following Tds.
			 */
			td-&gt;last = 1;
			td-&gt;ctrl &amp;= ~(Tdccmask &lt;&lt; Tdccshift);
			break;
		}
		/* else fall; it's an error */
	case Tdcrc:
	case Tdbitstuff:
	case Tdbadtog:
	case Tdstalled:
	case Tdtmout:
	case Tdpidchk:
	case Tdbadpid:
		bp-&gt;wp = bp-&gt;rp;	/* no bytes in xfer. */
		io-&gt;err = errmsg(err);
		if(debug || ep-&gt;debug){
			print("tdinterrupt: failed err %d (%s)\n", err, io-&gt;err);
			dumptd(td, "failed", ed-&gt;ctrl &amp; Ediso);
		}
		td-&gt;last = 1;
		break;
	default:
		panic("ohci: td cc %ud unknown", err);
	}

	if(td-&gt;last != 0){
		/*
		 * clear td list and halt flag.
		 */
		ed-&gt;head = (ed-&gt;head &amp; Edtoggle) | ed-&gt;tail;
		ed-&gt;tds = pa2ptr(ed-&gt;tail);
		io-&gt;state = Qdone;
		wakeup(io);
	}
}

/*
 * BUG: Iso input streams are not implemented.
 */
static void
isointerrupt(Ctlr *ctlr, Ep *ep, Qio *io, Td *td, int)
{
	Isoio *iso;
	Block *bp;
	Ed *ed;
	int err, isoerr;

	iso = ep-&gt;aux;
	ed = io-&gt;ed;
	if(io-&gt;state == Qclose)
		return;
	bp = td-&gt;bp;
	/*
	 * When we get more than half the frames consecutive errors
	 * we signal an actual error. Errors in the entire Td are
	 * more serious and are always singaled.
	 * Errors like overrun are not really errors. In fact, for
	 * output, errors cannot be really detected. The driver will
	 * hopefully notice I/O errors on input endpoints and detach the device.
	 */
	err = tderrs(td);
	isoerr = (td-&gt;offsets[0] &gt;&gt; Tdiccshift) &amp; Tdiccmask;
	if(isoerr == Tdok || isoerr == Tdnotacc)
		iso-&gt;nerrs = 0;
	else if(iso-&gt;nerrs++ &gt; iso-&gt;nframes/2)
		err = Tdstalled;
	if(err != Tdok &amp;&amp; err != Tddataovr){
		bp-&gt;wp = bp-&gt;rp;
		io-&gt;err = errmsg(err);
		if(debug || ep-&gt;debug){
			print("ohci: isointerrupt: ep%d.%d: err %d (%s) frnum 0x%lux\n",
				ep-&gt;dev-&gt;nb, ep-&gt;nb,
				err, errmsg(err), ctlr-&gt;ohci-&gt;fmnumber);
			dumptd(td, "failed", ed-&gt;ctrl &amp; Ediso);
		}
	}
	td-&gt;bp-&gt;wp = td-&gt;bp-&gt;rp;
	td-&gt;nbytes = 0;
	td-&gt;anext = iso-&gt;atds;
	iso-&gt;atds = td;
	iso-&gt;navail++;
	/*
	 * If almost all Tds are avail the user is not doing I/O at the
	 * required rate. We put another Td in place to keep the polling rate.
	 */
	if(iso-&gt;err == nil &amp;&amp; iso-&gt;navail &gt; iso-&gt;nframes - 10)
		isoadvance(ep, iso, pa2ptr(iso-&gt;ed-&gt;tail));
	/*
	 * If there's enough buffering futher I/O can be done.
	 */
	if(isocanwrite(iso))
		wakeup(iso);
}

static void
interrupt(Ureg *, void *arg)
{
	Td *td, *ntd;
	Hci *hp;
	Ctlr *ctlr;
	ulong status, curred;
	int i, frno;

	hp = arg;
	ctlr = hp-&gt;aux;
	ilock(ctlr);
	ctlr-&gt;ohci-&gt;intrdisable = Mie;
	coherence();
	status = ctlr-&gt;ohci-&gt;intrsts &amp; ctlr-&gt;ohci-&gt;intrenable;
	status &amp;= Oc|Rhsc|Fno|Ue|Rd|Sf|Wdh|So;
	frno = TRUNC(ctlr-&gt;ohci-&gt;fmnumber, Ntdframes);
	if(status &amp; Wdh){
		/* lsb of donehead has bit to flag other intrs.  */
		td = pa2ptr(ctlr-&gt;hcca-&gt;donehead &amp; ~0xF);

		for(i = 0; td != nil &amp;&amp; i &lt; 1024; i++){
			if(0)ddprint("ohci tdinterrupt: td %#p\n", td);
			ntd = pa2ptr(td-&gt;nexttd &amp; ~0xF);
			td-&gt;nexttd = 0;
			if(td-&gt;ep == nil || td-&gt;io == nil)
				panic("ohci: interrupt: ep %#p io %#p",
					td-&gt;ep, td-&gt;io);
			ohciinterrupts[td-&gt;ep-&gt;ttype]++;
			if(td-&gt;ep-&gt;ttype == Tiso)
				isointerrupt(ctlr, td-&gt;ep, td-&gt;io, td, frno);
			else
				qhinterrupt(ctlr, td-&gt;ep, td-&gt;io, td, frno);
			td = ntd;
		}
		if(i &gt;= 1024)
			print("ohci: bug: more than 1024 done Tds?\n");
		ctlr-&gt;hcca-&gt;donehead = 0;
	}

	ctlr-&gt;ohci-&gt;intrsts = status;
	status &amp;= ~Wdh;
	status &amp;= ~Sf;
	if(status &amp; So){
		print("ohci: sched overrun: too much load\n");
		ctlr-&gt;overrun++;
		status &amp;= ~So;
	}
	if((status &amp; Ue) != 0){
		curred = ctlr-&gt;ohci-&gt;periodcurred;
		print("ohci: unrecoverable error frame 0x%.8lux ed 0x%.8lux, "
			"ints %d %d %d %d\n",
			ctlr-&gt;ohci-&gt;fmnumber, curred,
			ohciinterrupts[Tctl], ohciinterrupts[Tintr],
			ohciinterrupts[Tbulk], ohciinterrupts[Tiso]);
		if(curred != 0)
			dumped(pa2ptr(curred));
		status &amp;= ~Ue;
	}
	if(status != 0)
		print("ohci interrupt: unhandled sts 0x%.8lux\n", status);
	ctlr-&gt;ohci-&gt;intrenable = Mie | Wdh | Ue;
	iunlock(ctlr);
}

/*
 * The old dummy Td is used to implement the new Td.
 * A new dummy is linked at the end of the old one and
 * returned, to link further Tds if needed.
 */
static Td*
epgettd(Ep *ep, Qio *io, Td **dtdp, int flags, void *a, int count)
{
	Td *td, *dtd;
	Block *bp;

	if(count &lt;= BY2PG)
		bp = allocb(count);
	else{
		if(count &gt; 2*BY2PG)
			panic("ohci: transfer &gt; two pages");
		/* maximum of one physical page crossing allowed */
		bp = allocb(count+BY2PG);
		bp-&gt;rp = (uchar*)PGROUND((uintptr)bp-&gt;rp);
		bp-&gt;wp = bp-&gt;rp;
	}
	dtd = *dtdp;
	td = dtd;
	td-&gt;bp = bp;
	if(count &gt; 0){
		td-&gt;cbp0 = td-&gt;cbp = ptr2pa(bp-&gt;wp);
		td-&gt;be = ptr2pa(bp-&gt;wp + count - 1);
		if(a != nil){
			/* validaddr((uintptr)a, count, 0); DEBUG */
			memmove(bp-&gt;wp, a, count);
		}
		bp-&gt;wp += count;
	}
	td-&gt;nbytes = count;
	td-&gt;ctrl = io-&gt;tok|Tdusetog|io-&gt;toggle|flags;
	if(io-&gt;toggle == Tddata0)
		io-&gt;toggle = Tddata1;
	else
		io-&gt;toggle = Tddata0;
	assert(td-&gt;ep == ep);
	td-&gt;io = io;
 	dtd = tdalloc();	/* new dummy */
	dtd-&gt;ep = ep;
	td-&gt;nexttd = ptr2pa(dtd);
	td-&gt;next = dtd;
	*dtdp = dtd;
	return td;
}

/*
 * Try to get them idle
 */
static void
aborttds(Qio *io)
{
	Ed *ed;
	Td *td;

	ed = io-&gt;ed;
	if(ed == nil)
		return;
	ed-&gt;ctrl |= Edskip;
	for(td = ed-&gt;tds; td != nil; td = td-&gt;next)
		if(td-&gt;bp != nil)
			td-&gt;bp-&gt;wp = td-&gt;bp-&gt;rp;
	ed-&gt;head = (ed-&gt;head&amp;0xF) | ed-&gt;tail;
	if((ed-&gt;ctrl &amp; Ediso) == 0)
		ed-&gt;tds = pa2ptr(ed-&gt;tail);
}

static int
epiodone(void *a)
{
	Qio *io;

	io = a;
	return io-&gt;state != Qrun;
}

static void
epiowait(Ctlr *ctlr, Qio *io, int tmout, ulong)
{
	Ed *ed;
	int timedout;

	ed = io-&gt;ed;
	if(0)ddqprint("ohci io %#p sleep on ed %#p state %s\n",
		io, ed, iosname[io-&gt;state]);
	timedout = 0;
	if(waserror()){
		dqprint("ohci io %#p ed %#p timed out\n", io, ed);
		timedout++;
	}else{
		if(tmout == 0)
			sleep(io, epiodone, io);
		else
			tsleep(io, epiodone, io, tmout);
		poperror();
	}
	ilock(ctlr);
	if(io-&gt;state == Qrun)
		timedout = 1;
	else if(io-&gt;state != Qdone &amp;&amp; io-&gt;state != Qclose)
		panic("epio: ed not done and not closed");
	if(timedout){
		aborttds(io);
		io-&gt;err = "request timed out";
		iunlock(ctlr);
		if(!waserror()){
			tsleep(&amp;up-&gt;sleep, return0, 0, Abortdelay);
			poperror();
		}
		ilock(ctlr);
	}
	if(io-&gt;state != Qclose)
		io-&gt;state = Qidle;
	iunlock(ctlr);
}

/*
 * Non iso I/O.
 * To make it work for control transfers, the caller may
 * lock the Qio for the entire control transfer.
 */
static long
epio(Ep *ep, Qio *io, void *a, long count, int mustlock)
{
	Ed *ed;
	Ctlr *ctlr;
	char buf[80];
	char *err;
	uchar *c;
	Td *td, *ltd, *ntd, *td0;
	int last, ntds, tmout;
	long tot, n;
	ulong load;

	ed = io-&gt;ed;
	ctlr = ep-&gt;hp-&gt;aux;
	io-&gt;debug = ep-&gt;debug;
	tmout = ep-&gt;tmout;
	ddeprint("ohci: %s ep%d.%d io %#p count %ld\n",
		io-&gt;tok == Tdtokin ? "in" : "out",
		ep-&gt;dev-&gt;nb, ep-&gt;nb, io, count);
	if((debug &gt; 1 || ep-&gt;debug &gt; 1) &amp;&amp; io-&gt;tok != Tdtokin){
		seprintdata(buf, buf+sizeof(buf), a, count);
		print("\t%s\n", buf);
	}
	if(mustlock){
		qlock(io);
		if(waserror()){
			qunlock(io);
			nexterror();
		}
	}
	io-&gt;err = nil;
	ilock(ctlr);
	if(io-&gt;state == Qclose){	/* Tds released by cancelio */
		iunlock(ctlr);
		error(io-&gt;err ? io-&gt;err : Eio);
	}
	if(io-&gt;state != Qidle)
		panic("epio: qio not idle");
	io-&gt;state = Qinstall;

	c = a;
	ltd = td0 = ed-&gt;tds;
	load = tot = 0;
	do{
		n = 2*BY2PG;
		if(count-tot &lt; n)
			n = count-tot;
		if(c != nil &amp;&amp; io-&gt;tok != Tdtokin)
			td = epgettd(ep, io, &amp;ltd, 0, c+tot, n);
		else
			td = epgettd(ep, io, &amp;ltd, 0, nil, n);
		tot += n;
		load += ep-&gt;load;
	}while(tot &lt; count);
	if(td0 == nil || ltd == nil || td0 == ltd)
		panic("epio: no td");
	td-&gt;last = 1;
	if(debug &gt; 2 || ep-&gt;debug &gt; 2)
		dumptds(td0, "put td", ep-&gt;ttype == Tiso);
	iunlock(ctlr);

	ilock(ctlr);
	if(io-&gt;state != Qclose){
		io-&gt;iotime = TK2MS(MACHP(0)-&gt;ticks);
		io-&gt;state = Qrun;
		ed-&gt;tail = ptr2pa(ltd);
		if(ep-&gt;ttype == Tctl)
			ctlr-&gt;ohci-&gt;cmdsts |= Sclf;
		else if(ep-&gt;ttype == Tbulk)
			ctlr-&gt;ohci-&gt;cmdsts |= Sblf;
	}
	iunlock(ctlr);

	epiowait(ctlr, io, tmout, load);
	ilock(ctlr);
	if(debug &gt; 1 || ep-&gt;debug &gt; 1)
		dumptds(td0, "got td", 0);
	iunlock(ctlr);

	tot = 0;
	c = a;
	ntds = last = 0;
	for(td = td0; td != ltd; td = ntd){
		ntds++;
		/*
		 * If the Td is flagged as last we must
		 * ignore any following Td. The block may
		 * seem to have bytes but interrupt has not seen
		 * those Tds through the done queue, and they are void.
		 */
		if(last == 0 &amp;&amp; tderrs(td) == Tdok){
			n = BLEN(td-&gt;bp);
			tot += n;
			if(c != nil &amp;&amp; tdtok(td) == Tdtokin &amp;&amp; n &gt; 0){
				memmove(c, td-&gt;bp-&gt;rp, n);
				c += n;
			}
		}
		last |= td-&gt;last;
		ntd = td-&gt;next;
		tdfree(td);
	}
	if(edtoggle(ed) == 0)
		io-&gt;toggle = Tddata0;
	else
		io-&gt;toggle = Tddata1;

	err = io-&gt;err;
	if(mustlock){
		qunlock(io);
		poperror();
	}
	ddeprint("ohci: io %#p: %d tds: return %ld err '%s'\n\n",
		io, ntds, tot, err);
	if(err != nil)
		error(err);
	if(tot &lt; 0)
		error(Eio);
	return tot;
}

/*
 * halt condition was cleared on the endpoint. update our toggles.
 */
static void
clrhalt(Ep *ep)
{
	Qio *io;

	ep-&gt;clrhalt = 0;
	switch(ep-&gt;ttype){
	case Tbulk:
	case Tintr:
		io = ep-&gt;aux;
		if(ep-&gt;mode != OREAD){
			qlock(&amp;io[OWRITE]);
			io[OWRITE].toggle = Tddata0;
			deprint("ep clrhalt for io %#p\n", io+OWRITE);
			qunlock(&amp;io[OWRITE]);
		}
		if(ep-&gt;mode != OWRITE){
			qlock(&amp;io[OREAD]);
			io[OREAD].toggle = Tddata0;
			deprint("ep clrhalt for io %#p\n", io+OREAD);
			qunlock(&amp;io[OREAD]);
		}
		break;
	}
}

static long
epread(Ep *ep, void *a, long count)
{
	Ctlio *cio;
	Qio *io;
	char buf[80];
	ulong delta;

	if(ep-&gt;aux == nil)
		panic("epread: not open");

	switch(ep-&gt;ttype){
	case Tctl:
		cio = ep-&gt;aux;
		qlock(cio);
		if(waserror()){
			qunlock(cio);
			nexterror();
		}
		ddeprint("epread ctl ndata %d\n", cio-&gt;ndata);
		if(cio-&gt;ndata &lt; 0)
			error("request expected");
		else if(cio-&gt;ndata == 0){
			cio-&gt;ndata = -1;
			count = 0;
		}else{
			if(count &gt; cio-&gt;ndata)
				count = cio-&gt;ndata;
			if(count &gt; 0)
				memmove(a, cio-&gt;data, count);
			/* BUG for big transfers */
			free(cio-&gt;data);
			cio-&gt;data = nil;
			cio-&gt;ndata = 0;	/* signal EOF next time */
		}
		qunlock(cio);
		poperror();
		if(debug&gt;1 || ep-&gt;debug){
			seprintdata(buf, buf+sizeof(buf), a, count);
			print("epread: %s\n", buf);
		}
		return count;
	case Tbulk:
		io = ep-&gt;aux;
		if(ep-&gt;clrhalt)
			clrhalt(ep);
		return epio(ep, &amp;io[OREAD], a, count, 1);
	case Tintr:
		io = ep-&gt;aux;
		delta = TK2MS(MACHP(0)-&gt;ticks) - io[OREAD].iotime + 1;
		if(delta &lt; ep-&gt;pollival / 2)
			tsleep(&amp;up-&gt;sleep, return0, 0, ep-&gt;pollival/2 - delta);
		if(ep-&gt;clrhalt)
			clrhalt(ep);
		return epio(ep, &amp;io[OREAD], a, count, 1);
	case Tiso:
		panic("ohci: iso read not implemented");
		break;
	default:
		panic("epread: bad ep ttype %d", ep-&gt;ttype);
	}
	return -1;
}

/*
 * Control transfers are one setup write (data0)
 * plus zero or more reads/writes (data1, data0, ...)
 * plus a final write/read with data1 to ack.
 * For both host to device and device to host we perform
 * the entire transfer when the user writes the request,
 * and keep any data read from the device for a later read.
 * We call epio three times instead of placing all Tds at
 * the same time because doing so leads to crc/tmout errors
 * for some devices.
 * Upon errors on the data phase we must still run the status
 * phase or the device may cease responding in the future.
 */
static long
epctlio(Ep *ep, Ctlio *cio, void *a, long count)
{
	uchar *c;
	long len;

	ddeprint("epctlio: cio %#p ep%d.%d count %ld\n",
		cio, ep-&gt;dev-&gt;nb, ep-&gt;nb, count);
	if(count &lt; Rsetuplen)
		error("short usb command");
	qlock(cio);
	free(cio-&gt;data);
	cio-&gt;data = nil;
	cio-&gt;ndata = 0;
	if(waserror()){
		qunlock(cio);
		free(cio-&gt;data);
		cio-&gt;data = nil;
		cio-&gt;ndata = 0;
		nexterror();
	}

	/* set the address if unset and out of configuration state */
	if(ep-&gt;dev-&gt;state != Dconfig &amp;&amp; ep-&gt;dev-&gt;state != Dreset)
		if(cio-&gt;usbid == 0){
			cio-&gt;usbid = (ep-&gt;nb&lt;&lt;7)|(ep-&gt;dev-&gt;nb &amp; Devmax);
			edsetaddr(cio-&gt;ed, cio-&gt;usbid);
		}
	/* adjust maxpkt if the user has learned a different one */
	if(edmaxpkt(cio-&gt;ed) != ep-&gt;maxpkt)
		edsetmaxpkt(cio-&gt;ed, ep-&gt;maxpkt);
	c = a;
	cio-&gt;tok = Tdtoksetup;
	cio-&gt;toggle = Tddata0;
	if(epio(ep, cio, a, Rsetuplen, 0) &lt; Rsetuplen)
		error(Eio);

	a = c + Rsetuplen;
	count -= Rsetuplen;

	cio-&gt;toggle = Tddata1;
	if(c[Rtype] &amp; Rd2h){
		cio-&gt;tok = Tdtokin;
		len = GET2(c+Rcount);
		if(len &lt;= 0)
			error("bad length in d2h request");
		if(len &gt; Maxctllen)
			error("d2h data too large to fit in ohci");
		a = cio-&gt;data = smalloc(len+1);
	}else{
		cio-&gt;tok = Tdtokout;
		len = count;
	}
	if(len &gt; 0)
		if(waserror())
			len = -1;
		else{
			len = epio(ep, cio, a, len, 0);
			poperror();
		}
	if(c[Rtype] &amp; Rd2h){
		count = Rsetuplen;
		cio-&gt;ndata = len;
		cio-&gt;tok = Tdtokout;
	}else{
		if(len &lt; 0)
			count = -1;
		else
			count = Rsetuplen + len;
		cio-&gt;tok = Tdtokin;
	}
	cio-&gt;toggle = Tddata1;
	epio(ep, cio, nil, 0, 0);
	qunlock(cio);
	poperror();
	ddeprint("epctlio cio %#p return %ld\n", cio, count);
	return count;
}

/*
 * Put new samples in the dummy Td.
 * BUG: This does only a transfer per Td. We could do up to 8.
 */
static long
putsamples(Ctlr *ctlr, Ep *ep, Isoio *iso, uchar *b, long count)
{
	Td *td;
	ulong n;

	td = pa2ptr(iso-&gt;ed-&gt;tail);
	n = count;
	if(n &gt; td-&gt;nbytes - BLEN(td-&gt;bp))
		n = td-&gt;nbytes - BLEN(td-&gt;bp);
	assert(td-&gt;bp-&gt;wp + n &lt;= td-&gt;bp-&gt;lim);
	memmove(td-&gt;bp-&gt;wp, b, n);
	td-&gt;bp-&gt;wp += n;
	if(BLEN(td-&gt;bp) == td-&gt;nbytes){	/* full Td: activate it */
		ilock(ctlr);
		isoadvance(ep, iso, td);
		iunlock(ctlr);
	}
	return n;
}

static long
episowrite(Ep *ep, void *a, long count)
{
	long tot, nw;
	char *err;
	uchar *b;
	Ctlr *ctlr;
	Isoio *iso;

	ctlr = ep-&gt;hp-&gt;aux;
	iso = ep-&gt;aux;
	iso-&gt;debug = ep-&gt;debug;

	qlock(iso);
	if(waserror()){
		qunlock(iso);
		nexterror();
	}
	diprint("ohci: episowrite: %#p ep%d.%d\n", iso, ep-&gt;dev-&gt;nb, ep-&gt;nb);
	ilock(ctlr);
	if(iso-&gt;state == Qclose){
		iunlock(ctlr);
		error(iso-&gt;err ? iso-&gt;err : Eio);
	}
	iso-&gt;state = Qrun;
	b = a;
	for(tot = 0; tot &lt; count; tot += nw){
		while(isocanwrite(iso) == 0){
			iunlock(ctlr);
			diprint("ohci: episowrite: %#p sleep\n", iso);
			if(waserror()){
				if(iso-&gt;err == nil)
					iso-&gt;err = "I/O timed out";
				ilock(ctlr);
				break;
			}
			tsleep(iso, isocanwrite, iso, ep-&gt;tmout);
			poperror();
			ilock(ctlr);
		}
		err = iso-&gt;err;
		iso-&gt;err = nil;
		if(iso-&gt;state == Qclose || err != nil){
			iunlock(ctlr);
			error(err ? err : Eio);
		}
		if(iso-&gt;state != Qrun)
			panic("episowrite: iso not running");
		iunlock(ctlr);		/* We could page fault here */
		nw = putsamples(ctlr, ep, iso, b+tot, count-tot);
		ilock(ctlr);
	}
	if(iso-&gt;state != Qclose)
		iso-&gt;state = Qdone;
	iunlock(ctlr);
	err = iso-&gt;err;		/* in case it failed early */
	iso-&gt;err = nil;
	qunlock(iso);
	poperror();
	if(err != nil)
		error(err);
	diprint("ohci: episowrite: %#p %ld bytes\n", iso, tot);
	return tot;
}

static long
epwrite(Ep *ep, void *a, long count)
{
	Qio *io;
	Ctlio *cio;
	ulong delta;
	uchar *b;
	long tot, nw;

	if(ep-&gt;aux == nil)
		panic("ohci: epwrite: not open");
	switch(ep-&gt;ttype){
	case Tctl:
		cio = ep-&gt;aux;
		return epctlio(ep, cio, a, count);
	case Tbulk:
		io = ep-&gt;aux;
		if(ep-&gt;clrhalt)
			clrhalt(ep);
		/*
		 * Put at most Tdatomic Tds (512 bytes) at a time.
		 * Otherwise some devices produce babble errors.
		 */
		b = a;
		assert(a != nil);
		for(tot = 0; tot &lt; count ; tot += nw){
			nw = count - tot;
			if(nw &gt; Tdatomic * ep-&gt;maxpkt)
				nw = Tdatomic * ep-&gt;maxpkt;
			nw = epio(ep, &amp;io[OWRITE], b+tot, nw, 1);
		}
		return tot;
	case Tintr:
		io = ep-&gt;aux;
		delta = TK2MS(MACHP(0)-&gt;ticks) - io[OWRITE].iotime + 1;
		if(delta &lt; ep-&gt;pollival)
			tsleep(&amp;up-&gt;sleep, return0, 0, ep-&gt;pollival - delta);
		if(ep-&gt;clrhalt)
			clrhalt(ep);
		return epio(ep, &amp;io[OWRITE], a, count, 1);
	case Tiso:
		return episowrite(ep, a, count);
	default:
		panic("ohci: epwrite: bad ep ttype %d", ep-&gt;ttype);
	}
	return -1;
}

static Ed*
newed(Ctlr *ctlr, Ep *ep, Qio *io, char *)
{
	Ed *ed;
	Td *td;

	ed = io-&gt;ed = edalloc();	/* no errors raised here, really */
	td = tdalloc();
	td-&gt;ep = ep;
	td-&gt;io = io;
	ed-&gt;tail =  ptr2pa(td);
	ed-&gt;head = ptr2pa(td);
	ed-&gt;tds = td;
	ed-&gt;ep = ep;
	ed-&gt;ctrl = (ep-&gt;maxpkt &amp; Edmpsmask) &lt;&lt; Edmpsshift;
	if(ep-&gt;ttype == Tiso)
		ed-&gt;ctrl |= Ediso;
	if(waserror()){
		edfree(ed);
		io-&gt;ed = nil;
		nexterror();
	}
	/* For setup endpoints we start with the config address */
	if(ep-&gt;ttype != Tctl)
		edsetaddr(io-&gt;ed, io-&gt;usbid);
	if(ep-&gt;dev-&gt;speed == Lowspeed)
		ed-&gt;ctrl |= Edlow;
	switch(io-&gt;tok){
	case Tdtokin:
		ed-&gt;ctrl |= Edin;
		break;
	case Tdtokout:
		ed-&gt;ctrl |= Edout;
		break;
	default:
		ed-&gt;ctrl |= Edtddir;	/* Td will say */
		break;
	}

	switch(ep-&gt;ttype){
	case Tctl:
		ilock(ctlr);
		edlinked(ed, ctlhd(ctlr));
		setctlhd(ctlr, ed);
		iunlock(ctlr);
		break;
	case Tbulk:
		ilock(ctlr);
		edlinked(ed, bulkhd(ctlr));
		setbulkhd(ctlr, ed);
		iunlock(ctlr);
		break;
	case Tintr:
	case Tiso:
		ilock(ctlr);
		schedq(ctlr, io, ep-&gt;pollival);
		iunlock(ctlr);
		break;
	default:
		panic("ohci: newed: bad ttype");
	}
	poperror();
	return ed;
}

static void
isoopen(Ctlr *ctlr, Ep *ep)
{
	Td *td, *edtds;
	Isoio *iso;
	int i;

	iso = ep-&gt;aux;
	iso-&gt;usbid = (ep-&gt;nb&lt;&lt;7)|(ep-&gt;dev-&gt;nb &amp; Devmax);
	iso-&gt;bw = ep-&gt;hz * ep-&gt;samplesz;	/* bytes/sec */
	if(ep-&gt;mode != OWRITE){
		print("ohci: bug: iso input streams not implemented\n");
		error("ohci iso input streams not implemented");
	}else
		iso-&gt;tok = Tdtokout;

	iso-&gt;left = 0;
	iso-&gt;nerrs = 0;
	iso-&gt;frno = TRUNC(ctlr-&gt;ohci-&gt;fmnumber + 10, Ntdframes);
	iso-&gt;nframes = 1000 / ep-&gt;pollival;
	if(iso-&gt;nframes &lt; 10){
		print("ohci: isoopen: less than 10 frames; using 10.\n");
		iso-&gt;nframes = 10;
	}
	iso-&gt;navail = iso-&gt;nframes;
	iso-&gt;atds = edtds = nil;
	for(i = 0; i &lt; iso-&gt;nframes-1; i++){	/* -1 for dummy */
		td = tdalloc();
		td-&gt;ep = ep;
		td-&gt;io = iso;
		td-&gt;bp = allocb(ep-&gt;maxpkt);
		td-&gt;anext = iso-&gt;atds;		/* link as avail */
		iso-&gt;atds = td;
		td-&gt;next = edtds;
		edtds = td;
	}
	newed(ctlr, ep, iso, "iso");		/* allocates a dummy td */
	iso-&gt;ed-&gt;tds-&gt;bp = allocb(ep-&gt;maxpkt);	/* but not its block */
	iso-&gt;ed-&gt;tds-&gt;next = edtds;
	isodtdinit(ep, iso, iso-&gt;ed-&gt;tds);
}

/*
 * Allocate the endpoint and set it up for I/O
 * in the controller. This must follow what's said
 * in Ep regarding configuration, including perhaps
 * the saved toggles (saved on a previous close of
 * the endpoint data file by epclose).
 */
static void
epopen(Ep *ep)
{
	Ctlr *ctlr;
	Qio *io;
	Ctlio *cio;
	ulong usbid;

	ctlr = ep-&gt;hp-&gt;aux;
	deprint("ohci: epopen ep%d.%d\n", ep-&gt;dev-&gt;nb, ep-&gt;nb);
	if(ep-&gt;aux != nil)
		panic("ohci: epopen called with open ep");
	if(waserror()){
		free(ep-&gt;aux);
		ep-&gt;aux = nil;
		nexterror();
	}
	switch(ep-&gt;ttype){
	case Tnone:
		error("endpoint not configured");
	case Tiso:
		ep-&gt;aux = smalloc(sizeof(Isoio));
		isoopen(ctlr, ep);
		break;
	case Tctl:
		cio = ep-&gt;aux = smalloc(sizeof(Ctlio));
		cio-&gt;debug = ep-&gt;debug;
		cio-&gt;ndata = -1;
		cio-&gt;data = nil;
		cio-&gt;tok = -1;	/* invalid; Tds will say */
		if(ep-&gt;dev-&gt;isroot != 0 &amp;&amp; ep-&gt;nb == 0)	/* root hub */
			break;
		newed(ctlr, ep, cio, "epc");
		break;
	case Tbulk:
		ep-&gt;pollival = 1;	/* assume this; doesn't really matter */
		/* and fall... */
	case Tintr:
		io = ep-&gt;aux = smalloc(sizeof(Qio)*2);
		io[OREAD].debug = io[OWRITE].debug = ep-&gt;debug;
		usbid = (ep-&gt;nb&lt;&lt;7)|(ep-&gt;dev-&gt;nb &amp; Devmax);
		if(ep-&gt;mode != OREAD){
			if(ep-&gt;toggle[OWRITE] != 0)
				io[OWRITE].toggle = Tddata1;
			else
				io[OWRITE].toggle = Tddata0;
			io[OWRITE].tok = Tdtokout;
			io[OWRITE].usbid = usbid;
			io[OWRITE].bw = ep-&gt;maxpkt*1000/ep-&gt;pollival; /* bytes/s */
			newed(ctlr, ep, io+OWRITE, "epw");
		}
		if(ep-&gt;mode != OWRITE){
			if(ep-&gt;toggle[OREAD] != 0)
				io[OREAD].toggle = Tddata1;
			else
				io[OREAD].toggle = Tddata0;
			io[OREAD].tok = Tdtokin;
			io[OREAD].usbid = usbid;
			io[OREAD].bw = ep-&gt;maxpkt*1000/ep-&gt;pollival; /* bytes/s */
			newed(ctlr, ep, io+OREAD, "epr");
		}
		break;
	}
	deprint("ohci: epopen done:\n");
	if(debug || ep-&gt;debug)
		dump(ep-&gt;hp);
	poperror();
}

static void
cancelio(Ep *ep, Qio *io)
{
	Ed *ed;
	Ctlr *ctlr;

	ctlr = ep-&gt;hp-&gt;aux;

	ilock(ctlr);
	if(io == nil || io-&gt;state == Qclose){
		assert(io == nil || io-&gt;ed == nil);
		iunlock(ctlr);
		return;
	}
	ed = io-&gt;ed;
	io-&gt;state = Qclose;
	io-&gt;err = Eio;
	aborttds(io);
	iunlock(ctlr);
	if(!waserror()){
		tsleep(&amp;up-&gt;sleep, return0, 0, Abortdelay);
		poperror();
	}

	wakeup(io);
	qlock(io);
	/* wait for epio if running */
	qunlock(io);

	ilock(ctlr);
	switch(ep-&gt;ttype){
	case Tctl:
		unlinkctl(ctlr, ed);
		break;
	case Tbulk:
		unlinkbulk(ctlr, ed);
		break;
	case Tintr:
	case Tiso:
		unschedq(ctlr, io);
		break;
	default:
		panic("ohci cancelio: bad ttype");
	}
	iunlock(ctlr);
	edfree(io-&gt;ed);
	io-&gt;ed = nil;
}

static void
epclose(Ep *ep)
{
	Ctlio *cio;
	Isoio *iso;
	Qio *io;

	deprint("ohci: epclose ep%d.%d\n", ep-&gt;dev-&gt;nb, ep-&gt;nb);
	if(ep-&gt;aux == nil)
		panic("ohci: epclose called with closed ep");
	switch(ep-&gt;ttype){
	case Tctl:
		cio = ep-&gt;aux;
		cancelio(ep, cio);
		free(cio-&gt;data);
		cio-&gt;data = nil;
		break;
	case Tbulk:
	case Tintr:
		io = ep-&gt;aux;
		if(ep-&gt;mode != OWRITE){
			cancelio(ep, &amp;io[OREAD]);
			if(io[OREAD].toggle == Tddata1)
				ep-&gt;toggle[OREAD] = 1;
		}
		if(ep-&gt;mode != OREAD){
			cancelio(ep, &amp;io[OWRITE]);
			if(io[OWRITE].toggle == Tddata1)
				ep-&gt;toggle[OWRITE] = 1;
		}
		break;
	case Tiso:
		iso = ep-&gt;aux;
		cancelio(ep, iso);
		break;
	default:
		panic("epclose: bad ttype %d", ep-&gt;ttype);
	}

	deprint("ohci: epclose ep%d.%d: done\n", ep-&gt;dev-&gt;nb, ep-&gt;nb);
	free(ep-&gt;aux);
	ep-&gt;aux = nil;
}

static int
portreset(Hci *hp, int port, int on)
{
	Ctlr *ctlr;
	Ohci *ohci;

	if(on == 0)
		return 0;

	ctlr = hp-&gt;aux;
	qlock(&amp;ctlr-&gt;resetl);
	if(waserror()){
		qunlock(&amp;ctlr-&gt;resetl);
		nexterror();
	}
	ilock(ctlr);
	ohci = ctlr-&gt;ohci;
	ohci-&gt;rhportsts[port - 1] = Spp;
	if((ohci-&gt;rhportsts[port - 1] &amp; Ccs) == 0){
		iunlock(ctlr);
		error("port not connected");
	}
	ohci-&gt;rhportsts[port - 1] = Spr;
	while((ohci-&gt;rhportsts[port - 1] &amp; Prsc) == 0){
		iunlock(ctlr);
		dprint("ohci: portreset, wait for reset complete\n");
		ilock(ctlr);
	}
	ohci-&gt;rhportsts[port - 1] = Prsc;
	iunlock(ctlr);
	poperror();
	qunlock(&amp;ctlr-&gt;resetl);
	return 0;
}

static int
portenable(Hci *hp, int port, int on)
{
	Ctlr *ctlr;

	ctlr = hp-&gt;aux;
	dprint("ohci: %#p port %d enable=%d\n", ctlr-&gt;ohci, port, on);
	qlock(&amp;ctlr-&gt;resetl);
	if(waserror()){
		qunlock(&amp;ctlr-&gt;resetl);
		nexterror();
	}
	ilock(ctlr);
	if(on)
		ctlr-&gt;ohci-&gt;rhportsts[port - 1] = Spe | Spp;
	else
		ctlr-&gt;ohci-&gt;rhportsts[port - 1] = Cpe;
	iunlock(ctlr);
	tsleep(&amp;up-&gt;sleep, return0, 0, Enabledelay);
	poperror();
	qunlock(&amp;ctlr-&gt;resetl);
	return 0;
}

static int
portstatus(Hci *hp, int port)
{
	int v;
	Ctlr *ub;
	ulong ohcistatus;

	/*
	 * We must return status bits as a
	 * get port status hub request would do.
	 */
	ub = hp-&gt;aux;
	ohcistatus = ub-&gt;ohci-&gt;rhportsts[port - 1];
	v = 0;
	if(ohcistatus &amp; Ccs)
		v |= HPpresent;
	if(ohcistatus &amp; Pes)
		v |= HPenable;
	if(ohcistatus &amp; Pss)
		v |= HPsuspend;
	if(ohcistatus &amp; Prs)
		v |= HPreset;
	else {
		/* port is not in reset; these potential writes are ok */
		if(ohcistatus &amp; Csc){
			v |= HPstatuschg;
			ub-&gt;ohci-&gt;rhportsts[port - 1] = Csc;
		}
		if(ohcistatus &amp; Pesc){
			v |= HPchange;
			ub-&gt;ohci-&gt;rhportsts[port - 1] = Pesc;
		}
	}
	if(ohcistatus &amp; Lsda)
		v |= HPslow;
	if(v &amp; (HPstatuschg|HPchange))
		ddprint("ohci port %d sts %#ulx hub sts %#x\n", port, ohcistatus, v);
	return v;
}

static void
dumpohci(Ctlr *ctlr)
{
	int i;
	ulong *ohci;

	ohci = &amp;ctlr-&gt;ohci-&gt;revision;
	print("ohci registers: \n");
	for(i = 0; i &lt; sizeof(Ohci)/sizeof(ulong); i++)
		if(i &lt; 3 || ohci[i] != 0)
			print("\t[%#2.2x]\t%#8.8ulx\n", i * 4, ohci[i]);
	print("\n");
}

static void
init(Hci *hp)
{
	Ctlr *ctlr;
	Ohci *ohci;
	int i;
	ulong ival, ctrl, fmi;

	ctlr = hp-&gt;aux;
	dprint("ohci %#p init\n", ctlr-&gt;ohci);
	ohci = ctlr-&gt;ohci;

	fmi =  ctlr-&gt;ohci-&gt;fminterval;
	ctlr-&gt;ohci-&gt;cmdsts = Shcr;         /* reset the block */
	while(ctlr-&gt;ohci-&gt;cmdsts &amp; Shcr)
		delay(1);  /* wait till reset complete, Ohci says 10us max. */
	ctlr-&gt;ohci-&gt;fminterval = fmi;

	/*
	 * now that soft reset is done we are in suspend state.
	 * Setup registers which take in suspend state
	 * (will only be here for 2ms).
	 */

	ctlr-&gt;ohci-&gt;hcca = ptr2pa(ctlr-&gt;hcca);
	setctlhd(ctlr, nil);
	ctlr-&gt;ohci-&gt;ctlcurred = 0;
	setbulkhd(ctlr, nil);
	ctlr-&gt;ohci-&gt;bulkcurred = 0;

	ohci-&gt;intrenable = Mie | Wdh | Ue;
	ohci-&gt;control |= Ccle | Cble | Cple | Cie | Cfsoper;

	/* set frame after operational */
	ohci-&gt;rhdesca = Nps;	/* no power switching */
	if(ohci-&gt;rhdesca &amp; Nps){
		dprint("ohci: ports are not power switched\n");
	}else{
		dprint("ohci: ports are power switched\n");
		ohci-&gt;rhdesca &amp;= ~Psm;
		ohci-&gt;rhsts &amp;= ~Lpsc;
	}
	for(i = 0; i &lt; ctlr-&gt;nports; i++)	/* paranoia */
		ohci-&gt;rhportsts[i] = 0;		/* this has no effect */
	delay(50);

	for(i = 0; i &lt; ctlr-&gt;nports; i++){
		ohci-&gt;rhportsts[i] =  Spp;
		if((ohci-&gt;rhportsts[i] &amp; Ccs) != 0)
			ohci-&gt;rhportsts[i] |= Spr;
	}
	delay(100);

	ctrl = ohci-&gt;control;
	if((ctrl &amp; Cfsmask) != Cfsoper){
		ctrl = (ctrl &amp; ~Cfsmask) | Cfsoper;
		ohci-&gt;control = ctrl;
		ohci-&gt;rhsts = Lpsc;
	}
	ival = ohci-&gt;fminterval &amp; ~(Fmaxpktmask &lt;&lt; Fmaxpktshift);
	ohci-&gt;fminterval = ival | (5120 &lt;&lt; Fmaxpktshift);

	if(debug &gt; 1)
		dumpohci(ctlr);
}

static void
scanpci(void)
{
	ulong mem;
	Ctlr *ctlr;
	Pcidev *p;
	int i;
	static int already = 0;

	if(already)
		return;
	already = 1;
	p = nil;
	while(p = pcimatch(p, 0, 0)) {
		/*
		 * Find Ohci controllers (Programming Interface = 0x10).
		 */
		if(p-&gt;ccrb != Pcibcserial || p-&gt;ccru != Pciscusb ||
		    p-&gt;ccrp != 0x10)
			continue;
		mem = p-&gt;mem[0].bar &amp; ~0x0F;
		dprint("ohci: %x/%x port 0x%lux size 0x%x irq %d\n",
			p-&gt;vid, p-&gt;did, mem, p-&gt;mem[0].size, p-&gt;intl);
		if(mem == 0){
			print("ohci: failed to map registers\n");
			continue;
		}
		if(p-&gt;intl == 0xFF || p-&gt;intl == 0) {
			print("ohci: no irq assigned for port %#lux\n", mem);
			continue;
		}

		ctlr = malloc(sizeof(Ctlr));
		if (ctlr == nil)
			panic("ohci: out of memory");
		ctlr-&gt;pcidev = p;
		ctlr-&gt;ohci = vmap(mem, p-&gt;mem[0].size);
		dprint("scanpci: ctlr %#p, ohci %#p\n", ctlr, ctlr-&gt;ohci);
		pcisetbme(p);
		pcisetpms(p, 0);
		for(i = 0; i &lt; Nhcis; i++)
			if(ctlrs[i] == nil){
				ctlrs[i] = ctlr;
				break;
			}
		if(i == Nhcis)
			print("ohci: bug: no more controllers\n");
	}
}

static void
usbdebug(Hci*, int d)
{
	debug = d;
}

/*
 * build the periodic scheduling tree:
 * framesize must be a multiple of the tree size
 */
static void
mkqhtree(Ctlr *ctlr)
{
	int i, n, d, o, leaf0, depth;
	Ed **tree;
	Qtree *qt;

	depth = flog2(32);
	n = (1 &lt;&lt; (depth+1)) - 1;
	qt = mallocz(sizeof(*qt), 1);
	if(qt == nil)
		panic("usb: can't allocate scheduling tree");
	qt-&gt;nel = n;
	qt-&gt;depth = depth;
	qt-&gt;bw = mallocz(n * sizeof(qt-&gt;bw), 1);
	qt-&gt;root = tree = mallocz(n * sizeof(Ed *), 1);
	if(qt-&gt;bw == nil || qt-&gt;root == nil)
		panic("usb: can't allocate scheduling tree");
	for(i = 0; i &lt; n; i++){
		if((tree[i] = edalloc()) == nil)
			panic("mkqhtree");
		tree[i]-&gt;ctrl = (8 &lt;&lt; Edmpsshift);	/* not needed */
		tree[i]-&gt;ctrl |= Edskip;

		if(i &gt; 0)
			edlinked(tree[i], tree[(i-1)/2]);
		else
			edlinked(tree[i], nil);
	}
	ctlr-&gt;ntree = i;
	dprint("ohci: tree: %d endpoints allocated\n", i);

	/* distribute leaves evenly round the frame list */
	leaf0 = n / 2;
	for(i = 0; i &lt; 32; i++){
		o = 0;
		for(d = 0; d &lt; depth; d++){
			o &lt;&lt;= 1;
			if(i &amp; (1 &lt;&lt; d))
				o |= 1;
		}
		if(leaf0 + o &gt;= n){
			print("leaf0=%d o=%d i=%d n=%d\n", leaf0, o, i, n);
			break;
		}
		ctlr-&gt;hcca-&gt;intrtable[i] = ptr2pa(tree[leaf0 + o]);
	}
	ctlr-&gt;tree = qt;
}

static void
ohcimeminit(Ctlr *ctlr)
{
	Hcca *hcca;

	edfree(edalloc());	/* allocate pools now */
	tdfree(tdalloc());

	hcca = xspanalloc(sizeof(Hcca), 256, 0);
	if(hcca == nil)
		panic("usbhreset: no memory for Hcca");
	memset(hcca, 0, sizeof(*hcca));
	ctlr-&gt;hcca = hcca;

	mkqhtree(ctlr);
}

static void
ohcireset(Ctlr *ctlr)
{
	ilock(ctlr);
	dprint("ohci %#p reset\n", ctlr-&gt;ohci);

	/*
	 * usually enter here in reset, wait till its through,
	 * then do our own so we are on known timing conditions.
	 * Is this needed?
	 */
	delay(100);
	ctlr-&gt;ohci-&gt;control = 0;
	delay(100);

	/* legacy support register: turn off lunacy mode */
	pcicfgw16(ctlr-&gt;pcidev, 0xc0, 0x2000);

	iunlock(ctlr);
}

static void
shutdown(Hci *hp)
{
	Ctlr *ctlr;

	ctlr = hp-&gt;aux;

	ilock(ctlr);
	ctlr-&gt;ohci-&gt;intrdisable = Mie;
	ctlr-&gt;ohci-&gt;control = 0;
	coherence();
	delay(100);
	iunlock(ctlr);
}

static int
reset(Hci *hp)
{
	int i;
	Ctlr *ctlr;
	Pcidev *p;
	static Lock resetlck;

	if(getconf("*nousbohci"))
		return -1;
	ilock(&amp;resetlck);
	scanpci();

	/*
	 * Any adapter matches if no hp-&gt;port is supplied,
	 * otherwise the ports must match.
	 */
	ctlr = nil;
	for(i = 0; i &lt; Nhcis &amp;&amp; ctlrs[i] != nil; i++){
		ctlr = ctlrs[i];
		if(ctlr-&gt;active == 0)
		if(hp-&gt;port == 0 || hp-&gt;port == (uintptr)ctlr-&gt;ohci){
			ctlr-&gt;active = 1;
			break;
		}
	}
	iunlock(&amp;resetlck);
	if(ctlrs[i] == nil || i == Nhcis)
		return -1;
	if(ctlr-&gt;ohci-&gt;control == ~0)
		return -1;


	p = ctlr-&gt;pcidev;
	hp-&gt;aux = ctlr;
	hp-&gt;port = (uintptr)ctlr-&gt;ohci;
	hp-&gt;irq = p-&gt;intl;
	hp-&gt;tbdf = p-&gt;tbdf;
	ctlr-&gt;nports = hp-&gt;nports = ctlr-&gt;ohci-&gt;rhdesca &amp; 0xff;

	ohcireset(ctlr);
	ohcimeminit(ctlr);

	/*
	 * Linkage to the generic HCI driver.
	 */
	hp-&gt;init = init;
	hp-&gt;dump = dump;
	hp-&gt;interrupt = interrupt;
	hp-&gt;epopen = epopen;
	hp-&gt;epclose = epclose;
	hp-&gt;epread = epread;
	hp-&gt;epwrite = epwrite;
	hp-&gt;seprintep = seprintep;
	hp-&gt;portenable = portenable;
	hp-&gt;portreset = portreset;
	hp-&gt;portstatus = portstatus;
	hp-&gt;shutdown = shutdown;
	hp-&gt;debug = usbdebug;
	hp-&gt;type = "ohci";
	return 0;
}

void
usbohcilink(void)
{
	addhcitype("ohci", reset);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

