<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/acme/elog.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/acme/elog.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;draw.h&gt;
#include &lt;thread.h&gt;
#include &lt;cursor.h&gt;
#include &lt;mouse.h&gt;
#include &lt;keyboard.h&gt;
#include &lt;frame.h&gt;
#include &lt;fcall.h&gt;
#include &lt;plumb.h&gt;
#include "dat.h"
#include "fns.h"
#include "edit.h"

static char Wsequence[] = "warning: changes out of sequence\n";
static int	warned = FALSE;

/*
 * Log of changes made by editing commands.  Three reasons for this:
 * 1) We want addresses in commands to apply to old file, not file-in-change.
 * 2) It's difficult to track changes correctly as things move, e.g. ,x m$
 * 3) This gives an opportunity to optimize by merging adjacent changes.
 * It's a little bit like the Undo/Redo log in Files, but Point 3) argues for a
 * separate implementation.  To do this well, we use Replace as well as
 * Insert and Delete
 */

typedef struct Buflog Buflog;
struct Buflog
{
	short	type;		/* Replace, Filename */
	uint		q0;		/* location of change (unused in f) */
	uint		nd;		/* # runes to delete */
	uint		nr;		/* # runes in string or file name */
};

enum
{
	Buflogsize = sizeof(Buflog)/sizeof(Rune),
};

/*
 * Minstring shouldn't be very big or we will do lots of I/O for small changes.
 * Maxstring is RBUFSIZE so we can fbufalloc() once and not realloc elog.r.
 */
enum
{
	Minstring = 16,		/* distance beneath which we merge changes */
	Maxstring = RBUFSIZE,	/* maximum length of change we will merge into one */
};

void
eloginit(File *f)
{
	if(f-&gt;elog.type != Empty)
		return;
	f-&gt;elog.type = Null;
	if(f-&gt;elogbuf == nil)
		f-&gt;elogbuf = emalloc(sizeof(Buffer));
	if(f-&gt;elog.r == nil)
		f-&gt;elog.r = fbufalloc();
	bufreset(f-&gt;elogbuf);
}

void
elogclose(File *f)
{
	if(f-&gt;elogbuf){
		bufclose(f-&gt;elogbuf);
		free(f-&gt;elogbuf);
		f-&gt;elogbuf = nil;
	}
}

void
elogreset(File *f)
{
	f-&gt;elog.type = Null;
	f-&gt;elog.nd = 0;
	f-&gt;elog.nr = 0;
}

void
elogterm(File *f)
{
	elogreset(f);
	if(f-&gt;elogbuf)
		bufreset(f-&gt;elogbuf);
	f-&gt;elog.type = Empty;
	fbuffree(f-&gt;elog.r);
	f-&gt;elog.r = nil;
	warned = FALSE;
}

void
elogflush(File *f)
{
	Buflog b;

	b.type = f-&gt;elog.type;
	b.q0 = f-&gt;elog.q0;
	b.nd = f-&gt;elog.nd;
	b.nr = f-&gt;elog.nr;
	switch(f-&gt;elog.type){
	default:
		warning(nil, "unknown elog type 0x%ux\n", f-&gt;elog.type);
		break;
	case Null:
		break;
	case Insert:
	case Replace:
		if(f-&gt;elog.nr &gt; 0)
			bufinsert(f-&gt;elogbuf, f-&gt;elogbuf-&gt;nc, f-&gt;elog.r, f-&gt;elog.nr);
		/* fall through */
	case Delete:
		bufinsert(f-&gt;elogbuf, f-&gt;elogbuf-&gt;nc, (Rune*)&amp;b, Buflogsize);
		break;
	}
	elogreset(f);
}

void
elogreplace(File *f, int q0, int q1, Rune *r, int nr)
{
	uint gap;

	if(q0==q1 &amp;&amp; nr==0)
		return;
	eloginit(f);
	if(f-&gt;elog.type!=Null &amp;&amp; q0&lt;f-&gt;elog.q0){
		if(warned++ == 0)
			warning(nil, Wsequence);
		elogflush(f);
	}
	/* try to merge with previous */
	gap = q0 - (f-&gt;elog.q0+f-&gt;elog.nd);	/* gap between previous and this */
	if(f-&gt;elog.type==Replace &amp;&amp; f-&gt;elog.nr+gap+nr&lt;Maxstring){
		if(gap &lt; Minstring){
			if(gap &gt; 0){
				bufread(f, f-&gt;elog.q0+f-&gt;elog.nd, f-&gt;elog.r+f-&gt;elog.nr, gap);
				f-&gt;elog.nr += gap;
			}
			f-&gt;elog.nd += gap + q1-q0;
			runemove(f-&gt;elog.r+f-&gt;elog.nr, r, nr);
			f-&gt;elog.nr += nr;
			return;
		}
	}
	elogflush(f);
	f-&gt;elog.type = Replace;
	f-&gt;elog.q0 = q0;
	f-&gt;elog.nd = q1-q0;
	f-&gt;elog.nr = nr;
	if(nr &gt; RBUFSIZE)
		editerror("internal error: replacement string too large(%d)", nr);
	runemove(f-&gt;elog.r, r, nr);
}

void
eloginsert(File *f, int q0, Rune *r, int nr)
{
	int n;

	if(nr == 0)
		return;
	eloginit(f);
	if(f-&gt;elog.type!=Null &amp;&amp; q0&lt;f-&gt;elog.q0){
		if(warned++ == 0)
			warning(nil, Wsequence);
		elogflush(f);
	}
	/* try to merge with previous */
	if(f-&gt;elog.type==Insert &amp;&amp; q0==f-&gt;elog.q0 &amp;&amp; f-&gt;elog.nr+nr&lt;Maxstring){
		runemove(f-&gt;elog.r+f-&gt;elog.nr, r, nr);
		f-&gt;elog.nr += nr;
		return;
	}
	while(nr &gt; 0){
		elogflush(f);
		f-&gt;elog.type = Insert;
		f-&gt;elog.q0 = q0;
		n = nr;
		if(n &gt; RBUFSIZE)
			n = RBUFSIZE;
		f-&gt;elog.nr = n;
		runemove(f-&gt;elog.r, r, n);
		r += n;
		nr -= n;
	}
}

void
elogdelete(File *f, int q0, int q1)
{
	if(q0 == q1)
		return;
	eloginit(f);
	if(f-&gt;elog.type!=Null &amp;&amp; q0&lt;f-&gt;elog.q0+f-&gt;elog.nd){
		if(warned++ == 0)
			warning(nil, Wsequence);
		elogflush(f);
	}
	/* try to merge with previous */
	if(f-&gt;elog.type==Delete &amp;&amp; f-&gt;elog.q0+f-&gt;elog.nd==q0){
		f-&gt;elog.nd += q1-q0;
		return;
	}
	elogflush(f);
	f-&gt;elog.type = Delete;
	f-&gt;elog.q0 = q0;
	f-&gt;elog.nd = q1-q0;
}

#define tracelog 0
void
elogapply(File *f)
{
	Buflog b;
	Rune *buf;
	uint i, n, up, mod;
	uint tq0, tq1;
	Buffer *log;
	Text *t;
	int owner;

	elogflush(f);
	log = f-&gt;elogbuf;
	t = f-&gt;curtext;

	buf = fbufalloc();
	mod = FALSE;

	owner = 0;
	if(t-&gt;w){
		owner = t-&gt;w-&gt;owner;
		if(owner == 0)
			t-&gt;w-&gt;owner = 'E';
	}

	/*
	 * The edit commands have already updated the selection in t-&gt;q0, t-&gt;q1,
	 * but using coordinates relative to the unmodified buffer.  As we apply the log,
	 * we have to update the coordinates to be relative to the modified buffer.
	 * Textinsert and textdelete will do this for us; our only work is to apply the
	 * convention that an insertion at t-&gt;q0==t-&gt;q1 is intended to select the 
	 * inserted text.
	 */

	/*
	 * We constrain the addresses in here (with textconstrain()) because
	 * overlapping changes will generate bogus addresses.   We will warn
	 * about changes out of sequence but proceed anyway; here we must
	 * keep things in range.
	 */

	while(log-&gt;nc &gt; 0){
		up = log-&gt;nc-Buflogsize;
		bufread(log, up, (Rune*)&amp;b, Buflogsize);
		switch(b.type){
		default:
			fprint(2, "elogapply: 0x%ux\n", b.type);
			abort();
			break;

		case Replace:
			if(tracelog)
				warning(nil, "elog replace %d %d (%d %d)\n",
					b.q0, b.q0+b.nd, t-&gt;q0, t-&gt;q1);
			if(!mod){
				mod = TRUE;
				filemark(f);
			}
			textconstrain(t, b.q0, b.q0+b.nd, &amp;tq0, &amp;tq1);
			textdelete(t, tq0, tq1, TRUE);
			up -= b.nr;
			for(i=0; i&lt;b.nr; i+=n){
				n = b.nr - i;
				if(n &gt; RBUFSIZE)
					n = RBUFSIZE;
				bufread(log, up+i, buf, n);
				textinsert(t, tq0+i, buf, n, TRUE);
			}
			if(t-&gt;q0 == b.q0 &amp;&amp; t-&gt;q1 == b.q0)
				t-&gt;q1 += b.nr;
			break;

		case Delete:
			if(tracelog)
				warning(nil, "elog delete %d %d (%d %d)\n",
					b.q0, b.q0+b.nd, t-&gt;q0, t-&gt;q1);
			if(!mod){
				mod = TRUE;
				filemark(f);
			}
			textconstrain(t, b.q0, b.q0+b.nd, &amp;tq0, &amp;tq1);
			textdelete(t, tq0, tq1, TRUE);
			break;

		case Insert:
			if(tracelog)
				warning(nil, "elog insert %d %d (%d %d)\n",
					b.q0, b.q0+b.nr, t-&gt;q0, t-&gt;q1);
			if(!mod){
				mod = TRUE;
				filemark(f);
			}
			textconstrain(t, b.q0, b.q0, &amp;tq0, &amp;tq1);
			up -= b.nr;
			for(i=0; i&lt;b.nr; i+=n){
				n = b.nr - i;
				if(n &gt; RBUFSIZE)
					n = RBUFSIZE;
				bufread(log, up+i, buf, n);
				textinsert(t, tq0+i, buf, n, TRUE);
			}
			if(t-&gt;q0 == b.q0 &amp;&amp; t-&gt;q1 == b.q0)
				t-&gt;q1 += b.nr;
			break;

/*		case Filename:
			f-&gt;seq = u.seq;
			fileunsetname(f, epsilon);
			f-&gt;mod = u.mod;
			up -= u.n;
			free(f-&gt;name);
			if(u.n == 0)
				f-&gt;name = nil;
			else
				f-&gt;name = runemalloc(u.n);
			bufread(delta, up, f-&gt;name, u.n);
			f-&gt;nname = u.n;
			break;
*/
		}
		bufdelete(log, up, log-&gt;nc);
	}
	fbuffree(buf);
	elogterm(f);

	/*
	 * Bad addresses will cause bufload to crash, so double check.
	 * If changes were out of order, we expect problems so don't complain further.
	 */
	if(t-&gt;q0 &gt; f-&gt;nc || t-&gt;q1 &gt; f-&gt;nc || t-&gt;q0 &gt; t-&gt;q1){
		if(!warned)
			warning(nil, "elogapply: can't happen %d %d %d\n", t-&gt;q0, t-&gt;q1, f-&gt;nc);
		t-&gt;q1 = min(t-&gt;q1, f-&gt;nc);
		t-&gt;q0 = min(t-&gt;q0, t-&gt;q1);
	}

	if(t-&gt;w)
		t-&gt;w-&gt;owner = owner;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

