<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/archfs.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/archfs.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * archfs - mount mkfs style archives
 */

#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;bio.h&gt;
#include &lt;auth.h&gt;
#include &lt;fcall.h&gt;
#include &lt;thread.h&gt;
#include &lt;9p.h&gt;

Tree *archtree;
Biobuf *b;
int verbose;

typedef struct Ahdr Ahdr;
struct Ahdr {
	char *name;
	Dir;
};

typedef struct Arch Arch;
struct Arch {
	vlong off;
	vlong length;
};

static void*
emalloc(long sz)
{
	void *v;

	v = malloc(sz);
	if(v == nil)
		sysfatal("malloc %lud fails", sz);
	memset(v, 0, sz);
	return v;
}

static char*
estrdup(char *s)
{
	s = strdup(s);
	if(s == nil)
		sysfatal("strdup (%.10s) fails", s);
	return s;
}

static char*
Bgetline(Biobuf *b)
{
	char *p;

	if(p = Brdline(b, '\n'))
		p[Blinelen(b)-1] = '\0';
	return p;
}

Ahdr*
gethdr(Biobuf *b)
{
	Ahdr *a;
	char *p, *f[10];

	if((p = Bgetline(b)) == nil)
		return nil;

	if(strcmp(p, "end of archive") == 0) {
		werrstr("");
		return nil;
	}

	if(tokenize(p, f, nelem(f)) != 6) {
		werrstr("bad format");
		return nil;
	}

	a = emalloc(sizeof(*a));
	a-&gt;name = estrdup(f[0]);
	a-&gt;mode = strtoul(f[1], 0, 8);
	a-&gt;uid = estrdup(f[2]);
	a-&gt;gid = estrdup(f[3]);
	a-&gt;mtime = strtoll(f[4], 0, 10);
	a-&gt;length = strtoll(f[5], 0, 10);
	return a;
}

static Arch*
newarch(vlong off, vlong length)
{
	static Arch *abuf;
	static int nabuf;

	if(nabuf == 0) {
		nabuf = 256;
		abuf = emalloc(sizeof(Arch)*nabuf);
	}

	nabuf--;
	abuf-&gt;off = off;
	abuf-&gt;length = length;
	return abuf++;
}

static File*
createpath(File *f, char *name, char *u, ulong m)
{
	char *p;
	File *nf;

	if(verbose)
		fprint(2, "createpath %s\n", name);
	incref(f);
	while(f &amp;&amp; (p = strchr(name, '/'))) {
		*p = '\0';
		if(strcmp(name, "") != 0 &amp;&amp; strcmp(name, ".") != 0){
			/* this would be a race if we were multithreaded */
			incref(f);	/* so walk doesn't kill it immediately on failure */
			if((nf = walkfile(f, name)) == nil)
				nf = createfile(f, name, u, DMDIR|0777, nil);
			decref(f);
			f = nf;
		}
		*p = '/';
		name = p+1;
	}
	if(f == nil)
		return nil;

	incref(f);
	if((nf = walkfile(f, name)) == nil)
		nf = createfile(f, name, u, m, nil);
	decref(f);
	return nf;
}

static void
archcreatefile(char *name, Arch *arch, Dir *d)
{
	File *f;
	f = createpath(archtree-&gt;root, name, d-&gt;uid, d-&gt;mode);
	if(f == nil)
		sysfatal("creating %s: %r", name);
	free(f-&gt;gid);
	f-&gt;gid = estrdup9p(d-&gt;gid);
	f-&gt;aux = arch;
	f-&gt;mtime = d-&gt;mtime;
	f-&gt;length = d-&gt;length;
	decref(f);
}

static void
fsread(Req *r)
{
	Arch *a;
	char err[ERRMAX];
	int n;

	a = r-&gt;fid-&gt;file-&gt;aux;
	if(a-&gt;length &lt;= r-&gt;ifcall.offset) 
		r-&gt;ifcall.count = 0;
	else if(a-&gt;length &lt;= r-&gt;ifcall.offset+r-&gt;ifcall.count)
		r-&gt;ifcall.count = a-&gt;length - r-&gt;ifcall.offset;

	werrstr("unknown error");
	if(Bseek(b, a-&gt;off+r-&gt;ifcall.offset, 0) &lt; 0 
	|| (n = Bread(b, r-&gt;ofcall.data, r-&gt;ifcall.count)) &lt; 0) {
		err[0] = '\0';
		errstr(err, sizeof err);
		respond(r, err);
	} else {
		r-&gt;ofcall.count = n;
		respond(r, nil);	
	}
}

Srv fs = {
	.read=	fsread,
};

static void
usage(void)
{
	fprint(2, "usage: archfs [-abcC] [-m mtpt] archfile\n");
	exits("usage");
}

void
main(int argc, char **argv)
{
	Ahdr *a;
	ulong flag;
	char *mtpt;
	char err[ERRMAX];

	flag = 0;
	mtpt = "/mnt/arch";
	ARGBEGIN{
	case 'D':
		chatty9p++;
		break;
	case 'a':
		flag |= MAFTER;
		break;
	case 'b':
		flag |= MBEFORE;
		break;
	case 'c':
		flag |= MCREATE;
		break;
	case 'C':
		flag |= MCACHE;
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	default:
		usage();
		break;
	}ARGEND;

	if(argc != 1)
		usage();

	if((b = Bopen(argv[0], OREAD)) == nil)
		sysfatal("open '%s': %r", argv[0]);

	archtree = fs.tree = alloctree("sys", "sys", DMDIR|0775, nil);
	while(a = gethdr(b)) {
		archcreatefile(a-&gt;name, newarch(Boffset(b), a-&gt;length), a);
		Bseek(b, a-&gt;length, 1);
	}

	err[0] = '\0';
	errstr(err, sizeof err);
	if(err[0])
		sysfatal("reading archive: %s", err);

	postmountsrv(&amp;fs, nil, mtpt, flag);
	exits(0);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

