<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/auth/factotum/rpc.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/auth/factotum/rpc.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * Factotum RPC
 *
 * Must be paired write/read cycles on /mnt/factotum/rpc.
 * The format of a request is verb, single space, data.
 * Data format is verb-dependent; in particular, it can be binary.
 * The format of a response is the same.  The write only sets up
 * the RPC.  The read tries to execute it.  If the /mnt/factotum/key
 * file is open, we ask for new keys using that instead of returning
 * an error in the RPC.  This means the read blocks.
 * Textual arguments are parsed with tokenize, so rc-style quoting
 * rules apply.
 *
 * Only authentication protocol messages go here.  Configuration
 * is still via ctl (below).
 *
 * Return values are:
 *	error message - an error happened.
 *	ok [data] - success, possible data is request dependent.
 *	needkey proto domain user - request aborted, get me this key and try again
 *	badkey proto domain user - request aborted, this key might be bad
 *	done [haveai] - authentication is done [haveai: you can get an ai with authinfo]
 * Request RPCs are:
 *	start attrs - initializes protocol for authentication, can fail.
 *		returns "ok read" or "ok write" on success.
 *	read - execute protocol read
 *	write - execute protocol write
 *	authinfo - if the protocol is finished, return the AI if any
 *	attr - return protocol information
 */

#include "dat.h"

Req *rpcwait;

typedef struct Verb Verb;
struct Verb {
	char *verb;
	int iverb;
};

enum {
	Vunknown = -1,
	Vauthinfo = 0,
	Vread,
	Vstart,
	Vwrite,
	Vattr,
};

Verb rpctab[] = {
	"authinfo",	Vauthinfo,
	"read",		Vread,
	"start",		Vstart,
	"write",		Vwrite,
	"attr",		Vattr,
};

static int
classify(char *s, Verb *verbtab, int nverbtab)
{
	int i;

	for(i=0; i&lt;nverbtab; i++)
		if(strcmp(s, verbtab[i].verb) == 0)
			return verbtab[i].iverb;
	return Vunknown;
}

void
rpcwrite(Req *r)
{
	Fsstate *fss;

	if(r-&gt;ifcall.count &gt;= Maxrpc){
		respond(r, Etoolarge);
		return;
	}
	fss = r-&gt;fid-&gt;aux;
	if(fss-&gt;pending){
		respond(r, "rpc already pending; read to clear");
		return;
	}
	memmove(fss-&gt;rpc.buf, r-&gt;ifcall.data, r-&gt;ifcall.count);
	fss-&gt;rpc.buf[r-&gt;ifcall.count] = '\0';
	fss-&gt;rpc.verb = fss-&gt;rpc.buf;
	if(fss-&gt;rpc.arg = strchr(fss-&gt;rpc.buf, ' ')){
		*fss-&gt;rpc.arg++ = '\0';
		fss-&gt;rpc.narg = r-&gt;ifcall.count - (fss-&gt;rpc.arg - fss-&gt;rpc.buf);
	}else{
		fss-&gt;rpc.arg = "";
		fss-&gt;rpc.narg = 0;
	}
	fss-&gt;rpc.iverb = classify(fss-&gt;rpc.verb, rpctab, nelem(rpctab));
	r-&gt;ofcall.count = r-&gt;ifcall.count;
	fss-&gt;pending = 1;
	respond(r, nil);
}

static void
retstring(Req *r, Fsstate *fss, char *s)
{
	int n;

	n = strlen(s);
	if(n &gt; r-&gt;ifcall.count)
		n = r-&gt;ifcall.count;
	memmove(r-&gt;ofcall.data, s, n);
	r-&gt;ofcall.count = n;
	fss-&gt;pending = 0;
	respond(r, nil);
	return;
}

static void
retrpc(Req *r, int ret, Fsstate *fss)
{
	switch(ret){
	default:
		snprint(fss-&gt;rpc.buf, Maxrpc, "internal error %d", ret);
		retstring(r, fss, fss-&gt;rpc.buf);
		return;
	case RpcErrstr:
		snprint(fss-&gt;rpc.buf, Maxrpc, "error %r");
		retstring(r, fss, fss-&gt;rpc.buf);
		return;
	case RpcFailure:
		snprint(fss-&gt;rpc.buf, Maxrpc, "error %s", fss-&gt;err);
		retstring(r, fss, fss-&gt;rpc.buf);
		return;
	case RpcNeedkey:
		if(needkeyqueue(r, fss) &lt; 0){
			snprint(fss-&gt;rpc.buf, Maxrpc, "needkey %s", fss-&gt;keyinfo);
			retstring(r, fss, fss-&gt;rpc.buf);
		}
		return;
	case RpcOk:
		retstring(r, fss, "ok");
		return;
	case RpcToosmall:
		snprint(fss-&gt;rpc.buf, Maxrpc, "toosmall %d", fss-&gt;rpc.nwant);
		retstring(r, fss, fss-&gt;rpc.buf);
		return;
	case RpcPhase:
		snprint(fss-&gt;rpc.buf, Maxrpc, "phase %r");
		retstring(r, fss, fss-&gt;rpc.buf);
		return;
	case RpcConfirm:
		confirmqueue(r, fss);
		return;
	}
}

int
rdwrcheck(Req *r, Fsstate *fss)
{
	if(fss-&gt;ps == nil){
		retstring(r, fss, "error no current protocol");
		return -1;
	}
	if(fss-&gt;phase == Notstarted){
		retstring(r, fss, "protocol not started");
		return -1;
	}
	if(fss-&gt;phase == Broken){
		snprint(fss-&gt;rpc.buf, Maxrpc, "error %s", fss-&gt;err);
		retstring(r, fss, fss-&gt;rpc.buf);
		return -1;
	}
	if(fss-&gt;phase == Established){
		if(fss-&gt;haveai)
			retstring(r, fss, "done haveai");
		else
			retstring(r, fss, "done");
		return -1;
	}
	return 0;
}

static void
logret(char *pre, Fsstate *fss, int ret)
{
	switch(ret){
	default:
		flog("%s: code %d", pre, ret);
		break;
	case RpcErrstr:
		flog("%s: error %r", pre);
		break;
	case RpcFailure:
		flog("%s: failure %s", pre, fss-&gt;err);
		break;
	case RpcNeedkey:
		flog("%s: needkey %s", pre, fss-&gt;keyinfo);
		break;
	case RpcOk:
		flog("%s: ok", pre);
		break;
	case RpcToosmall:
		flog("%s: toosmall %d", pre, fss-&gt;rpc.nwant);
		break;
	case RpcPhase:
		flog("%s: phase: %r", pre);
		break;
	case RpcConfirm:
		flog("%s: waiting for confirm", pre);
		break;
	}
}

void
rpcrdwrlog(Fsstate *fss, char *rdwr, uint n, int ophase, int ret)
{
	char buf0[40], buf1[40], pre[300];

	if(!debug)
		return;
	snprint(pre, sizeof pre, "%d: %s %ud in phase %s yields phase %s",
		fss-&gt;seqnum, rdwr, n, phasename(fss, ophase, buf0), phasename(fss, fss-&gt;phase, buf1));
	logret(pre, fss, ret);
}

void
rpcstartlog(Attr *attr, Fsstate *fss, int ret)
{
	char pre[300], tmp[40];

	if(!debug)
		return;
	snprint(pre, sizeof pre, "%d: start %A yields phase %s", fss-&gt;seqnum,
		attr, phasename(fss, fss-&gt;phase, tmp));
	logret(pre, fss, ret);		
}

int seqnum;

void
rpcread(Req *r)
{
	Attr *attr;
	char *p;
	int ophase, ret;
	uchar *e;
	uint count;
	Fsstate *fss;
	Proto *proto;

	if(r-&gt;ifcall.count &lt; 64){
		respond(r, "rpc read too small");
		return;
	}
	fss = r-&gt;fid-&gt;aux;
	if(!fss-&gt;pending){
		respond(r, "no rpc pending");
		return;
	}
	switch(fss-&gt;rpc.iverb){
	default:
	case Vunknown:
		retstring(r, fss, "error unknown verb");
		break;

	case Vstart:
		if(fss-&gt;phase != Notstarted){
			flog("%d: implicit close due to second start; old attr '%A'", fss-&gt;seqnum, fss-&gt;attr);
			if(fss-&gt;proto &amp;&amp; fss-&gt;ps)
				(*fss-&gt;proto-&gt;close)(fss);
			fss-&gt;ps = nil;
			fss-&gt;proto = nil;
			_freeattr(fss-&gt;attr);
			fss-&gt;attr = nil;
			fss-&gt;phase = Notstarted;
		}	
		attr = _parseattr(fss-&gt;rpc.arg);
		if((p = _strfindattr(attr, "proto")) == nil){
			retstring(r, fss, "error did not specify proto");
			_freeattr(attr);
			break;
		}
		if((proto = findproto(p)) == nil){
			snprint(fss-&gt;rpc.buf, Maxrpc, "error unknown protocol %q", p);
			retstring(r, fss, fss-&gt;rpc.buf);
			_freeattr(attr);
			break;
		}
		fss-&gt;attr = attr;
		fss-&gt;proto = proto;
		fss-&gt;seqnum = ++seqnum;
		ret = (*proto-&gt;init)(proto, fss);
		rpcstartlog(attr, fss, ret);
		if(ret != RpcOk){
			_freeattr(fss-&gt;attr);
			fss-&gt;attr = nil;
			fss-&gt;phase = Notstarted;
		}
		retrpc(r, ret, fss);
		break;

	case Vread:
		if(fss-&gt;rpc.arg &amp;&amp; fss-&gt;rpc.arg[0]){
			retstring(r, fss, "error read needs no parameters");
			break;
		}
		if(rdwrcheck(r, fss) &lt; 0)
			break;
		count = r-&gt;ifcall.count - 3;
		ophase = fss-&gt;phase;
		ret = fss-&gt;proto-&gt;read(fss, (uchar*)r-&gt;ofcall.data+3, &amp;count);
		rpcrdwrlog(fss, "read", count, ophase, ret);
		if(ret == RpcOk){
			memmove(r-&gt;ofcall.data, "ok ", 3);
			if(count == 0)
				r-&gt;ofcall.count = 2;
			else
				r-&gt;ofcall.count = 3+count;
			fss-&gt;pending = 0;
			respond(r, nil);
		}else
			retrpc(r, ret, fss);
		break;

	case Vwrite:
		if(rdwrcheck(r, fss) &lt; 0)
			break;
		ophase = fss-&gt;phase;
		ret = fss-&gt;proto-&gt;write(fss, fss-&gt;rpc.arg, fss-&gt;rpc.narg);
		rpcrdwrlog(fss, "write", fss-&gt;rpc.narg, ophase, ret);
		retrpc(r, ret, fss);
		break;

	case Vauthinfo:
		if(fss-&gt;phase != Established){
			retstring(r, fss, "error authentication unfinished");
			break;
		}
		if(!fss-&gt;haveai){
			retstring(r, fss, "error no authinfo available");
			break;
		}
		memmove(r-&gt;ofcall.data, "ok ", 3);
		fss-&gt;ai.cap = mkcap(r-&gt;fid-&gt;uid, fss-&gt;ai.suid);
		e = convAI2M(&amp;fss-&gt;ai, (uchar*)r-&gt;ofcall.data+3, r-&gt;ifcall.count-3);
		free(fss-&gt;ai.cap);
		fss-&gt;ai.cap = nil;
		if(e == nil){
			retstring(r, fss, "error read too small");
			break;
		}
		r-&gt;ofcall.count = e - (uchar*)r-&gt;ofcall.data;
		fss-&gt;pending = 0;
		respond(r, nil);
		break;

	case Vattr:
		snprint(fss-&gt;rpc.buf, Maxrpc, "ok %A", fss-&gt;attr);
		retstring(r, fss, fss-&gt;rpc.buf);
		break;
	}
}

enum {
	Vdelkey,
	Vaddkey,
	Vdebug,
};

Verb ctltab[] = {
	"delkey",		Vdelkey,
	"key",		Vaddkey,
	"debug",	Vdebug,
};

/*
 *	key attr=val... - add a key
 *		the attr=val pairs are protocol-specific.
 *		for example, both of these are valid:
 *			key p9sk1 gre cs.bell-labs.com mysecret
 *			key p9sk1 gre cs.bell-labs.com 11223344556677 fmt=des7hex
 *	delkey ... - delete a key
 *		if given, the attr=val pairs are used to narrow the search
 *		[maybe should require a password?]
 */

int
ctlwrite(char *a, int atzero)
{
	char *p;
	int i, nmatch, ret;
	Attr *attr, **l, **lpriv, **lprotos, *pa, *priv, *protos;
	Key *k;
	Proto *proto;

	if(a[0] == '#' || a[0] == '\0')
		return 0;

	/*
	 * it would be nice to emit a warning of some sort here.
	 * we ignore all but the first line of the write.  this helps
	 * both with things like "echo delkey &gt;/mnt/factotum/ctl"
	 * and writes that (incorrectly) contain multiple key lines.
	 */
	if(p = strchr(a, '\n')){
		if(p[1] != '\0'){
			werrstr("multiline write not allowed");
			return -1;
		}
		*p = '\0';
	}

	if((p = strchr(a, ' ')) == nil)
		p = "";
	else
		*p++ = '\0';
	switch(classify(a, ctltab, nelem(ctltab))){
	default:
	case Vunknown:
		werrstr("unknown verb");
		return -1;
	case Vdebug:
		debug ^= 1;
		return 0;
	case Vdelkey:
		nmatch = 0;
		attr = _parseattr(p);
		for(pa=attr; pa; pa=pa-&gt;next){
			if(pa-&gt;type != AttrQuery &amp;&amp; pa-&gt;name[0]=='!'){
				werrstr("only !private? patterns are allowed for private fields");
				_freeattr(attr);
				return -1;
			}
		}
		for(i=0; i&lt;ring-&gt;nkey; ){
			if(matchattr(attr, ring-&gt;key[i]-&gt;attr, ring-&gt;key[i]-&gt;privattr)){
				nmatch++;
				closekey(ring-&gt;key[i]);
				ring-&gt;nkey--;
				memmove(&amp;ring-&gt;key[i], &amp;ring-&gt;key[i+1], (ring-&gt;nkey-i)*sizeof(ring-&gt;key[0]));
			}else
				i++;
		}
		_freeattr(attr);
		if(nmatch == 0){
			werrstr("found no keys to delete");
			return -1;
		}
		return 0;
	case Vaddkey:
		attr = _parseattr(p);
		/* separate out proto= attributes */
		lprotos = &amp;protos;
		for(l=&amp;attr; (*l); ){
			if(strcmp((*l)-&gt;name, "proto") == 0){
				*lprotos = *l;
				lprotos = &amp;(*l)-&gt;next;
				*l = (*l)-&gt;next;
			}else
				l = &amp;(*l)-&gt;next;
		}
		*lprotos = nil;
		if(protos == nil){
			werrstr("key without protos");
			_freeattr(attr);
			return -1;
		}

		/* separate out private attributes */
		lpriv = &amp;priv;
		for(l=&amp;attr; (*l); ){
			if((*l)-&gt;name[0] == '!'){
				*lpriv = *l;
				lpriv = &amp;(*l)-&gt;next;
				*l = (*l)-&gt;next;
			}else
				l = &amp;(*l)-&gt;next;
		}
		*lpriv = nil;

		/* add keys */
		ret = 0;
		for(pa=protos; pa; pa=pa-&gt;next){
			if((proto = findproto(pa-&gt;val)) == nil){
				werrstr("unknown proto %s", pa-&gt;val);
				ret = -1;
				continue;
			}
			if(proto-&gt;addkey == nil){
				werrstr("proto %s doesn't take keys", proto-&gt;name);
				ret = -1;
				continue;
			}
			k = emalloc(sizeof(Key));
			k-&gt;attr = _mkattr(AttrNameval, "proto", proto-&gt;name, _copyattr(attr));
			k-&gt;privattr = _copyattr(priv);
			k-&gt;ref = 1;
			k-&gt;proto = proto;
			if(proto-&gt;addkey(k, atzero) &lt; 0){
				ret = -1;
				closekey(k);
				continue;
			}
			closekey(k);
		}
		_freeattr(attr);
		_freeattr(priv);
		_freeattr(protos);
		return ret;
	}
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

