<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/cfs/bcache.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/cfs/bcache.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include "cformat.h"
#include "lru.h"
#include "bcache.h"

int
bcinit(Bcache *bc, int f, int bsize)
{
	Bbuf *b;

	/*
	 *  allocate space for all buffers
	 *  point all buffers into outer space
	 */
	bc-&gt;dfirst = 0;
	bc-&gt;bsize = bsize;
	bc-&gt;f = f;
	lruinit(bc);
	for(b = bc-&gt;bb; b &lt; &amp;bc-&gt;bb[Nbcache]; b++){
		b-&gt;inuse = 0;
		b-&gt;next = 0;
		b-&gt;dirty = 0;
		if(b-&gt;data == 0)
			b-&gt;data = (char *)malloc(bc-&gt;bsize);
		if(b-&gt;data == 0)
			return -1;
		lruadd(bc, b);
	}

	return 0;
}

/*
 *  Find a buffer for block b.  If it's dirty, write it out.
 */
Bbuf *
bcfind(Bcache *bc, ulong bno)
{
	Bbuf *b;

	if(bno == Notabno)
		error("bcfind: Notabno");
	bno &amp;= ~Indbno;

	/*
	 *  if we already have a buffer for this bno, use it
	 */
	for(b = bc-&gt;bb; b &lt; &amp;bc-&gt;bb[Nbcache]; b++)
		if(b-&gt;inuse &amp;&amp; b-&gt;bno==bno)
			goto out;

	/*
	 *  get least recently used block
	 */
	b = (Bbuf*)bc-&gt;lnext;
out:
	/*
	 *  if dirty, write it out
	 */
	if(b-&gt;dirty)
		if(bcwrite(bc, b) &lt; 0)
			warning("writing dirty page");
	lruref(bc, b);
	return b;
}

/*
 *  allocate a buffer block for a block.  it's guaranteed to be there till
 *  the next Nbcache bcread's.
 */
Bbuf *
bcalloc(Bcache *bc, ulong bno)
{
	Bbuf *b;

	b = bcfind(bc, bno);
	bno &amp;= ~Indbno;
	b-&gt;bno = bno;
	b-&gt;inuse = 1;
	return b;
}

/*
 *  read a block into a buffer cache.  it's guaranteed to be there till
 *  the next Nbcache bcread's.
 */
Bbuf *
bcread(Bcache *bc, ulong bno)
{
	Bbuf *b;

	b = bcfind(bc, bno);
	bno &amp;= ~Indbno;
	if(b-&gt;bno!=bno || !b-&gt;inuse)
		/*
		 *  read in the one we really want
		 */
		if(bread(bc, bno, b-&gt;data) &lt; 0){
			b-&gt;inuse = 0;
			return 0;
		}
	b-&gt;bno = bno;
	b-&gt;inuse = 1;
	return b;
}

/*
 *  mark a page dirty, if it's already dirty force a write
 *
 *	N.B: ordering is important.
 */
void
bcmark(Bcache *bc, Bbuf *b)
{
	lruref(bc, b);

	if(b-&gt;dirty){
		bcwrite(bc, b);
		return;
	}

	b-&gt;dirty = 1;
	if(bc-&gt;dfirst)
		bc-&gt;dlast-&gt;next = b;
	else
		bc-&gt;dfirst = b;
	bc-&gt;dlast = b;
}

/*
 *  write out a page (and all preceding dirty ones)
 */
int
bcwrite(Bcache *bc, Bbuf *b)
{
	Bbuf *nb;

	/*
	 *  write out all preceding pages
	 */
	while(nb = bc-&gt;dfirst){
		if(bwrite(bc, nb-&gt;bno, nb-&gt;data) &lt; 0)
			return -1;
		nb-&gt;dirty = 0;
		bc-&gt;dfirst = nb-&gt;next;
		nb-&gt;next = 0;
		if(nb == b)
			return 0;
	}

	/*
	 *  write out this page
	 */
	if(bwrite(bc, b-&gt;bno, b-&gt;data) &lt; 0)
		return -1;
	b-&gt;dirty = 0;
	b-&gt;next = 0;
	return 0;
}

/*
 *  write out all dirty pages (in order)
 */
int
bcsync(Bcache *bc)
{
	if(bc-&gt;dfirst)
		return bcwrite(bc, bc-&gt;dlast);
	return 0;
}

/*
 *  read a block from disk
 */
int
bread(Bcache *bc, ulong bno, void *buf)
{
	uvlong x = (uvlong)bno * bc-&gt;bsize;

	if(pread(bc-&gt;f, buf, bc-&gt;bsize, x) != bc-&gt;bsize)
		return -1;
	return 0;
}

/*
 *  write a block to disk
 */
int
bwrite(Bcache *bc, ulong bno, void *buf)
{
	uvlong x = (uvlong)bno * bc-&gt;bsize;

	if(pwrite(bc-&gt;f, buf, bc-&gt;bsize, x) != bc-&gt;bsize)
		return -1;
	return 0;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

