<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/cifs/dfs.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/cifs/dfs.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * DNS referrals give two main fields: the path to connect to in
 * /Netbios-host-name/share-name/path... form and a network
 * address of how to find this path of the form domain.dom.
 *
 * The domain.dom is resolved in XP/Win2k etc using AD to do
 * a lookup (this is a consensus view, I don't think anyone
 * has proved it).  I cannot do this as AD needs Kerberos and
 * LDAP which I don't have.
 *
 * Instead I just use the NetBios names passed in the paths
 * and assume that the servers are in the same DNS domain as me
 * and have their DNS hostname set the same as their netbios
 * called-name; thankfully this always seems to be the case (so far).
 *
 * I have not added support for starting another instance of
 * cifs to connect to other servers referenced in DFS links,
 * this is not a problem for me and I think it hides a load
 * of problems of its own wrt plan9's private namespaces.
 *
 * The proximity of my test server (AD enabled) is always 0 but some
 * systems may report more meaningful values.  The expiry time is
 * similarly zero, so I guess at 5 mins.
 *
 * If the redirection points to a "hidden" share (i.e., its name
 * ends in a $) then the type of the redirection is 0 (unknown) even
 * though it is a CIFS share.
 *
 * It would be nice to add a check for which subnet a server is on
 * so our first choice is always the the server on the same subnet
 * as us which replies to a ping (i.e., is up).  This could short-
 * circuit the tests as the a server on the same subnet will always
 * be the fastest to get to.
 *
 * If I set Flags2_DFS then I don't see DFS links, I just get
 * path not found (?!).
 *
 * If I do a QueryFileInfo of a DFS link point (IE when I'am doing a walk)
 * Then I just see a directory, its not until I try to walk another level
 * That I get  "IO reparse tag not handled" error rather than
 * "Path not covered".
 *
 * If I check the extended attributes of the QueryFileInfo in walk() then I can
 * see this is a reparse point and so I can get the referral.  The only
 * problem here is that samba and the like may not support this.
 */
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;fcall.h&gt;
#include &lt;thread.h&gt;
#include &lt;libsec.h&gt;
#include &lt;ctype.h&gt;
#include &lt;9p.h&gt;
#include "cifs.h"

enum {
	Nomatch,	/* not found in cache */
	Exactmatch,	/* perfect match found */
	Badmatch	/* matched but wrong case */
};

#define SINT_MAX	0x7fffffff

typedef struct Dfscache Dfscache;
struct Dfscache {
	Dfscache*next;		/* next entry */
	char	*src;
	char	*host;
	char	*share;
	char	*path;
	long	expiry;		/* expiry time in sec */
	long	rtt;		/* round trip time, nsec */
	int	prox;		/* proximity, lower = closer */
};

Dfscache *Cache;

int
dfscacheinfo(Fmt *f)
{
	long ex;
	Dfscache *cp;

	for(cp = Cache; cp; cp = cp-&gt;next){
		ex = cp-&gt;expiry - time(nil);
		if(ex &lt; 0)
			ex = -1;
		fmtprint(f, "%-42s %6ld %8.1f %4d %-16s %-24s %s\n",
			cp-&gt;src, ex, (double)cp-&gt;rtt/1000.0L, cp-&gt;prox,
			cp-&gt;host, cp-&gt;share, cp-&gt;path);
	}
	return 0;
}

char *
trimshare(char *s)
{
	char *p;
	static char name[128];

	strncpy(name, s, sizeof(name));
	name[sizeof(name)-1] = 0;
	if((p = strrchr(name, '$')) != nil &amp;&amp; p[1] == 0)
		*p = 0;
	return name;
}

static Dfscache *
lookup(char *path, int *match)
{
	int len, n, m;
	Dfscache *cp, *best;

	if(match)
		*match = Nomatch;

	len = 0;
	best = nil;
	m = strlen(path);
	for(cp = Cache; cp; cp = cp-&gt;next){
		n = strlen(cp-&gt;src);
		if(n &lt; len)
			continue;
		if(strncmp(path, cp-&gt;src, n) != 0)
			continue;
		if(path[n] != 0 &amp;&amp; path[n] != '/')
			continue;
		best = cp;
		len = n;
		if(n == m){
			if(match)
				*match = Exactmatch;
			break;
		}
	}
	return best;
}

char *
mapfile(char *opath)
{
	int exact;
	Dfscache *cp;
	char *p, *path;
	static char npath[MAX_DFS_PATH];

	path = opath;
	if((cp = lookup(path, &amp;exact)) != nil){
		snprint(npath, sizeof npath, "/%s%s%s%s", cp-&gt;share,
			*cp-&gt;path? "/": "", cp-&gt;path, path + strlen(cp-&gt;src));
		path = npath;
	}

	if((p = strchr(path+1, '/')) == nil)
		p = "/";
	if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
		print("mapfile src=%q =&gt; dst=%q\n", opath, p);
	return p;
}

int
mapshare(char *path, Share **osp)
{
	int i;
	Share *sp;
	Dfscache *cp;
	char *s, *try;
	char *tail[] = { "", "$" };

	if((cp = lookup(path, nil)) == nil)
		return 0;

	for(sp = Shares; sp &lt; Shares+Nshares; sp++){
		s = trimshare(sp-&gt;name);
		if(cistrcmp(cp-&gt;share, s) != 0)
			continue;
		if(Checkcase &amp;&amp; strcmp(cp-&gt;share, s) != 0)
			continue;
		if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
			print("mapshare, already connected, src=%q =&gt; dst=%q\n", path, sp-&gt;name);
		*osp = sp;
		return 0;
	}
	/*
	 * Try to autoconnect to share if it is not known.  Note even if you
	 * didn't specify any shares and let the system autoconnect you may
	 * not already have the share you need as RAP (which we use) throws
	 * away names &gt; 12 chars long.  If we where to use RPC then this block
	 * of code would be less important, though it would still be useful
	 * to catch Shares added since cifs(1) was started.
	 */
	sp = Shares + Nshares;
	for(i = 0; i &lt; 2; i++){
		try = smprint("%s%s", cp-&gt;share, tail[i]);
		if(CIFStreeconnect(Sess, Sess-&gt;cname, try, sp) == 0){
			sp-&gt;name = try;
			*osp = sp;
			Nshares++;
			if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
				print("mapshare connected, src=%q dst=%q\n",
					path, cp-&gt;share);
			return 0;
		}
		free(try);
	}

	if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
		print("mapshare failed src=%s\n", path);
	werrstr("not found");
	return -1;
}

/*
 * Rtt_tol is the fractional tollerance for RTT comparisons.
 * If a later (further down the list) host's RTT is less than
 * 1/Rtt_tol better than my current best then I don't bother
 * with it.  This biases me towards entries at the top of the list
 * which Active Directory has already chosen for me and prevents
 * noise in RTTs from pushing me to more distant machines.
 */
static int
remap(Dfscache *cp, Refer *re)
{
	int n;
	long rtt;
	char *p, *a[4];
	enum {
		Hostname = 1,
		Sharename = 2,
		Pathname = 3,

		Rtt_tol = 10
	};

	if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
		print("	remap %s\n", re-&gt;addr);

	for(p = re-&gt;addr; *p; p++)
		if(*p == '\\')
			*p = '/';

	if(cp-&gt;prox &lt; re-&gt;prox){
		if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
			print("	remap %d &lt; %d\n", cp-&gt;prox, re-&gt;prox);
		return -1;
	}
	if((n = getfields(re-&gt;addr, a, sizeof(a), 0, "/")) &lt; 3){
		if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
			print("	remap nfields=%d\n", n);
		return -1;
	}
	if((rtt = ping(a[Hostname], Dfstout)) == -1){
		if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
			print("	remap ping failed\n");
		return -1;
	}
	if(cp-&gt;rtt &lt; rtt &amp;&amp; (rtt/labs(rtt-cp-&gt;rtt)) &lt; Rtt_tol){
		if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
			print("	remap bad ping %ld &lt; %ld &amp;&amp; %ld &lt; %d\n",
				cp-&gt;rtt, rtt, (rtt/labs(rtt-cp-&gt;rtt)), Rtt_tol);
		return -1;
	}

	if(n &lt; 4)
		a[Pathname] = "";
	if(re-&gt;ttl == 0)
		re-&gt;ttl = 60*5;

	free(cp-&gt;host);
	free(cp-&gt;share);
	free(cp-&gt;path);
	cp-&gt;rtt = rtt;
	cp-&gt;prox = re-&gt;prox;
	cp-&gt;expiry = time(nil)+re-&gt;ttl;
	cp-&gt;host = estrdup9p(a[Hostname]);
	cp-&gt;share = estrdup9p(trimshare(a[Sharename]));
	cp-&gt;path = estrdup9p(a[Pathname]);
	if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
		print("	remap ping OK prox=%d host=%s share=%s path=%s\n",
			cp-&gt;prox, cp-&gt;host, cp-&gt;share, cp-&gt;path);
	return 0;
}

static int
redir1(Session *s, char *path, Dfscache *cp, int level)
{
	Refer retab[16], *re;
	int n, gflags, used, found;

	if(level &gt; 8)
		return -1;

	if((n = T2getdfsreferral(s, &amp;Ipc, path, &amp;gflags, &amp;used, retab,
	    nelem(retab))) == -1)
		return -1;

	if(! (gflags &amp; DFS_HEADER_ROOT))
		used = SINT_MAX;

	found = 0;
	for(re = retab; re &lt; retab+n; re++){
		if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
			print("referal level=%d prox=%d path=%q addr=%q\n",
				level, re-&gt;prox, re-&gt;path, re-&gt;addr);

		if(gflags &amp; DFS_HEADER_STORAGE){
			if(remap(cp, re) == 0)
				found = 1;
		} else{
			if(redir1(s, re-&gt;addr, cp, level+1) != -1)  /* ???? */
				found = 1;
		}
		free(re-&gt;addr);
		free(re-&gt;path);
	}

	if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
		print("referal level=%d path=%q found=%d used=%d\n",
			level, path, found, used);
	if(!found)
		return -1;
	return used;
}

/*
 * We can afford to ignore the used count returned by redir
 * because of the semantics of 9p - we always walk to directories
 * ome and we a time and we always walk before any other file operations
 */
int
redirect(Session *s, Share *sp, char *path)
{
	int match;
	char *unc;
	Dfscache *cp;

	if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
		print("redirect name=%q path=%q\n", sp-&gt;name, path);

	cp = lookup(path, &amp;match);
	if(match == Badmatch)
		return -1;

	if(cp &amp;&amp; match == Exactmatch){
		if(cp-&gt;expiry &gt;= time(nil)){		/* cache hit */
			if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
				print("redirect cache=hit src=%q =&gt; share=%q path=%q\n",
					cp-&gt;src, cp-&gt;share, cp-&gt;path);
			return 0;

		} else{				/* cache hit, but entry stale */
			cp-&gt;rtt = SINT_MAX;
			cp-&gt;prox = SINT_MAX;

			unc = smprint("//%s/%s/%s%s%s", s-&gt;auth-&gt;windom,
				cp-&gt;share, cp-&gt;path, *cp-&gt;path? "/": "",
				path + strlen(cp-&gt;src) + 1);
			if(unc == nil)
				sysfatal("no memory: %r");
			if(redir1(s, unc, cp, 1) == -1){
				if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
					print("redirect refresh failed unc=%q\n",
						unc);
				free(unc);
				return -1;
			}
			free(unc);
			if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
				print("redirect refresh cache=stale src=%q =&gt; share=%q path=%q\n",
					cp-&gt;src, cp-&gt;share, cp-&gt;path);
			return 0;
		}
	}


	/* in-exact match or complete miss */
	if(cp)
		unc = smprint("//%s/%s/%s%s%s", s-&gt;auth-&gt;windom, cp-&gt;share,
			cp-&gt;path, *cp-&gt;path? "/": "", path + strlen(cp-&gt;src) + 1);
	else
		unc = smprint("//%s%s", s-&gt;auth-&gt;windom, path);
	if(unc == nil)
		sysfatal("no memory: %r");

	cp = emalloc9p(sizeof(Dfscache));
	memset(cp, 0, sizeof(Dfscache));
	cp-&gt;rtt = SINT_MAX;
	cp-&gt;prox = SINT_MAX;

	if(redir1(s, unc, cp, 1) == -1){
		if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
			print("redirect new failed unc=%q\n", unc);
		free(unc);
		free(cp);
		return -1;
	}
	free(unc);

	cp-&gt;src = estrdup9p(path);
	cp-&gt;next = Cache;
	Cache = cp;
	if(Debug &amp;&amp; strstr(Debug, "dfs") != nil)
		print("redirect cache=miss src=%q =&gt; share=%q path=%q\n",
			cp-&gt;src, cp-&gt;share, cp-&gt;path);
	return 0;
}

<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

