<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/fossil/file.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/fossil/file.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include "stdinc.h"
#include "9.h"			/* for consPrint */
#include "dat.h"
#include "fns.h"
#include "error.h"

/*
 * locking order is upwards.  A thread can hold the lock for a File
 * and then acquire the lock of its parent
 */

struct File {
	Fs	*fs;		/* immutable */

	/* meta data for file: protected by the lk in the parent */
	int	ref;		/* holds this data structure up */

	int	partial;	/* file was never really open */
	int	removed;	/* file has been removed */
	int	dirty;	/* dir is dirty with respect to meta data in block */
	u32int	boff;	/* block offset within msource for this file's meta data */

	DirEntry dir;	/* meta data for this file, including component name */

	File	*up;		/* parent file (directory) */
	File	*next;		/* sibling */

	/* data for file */
	VtLock	*lk;		/* lock for the following */
	Source	*source;
	Source	*msource;	/* for directories: meta data for children */
	File	*down;		/* children */

	int	mode;
	int	issnapshot;
};

static int fileMetaFlush2(File*, char*);
static u32int fileMetaAlloc(File*, DirEntry*, u32int);
static int fileRLock(File*);
static void fileRUnlock(File*);
static int fileLock(File*);
static void fileUnlock(File*);
static void fileMetaLock(File*);
static void fileMetaUnlock(File*);
static void fileRAccess(File*);
static void fileWAccess(File*, char*);

static File *
fileAlloc(Fs *fs)
{
	File *f;

	f = vtMemAllocZ(sizeof(File));
	f-&gt;lk = vtLockAlloc();
	f-&gt;ref = 1;
	f-&gt;fs = fs;
	f-&gt;boff = NilBlock;
	f-&gt;mode = fs-&gt;mode;
	return f;
}

static void
fileFree(File *f)
{
	sourceClose(f-&gt;source);
	vtLockFree(f-&gt;lk);
	sourceClose(f-&gt;msource);
	deCleanup(&amp;f-&gt;dir);

	memset(f, ~0, sizeof(File));
	vtMemFree(f);
}

/*
 * the file is locked already
 * f-&gt;msource is unlocked
 */
static File *
dirLookup(File *f, char *elem)
{
	int i;
	MetaBlock mb;
	MetaEntry me;
	Block *b;
	Source *meta;
	File *ff;
	u32int bo, nb;

	meta = f-&gt;msource;
	b = nil;
	if(!sourceLock(meta, -1))
		return nil;
	nb = (sourceGetSize(meta)+meta-&gt;dsize-1)/meta-&gt;dsize;
	for(bo=0; bo&lt;nb; bo++){
		b = sourceBlock(meta, bo, OReadOnly);
		if(b == nil)
			goto Err;
		if(!mbUnpack(&amp;mb, b-&gt;data, meta-&gt;dsize))
			goto Err;
		if(mbSearch(&amp;mb, elem, &amp;i, &amp;me)){
			ff = fileAlloc(f-&gt;fs);
			if(!deUnpack(&amp;ff-&gt;dir, &amp;me)){
				fileFree(ff);
				goto Err;
			}
			sourceUnlock(meta);
			blockPut(b);
			ff-&gt;boff = bo;
			ff-&gt;mode = f-&gt;mode;
			ff-&gt;issnapshot = f-&gt;issnapshot;
			return ff;
		}

		blockPut(b);
		b = nil;
	}
	vtSetError(ENoFile);
	/* fall through */
Err:
	sourceUnlock(meta);
	blockPut(b);
	return nil;
}

File *
fileRoot(Source *r)
{
	Block *b;
	Source *r0, *r1, *r2;
	MetaBlock mb;
	MetaEntry me;
	File *root, *mr;
	Fs *fs;

	b = nil;
	root = nil;
	mr = nil;
	r1 = nil;
	r2 = nil;

	fs = r-&gt;fs;
	if(!sourceLock(r, -1))
		return nil;
	r0 = sourceOpen(r, 0, fs-&gt;mode, 0);
	if(r0 == nil)
		goto Err;
	r1 = sourceOpen(r, 1, fs-&gt;mode, 0);
	if(r1 == nil)
		goto Err;
	r2 = sourceOpen(r, 2, fs-&gt;mode, 0);
	if(r2 == nil)
		goto Err;

	mr = fileAlloc(fs);
	mr-&gt;msource = r2;
	r2 = nil;

	root = fileAlloc(fs);
	root-&gt;boff = 0;
	root-&gt;up = mr;
	root-&gt;source = r0;
	r0-&gt;file = root;			/* point back to source */
	r0 = nil;
	root-&gt;msource = r1;
	r1 = nil;

	mr-&gt;down = root;

	if(!sourceLock(mr-&gt;msource, -1))
		goto Err;
	b = sourceBlock(mr-&gt;msource, 0, OReadOnly);
	sourceUnlock(mr-&gt;msource);
	if(b == nil)
		goto Err;

	if(!mbUnpack(&amp;mb, b-&gt;data, mr-&gt;msource-&gt;dsize))
		goto Err;

	meUnpack(&amp;me, &amp;mb, 0);
	if(!deUnpack(&amp;root-&gt;dir, &amp;me))
		goto Err;
	blockPut(b);
	sourceUnlock(r);
	fileRAccess(root);

	return root;
Err:
	blockPut(b);
	if(r0)
		sourceClose(r0);
	if(r1)
		sourceClose(r1);
	if(r2)
		sourceClose(r2);
	if(mr)
		fileFree(mr);
	if(root)
		fileFree(root);
	sourceUnlock(r);

	return nil;
}

static Source *
fileOpenSource(File *f, u32int offset, u32int gen, int dir, uint mode,
	int issnapshot)
{
	char *rname, *fname;
	Source *r;

	if(!sourceLock(f-&gt;source, mode))
		return nil;
	r = sourceOpen(f-&gt;source, offset, mode, issnapshot);
	sourceUnlock(f-&gt;source);
	if(r == nil)
		return nil;
	if(r-&gt;gen != gen){
		vtSetError(ERemoved);
		goto Err;
	}
	if(r-&gt;dir != dir &amp;&amp; r-&gt;mode != -1){
		/* this hasn't been as useful as we hoped it would be. */
		rname = sourceName(r);
		fname = fileName(f);
		consPrint("%s: source %s for file %s: fileOpenSource: "
			"dir mismatch %d %d\n",
			f-&gt;source-&gt;fs-&gt;name, rname, fname, r-&gt;dir, dir);
		free(rname);
		free(fname);

		vtSetError(EBadMeta);
		goto Err;
	}
	return r;
Err:
	sourceClose(r);
	return nil;
}

File *
_fileWalk(File *f, char *elem, int partial)
{
	File *ff;

	fileRAccess(f);

	if(elem[0] == 0){
		vtSetError(EBadPath);
		return nil;
	}

	if(!fileIsDir(f)){
		vtSetError(ENotDir);
		return nil;
	}

	if(strcmp(elem, ".") == 0){
		return fileIncRef(f);
	}

	if(strcmp(elem, "..") == 0){
		if(fileIsRoot(f))
			return fileIncRef(f);
		return fileIncRef(f-&gt;up);
	}

	if(!fileLock(f))
		return nil;

	for(ff = f-&gt;down; ff; ff=ff-&gt;next){
		if(strcmp(elem, ff-&gt;dir.elem) == 0 &amp;&amp; !ff-&gt;removed){
			ff-&gt;ref++;
			goto Exit;
		}
	}

	ff = dirLookup(f, elem);
	if(ff == nil)
		goto Err;

	if(ff-&gt;dir.mode &amp; ModeSnapshot){
		ff-&gt;mode = OReadOnly;
		ff-&gt;issnapshot = 1;
	}

	if(partial){
		/*
		 * Do nothing.  We're opening this file only so we can clri it.
		 * Usually the sources can't be opened, hence we won't even bother.
		 * Be VERY careful with the returned file.  If you hand it to a routine
		 * expecting ff-&gt;source and/or ff-&gt;msource to be non-nil, we're
		 * likely to dereference nil.  FileClri should be the only routine
		 * setting partial.
		 */
		ff-&gt;partial = 1;
	}else if(ff-&gt;dir.mode &amp; ModeDir){
		ff-&gt;source = fileOpenSource(f, ff-&gt;dir.entry, ff-&gt;dir.gen,
			1, ff-&gt;mode, ff-&gt;issnapshot);
		ff-&gt;msource = fileOpenSource(f, ff-&gt;dir.mentry, ff-&gt;dir.mgen,
			0, ff-&gt;mode, ff-&gt;issnapshot);
		if(ff-&gt;source == nil || ff-&gt;msource == nil)
			goto Err;
	}else{
		ff-&gt;source = fileOpenSource(f, ff-&gt;dir.entry, ff-&gt;dir.gen,
			0, ff-&gt;mode, ff-&gt;issnapshot);
		if(ff-&gt;source == nil)
			goto Err;
	}

	/* link in and up parent ref count */
	if (ff-&gt;source)
		ff-&gt;source-&gt;file = ff;		/* point back */
	ff-&gt;next = f-&gt;down;
	f-&gt;down = ff;
	ff-&gt;up = f;
	fileIncRef(f);
Exit:
	fileUnlock(f);
	return ff;
Err:
	fileUnlock(f);
	if(ff != nil)
		fileDecRef(ff);
	return nil;
}

File *
fileWalk(File *f, char *elem)
{
	return _fileWalk(f, elem, 0);
}

File *
_fileOpen(Fs *fs, char *path, int partial)
{
	File *f, *ff;
	char *p, elem[VtMaxStringSize], *opath;
	int n;

	f = fs-&gt;file;
	fileIncRef(f);
	opath = path;
	while(*path != 0){
		for(p = path; *p &amp;&amp; *p != '/'; p++)
			;
		n = p - path;
		if(n &gt; 0){
			if(n &gt; VtMaxStringSize){
				vtSetError("%s: element too long", EBadPath);
				goto Err;
			}
			memmove(elem, path, n);
			elem[n] = 0;
			ff = _fileWalk(f, elem, partial &amp;&amp; *p=='\0');
			if(ff == nil){
				vtSetError("%.*s: %R", utfnlen(opath, p-opath),
					opath);
				goto Err;
			}
			fileDecRef(f);
			f = ff;
		}
		if(*p == '/')
			p++;
		path = p;
	}
	return f;
Err:
	fileDecRef(f);
	return nil;
}

File*
fileOpen(Fs *fs, char *path)
{
	return _fileOpen(fs, path, 0);
}

static void
fileSetTmp(File *f, int istmp)
{
	int i;
	Entry e;
	Source *r;

	for(i=0; i&lt;2; i++){
		if(i==0)
			r = f-&gt;source;
		else
			r = f-&gt;msource;
		if(r == nil)
			continue;
		if(!sourceGetEntry(r, &amp;e)){
			fprint(2, "sourceGetEntry failed (cannot happen): %r\n");
			continue;
		}
		if(istmp)
			e.flags |= VtEntryNoArchive;
		else
			e.flags &amp;= ~VtEntryNoArchive;
		if(!sourceSetEntry(r, &amp;e)){
			fprint(2, "sourceSetEntry failed (cannot happen): %r\n");
			continue;
		}
	}
}

File *
fileCreate(File *f, char *elem, ulong mode, char *uid)
{
	File *ff;
	DirEntry *dir;
	Source *pr, *r, *mr;
	int isdir;

	if(!fileLock(f))
		return nil;

	r = nil;
	mr = nil;
	for(ff = f-&gt;down; ff; ff=ff-&gt;next){
		if(strcmp(elem, ff-&gt;dir.elem) == 0 &amp;&amp; !ff-&gt;removed){
			ff = nil;
			vtSetError(EExists);
			goto Err1;
		}
	}

	ff = dirLookup(f, elem);
	if(ff != nil){
		vtSetError(EExists);
		goto Err1;
	}

	pr = f-&gt;source;
	if(pr-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		goto Err1;
	}

	if(!sourceLock2(f-&gt;source, f-&gt;msource, -1))
		goto Err1;

	ff = fileAlloc(f-&gt;fs);
	isdir = mode &amp; ModeDir;

	r = sourceCreate(pr, pr-&gt;dsize, isdir, 0);
	if(r == nil)
		goto Err;
	if(isdir){
		mr = sourceCreate(pr, pr-&gt;dsize, 0, r-&gt;offset);
		if(mr == nil)
			goto Err;
	}

	dir = &amp;ff-&gt;dir;
	dir-&gt;elem = vtStrDup(elem);
	dir-&gt;entry = r-&gt;offset;
	dir-&gt;gen = r-&gt;gen;
	if(isdir){
		dir-&gt;mentry = mr-&gt;offset;
		dir-&gt;mgen = mr-&gt;gen;
	}
	dir-&gt;size = 0;
	if(!fsNextQid(f-&gt;fs, &amp;dir-&gt;qid))
		goto Err;
	dir-&gt;uid = vtStrDup(uid);
	dir-&gt;gid = vtStrDup(f-&gt;dir.gid);
	dir-&gt;mid = vtStrDup(uid);
	dir-&gt;mtime = time(0L);
	dir-&gt;mcount = 0;
	dir-&gt;ctime = dir-&gt;mtime;
	dir-&gt;atime = dir-&gt;mtime;
	dir-&gt;mode = mode;

	ff-&gt;boff = fileMetaAlloc(f, dir, 0);
	if(ff-&gt;boff == NilBlock)
		goto Err;

	sourceUnlock(f-&gt;source);
	sourceUnlock(f-&gt;msource);

	ff-&gt;source = r;
	r-&gt;file = ff;			/* point back */
	ff-&gt;msource = mr;

	if(mode&amp;ModeTemporary){
		if(!sourceLock2(r, mr, -1))
			goto Err1;
		fileSetTmp(ff, 1);
		sourceUnlock(r);
		if(mr)
			sourceUnlock(mr);
	}

	/* committed */

	/* link in and up parent ref count */
	ff-&gt;next = f-&gt;down;
	f-&gt;down = ff;
	ff-&gt;up = f;
	fileIncRef(f);

	fileWAccess(f, uid);

	fileUnlock(f);
	return ff;

Err:
	sourceUnlock(f-&gt;source);
	sourceUnlock(f-&gt;msource);
Err1:
	if(r){
		sourceLock(r, -1);
		sourceRemove(r);
	}
	if(mr){
		sourceLock(mr, -1);
		sourceRemove(mr);
	}
	if(ff)
		fileDecRef(ff);
	fileUnlock(f);
	return 0;
}

int
fileRead(File *f, void *buf, int cnt, vlong offset)
{
	Source *s;
	uvlong size;
	u32int bn;
	int off, dsize, n, nn;
	Block *b;
	uchar *p;

if(0)fprint(2, "fileRead: %s %d, %lld\n", f-&gt;dir.elem, cnt, offset);

	if(!fileRLock(f))
		return -1;

	if(offset &lt; 0){
		vtSetError(EBadOffset);
		goto Err1;
	}

	fileRAccess(f);

	if(!sourceLock(f-&gt;source, OReadOnly))
		goto Err1;

	s = f-&gt;source;
	dsize = s-&gt;dsize;
	size = sourceGetSize(s);

	if(offset &gt;= size)
		offset = size;

	if(cnt &gt; size-offset)
		cnt = size-offset;
	bn = offset/dsize;
	off = offset%dsize;
	p = buf;
	while(cnt &gt; 0){
		b = sourceBlock(s, bn, OReadOnly);
		if(b == nil)
			goto Err;
		n = cnt;
		if(n &gt; dsize-off)
			n = dsize-off;
		nn = dsize-off;
		if(nn &gt; n)
			nn = n;
		memmove(p, b-&gt;data+off, nn);
		memset(p+nn, 0, nn-n);
		off = 0;
		bn++;
		cnt -= n;
		p += n;
		blockPut(b);
	}
	sourceUnlock(s);
	fileRUnlock(f);
	return p-(uchar*)buf;

Err:
	sourceUnlock(s);
Err1:
	fileRUnlock(f);
	return -1;
}

/*
 * Changes the file block bn to be the given block score.
 * Very sneaky.  Only used by flfmt.
 */
int
fileMapBlock(File *f, ulong bn, uchar score[VtScoreSize], ulong tag)
{
	Block *b;
	Entry e;
	Source *s;

	if(!fileLock(f))
		return 0;

	s = nil;
	if(f-&gt;dir.mode &amp; ModeDir){
		vtSetError(ENotFile);
		goto Err;
	}

	if(f-&gt;source-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		goto Err;
	}

	if(!sourceLock(f-&gt;source, -1))
		goto Err;

	s = f-&gt;source;
	b = _sourceBlock(s, bn, OReadWrite, 1, tag);
	if(b == nil)
		goto Err;

	if(!sourceGetEntry(s, &amp;e))
		goto Err;
	if(b-&gt;l.type == BtDir){
		memmove(e.score, score, VtScoreSize);
		assert(e.tag == tag || e.tag == 0);
		e.tag = tag;
		e.flags |= VtEntryLocal;
		entryPack(&amp;e, b-&gt;data, f-&gt;source-&gt;offset % f-&gt;source-&gt;epb);
	}else
		memmove(b-&gt;data + (bn%(e.psize/VtScoreSize))*VtScoreSize, score, VtScoreSize);
	blockDirty(b);
	blockPut(b);
	sourceUnlock(s);
	fileUnlock(f);
	return 1;

Err:
	if(s)
		sourceUnlock(s);
	fileUnlock(f);
	return 0;
}

int
fileSetSize(File *f, uvlong size)
{
	int r;

	if(!fileLock(f))
		return 0;
	r = 0;
	if(f-&gt;dir.mode &amp; ModeDir){
		vtSetError(ENotFile);
		goto Err;
	}
	if(f-&gt;source-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		goto Err;
	}
	if(!sourceLock(f-&gt;source, -1))
		goto Err;
	r = sourceSetSize(f-&gt;source, size);
	sourceUnlock(f-&gt;source);
Err:
	fileUnlock(f);
	return r;
}

int
fileWrite(File *f, void *buf, int cnt, vlong offset, char *uid)
{
	Source *s;
	ulong bn;
	int off, dsize, n;
	Block *b;
	uchar *p;
	vlong eof;

if(0)fprint(2, "fileWrite: %s %d, %lld\n", f-&gt;dir.elem, cnt, offset);

	if(!fileLock(f))
		return -1;

	s = nil;
	if(f-&gt;dir.mode &amp; ModeDir){
		vtSetError(ENotFile);
		goto Err;
	}

	if(f-&gt;source-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		goto Err;
	}
	if(offset &lt; 0){
		vtSetError(EBadOffset);
		goto Err;
	}

	fileWAccess(f, uid);

	if(!sourceLock(f-&gt;source, -1))
		goto Err;
	s = f-&gt;source;
	dsize = s-&gt;dsize;

	eof = sourceGetSize(s);
	if(f-&gt;dir.mode &amp; ModeAppend)
		offset = eof;
	bn = offset/dsize;
	off = offset%dsize;
	p = buf;
	while(cnt &gt; 0){
		n = cnt;
		if(n &gt; dsize-off)
			n = dsize-off;
		b = sourceBlock(s, bn, n&lt;dsize?OReadWrite:OOverWrite);
		if(b == nil){
			if(offset &gt; eof)
				sourceSetSize(s, offset);
			goto Err;
		}
		memmove(b-&gt;data+off, p, n);
		off = 0;
		cnt -= n;
		p += n;
		offset += n;
		bn++;
		blockDirty(b);
		blockPut(b);
	}
	if(offset &gt; eof &amp;&amp; !sourceSetSize(s, offset))
		goto Err;
	sourceUnlock(s);
	fileUnlock(f);
	return p-(uchar*)buf;
Err:
	if(s)
		sourceUnlock(s);
	fileUnlock(f);
	return -1;
}

int
fileGetDir(File *f, DirEntry *dir)
{
	if(!fileRLock(f))
		return 0;

	fileMetaLock(f);
	deCopy(dir, &amp;f-&gt;dir);
	fileMetaUnlock(f);

	if(!fileIsDir(f)){
		if(!sourceLock(f-&gt;source, OReadOnly)){
			fileRUnlock(f);
			return 0;
		}
		dir-&gt;size = sourceGetSize(f-&gt;source);
		sourceUnlock(f-&gt;source);
	}
	fileRUnlock(f);

	return 1;
}

int
fileTruncate(File *f, char *uid)
{
	if(fileIsDir(f)){
		vtSetError(ENotFile);
		return 0;
	}

	if(!fileLock(f))
		return 0;

	if(f-&gt;source-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		fileUnlock(f);
		return 0;
	}
	if(!sourceLock(f-&gt;source, -1)){
		fileUnlock(f);
		return 0;
	}
	if(!sourceTruncate(f-&gt;source)){
		sourceUnlock(f-&gt;source);
		fileUnlock(f);
		return 0;
	}
	sourceUnlock(f-&gt;source);
	fileUnlock(f);

	fileWAccess(f, uid);

	return 1;
}

int
fileSetDir(File *f, DirEntry *dir, char *uid)
{
	File *ff;
	char *oelem;
	u32int mask;
	u64int size;

	/* can not set permissions for the root */
	if(fileIsRoot(f)){
		vtSetError(ERoot);
		return 0;
	}

	if(!fileLock(f))
		return 0;

	if(f-&gt;source-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		fileUnlock(f);
		return 0;
	}

	fileMetaLock(f);

	/* check new name does not already exist */
	if(strcmp(f-&gt;dir.elem, dir-&gt;elem) != 0){
		for(ff = f-&gt;up-&gt;down; ff; ff=ff-&gt;next){
			if(strcmp(dir-&gt;elem, ff-&gt;dir.elem) == 0 &amp;&amp; !ff-&gt;removed){
				vtSetError(EExists);
				goto Err;
			}
		}

		ff = dirLookup(f-&gt;up, dir-&gt;elem);
		if(ff != nil){
			fileDecRef(ff);
			vtSetError(EExists);
			goto Err;
		}
	}

	if(!sourceLock2(f-&gt;source, f-&gt;msource, -1))
		goto Err;
	if(!fileIsDir(f)){
		size = sourceGetSize(f-&gt;source);
		if(size != dir-&gt;size){
			if(!sourceSetSize(f-&gt;source, dir-&gt;size)){
				sourceUnlock(f-&gt;source);
				if(f-&gt;msource)
					sourceUnlock(f-&gt;msource);
				goto Err;
			}
			/* commited to changing it now */
		}
	}
	/* commited to changing it now */
	if((f-&gt;dir.mode&amp;ModeTemporary) != (dir-&gt;mode&amp;ModeTemporary))
		fileSetTmp(f, dir-&gt;mode&amp;ModeTemporary);
	sourceUnlock(f-&gt;source);
	if(f-&gt;msource)
		sourceUnlock(f-&gt;msource);

	oelem = nil;
	if(strcmp(f-&gt;dir.elem, dir-&gt;elem) != 0){
		oelem = f-&gt;dir.elem;
		f-&gt;dir.elem = vtStrDup(dir-&gt;elem);
	}

	if(strcmp(f-&gt;dir.uid, dir-&gt;uid) != 0){
		vtMemFree(f-&gt;dir.uid);
		f-&gt;dir.uid = vtStrDup(dir-&gt;uid);
	}

	if(strcmp(f-&gt;dir.gid, dir-&gt;gid) != 0){
		vtMemFree(f-&gt;dir.gid);
		f-&gt;dir.gid = vtStrDup(dir-&gt;gid);
	}

	f-&gt;dir.mtime = dir-&gt;mtime;
	f-&gt;dir.atime = dir-&gt;atime;

//fprint(2, "mode %x %x ", f-&gt;dir.mode, dir-&gt;mode);
	mask = ~(ModeDir|ModeSnapshot);
	f-&gt;dir.mode &amp;= ~mask;
	f-&gt;dir.mode |= mask &amp; dir-&gt;mode;
	f-&gt;dirty = 1;
//fprint(2, "-&gt;%x\n", f-&gt;dir.mode);

	fileMetaFlush2(f, oelem);
	vtMemFree(oelem);

	fileMetaUnlock(f);
	fileUnlock(f);

	fileWAccess(f-&gt;up, uid);

	return 1;
Err:
	fileMetaUnlock(f);
	fileUnlock(f);
	return 0;
}

int
fileSetQidSpace(File *f, u64int offset, u64int max)
{
	int ret;

	if(!fileLock(f))
		return 0;
	fileMetaLock(f);
	f-&gt;dir.qidSpace = 1;
	f-&gt;dir.qidOffset = offset;
	f-&gt;dir.qidMax = max;
	ret = fileMetaFlush2(f, nil)&gt;=0;
	fileMetaUnlock(f);
	fileUnlock(f);
	return ret;
}


uvlong
fileGetId(File *f)
{
	/* immutable */
	return f-&gt;dir.qid;
}

ulong
fileGetMcount(File *f)
{
	ulong mcount;

	fileMetaLock(f);
	mcount = f-&gt;dir.mcount;
	fileMetaUnlock(f);
	return mcount;
}

ulong
fileGetMode(File *f)
{
	ulong mode;

	fileMetaLock(f);
	mode = f-&gt;dir.mode;
	fileMetaUnlock(f);
	return mode;
}

int
fileIsDir(File *f)
{
	/* immutable */
	return (f-&gt;dir.mode &amp; ModeDir) != 0;
}

int
fileIsAppend(File *f)
{
	return (f-&gt;dir.mode &amp; ModeAppend) != 0;
}

int
fileIsExclusive(File *f)
{
	return (f-&gt;dir.mode &amp; ModeExclusive) != 0;
}

int
fileIsTemporary(File *f)
{
	return (f-&gt;dir.mode &amp; ModeTemporary) != 0;
}

int
fileIsRoot(File *f)
{
	return f == f-&gt;fs-&gt;file;
}

int
fileIsRoFs(File *f)
{
	return f-&gt;fs-&gt;mode == OReadOnly;
}

int
fileGetSize(File *f, uvlong *size)
{
	if(!fileRLock(f))
		return 0;
	if(!sourceLock(f-&gt;source, OReadOnly)){
		fileRUnlock(f);
		return 0;
	}
	*size = sourceGetSize(f-&gt;source);
	sourceUnlock(f-&gt;source);
	fileRUnlock(f);

	return 1;
}

int
fileMetaFlush(File *f, int rec)
{
	File **kids, *p;
	int nkids;
	int i, rv;

	fileMetaLock(f);
	rv = fileMetaFlush2(f, nil);
	fileMetaUnlock(f);

	if(!rec || !fileIsDir(f))
		return rv;

	if(!fileLock(f))
		return rv;
	nkids = 0;
	for(p=f-&gt;down; p; p=p-&gt;next)
		nkids++;
	kids = vtMemAlloc(nkids*sizeof(File*));
	i = 0;
	for(p=f-&gt;down; p; p=p-&gt;next){
		kids[i++] = p;
		p-&gt;ref++;
	}
	fileUnlock(f);

	for(i=0; i&lt;nkids; i++){
		rv |= fileMetaFlush(kids[i], 1);
		fileDecRef(kids[i]);
	}
	vtMemFree(kids);
	return rv;
}

/* assumes metaLock is held */
static int
fileMetaFlush2(File *f, char *oelem)
{
	File *fp;
	Block *b, *bb;
	MetaBlock mb;
	MetaEntry me, me2;
	int i, n;
	u32int boff;

	if(!f-&gt;dirty)
		return 0;

	if(oelem == nil)
		oelem = f-&gt;dir.elem;

//print("fileMetaFlush %s-&gt;%s\n", oelem, f-&gt;dir.elem);

	fp = f-&gt;up;

	if(!sourceLock(fp-&gt;msource, -1))
		return -1;
	/* can happen if source is clri'ed out from under us */
	if(f-&gt;boff == NilBlock)
		goto Err1;
	b = sourceBlock(fp-&gt;msource, f-&gt;boff, OReadWrite);
	if(b == nil)
		goto Err1;

	if(!mbUnpack(&amp;mb, b-&gt;data, fp-&gt;msource-&gt;dsize))
		goto Err;
	if(!mbSearch(&amp;mb, oelem, &amp;i, &amp;me))
		goto Err;

	n = deSize(&amp;f-&gt;dir);
if(0)fprint(2, "old size %d new size %d\n", me.size, n);

	if(mbResize(&amp;mb, &amp;me, n)){
		/* fits in the block */
		mbDelete(&amp;mb, i);
		if(strcmp(f-&gt;dir.elem, oelem) != 0)
			mbSearch(&amp;mb, f-&gt;dir.elem, &amp;i, &amp;me2);
		dePack(&amp;f-&gt;dir, &amp;me);
		mbInsert(&amp;mb, i, &amp;me);
		mbPack(&amp;mb);
		blockDirty(b);
		blockPut(b);
		sourceUnlock(fp-&gt;msource);
		f-&gt;dirty = 0;

		return 1;
	}

	/*
	 * moving entry to another block
	 * it is feasible for the fs to crash leaving two copies
	 * of the directory entry.  This is just too much work to
	 * fix.  Given that entries are only allocated in a block that
	 * is less than PercentageFull, most modifications of meta data
	 * will fit within the block.  i.e. this code should almost
	 * never be executed.
	 */
	boff = fileMetaAlloc(fp, &amp;f-&gt;dir, f-&gt;boff+1);
	if(boff == NilBlock){
		/* mbResize might have modified block */
		mbPack(&amp;mb);
		blockDirty(b);
		goto Err;
	}
fprint(2, "fileMetaFlush moving entry from %ud -&gt; %ud\n", f-&gt;boff, boff);
	f-&gt;boff = boff;

	/* make sure deletion goes to disk after new entry */
	bb = sourceBlock(fp-&gt;msource, f-&gt;boff, OReadWrite);
	mbDelete(&amp;mb, i);
	mbPack(&amp;mb);
	blockDependency(b, bb, -1, nil, nil);
	blockPut(bb);
	blockDirty(b);
	blockPut(b);
	sourceUnlock(fp-&gt;msource);

	f-&gt;dirty = 0;

	return 1;

Err:
	blockPut(b);
Err1:
	sourceUnlock(fp-&gt;msource);
	return -1;
}

static int
fileMetaRemove(File *f, char *uid)
{
	Block *b;
	MetaBlock mb;
	MetaEntry me;
	int i;
	File *up;

	up = f-&gt;up;

	fileWAccess(up, uid);

	fileMetaLock(f);

	sourceLock(up-&gt;msource, OReadWrite);
	b = sourceBlock(up-&gt;msource, f-&gt;boff, OReadWrite);
	if(b == nil)
		goto Err;

	if(!mbUnpack(&amp;mb, b-&gt;data, up-&gt;msource-&gt;dsize))
{
fprint(2, "U\n");
		goto Err;
}
	if(!mbSearch(&amp;mb, f-&gt;dir.elem, &amp;i, &amp;me))
{
fprint(2, "S\n");
		goto Err;
}
	mbDelete(&amp;mb, i);
	mbPack(&amp;mb);
	sourceUnlock(up-&gt;msource);

	blockDirty(b);
	blockPut(b);

	f-&gt;removed = 1;
	f-&gt;boff = NilBlock;
	f-&gt;dirty = 0;

	fileMetaUnlock(f);
	return 1;

Err:
	sourceUnlock(up-&gt;msource);
	blockPut(b);
	fileMetaUnlock(f);
	return 0;
}

/* assume file is locked, assume f-&gt;msource is locked */
static int
fileCheckEmpty(File *f)
{
	u32int i, n;
	Block *b;
	MetaBlock mb;
	Source *r;

	r = f-&gt;msource;
	n = (sourceGetSize(r)+r-&gt;dsize-1)/r-&gt;dsize;
	for(i=0; i&lt;n; i++){
		b = sourceBlock(r, i, OReadOnly);
		if(b == nil)
			goto Err;
		if(!mbUnpack(&amp;mb, b-&gt;data, r-&gt;dsize))
			goto Err;
		if(mb.nindex &gt; 0){
			vtSetError(ENotEmpty);
			goto Err;
		}
		blockPut(b);
	}
	return 1;
Err:
	blockPut(b);
	return 0;
}

int
fileRemove(File *f, char *uid)
{
	File *ff;

	/* can not remove the root */
	if(fileIsRoot(f)){
		vtSetError(ERoot);
		return 0;
	}

	if(!fileLock(f))
		return 0;

	if(f-&gt;source-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		goto Err1;
	}
	if(!sourceLock2(f-&gt;source, f-&gt;msource, -1))
		goto Err1;
	if(fileIsDir(f) &amp;&amp; !fileCheckEmpty(f))
		goto Err;

	for(ff=f-&gt;down; ff; ff=ff-&gt;next)
		assert(ff-&gt;removed);

	sourceRemove(f-&gt;source);
	f-&gt;source-&gt;file = nil;		/* erase back pointer */
	f-&gt;source = nil;
	if(f-&gt;msource){
		sourceRemove(f-&gt;msource);
		f-&gt;msource = nil;
	}

	fileUnlock(f);

	if(!fileMetaRemove(f, uid))
		return 0;

	return 1;

Err:
	sourceUnlock(f-&gt;source);
	if(f-&gt;msource)
		sourceUnlock(f-&gt;msource);
Err1:
	fileUnlock(f);
	return 0;
}

static int
clri(File *f, char *uid)
{
	int r;

	if(f == nil)
		return 0;
	if(f-&gt;up-&gt;source-&gt;mode != OReadWrite){
		vtSetError(EReadOnly);
		fileDecRef(f);
		return 0;
	}
	r = fileMetaRemove(f, uid);
	fileDecRef(f);
	return r;
}

int
fileClriPath(Fs *fs, char *path, char *uid)
{
	return clri(_fileOpen(fs, path, 1), uid);
}

int
fileClri(File *dir, char *elem, char *uid)
{
	return clri(_fileWalk(dir, elem, 1), uid);
}

File *
fileIncRef(File *vf)
{
	fileMetaLock(vf);
	assert(vf-&gt;ref &gt; 0);
	vf-&gt;ref++;
	fileMetaUnlock(vf);
	return vf;
}

int
fileDecRef(File *f)
{
	File *p, *q, **qq;

	if(f-&gt;up == nil){
		/* never linked in */
		assert(f-&gt;ref == 1);
		fileFree(f);
		return 1;
	}

	fileMetaLock(f);
	f-&gt;ref--;
	if(f-&gt;ref &gt; 0){
		fileMetaUnlock(f);
		return 0;
	}
	assert(f-&gt;ref == 0);
	assert(f-&gt;down == nil);

	fileMetaFlush2(f, nil);

	p = f-&gt;up;
	qq = &amp;p-&gt;down;
	for(q = *qq; q; q = *qq){
		if(q == f)
			break;
		qq = &amp;q-&gt;next;
	}
	assert(q != nil);
	*qq = f-&gt;next;

	fileMetaUnlock(f);
	fileFree(f);

	fileDecRef(p);
	return 1;
}

File *
fileGetParent(File *f)
{
	if(fileIsRoot(f))
		return fileIncRef(f);
	return fileIncRef(f-&gt;up);
}

DirEntryEnum *
deeOpen(File *f)
{
	DirEntryEnum *dee;
	File *p;

	if(!fileIsDir(f)){
		vtSetError(ENotDir);
		fileDecRef(f);
		return nil;
	}

	/* flush out meta data */
	if(!fileLock(f))
		return nil;
	for(p=f-&gt;down; p; p=p-&gt;next)
		fileMetaFlush2(p, nil);
	fileUnlock(f);

	dee = vtMemAllocZ(sizeof(DirEntryEnum));
	dee-&gt;file = fileIncRef(f);

	return dee;
}

static int
dirEntrySize(Source *s, ulong elem, ulong gen, uvlong *size)
{
	Block *b;
	ulong bn;
	Entry e;
	int epb;

	epb = s-&gt;dsize/VtEntrySize;
	bn = elem/epb;
	elem -= bn*epb;

	b = sourceBlock(s, bn, OReadOnly);
	if(b == nil)
		goto Err;
	if(!entryUnpack(&amp;e, b-&gt;data, elem))
		goto Err;

	/* hanging entries are returned as zero size */
	if(!(e.flags &amp; VtEntryActive) || e.gen != gen)
		*size = 0;
	else
		*size = e.size;
	blockPut(b);
	return 1;

Err:
	blockPut(b);
	return 0;
}

static int
deeFill(DirEntryEnum *dee)
{
	int i, n;
	Source *meta, *source;
	MetaBlock mb;
	MetaEntry me;
	File *f;
	Block *b;
	DirEntry *de;

	/* clean up first */
	for(i=dee-&gt;i; i&lt;dee-&gt;n; i++)
		deCleanup(dee-&gt;buf+i);
	vtMemFree(dee-&gt;buf);
	dee-&gt;buf = nil;
	dee-&gt;i = 0;
	dee-&gt;n = 0;

	f = dee-&gt;file;

	source = f-&gt;source;
	meta = f-&gt;msource;

	b = sourceBlock(meta, dee-&gt;boff, OReadOnly);
	if(b == nil)
		goto Err;
	if(!mbUnpack(&amp;mb, b-&gt;data, meta-&gt;dsize))
		goto Err;

	n = mb.nindex;
	dee-&gt;buf = vtMemAlloc(n * sizeof(DirEntry));

	for(i=0; i&lt;n; i++){
		de = dee-&gt;buf + i;
		meUnpack(&amp;me, &amp;mb, i);
		if(!deUnpack(de, &amp;me))
			goto Err;
		dee-&gt;n++;
		if(!(de-&gt;mode &amp; ModeDir))
		if(!dirEntrySize(source, de-&gt;entry, de-&gt;gen, &amp;de-&gt;size))
			goto Err;
	}
	dee-&gt;boff++;
	blockPut(b);
	return 1;
Err:
	blockPut(b);
	return 0;
}

int
deeRead(DirEntryEnum *dee, DirEntry *de)
{
	int ret, didread;
	File *f;
	u32int nb;

	if(dee == nil){
		vtSetError("cannot happen in deeRead");
		return -1;
	}

	f = dee-&gt;file;
	if(!fileRLock(f))
		return -1;

	if(!sourceLock2(f-&gt;source, f-&gt;msource, OReadOnly)){
		fileRUnlock(f);
		return -1;
	}

	nb = (sourceGetSize(f-&gt;msource)+f-&gt;msource-&gt;dsize-1)/f-&gt;msource-&gt;dsize;

	didread = 0;
	while(dee-&gt;i &gt;= dee-&gt;n){
		if(dee-&gt;boff &gt;= nb){
			ret = 0;
			goto Return;
		}
		didread = 1;
		if(!deeFill(dee)){
			ret = -1;
			goto Return;
		}
	}

	memmove(de, dee-&gt;buf + dee-&gt;i, sizeof(DirEntry));
	dee-&gt;i++;
	ret = 1;

Return:
	sourceUnlock(f-&gt;source);
	sourceUnlock(f-&gt;msource);
	fileRUnlock(f);

	if(didread)
		fileRAccess(f);
	return ret;
}

void
deeClose(DirEntryEnum *dee)
{
	int i;
	if(dee == nil)
		return;
	for(i=dee-&gt;i; i&lt;dee-&gt;n; i++)
		deCleanup(dee-&gt;buf+i);
	vtMemFree(dee-&gt;buf);
	fileDecRef(dee-&gt;file);
	vtMemFree(dee);
}

/*
 * caller must lock f-&gt;source and f-&gt;msource
 * caller must NOT lock the source and msource
 * referenced by dir.
 */
static u32int
fileMetaAlloc(File *f, DirEntry *dir, u32int start)
{
	u32int nb, bo;
	Block *b, *bb;
	MetaBlock mb;
	int nn;
	uchar *p;
	int i, n, epb;
	MetaEntry me;
	Source *s, *ms;

	s = f-&gt;source;
	ms = f-&gt;msource;

	n = deSize(dir);
	nb = (sourceGetSize(ms)+ms-&gt;dsize-1)/ms-&gt;dsize;
	b = nil;
	if(start &gt; nb)
		start = nb;
	for(bo=start; bo&lt;nb; bo++){
		b = sourceBlock(ms, bo, OReadWrite);
		if(b == nil)
			goto Err;
		if(!mbUnpack(&amp;mb, b-&gt;data, ms-&gt;dsize))
			goto Err;
		nn = (mb.maxsize*FullPercentage/100) - mb.size + mb.free;
		if(n &lt;= nn &amp;&amp; mb.nindex &lt; mb.maxindex)
			break;
		blockPut(b);
		b = nil;
	}

	/* add block to meta file */
	if(b == nil){
		b = sourceBlock(ms, bo, OReadWrite);
		if(b == nil)
			goto Err;
		sourceSetSize(ms, (nb+1)*ms-&gt;dsize);
		mbInit(&amp;mb, b-&gt;data, ms-&gt;dsize, ms-&gt;dsize/BytesPerEntry);
	}

	p = mbAlloc(&amp;mb, n);
	if(p == nil){
		/* mbAlloc might have changed block */
		mbPack(&amp;mb);
		blockDirty(b);
		vtSetError(EBadMeta);
		goto Err;
	}

	mbSearch(&amp;mb, dir-&gt;elem, &amp;i, &amp;me);
	assert(me.p == nil);
	me.p = p;
	me.size = n;
	dePack(dir, &amp;me);
	mbInsert(&amp;mb, i, &amp;me);
	mbPack(&amp;mb);

	/* meta block depends on super block for qid ... */
	bb = cacheLocal(b-&gt;c, PartSuper, 0, OReadOnly);
	blockDependency(b, bb, -1, nil, nil);
	blockPut(bb);

	/* ... and one or two dir entries */
	epb = s-&gt;dsize/VtEntrySize;
	bb = sourceBlock(s, dir-&gt;entry/epb, OReadOnly);
	blockDependency(b, bb, -1, nil, nil);
	blockPut(bb);
	if(dir-&gt;mode &amp; ModeDir){
		bb = sourceBlock(s, dir-&gt;mentry/epb, OReadOnly);
		blockDependency(b, bb, -1, nil, nil);
		blockPut(bb);
	}

	blockDirty(b);
	blockPut(b);
	return bo;
Err:
	blockPut(b);
	return NilBlock;
}

static int
chkSource(File *f)
{
	if(f-&gt;partial)
		return 1;

	if(f-&gt;source == nil || (f-&gt;dir.mode &amp; ModeDir) &amp;&amp; f-&gt;msource == nil){
		vtSetError(ERemoved);
		return 0;
	}
	return 1;
}

static int
fileRLock(File *f)
{
	assert(!vtCanLock(f-&gt;fs-&gt;elk));
	vtRLock(f-&gt;lk);
	if(!chkSource(f)){
		fileRUnlock(f);
		return 0;
	}
	return 1;
}

static void
fileRUnlock(File *f)
{
	vtRUnlock(f-&gt;lk);
}

static int
fileLock(File *f)
{
	assert(!vtCanLock(f-&gt;fs-&gt;elk));
	vtLock(f-&gt;lk);
	if(!chkSource(f)){
		fileUnlock(f);
		return 0;
	}
	return 1;
}

static void
fileUnlock(File *f)
{
	vtUnlock(f-&gt;lk);
}

/*
 * f-&gt;source and f-&gt;msource must NOT be locked.
 * fileMetaFlush locks the fileMeta and then the source (in fileMetaFlush2).
 * We have to respect that ordering.
 */
static void
fileMetaLock(File *f)
{
if(f-&gt;up == nil)
fprint(2, "f-&gt;elem = %s\n", f-&gt;dir.elem);
	assert(f-&gt;up != nil);
	assert(!vtCanLock(f-&gt;fs-&gt;elk));
	vtLock(f-&gt;up-&gt;lk);
}

static void
fileMetaUnlock(File *f)
{
	vtUnlock(f-&gt;up-&gt;lk);
}

/*
 * f-&gt;source and f-&gt;msource must NOT be locked.
 * see fileMetaLock.
 */
static void
fileRAccess(File* f)
{
	if(f-&gt;mode == OReadOnly || f-&gt;fs-&gt;noatimeupd)
		return;

	fileMetaLock(f);
	f-&gt;dir.atime = time(0L);
	f-&gt;dirty = 1;
	fileMetaUnlock(f);
}

/*
 * f-&gt;source and f-&gt;msource must NOT be locked.
 * see fileMetaLock.
 */
static void
fileWAccess(File* f, char *mid)
{
	if(f-&gt;mode == OReadOnly)
		return;

	fileMetaLock(f);
	f-&gt;dir.atime = f-&gt;dir.mtime = time(0L);
	if(strcmp(f-&gt;dir.mid, mid) != 0){
		vtMemFree(f-&gt;dir.mid);
		f-&gt;dir.mid = vtStrDup(mid);
	}
	f-&gt;dir.mcount++;
	f-&gt;dirty = 1;
	fileMetaUnlock(f);

/*RSC: let's try this */
/*presotto - lets not
	if(f-&gt;up)
		fileWAccess(f-&gt;up, mid);
*/
}

static int
getEntry(Source *r, Entry *e, int checkepoch)
{
	u32int epoch;
	Block *b;

	if(r == nil){
		memset(&amp;e, 0, sizeof e);
		return 1;
	}

	b = cacheGlobal(r-&gt;fs-&gt;cache, r-&gt;score, BtDir, r-&gt;tag, OReadOnly);
	if(b == nil)
		return 0;
	if(!entryUnpack(e, b-&gt;data, r-&gt;offset % r-&gt;epb)){
		blockPut(b);
		return 0;
	}
	epoch = b-&gt;l.epoch;
	blockPut(b);

	if(checkepoch){
		b = cacheGlobal(r-&gt;fs-&gt;cache, e-&gt;score, entryType(e), e-&gt;tag, OReadOnly);
		if(b){
			if(b-&gt;l.epoch &gt;= epoch)
				fprint(2, "warning: entry %p epoch not older %#.8ux/%d %V/%d in getEntry\n",
					r, b-&gt;addr, b-&gt;l.epoch, r-&gt;score, epoch);
			blockPut(b);
		}
	}

	return 1;
}

static int
setEntry(Source *r, Entry *e)
{
	Block *b;
	Entry oe;

	b = cacheGlobal(r-&gt;fs-&gt;cache, r-&gt;score, BtDir, r-&gt;tag, OReadWrite);
	if(0) fprint(2, "setEntry: b %#ux %d score=%V\n", b-&gt;addr, r-&gt;offset % r-&gt;epb, e-&gt;score);
	if(b == nil)
		return 0;
	if(!entryUnpack(&amp;oe, b-&gt;data, r-&gt;offset % r-&gt;epb)){
		blockPut(b);
		return 0;
	}
	e-&gt;gen = oe.gen;
	entryPack(e, b-&gt;data, r-&gt;offset % r-&gt;epb);

	/* BUG b should depend on the entry pointer */

	blockDirty(b);
	blockPut(b);
	return 1;
}

/* assumes hold elk */
int
fileSnapshot(File *dst, File *src, u32int epoch, int doarchive)
{
	Entry e, ee;

	/* add link to snapshot */
	if(!getEntry(src-&gt;source, &amp;e, 1) || !getEntry(src-&gt;msource, &amp;ee, 1))
		return 0;

	e.snap = epoch;
	e.archive = doarchive;
	ee.snap = epoch;
	ee.archive = doarchive;

	if(!setEntry(dst-&gt;source, &amp;e) || !setEntry(dst-&gt;msource, &amp;ee))
		return 0;
	return 1;
}

int
fileGetSources(File *f, Entry *e, Entry *ee)
{
	if(!getEntry(f-&gt;source, e, 0)
	|| !getEntry(f-&gt;msource, ee, 0))
		return 0;
	return 1;
}

/*
 * Walk down to the block(s) containing the Entries
 * for f-&gt;source and f-&gt;msource, copying as we go.
 */
int
fileWalkSources(File *f)
{
	if(f-&gt;mode == OReadOnly){
		fprint(2, "readonly in fileWalkSources\n");
		return 1;
	}
	if(!sourceLock2(f-&gt;source, f-&gt;msource, OReadWrite)){
		fprint(2, "sourceLock2 failed in fileWalkSources\n");
		return 0;
	}
	sourceUnlock(f-&gt;source);
	sourceUnlock(f-&gt;msource);
	return 1;
}

/*
 * convert File* to full path name in malloced string.
 * this hasn't been as useful as we hoped it would be.
 */
char *
fileName(File *f)
{
	char *name, *pname;
	File *p;
	static char root[] = "/";

	if (f == nil)
		return vtStrDup("/**GOK**");

	p = fileGetParent(f);
	if (p == f)
		name = vtStrDup(root);
	else {
		pname = fileName(p);
		if (strcmp(pname, root) == 0)
			name = smprint("/%s", f-&gt;dir.elem);
		else
			name = smprint("%s/%s", pname, f-&gt;dir.elem);
		free(pname);
	}
	fileDecRef(p);
	return name;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

