<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/ip/imap4d/mbox.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/ip/imap4d/mbox.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;bio.h&gt;
#include &lt;auth.h&gt;
#include "imap4d.h"

static NamedInt	flagChars[NFlags] =
{
	{"s",	MSeen},
	{"a",	MAnswered},
	{"f",	MFlagged},
	{"D",	MDeleted},
	{"d",	MDraft},
	{"r",	MRecent},
};

static	int	fsCtl = -1;

static	void	boxFlags(Box *box);
static	int	createImp(Box *box, Qid *qid);
static	void	fsInit(void);
static	void	mboxGone(Box *box);
static	MbLock	*openImp(Box *box, int new);
static	int	parseImp(Biobuf *b, Box *box);
static	int	readBox(Box *box);
static	ulong	uidRenumber(Msg *m, ulong uid, int force);
static	int	impFlags(Box *box, Msg *m, char *flags);

/*
 * strategy:
 * every mailbox file has an associated .imp file
 * which maps upas/fs message digests to uids &amp; message flags.
 *
 * the .imp files are locked by /mail/fs/usename/L.mbox.
 * whenever the flags can be modified, the lock file
 * should be opened, thereby locking the uid &amp; flag state.
 * for example, whenever new uids are assigned to messages,
 * and whenever flags are changed internally, the lock file
 * should be open and locked.  this means the file must be
 * opened during store command, and when changing the \seen
 * flag for the fetch command.
 *
 * if no .imp file exists, a null one must be created before
 * assigning uids.
 *
 * the .imp file has the following format
 * imp		: "imap internal mailbox description\n"
 * 			uidvalidity " " uidnext "\n"
 *			messageLines
 *
 * messageLines	:
 *		| messageLines digest " " uid " " flags "\n"
 *
 * uid, uidnext, and uidvalidity are 32 bit decimal numbers
 * printed right justified in a field NUid characters long.
 * the 0 uid implies that no uid has been assigned to the message,
 * but the flags are valid. note that message lines are in mailbox
 * order, except possibly for 0 uid messages.
 *
 * digest is an ascii hex string NDigest characters long.
 *
 * flags has a character for each of NFlag flag fields.
 * if the flag is clear, it is represented by a "-".
 * set flags are represented as a unique single ascii character.
 * the currently assigned flags are, in order:
 *	MSeen		s
 *	MAnswered	a
 *	MFlagged	f
 *	MDeleted	D
 *	MDraft		d
 */
Box*
openBox(char *name, char *fsname, int writable)
{
	Box *box;
	MbLock *ml;
	int n, new;

	if(cistrcmp(name, "inbox") == 0)
		if(access("msgs", AEXIST) == 0)
			name = "msgs";
		else
			name = "mbox";
	fsInit();
	debuglog("imap4d open %s %s\n", name, fsname);

	if(fprint(fsCtl, "open '/mail/box/%s/%s' %s", username, name, fsname) &lt; 0){
//ZZZ
		char err[ERRMAX];

		rerrstr(err, sizeof err);
		if(strstr(err, "file does not exist") == nil)
			fprint(2,
		"imap4d at %lud: upas/fs open %s/%s as %s failed: '%s' %s",
			time(nil), username, name, fsname, err,
			ctime(time(nil)));  /* NB: ctime result ends with \n */
		fprint(fsCtl, "close %s", fsname);
		return nil;
	}

	/*
	 * read box to find all messages
	 * each one has a directory, and is in numerical order
	 */
	box = MKZ(Box);
	box-&gt;writable = writable;

	n = strlen(name) + 1;
	box-&gt;name = emalloc(n);
	strcpy(box-&gt;name, name);

	n += STRLEN(".imp");
	box-&gt;imp = emalloc(n);
	snprint(box-&gt;imp, n, "%s.imp", name);

	n = strlen(fsname) + 1;
	box-&gt;fs = emalloc(n);
	strcpy(box-&gt;fs, fsname);

	n = STRLEN("/mail/fs/") + strlen(fsname) + 1;
	box-&gt;fsDir = emalloc(n);
	snprint(box-&gt;fsDir, n, "/mail/fs/%s", fsname);

	box-&gt;uidnext = 1;
	new = readBox(box);
	if(new &gt;= 0){
		ml = openImp(box, new);
		if(ml != nil){
			closeImp(box, ml);
			return box;
		}
	}
	closeBox(box, 0);
	return nil;
}

/*
 * check mailbox
 * returns fd of open .imp file if imped.
 * otherwise, return value is insignificant
 *
 * careful: called by idle polling proc
 */
MbLock*
checkBox(Box *box, int imped)
{
	MbLock *ml;
	Dir *d;
	int new;

	if(box == nil)
		return nil;

	/*
	 * if stat fails, mailbox must be gone
	 */
	d = cdDirstat(box-&gt;fsDir, ".");
	if(d == nil){
		mboxGone(box);
		return nil;
	}
	new = 0;
	if(box-&gt;qid.path != d-&gt;qid.path || box-&gt;qid.vers != d-&gt;qid.vers
	|| box-&gt;mtime != d-&gt;mtime){
		new = readBox(box);
		if(new &lt; 0){
			free(d);
			return nil;
		}
	}
	free(d);
	ml = openImp(box, new);
	if(ml == nil)
		box-&gt;writable = 0;
	else if(!imped){
		closeImp(box, ml);
		ml = nil;
	}
	return ml;
}

/*
 * mailbox is unreachable, so mark all messages expunged
 * clean up .imp files as well.
 */
static void
mboxGone(Box *box)
{
	Msg *m;

	if(cdExists(mboxDir, box-&gt;name) &lt; 0)
		cdRemove(mboxDir, box-&gt;imp);
	for(m = box-&gt;msgs; m != nil; m = m-&gt;next)
		m-&gt;expunged = 1;
	box-&gt;writable = 0;
}

/*
 * read messages in the mailbox
 * mark message that no longer exist as expunged
 * returns -1 for failure, 0 if no new messages, 1 if new messages.
 */
static int
readBox(Box *box)
{
	Msg *msgs, *m, *last;
	Dir *d;
	char *s;
	long max, id;
	int i, nd, fd, new;

	fd = cdOpen(box-&gt;fsDir, ".", OREAD);
	if(fd &lt; 0){
		syslog(0, "mail",
		    "imap4d at %lud: upas/fs stat of %s/%s aka %s failed: %r",
			time(nil), username, box-&gt;name, box-&gt;fsDir);
		mboxGone(box);
		return -1;
	}

	/*
	 * read box to find all messages
	 * each one has a directory, and is in numerical order
	 */
	d = dirfstat(fd);
	if(d == nil){
		close(fd);
		return -1;
	}
	box-&gt;mtime = d-&gt;mtime;
	box-&gt;qid = d-&gt;qid;
	last = nil;
	msgs = box-&gt;msgs;
	max = 0;
	new = 0;
	free(d);
	while((nd = dirread(fd, &amp;d)) &gt; 0){
		for(i = 0; i &lt; nd; i++){
			s = d[i].name;
			id = strtol(s, &amp;s, 10);
			if(id &lt;= max || *s != '\0'
			|| (d[i].mode &amp; DMDIR) != DMDIR)
				continue;

			max = id;

			while(msgs != nil){
				last = msgs;
				msgs = msgs-&gt;next;
				if(last-&gt;id == id)
					goto continueDir;
				last-&gt;expunged = 1;
			}

			new = 1;
			m = MKZ(Msg);
			m-&gt;id = id;
			m-&gt;fsDir = box-&gt;fsDir;
			m-&gt;fs = emalloc(2 * (MsgNameLen + 1));
			m-&gt;efs = seprint(m-&gt;fs, m-&gt;fs + (MsgNameLen + 1), "%lud/", id);
			m-&gt;size = ~0UL;
			m-&gt;lines = ~0UL;
			m-&gt;prev = last;
			m-&gt;flags = MRecent;
			if(!msgInfo(m))
				freeMsg(m);
			else{
				if(last == nil)
					box-&gt;msgs = m;
				else
					last-&gt;next = m;
				last = m;
			}
	continueDir:;
		}
		free(d);
	}
	close(fd);
	for(; msgs != nil; msgs = msgs-&gt;next)
		msgs-&gt;expunged = 1;

	/*
	 * make up the imap message sequence numbers
	 */
	id = 1;
	for(m = box-&gt;msgs; m != nil; m = m-&gt;next){
		if(m-&gt;seq &amp;&amp; m-&gt;seq != id)
			bye("internal error assigning message numbers");
		m-&gt;seq = id++;
	}
	box-&gt;max = id - 1;

	return new;
}

/*
 * read in the .imp file, or make one if it doesn't exist.
 * make sure all flags and uids are consistent.
 * return the mailbox lock.
 */
#define IMPMAGIC	"imap internal mailbox description\n"
static MbLock*
openImp(Box *box, int new)
{
	Qid qid;
	Biobuf b;
	MbLock *ml;
	int fd;
//ZZZZ
	int once;

	ml = mbLock();
	if(ml == nil)
		return nil;
	fd = cdOpen(mboxDir, box-&gt;imp, OREAD);
	once = 0;
ZZZhack:
	if(fd &lt; 0 || fqid(fd, &amp;qid) &lt; 0){
		if(fd &lt; 0){
			char buf[ERRMAX];

			errstr(buf, sizeof buf);
			if(cistrstr(buf, "does not exist") == nil)
				fprint(2, "imap4d at %lud: imp open failed: %s\n", time(nil), buf);
			if(!once &amp;&amp; cistrstr(buf, "locked") != nil){
				once = 1;
				fprint(2, "imap4d at %lud: imp %s/%s %s locked when it shouldn't be; spinning\n", time(nil), username, box-&gt;name, box-&gt;imp);
				fd = openLocked(mboxDir, box-&gt;imp, OREAD);
				goto ZZZhack;
			}
		}
		if(fd &gt;= 0)
			close(fd);
		fd = createImp(box, &amp;qid);
		if(fd &lt; 0){
			mbUnlock(ml);
			return nil;
		}
		box-&gt;dirtyImp = 1;
		if(box-&gt;uidvalidity == 0)
			box-&gt;uidvalidity = box-&gt;mtime;
		box-&gt;impQid = qid;
		new = 1;
	}else if(qid.path != box-&gt;impQid.path || qid.vers != box-&gt;impQid.vers){
		Binit(&amp;b, fd, OREAD);
		if(!parseImp(&amp;b, box)){
			box-&gt;dirtyImp = 1;
			if(box-&gt;uidvalidity == 0)
				box-&gt;uidvalidity = box-&gt;mtime;
		}
		Bterm(&amp;b);
		box-&gt;impQid = qid;
		new = 1;
	}
	if(new)
		boxFlags(box);
	close(fd);
	return ml;
}

/*
 * close the .imp file, after writing out any changes
 */
void
closeImp(Box *box, MbLock *ml)
{
	Msg *m;
	Qid qid;
	Biobuf b;
	char buf[NFlags+1];
	int fd;

	if(ml == nil)
		return;
	if(!box-&gt;dirtyImp){
		mbUnlock(ml);
		return;
	}

	fd = cdCreate(mboxDir, box-&gt;imp, OWRITE, 0664);
	if(fd &lt; 0){
		mbUnlock(ml);
		return;
	}
	Binit(&amp;b, fd, OWRITE);

	box-&gt;dirtyImp = 0;
	Bprint(&amp;b, "%s", IMPMAGIC);
	Bprint(&amp;b, "%.*lud %.*lud\n", NUid, box-&gt;uidvalidity, NUid, box-&gt;uidnext);
	for(m = box-&gt;msgs; m != nil; m = m-&gt;next){
		if(m-&gt;expunged)
			continue;
		wrImpFlags(buf, m-&gt;flags, strcmp(box-&gt;fs, "imap") == 0);
		Bprint(&amp;b, "%.*s %.*lud %s\n", NDigest, m-&gt;info[IDigest], NUid, m-&gt;uid, buf);
	}
	Bterm(&amp;b);

	if(fqid(fd, &amp;qid) &gt;= 0)
		box-&gt;impQid = qid;
	close(fd);
	mbUnlock(ml);
}

void
wrImpFlags(char *buf, int flags, int killRecent)
{
	int i;

	for(i = 0; i &lt; NFlags; i++){
		if((flags &amp; flagChars[i].v)
		&amp;&amp; (flagChars[i].v != MRecent || !killRecent))
			buf[i] = flagChars[i].name[0];
		else
			buf[i] = '-';
	}
	buf[i] = '\0';
}

int
emptyImp(char *mbox)
{
	Dir *d;
	long mode;
	int fd;

	fd = cdCreate(mboxDir, impName(mbox), OWRITE, 0664);
	if(fd &lt; 0)
		return -1;
	d = cdDirstat(mboxDir, mbox);
	if(d == nil){
		close(fd);
		return -1;
	}
	fprint(fd, "%s%.*lud %.*lud\n", IMPMAGIC, NUid, d-&gt;mtime, NUid, 1UL);
	mode = d-&gt;mode &amp; 0777;
	nulldir(d);
	d-&gt;mode = mode;
	dirfwstat(fd, d);
	free(d);
	return fd;
}

/*
 * try to match permissions with mbox
 */
static int
createImp(Box *box, Qid *qid)
{
	Dir *d;
	long mode;
	int fd;

	fd = cdCreate(mboxDir, box-&gt;imp, OREAD, 0664);
	if(fd &lt; 0)
		return -1;
	d = cdDirstat(mboxDir, box-&gt;name);
	if(d != nil){
		mode = d-&gt;mode &amp; 0777;
		nulldir(d);
		d-&gt;mode = mode;
		dirfwstat(fd, d);
		free(d);
	}
	if(fqid(fd, qid) &lt; 0){
		close(fd);
		return -1;
	}

	return fd;
}

/*
 * read or re-read a .imp file.
 * this is tricky:
 *	messages can be deleted by another agent
 *	we might still have a Msg for an expunged message,
 *		because we haven't told the client yet.
 *	we can have a Msg without a .imp entry.
 *	flag information is added at the end of the .imp by copy &amp; append
 *	there can be duplicate messages (same digests).
 *
 * look up existing messages based on uid.
 * look up new messages based on in order digest matching.
 *
 * note: in the face of duplicate messages, one of which is deleted,
 * two active servers may decide different ones are valid, and so return
 * different uids for the messages.  this situation will stablize when the servers exit.
 */
static int
parseImp(Biobuf *b, Box *box)
{
	Msg *m, *mm;
	char *s, *t, *toks[3];
	ulong uid, u;
	int match, n;

	m = box-&gt;msgs;
	s = Brdline(b, '\n');
	if(s == nil || Blinelen(b) != STRLEN(IMPMAGIC)
	|| strncmp(s, IMPMAGIC, STRLEN(IMPMAGIC)) != 0)
		return 0;

	s = Brdline(b, '\n');
	if(s == nil || Blinelen(b) != 2*NUid + 2)
		return 0;
	s[2*NUid + 1] = '\0';
	u = strtoul(s, &amp;t, 10);
	if(u != box-&gt;uidvalidity &amp;&amp; box-&gt;uidvalidity != 0)
		return 0;
	box-&gt;uidvalidity = u;
	if(*t != ' ' || t != s + NUid)
		return 0;
	t++;
	u = strtoul(t, &amp;t, 10);
	if(box-&gt;uidnext &gt; u)
		return 0;
	box-&gt;uidnext = u;
	if(t != s + 2*NUid+1 || box-&gt;uidnext == 0)
		return 0;

	uid = ~0;
	while(m != nil){
		s = Brdline(b, '\n');
		if(s == nil)
			break;
		n = Blinelen(b) - 1;
		if(n != NDigest + NUid + NFlags + 2
		|| s[NDigest] != ' ' || s[NDigest + NUid + 1] != ' ')
			return 0;
		toks[0] = s;
		s[NDigest] = '\0';
		toks[1] = s + NDigest + 1;
		s[NDigest + NUid + 1] = '\0';
		toks[2] = s + NDigest + NUid + 2;
		s[n] = '\0';
		t = toks[1];
		u = strtoul(t, &amp;t, 10);
		if(*t != '\0' || uid != ~0 &amp;&amp; (uid &gt;= u &amp;&amp; u || u &amp;&amp; !uid))
			return 0;
		uid = u;

		/*
		 * zero uid =&gt; added by append or copy, only flags valid
		 * can only match messages without uids, but this message
		 * may not be the next one, and may have been deleted.
		 */
		if(!uid){
			for(; m != nil &amp;&amp; m-&gt;uid; m = m-&gt;next)
				;
			for(mm = m; mm != nil; mm = mm-&gt;next){
				if(mm-&gt;info[IDigest] != nil &amp;&amp;
				    strcmp(mm-&gt;info[IDigest], toks[0]) == 0){
					if(!mm-&gt;uid)
						mm-&gt;flags = 0;
					if(!impFlags(box, mm, toks[2]))
						return 0;
					m = mm-&gt;next;
					break;
				}
			}
			continue;
		}

		/*
		 * ignore expunged messages,
		 * and messages already assigned uids which don't match this uid.
		 * such messages must have been deleted by another imap server,
		 * which updated the mailbox and .imp file since we read the mailbox,
		 * or because upas/fs got confused by consecutive duplicate messages,
		 * the first of which was deleted by another imap server.
		 */
		for(; m != nil &amp;&amp; (m-&gt;expunged || m-&gt;uid &amp;&amp; m-&gt;uid &lt; uid); m = m-&gt;next)
			;
		if(m == nil)
			break;

		/*
		 * only check for digest match on the next message,
		 * since it comes before all other messages, and therefore
		 * must be in the .imp file if they should be.
		 */
		match = m-&gt;info[IDigest] != nil &amp;&amp;
			strcmp(m-&gt;info[IDigest], toks[0]) == 0;
		if(uid &amp;&amp; (m-&gt;uid == uid || !m-&gt;uid &amp;&amp; match)){
			if(!match)
				bye("inconsistent uid");

			/*
			 * wipe out recent flag if some other server saw this new message.
			 * it will be read from the .imp file if is really should be set,
			 * ie the message was only seen by a status command.
			 */
			if(!m-&gt;uid)
				m-&gt;flags = 0;

			if(!impFlags(box, m, toks[2]))
				return 0;
			m-&gt;uid = uid;
			m = m-&gt;next;
		}
	}
	return 1;
}

/*
 * parse .imp flags
 */
static int
impFlags(Box *box, Msg *m, char *flags)
{
	int i, f;

	f = 0;
	for(i = 0; i &lt; NFlags; i++){
		if(flags[i] == '-')
			continue;
		if(flags[i] != flagChars[i].name[0])
			return 0;
		f |= flagChars[i].v;
	}

	/*
	 * recent flags are set until the first time message's box is selected or examined.
	 * it may be stored in the file as a side effect of a status or subscribe command;
	 * if so, clear it out.
	 */
	if((f &amp; MRecent) &amp;&amp; strcmp(box-&gt;fs, "imap") == 0)
		box-&gt;dirtyImp = 1;
	f |= m-&gt;flags &amp; MRecent;

	/*
	 * all old messages with changed flags should be reported to the client
	 */
	if(m-&gt;uid &amp;&amp; m-&gt;flags != f){
		box-&gt;sendFlags = 1;
		m-&gt;sendFlags = 1;
	}
	m-&gt;flags = f;
	return 1;
}

/*
 * assign uids to any new messages
 * which aren't already in the .imp file.
 * sum up totals for flag values.
 */
static void
boxFlags(Box *box)
{
	Msg *m;

	box-&gt;recent = 0;
	for(m = box-&gt;msgs; m != nil; m = m-&gt;next){
		if(m-&gt;uid == 0){
			box-&gt;dirtyImp = 1;
			box-&gt;uidnext = uidRenumber(m, box-&gt;uidnext, 0);
		}
		if(m-&gt;flags &amp; MRecent)
			box-&gt;recent++;
	}
}

static ulong
uidRenumber(Msg *m, ulong uid, int force)
{
	for(; m != nil; m = m-&gt;next){
		if(!force &amp;&amp; m-&gt;uid != 0)
			bye("uid renumbering with a valid uid");
		m-&gt;uid = uid++;
	}
	return uid;
}

void
closeBox(Box *box, int opened)
{
	Msg *m, *next;

	/*
	 * make sure to leave the mailbox directory so upas/fs can close the mailbox
	 */
	myChdir(mboxDir);

	if(box-&gt;writable){
		deleteMsgs(box);
		if(expungeMsgs(box, 0))
			closeImp(box, checkBox(box, 1));
	}

	if(fprint(fsCtl, "close %s", box-&gt;fs) &lt; 0 &amp;&amp; opened)
		bye("can't talk to mail server");
	for(m = box-&gt;msgs; m != nil; m = next){
		next = m-&gt;next;
		freeMsg(m);
	}
	free(box-&gt;name);
	free(box-&gt;fs);
	free(box-&gt;fsDir);
	free(box-&gt;imp);
	free(box);
}

int
deleteMsgs(Box *box)
{
	Msg *m;
	char buf[BufSize], *p, *start;
	int ok;

	if(!box-&gt;writable)
		return 0;

	/*
	 * first pass: delete messages; gang the writes together for speed.
	 */
	ok = 1;
	start = seprint(buf, buf + sizeof(buf), "delete %s", box-&gt;fs);
	p = start;
	for(m = box-&gt;msgs; m != nil; m = m-&gt;next){
		if((m-&gt;flags &amp; MDeleted) &amp;&amp; !m-&gt;expunged){
			m-&gt;expunged = 1;
			p = seprint(p, buf + sizeof(buf), " %lud", m-&gt;id);
			if(p + 32 &gt;= buf + sizeof(buf)){
				if(write(fsCtl, buf, p - buf) &lt; 0)
					bye("can't talk to mail server");
				p = start;
			}
		}
	}
	if(p != start &amp;&amp; write(fsCtl, buf, p - buf) &lt; 0)
		bye("can't talk to mail server");

	return ok;
}

/*
 * second pass: remove the message structure,
 * and renumber message sequence numbers.
 * update messages counts in mailbox.
 * returns true if anything changed.
 */
int
expungeMsgs(Box *box, int send)
{
	Msg *m, *next, *last;
	ulong n;

	n = 0;
	last = nil;
	for(m = box-&gt;msgs; m != nil; m = next){
		m-&gt;seq -= n;
		next = m-&gt;next;
		if(m-&gt;expunged){
			if(send)
				Bprint(&amp;bout, "* %lud expunge\r\n", m-&gt;seq);
			if(m-&gt;flags &amp; MRecent)
				box-&gt;recent--;
			n++;
			if(last == nil)
				box-&gt;msgs = next;
			else
				last-&gt;next = next;
			freeMsg(m);
		}else
			last = m;
	}
	if(n){
		box-&gt;max -= n;
		box-&gt;dirtyImp = 1;
	}
	return n;
}

static void
fsInit(void)
{
	if(fsCtl &gt;= 0)
		return;
	fsCtl = open("/mail/fs/ctl", ORDWR);
	if(fsCtl &lt; 0)
		bye("can't open mail file system");
	if(fprint(fsCtl, "close mbox") &lt; 0)
		bye("can't initialize mail file system");
}

static char *stoplist[] =
{
	"mbox",
	"pipeto",
	"forward",
	"names",
	"pipefrom",
	"headers",
	"imap.ok",
	0
};

enum {
	Maxokbytes	= 4096,
	Maxfolders	= Maxokbytes / 4,
};

static char *folders[Maxfolders];
static char *folderbuff;

static void
readokfolders(void)
{
	int fd, nr;

	fd = open("imap.ok", OREAD);
	if(fd &lt; 0)
		return;
	folderbuff = malloc(Maxokbytes);
	if(folderbuff == nil) {
		close(fd);
		return;
	}
	nr = read(fd, folderbuff, Maxokbytes-1);	/* once is ok */
	close(fd);
	if(nr &lt; 0){
		free(folderbuff);
		folderbuff = nil;
		return;
	}
	folderbuff[nr] = 0;
	tokenize(folderbuff, folders, nelem(folders));
}

/*
 * reject bad mailboxes based on mailbox name
 */
int
okMbox(char *path)
{
	char *name;
	int i;

	if(folderbuff == nil &amp;&amp; access("imap.ok", AREAD) == 0)
		readokfolders();
	name = strrchr(path, '/');
	if(name == nil)
		name = path;
	else
		name++;
	if(folderbuff != nil){
		for(i = 0; i &lt; nelem(folders) &amp;&amp; folders[i] != nil; i++)
			if(cistrcmp(folders[i], name) == 0)
				return 1;
		return 0;
	}
	if(strlen(name) + STRLEN(".imp") &gt;= MboxNameLen)
		return 0;
	for(i = 0; stoplist[i]; i++)
		if(strcmp(name, stoplist[i]) == 0)
			return 0;
	if(isprefix("L.", name) || isprefix("imap-tmp.", name)
	|| issuffix(".imp", name)
	|| strcmp("imap.subscribed", name) == 0
	|| isdotdot(name) || name[0] == '/')
		return 0;
	return 1;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

