<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/map/libmap/tetra.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/map/libmap/tetra.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include "map.h"

/*
 *	conformal map of earth onto tetrahedron
 *	the stages of mapping are
 *	(a) stereo projection of tetrahedral face onto
 *	    isosceles curvilinear triangle with 3 120-degree
 *	    angles and one straight side
 *	(b) map of this triangle onto half plane cut along
 *	    3 rays from the roots of unity to infinity
 *		formula (z^4+2*3^.5*z^2-1)/(z^4-2*3^.5*z^2-1)
 *	(c) do 3 times for  each sector of plane:
 *	    map of |arg z|&lt;=pi/6, cut along z&gt;1 into
 *	    triangle |arg z|&lt;=pi/6, Re z&lt;=const,
 *	    with upper side of cut going into upper half of
 *	    of vertical side of triangle and lowere into lower
 *		formula int from 0 to z dz/sqrt(1-z^3)
 *
 *	int from u to 1 3^.25*du/sqrt(1-u^3) =
		F(acos((rt3-1+u)/(rt3+1-u)),sqrt(1/2+rt3/4))
 *	int from 1 to u 3^.25*du/sqrt(u^3-1) =
 *		F(acos((rt3+1-u)/(rt3-1+u)),sqrt(1/2-rt3/4))
 *	this latter formula extends analytically down to
 *	u=0 and is the basis of this routine, with the
 *	argument of complex elliptic integral elco2
 *	being tan(acos...)
 *	the formula F(pi-x,k) = 2*F(pi/2,k)-F(x,k) is
 *	used to cross over into the region where Re(acos...)&gt;pi/2
 *		f0 and fpi are suitably scaled complete integrals
*/

#define TFUZZ 0.00001

static struct place tpole[4];	/* point of tangency of tetrahedron face*/
static double tpoleinit[4][2] = {
	1.,	0.,
	1.,	180.,
	-1.,	90.,
	-1.,	-90.
};
static struct tproj {
	double tlat,tlon;	/* center of stereo projection*/
	double ttwist;		/* rotatn before stereo*/
	double trot;		/*rotate after projection*/
	struct place projpl;	/*same as tlat,tlon*/
	struct coord projtw;	/*same as ttwist*/
	struct coord postrot;	/*same as trot*/
} tproj[4][4] = {
{/*00*/	{0.},
 /*01*/	{90.,	0.,	90.,	-90.},
 /*02*/	{0.,	45.,	-45.,	150.},
 /*03*/	{0.,	-45.,	-135.,	30.}
},
{/*10*/	{90.,	0.,	-90.,	90.},
 /*11*/ {0.},
 /*12*/ {0.,	135.,	-135.,	-150.},
 /*13*/	{0.,	-135.,	-45.,	-30.}
},
{/*20*/	{0.,	45.,	135.,	-30.},
 /*21*/	{0.,	135.,	45.,	-150.},
 /*22*/	{0.},
 /*23*/	{-90.,	0.,	180.,	90.}
},
{/*30*/	{0.,	-45.,	45.,	-150.},
 /*31*/ {0.,	-135.,	135.,	-30.},
 /*32*/	{-90.,	0.,	0.,	90.},
 /*33*/ {0.} 
}};
static double tx[4] = {	/*where to move facet after final rotation*/
	0.,	0.,	-1.,	1.	/*-1,1 to be sqrt(3)*/
};
static double ty[4] = {
	0.,	2.,	-1.,	-1.
};
static double root3;
static double rt3inv;
static double two_rt3;
static double tkc,tk,tcon;
static double f0r,f0i,fpir,fpii;

static void
twhichp(struct place *g, int *p, int *q)
{
	int i,j,k;
	double cosdist[4];
	struct place *tp;
	for(i=0;i&lt;4;i++) {
		tp = &amp;tpole[i];
		cosdist[i] = g-&gt;nlat.s*tp-&gt;nlat.s +
			  g-&gt;nlat.c*tp-&gt;nlat.c*(
			  g-&gt;wlon.s*tp-&gt;wlon.s +
			  g-&gt;wlon.c*tp-&gt;wlon.c);
	}
	j = 0;
	for(i=1;i&lt;4;i++)
		if(cosdist[i] &gt; cosdist[j])
			j = i;
	*p = j;
	k = j==0?1:0;
	for(i=0;i&lt;4;i++)
		if(i!=j&amp;&amp;cosdist[i]&gt;cosdist[k])
			k = i;
	*q = k;
}

int
Xtetra(struct place *place, double *x, double *y)
{
	int i,j;
	struct place pl;
	register struct tproj *tpp;
	double vr, vi;
	double br, bi;
	double zr,zi,z2r,z2i,z4r,z4i,sr,si,tr,ti;
	twhichp(place,&amp;i,&amp;j);
	copyplace(place,&amp;pl);
	norm(&amp;pl,&amp;tproj[i][j].projpl,&amp;tproj[i][j].projtw);
	Xstereographic(&amp;pl,&amp;vr,&amp;vi);
	zr = vr/2;
	zi = vi/2;
	if(zr&lt;=TFUZZ)
		zr = TFUZZ;
	csq(zr,zi,&amp;z2r,&amp;z2i);
	csq(z2r,z2i,&amp;z4r,&amp;z4i);
	z2r *= two_rt3;
	z2i *= two_rt3;
	cdiv(z4r+z2r-1,z4i+z2i,z4r-z2r-1,z4i-z2i,&amp;sr,&amp;si);
	csqrt(sr-1,si,&amp;tr,&amp;ti);
	cdiv(tcon*tr,tcon*ti,root3+1-sr,-si,&amp;br,&amp;bi);
	if(br&lt;0) {
		br = -br;
		bi = -bi;
		if(!elco2(br,bi,tk,1.,1.,&amp;vr,&amp;vi))
			return 0;
		vr = fpir - vr;
		vi = fpii - vi;
	} else 
		if(!elco2(br,bi,tk,1.,1.,&amp;vr,&amp;vi))
			return 0;
	if(si&gt;=0) {
		tr = f0r - vi;
		ti = f0i + vr;
	} else {
		tr = f0r + vi;
		ti = f0i - vr;
	}
	tpp = &amp;tproj[i][j];
	*x = tr*tpp-&gt;postrot.c +
	     ti*tpp-&gt;postrot.s + tx[i];
	*y = ti*tpp-&gt;postrot.c -
	     tr*tpp-&gt;postrot.s + ty[i];
	return(1);
}

int
tetracut(struct place *g, struct place *og, double *cutlon)
{
	int i,j,k;
	if((g-&gt;nlat.s&lt;=-rt3inv&amp;&amp;og-&gt;nlat.s&lt;=-rt3inv) &amp;&amp; 
	   (ckcut(g,og,*cutlon=0.)==2||ckcut(g,og,*cutlon=PI)==2))
		return(2);
	twhichp(g,&amp;i,&amp;k);
	twhichp(og,&amp;j,&amp;k);
	if(i==j||i==0||j==0)
		return(1);
	return(0);
}

proj
tetra(void)
{
	register i;
	int j;
	register struct place *tp;
	register struct tproj *tpp;
	double t;
	root3 = sqrt(3.);
	rt3inv = 1/root3;
	two_rt3 = 2*root3;
	tkc = sqrt(.5-.25*root3);
	tk = sqrt(.5+.25*root3);
	tcon = 2*sqrt(root3);
	elco2(tcon/(root3-1),0.,tkc,1.,1.,&amp;f0r,&amp;f0i);
	elco2(1.e15,0.,tk,1.,1.,&amp;fpir,&amp;fpii);
	fpir *= 2;
	fpii *= 2;
	for(i=0;i&lt;4;i++) {
		tx[i] *= f0r*root3;
		ty[i] *= f0r;
		tp = &amp;tpole[i];
		t = tp-&gt;nlat.s = tpoleinit[i][0]/root3;
		tp-&gt;nlat.c = sqrt(1 - t*t);
		tp-&gt;nlat.l = atan2(tp-&gt;nlat.s,tp-&gt;nlat.c);
		deg2rad(tpoleinit[i][1],&amp;tp-&gt;wlon);
		for(j=0;j&lt;4;j++) {
			tpp = &amp;tproj[i][j];
			latlon(tpp-&gt;tlat,tpp-&gt;tlon,&amp;tpp-&gt;projpl);
			deg2rad(tpp-&gt;ttwist,&amp;tpp-&gt;projtw);
			deg2rad(tpp-&gt;trot,&amp;tpp-&gt;postrot);
		}
	}
	return(Xtetra);
}

<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

