<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/unix/drawterm/kern/devaudio-sun.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/unix/drawterm/kern/devaudio-sun.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * Sun
 */
#include &lt;sys/ioctl.h&gt;
#include &lt;sys/audio.h&gt;
#include	"u.h"
#include	"lib.h"
#include	"dat.h"
#include	"fns.h"
#include	"error.h"
#include	"devaudio.h"

enum
{
	Channels = 2,
	Rate = 44100,
	Bits = 16,
};

static char* afn = 0;
static char* cfn = 0;
static int afd = -1;
static int cfd = -1;
static int speed = Rate;
static int needswap = -1;

static void
audiodevinit(void)
{
	uchar *p;
	ushort leorder;

	if ((afn = getenv("AUDIODEV")) == nil)
		afn = "/dev/audio";
	cfn = (char*)malloc(strlen(afn) + 3 + 1);
	if(cfn == nil)
		panic("out of memory");
	strcpy(cfn, afn);
	strcat(cfn, "ctl");

	/*
	 * Plan 9 /dev/audio is always little endian;
	 * solaris /dev/audio seems to expect native byte order,
	 * so on big endian machine (like sparc) we have to swap.
	 */
	leorder = (ushort) 0x0100;
	p = (uchar*)&amp;leorder;
	if (p[0] == 0 &amp;&amp; p[1] == 1) {
		/* little-endian: nothing to do */
		needswap = 0;
	} else {
		/* big-endian: translate Plan 9 little-endian */
		needswap = 1;
	}
}

/* maybe this should return -1 instead of sysfatal */
void
audiodevopen(void)
{
	audio_info_t info;
	struct audio_device ad;

	if (afn == nil || cfn == nil)
		audiodevinit();
	if((afd = open(afn, O_WRONLY)) &lt; 0)
		goto err;
	if(cfd &lt; 0 &amp;&amp; (cfd = open(cfn, O_RDWR)) &lt; 0)
		goto err;

	AUDIO_INITINFO(&amp;info);
	info.play.precision = Bits;
	info.play.channels = Channels;
	info.play.sample_rate = speed;
	info.play.encoding = AUDIO_ENCODING_LINEAR;
	if(ioctl(afd, AUDIO_SETINFO, &amp;info) &lt; 0)
		goto err;

	return;

err:
	if(afd &gt;= 0)
		close(afd);
	afd = -1;
	if(cfd &gt;= 0)
		close(cfd);
	cfd = -1;
	oserror();
}

void
audiodevclose(void)
{
	if(afd &gt;= 0)
		close(afd);
	if(cfd &gt;= 0)
		close(cfd);
	afd = -1;
	cfd = -1;
}

static double
fromsun(double val, double min, double max)
{
	return (val-min) / (max-min);
}

static double
tosun(double val, double min, double max)
{
	return val*(max-min) + min;
}

static void
setvolbal(double left, double right)
{
	audio_info_t info;
	double vol, bal;

	if (left &lt; 0 || right &lt; 0) {
		/* should not happen */
		return;
	} else if (left == right) {
		vol = tosun(left/100.0, AUDIO_MIN_GAIN, AUDIO_MAX_GAIN);
		bal = AUDIO_MID_BALANCE;
	} else if (left &lt; right) {
		vol = tosun(right/100.0, AUDIO_MIN_GAIN, AUDIO_MAX_GAIN);
		bal = tosun(1.0 - left/right, AUDIO_MID_BALANCE, AUDIO_RIGHT_BALANCE);
	} else if (right &lt; left) {
		vol = tosun(left/100.0, AUDIO_MIN_GAIN, AUDIO_MAX_GAIN);
		bal = tosun(1.0 - right/left, AUDIO_MID_BALANCE, AUDIO_LEFT_BALANCE);
	}
	AUDIO_INITINFO(&amp;info);
	info.play.gain = (long)(vol+0.5);
	info.play.balance = (long)(bal+0.5);
	if(ioctl(cfd, AUDIO_SETINFO, &amp;info) &lt; 0)
		oserror();
}

static void
getvolbal(int *left, int *right)
{
	audio_info_t info;
	double gain, bal, vol, l, r;

	AUDIO_INITINFO(&amp;info);
	if (ioctl(cfd, AUDIO_GETINFO, &amp;info) &lt; 0)
		oserror();

	gain = info.play.gain;
	bal = info.play.balance;
	vol = fromsun(gain, AUDIO_MIN_GAIN, AUDIO_MAX_GAIN) * 100.0;

	if (bal == AUDIO_MID_BALANCE) {
		l = r = vol;
	} else if (bal &lt; AUDIO_MID_BALANCE) {
		l = vol;
		r = vol * (1.0 - fromsun(bal, AUDIO_MID_BALANCE, AUDIO_LEFT_BALANCE));
	} else {
		r = vol;
		l = vol * (1.0 - fromsun(bal, AUDIO_MID_BALANCE, AUDIO_RIGHT_BALANCE));
	}
	*left = (long)(l+0.5);
	*right = (long)(r+0.5);
	return;
}

void
audiodevsetvol(int what, int left, int right)
{
	audio_info_t info;
	ulong x;
	int l, r;
	
	if (afn == nil || cfn == nil)
		audiodevinit();
	if(cfd &lt; 0 &amp;&amp; (cfd = open(cfn, O_RDWR)) &lt; 0) {
		cfd = -1;
		oserror();
	}

	if(what == Vspeed){
		x = left;
		AUDIO_INITINFO(&amp;info);
		info.play.sample_rate = x;
		if(ioctl(cfd, AUDIO_SETINFO, &amp;info) &lt; 0)
			oserror();
		speed = x;
		return;
	}
	if(what == Vaudio){
		getvolbal(&amp;l, &amp;r);
		if (left &lt; 0)
			setvolbal(l, right);
		else if (right &lt; 0)
			setvolbal(left, r);
		else 
			setvolbal(left, right);
		return;
	}
}

void
audiodevgetvol(int what, int *left, int *right)
{
	audio_info_t info;

	if (afn == nil || cfn == nil)
		audiodevinit();
	if(cfd &lt; 0 &amp;&amp; (cfd = open(cfn, O_RDWR)) &lt; 0) {
		cfd = -1;
		oserror();
	}
	switch(what) {
	case Vspeed:
		*left = *right = speed;
		break;
	case Vaudio:
		getvolbal(left, right);
		break;
	case Vtreb:
	case Vbass:
		*left = *right = 50;
		break;
	default:
		*left = *right = 0;
	}
}


static uchar *buf = 0;
static int nbuf = 0;

int
audiodevwrite(void *v, int n)
{
	int i, m, tot;
	uchar *p;

	if (needswap) {
		if (nbuf &lt; n) {
			buf = (uchar*)erealloc(buf, n);
			if(buf == nil)
				panic("out of memory");
			nbuf = n;
		}

		p = (uchar*)v;
		for(i=0; i+1&lt;n; i+=2) {
			buf[i] = p[i+1];
			buf[i+1] = p[i];
		}
		p = buf;
	} else
		p = (uchar*)v;
	
	for(tot=0; tot&lt;n; tot+=m)
		if((m = write(afd, p+tot, n-tot)) &lt;= 0)
			oserror();
	return tot;
}

int
audiodevread(void *v, int n)
{
	error("no reading");
	return -1;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

