<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/upas/smtp/greylist.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/upas/smtp/greylist.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * greylisting is the practice of making unknown callers call twice, with
 * a pause between them, before accepting their mail and adding them to a
 * whitelist of known callers.
 *
 * There's a bit of a problem with yahoo and other large sources of mail;
 * they have a vast pool of machines that all run the same queue(s), so a
 * 451 retry can come from a different IP address for many, many retries,
 * and it can take ~5 hours for the same IP to call us back.  To cope
 * better with this, we immediately accept mail from any system on the
 * same class C subnet (IPv4 /24) as anybody on our whitelist, since the
 * mail-sending machines tend to be clustered within a class C subnet.
 *
 * Various other goofballs, notably the IEEE, try to send mail just
 * before 9 AM, then refuse to try again until after 5 PM. D'oh!
 */
#include "common.h"
#include "smtpd.h"
#include "smtp.h"
#include &lt;ctype.h&gt;
#include &lt;ip.h&gt;
#include &lt;ndb.h&gt;

enum {
	Nonspammax = 14*60*60,  /* must call back within this time if real */
	Nonspammin = 5*60,	/* must wait this long to retry */
};

typedef struct {
	int	existed;	/* these two are distinct to cope with errors */
	int	created;
	int	noperm;
	long	mtime;		/* mod time, iff it already existed */
} Greysts;

static char whitelist[] = "/mail/grey/whitelist";

/*
 * matches ip addresses or subnets in whitelist against nci-&gt;rsys.
 * ignores comments and blank lines in /mail/grey/whitelist.
 */
static int
onwhitelist(void)
{
	int lnlen;
	char *line, *parse, *p;
	char input[128];
	uchar *mask;
	uchar mask4[IPaddrlen], addr4[IPaddrlen];
	uchar rmask[IPaddrlen], addr[IPaddrlen];
	uchar ipmasked[IPaddrlen], addrmasked[IPaddrlen];
	Biobuf *wl;

	wl = Bopen(whitelist, OREAD);
	if (wl == nil)
		return 1;
	while ((line = Brdline(wl, '\n')) != nil) {
		lnlen = Blinelen(wl);
		line[lnlen-1] = '\0';		/* clobber newline */

		p = strpbrk(line, " \t");
		if (p)
			*p = 0;
		if (line[0] == '#' || line[0] == 0)
			continue;

		/* default mask is /24 (v4) or /128 (v6) for bare IP */
		parse = line;
		if (strchr(line, '/') == nil) {
			strecpy(input, input + sizeof input - 5, line);
			if (strchr(line, ':') != nil)	/* v6? */
				strcat(input, "/128");
			else if (strchr(line, '.') != nil)
				strcat(input, "/24");	/* was /32 */
			parse = input;
		}
		mask = rmask;
		if (strchr(line, ':') != nil) {		/* v6? */
			parseip(addr, parse);
			p = strchr(parse, '/');
			if (p != nil)
				parseipmask(mask, p);
			else
				mask = IPallbits;
		} else {
			v4parsecidr(addr4, mask4, parse);
			v4tov6(addr, addr4);
			v4tov6(mask, mask4);
		}
		maskip(addr, mask, addrmasked);
		maskip(rsysip, mask, ipmasked);
		if (equivip6(ipmasked, addrmasked))
			break;
	}
	Bterm(wl);
	return line != nil;
}

static int mkdirs(char *);

/*
 * if any directories leading up to path don't exist, create them.
 * modifies but restores path.
 */
static int
mkpdirs(char *path)
{
	int rv = 0;
	char *sl = strrchr(path, '/');

	if (sl != nil) {
		*sl = '\0';
		rv = mkdirs(path);
		*sl = '/';
	}
	return rv;
}

/*
 * if path or any directories leading up to it don't exist, create them.
 * modifies but restores path.
 */
static int
mkdirs(char *path)
{
	int fd;

	if (access(path, AEXIST) &gt;= 0)
		return 0;

	/* make presumed-missing intermediate directories */
	if (mkpdirs(path) &lt; 0)
		return -1;

	/* make final directory */
	fd = create(path, OREAD, 0777|DMDIR);
	if (fd &lt; 0)
		/*
		 * we may have lost a race; if the directory now exists,
		 * it's okay.
		 */
		return access(path, AEXIST) &lt; 0? -1: 0;
	close(fd);
	return 0;
}

static long
getmtime(char *file)
{
	int fd;
	long mtime = -1;
	Dir *ds;

	fd = open(file, ORDWR);
	if (fd &lt; 0)
		return mtime;
	ds = dirfstat(fd);
	if (ds != nil) {
		mtime = ds-&gt;mtime;
		/*
		 * new twist: update file's mtime after reading it,
		 * so each call resets the future time after which
		 * we'll accept calls.  thus spammers who keep pounding
		 * us lose, but just pausing for a few minutes and retrying
		 * will succeed.
		 */
		if (0) {
			/*
			 * apparently none can't do this wstat
			 * (permission denied);
			 * more undocumented whacky none behaviour.
			 */
			ds-&gt;mtime = time(0);
			if (dirfwstat(fd, ds) &lt; 0)
				syslog(0, "smtpd", "dirfwstat %s: %r", file);
		}
		free(ds);
		write(fd, "x", 1);
	}
	close(fd);
	return mtime;
}

static void
tryaddgrey(char *file, Greysts *gsp)
{
	int fd = create(file, OWRITE|OEXCL, 0666);

	gsp-&gt;created = (fd &gt;= 0);
	if (fd &gt;= 0) {
		close(fd);
		gsp-&gt;existed = 0;  /* just created; couldn't have existed */
		gsp-&gt;mtime = time(0);
	} else {
		/*
		 * why couldn't we create file? it must have existed
		 * (or we were denied perm on parent dir.).
		 * if it existed, fill in gsp-&gt;mtime; otherwise
		 * make presumed-missing intermediate directories.
		 */
		gsp-&gt;existed = access(file, AEXIST) &gt;= 0;
		if (gsp-&gt;existed)
			gsp-&gt;mtime = getmtime(file);
		else if (mkpdirs(file) &lt; 0)
			gsp-&gt;noperm = 1;
	}
}

static void
addgreylist(char *file, Greysts *gsp)
{
	tryaddgrey(file, gsp);
	if (!gsp-&gt;created &amp;&amp; !gsp-&gt;existed &amp;&amp; !gsp-&gt;noperm)
		/* retry the greylist entry with parent dirs created */
		tryaddgrey(file, gsp);
}

static int
recentcall(Greysts *gsp)
{
	long delay = time(0) - gsp-&gt;mtime;

	if (!gsp-&gt;existed)
		return 0;
	/* reject immediate call-back; spammers are doing that now */
	return delay &gt;= Nonspammin &amp;&amp; delay &lt;= Nonspammax;
}

/*
 * policy: if (caller-IP, my-IP, rcpt) is not on the greylist,
 * reject this message as "451 temporary failure".  if the caller is real,
 * he'll retry soon, otherwise he's a spammer.
 * at the first rejection, create a greylist entry for (my-ip, caller-ip,
 * rcpt, time), where time is the file's mtime.  if they call back and there's
 * already a greylist entry, and it's within the allowed interval,
 * add their IP to the append-only whitelist.
 *
 * greylist files can be removed at will; at worst they'll cause a few
 * extra retries.
 */

static int
isrcptrecent(char *rcpt)
{
	char *user;
	char file[256];
	Greysts gs;
	Greysts *gsp = &amp;gs;

	if (rcpt[0] == '\0' || strchr(rcpt, '/') != nil ||
	    strcmp(rcpt, ".") == 0 || strcmp(rcpt, "..") == 0)
		return 0;

	/* shorten names to fit pre-fossil or pre-9p2000 file servers */
	user = strrchr(rcpt, '!');
	if (user == nil)
		user = rcpt;
	else
		user++;

	/* check &amp; try to update the grey list entry */
	snprint(file, sizeof file, "/mail/grey/tmp/%s/%s/%s",
		nci-&gt;lsys, nci-&gt;rsys, user);
	memset(gsp, 0, sizeof *gsp);
	addgreylist(file, gsp);

	/* if on greylist already and prior call was recent, add to whitelist */
	if (gsp-&gt;existed &amp;&amp; recentcall(gsp)) {
		syslog(0, "smtpd",
			"%s/%s was grey; adding IP to white", nci-&gt;rsys, rcpt);
		return 1;
	} else if (gsp-&gt;existed)
		syslog(0, "smtpd", "call for %s/%s was just minutes ago "
			"or long ago", nci-&gt;rsys, rcpt);
	else
		syslog(0, "smtpd", "no call registered for %s/%s; registering",
			nci-&gt;rsys, rcpt);
	return 0;
}

void
vfysenderhostok(void)
{
	char *fqdn;
	int recent = 0;
	Link *l;

	if (onwhitelist())
		return;

	for (l = rcvers.first; l; l = l-&gt;next)
		if (isrcptrecent(s_to_c(l-&gt;p)))
			recent = 1;

	/* if on greylist already and prior call was recent, add to whitelist */
	if (recent) {
		int fd = create(whitelist, OWRITE, 0666|DMAPPEND);

		if (fd &gt;= 0) {
			seek(fd, 0, 2);			/* paranoia */
			fqdn = csgetvalue(nci-&gt;root, "ip", nci-&gt;rsys, "dom",
				nil);
			if (fqdn != nil)
				fprint(fd, "%s %s\n", nci-&gt;rsys, fqdn);
			else
				fprint(fd, "%s\n", nci-&gt;rsys);
			free(fqdn);
			close(fd);
		}
	} else {
		syslog(0, "smtpd",
	"no recent call from %s for a rcpt; rejecting with temporary failure",
			nci-&gt;rsys);
		reply("451 please try again soon from the same IP.\r\n");
		exits("no recent call for a rcpt");
	}
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

