<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/usb/disk/disk.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/usb/disk/disk.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * usb/disk - usb mass storage file server
 *
 * supports only the scsi command interface, not ata.
 */

#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;ctype.h&gt;
#include &lt;fcall.h&gt;
#include &lt;thread.h&gt;
#include &lt;disk.h&gt;
#include "scsireq.h"
#include "usb.h"
#include "usbfs.h"
#include "ums.h"

enum
{
	Qdir = 0,
	Qctl,
	Qraw,
	Qdata,
	Qmax,
};

typedef struct Dirtab Dirtab;
struct Dirtab
{
	char	*name;
	int	mode;
};

static Dirtab dirtab[] =
{
	[Qdir]	"/",	DMDIR|0555,
	[Qctl]	"ctl",	0664,		/* nothing secret here */
	[Qraw]	"raw",	0640,
	[Qdata]	"data",	0640,
};

/*
 * These are used by scuzz scsireq
 */
int exabyte, force6bytecmds;

int diskdebug;

static void
ding(void *, char *msg)
{
	if(strstr(msg, "alarm") != nil)
		noted(NCONT);
	noted(NDFLT);
}

static int
getmaxlun(Dev *dev)
{
	uchar max;
	int r;

	max = 0;
	r = Rd2h|Rclass|Riface;
	if(usbcmd(dev, r, Getmaxlun, 0, 0, &amp;max, 1) &lt; 0){
		dprint(2, "disk: %s: getmaxlun failed: %r\n", dev-&gt;dir);
	}else{
		max &amp;= 017;			/* 15 is the max. allowed */
		dprint(2, "disk: %s: maxlun %d\n", dev-&gt;dir, max);
	}
	return max;
}

static int
umsreset(Ums *ums)
{
	int r;

	r = Rh2d|Rclass|Riface;
	if(usbcmd(ums-&gt;dev, r, Umsreset, 0, 0, nil, 0) &lt; 0){
		fprint(2, "disk: reset: %r\n");
		return -1;
	}
	return 0;
}

static int
umsrecover(Ums *ums)
{
	if(umsreset(ums) &lt; 0)
		return -1;
	if(unstall(ums-&gt;dev, ums-&gt;epin, Ein) &lt; 0)
		dprint(2, "disk: unstall epin: %r\n");

	/* do we need this when epin == epout? */
	if(unstall(ums-&gt;dev, ums-&gt;epout, Eout) &lt; 0)
		dprint(2, "disk: unstall epout: %r\n");
	return 0;
}

static void
umsfatal(Ums *ums)
{
	int i;

	devctl(ums-&gt;dev, "detach");
	for(i = 0; i &lt; ums-&gt;maxlun; i++)
		usbfsdel(&amp;ums-&gt;lun[i].fs);
}

static int
ispow2(uvlong ul)
{
	return (ul &amp; (ul - 1)) == 0;
}

/*
 * return smallest power of 2 &gt;= n
 */
static int
log2(int n)
{
	int i;

	for(i = 0; (1 &lt;&lt; i) &lt; n; i++)
		;
	return i;
}

static int
umscapacity(Umsc *lun)
{
	uchar data[32];

	lun-&gt;blocks = 0;
	lun-&gt;capacity = 0;
	lun-&gt;lbsize = 0;
	memset(data, 0, sizeof data);
	if(SRrcapacity(lun, data) &lt; 0 &amp;&amp; SRrcapacity(lun, data)  &lt; 0)
		return -1;
	lun-&gt;blocks = GETBELONG(data);
	lun-&gt;lbsize = GETBELONG(data+4);
	if(lun-&gt;blocks == 0xFFFFFFFF){
		if(SRrcapacity16(lun, data) &lt; 0){
			lun-&gt;lbsize = 0;
			lun-&gt;blocks = 0;
			return -1;
		}else{
			lun-&gt;lbsize = GETBELONG(data + 8);
			lun-&gt;blocks = (uvlong)GETBELONG(data)&lt;&lt;32 |
				GETBELONG(data + 4);
		}
	}
	lun-&gt;blocks++; /* SRcapacity returns LBA of last block */
	lun-&gt;capacity = (vlong)lun-&gt;blocks * lun-&gt;lbsize;
	if(diskdebug)
		fprint(2, "disk: logical block size %lud, # blocks %llud\n",
			lun-&gt;lbsize, lun-&gt;blocks);
	return 0;
}

static int
umsinit(Ums *ums)
{
	uchar i;
	Umsc *lun;
	int some;

	umsreset(ums);
	ums-&gt;maxlun = getmaxlun(ums-&gt;dev);
	ums-&gt;lun = mallocz((ums-&gt;maxlun+1) * sizeof(*ums-&gt;lun), 1);
	some = 0;
	for(i = 0; i &lt;= ums-&gt;maxlun; i++){
		lun = &amp;ums-&gt;lun[i];
		lun-&gt;ums = ums;
		lun-&gt;umsc = lun;
		lun-&gt;lun = i;
		lun-&gt;flags = Fopen | Fusb | Frw10;
		if(SRinquiry(lun) &lt; 0 &amp;&amp; SRinquiry(lun) &lt; 0){
			dprint(2, "disk: lun %d inquiry failed\n", i);
			continue;
		}
		switch(lun-&gt;inquiry[0]){
		case Devdir:
		case Devworm:		/* a little different than the others */
		case Devcd:
		case Devmo:
			break;
		default:
			fprint(2, "disk: lun %d is not a disk (type %#02x)\n",
				i, lun-&gt;inquiry[0]);
			continue;
		}
		SRstart(lun, 1);
		/*
		 * we ignore the device type reported by inquiry.
		 * Some devices return a wrong value but would still work.
		 */
		some++;
		lun-&gt;inq = smprint("%.48s", (char *)lun-&gt;inquiry+8);
		umscapacity(lun);
	}
	if(some == 0){
		dprint(2, "disk: all luns failed\n");
		devctl(ums-&gt;dev, "detach");
		return -1;
	}
	return 0;
}


/*
 * called by SR*() commands provided by scuzz's scsireq
 */
long
umsrequest(Umsc *umsc, ScsiPtr *cmd, ScsiPtr *data, int *status)
{
	Cbw cbw;
	Csw csw;
	int n, nio, left;
	Ums *ums;

	ums = umsc-&gt;ums;

	memcpy(cbw.signature, "USBC", 4);
	cbw.tag = ++ums-&gt;seq;
	cbw.datalen = data-&gt;count;
	cbw.flags = data-&gt;write? CbwDataOut: CbwDataIn;
	cbw.lun = umsc-&gt;lun;
	if(cmd-&gt;count &lt; 1 || cmd-&gt;count &gt; 16)
		fprint(2, "disk: umsrequest: bad cmd count: %ld\n", cmd-&gt;count);

	cbw.len = cmd-&gt;count;
	assert(cmd-&gt;count &lt;= sizeof(cbw.command));
	memcpy(cbw.command, cmd-&gt;p, cmd-&gt;count);
	memset(cbw.command + cmd-&gt;count, 0, sizeof(cbw.command) - cmd-&gt;count);

	werrstr("");		/* we use %r later even for n == 0 */
	if(diskdebug){
		fprint(2, "disk: cmd: tag %#lx: ", cbw.tag);
		for(n = 0; n &lt; cbw.len; n++)
			fprint(2, " %2.2x", cbw.command[n]&amp;0xFF);
		fprint(2, " datalen: %ld\n", cbw.datalen);
	}

	/* issue tunnelled scsi command */
	if(write(ums-&gt;epout-&gt;dfd, &amp;cbw, CbwLen) != CbwLen){
		fprint(2, "disk: cmd: %r\n");
		goto Fail;
	}

	/* transfer the data */
	nio = data-&gt;count;
	if(nio != 0){
		if(data-&gt;write)
			n = write(ums-&gt;epout-&gt;dfd, data-&gt;p, nio);
		else{
			n = read(ums-&gt;epin-&gt;dfd, data-&gt;p, nio);
			left = nio - n;
			if (n &gt;= 0 &amp;&amp; left &gt; 0)	/* didn't fill data-&gt;p? */
				memset(data-&gt;p + n, 0, left);
		}
		nio = n;
		if(diskdebug)
			if(n &lt; 0)
				fprint(2, "disk: data: %r\n");
			else
				fprint(2, "disk: data: %d bytes\n", n);
		if(n &lt;= 0)
			if(data-&gt;write == 0)
				unstall(ums-&gt;dev, ums-&gt;epin, Ein);
	}

	/* read the transfer's status */
	n = read(ums-&gt;epin-&gt;dfd, &amp;csw, CswLen);
	if(n &lt;= 0){
		/* n == 0 means "stalled" */
		unstall(ums-&gt;dev, ums-&gt;epin, Ein);
		n = read(ums-&gt;epin-&gt;dfd, &amp;csw, CswLen);
	}

	if(n != CswLen || strncmp(csw.signature, "USBS", 4) != 0){
		dprint(2, "disk: read n=%d: status: %r\n", n);
		goto Fail;
	}
	if(csw.tag != cbw.tag){
		dprint(2, "disk: status tag mismatch\n");
		goto Fail;
	}
	if(csw.status &gt;= CswPhaseErr){
		dprint(2, "disk: phase error\n");
		goto Fail;
	}
	if(csw.dataresidue == 0 || ums-&gt;wrongresidues)
		csw.dataresidue = data-&gt;count - nio;
	if(diskdebug){
		fprint(2, "disk: status: %2.2ux residue: %ld\n",
			csw.status, csw.dataresidue);
		if(cbw.command[0] == ScmdRsense){
			fprint(2, "sense data:");
			for(n = 0; n &lt; data-&gt;count - csw.dataresidue; n++)
				fprint(2, " %2.2x", data-&gt;p[n]);
			fprint(2, "\n");
		}
	}
	switch(csw.status){
	case CswOk:
		*status = STok;
		break;
	case CswFailed:
		*status = STcheck;
		break;
	default:
		dprint(2, "disk: phase error\n");
		goto Fail;
	}
	ums-&gt;nerrs = 0;
	return data-&gt;count - csw.dataresidue;

Fail:
	*status = STharderr;
	if(ums-&gt;nerrs++ &gt; 15){
		fprint(2, "disk: %s: too many errors: device detached\n", ums-&gt;dev-&gt;dir);
		umsfatal(ums);
	}else
		umsrecover(ums);
	return -1;
}

static int
dwalk(Usbfs *fs, Fid *fid, char *name)
{
	int i;
	Qid qid;

	qid = fid-&gt;qid;
	if((qid.type &amp; QTDIR) == 0){
		werrstr("walk in non-directory");
		return -1;
	}

	if(strcmp(name, "..") == 0)
		return 0;

	for(i = 1; i &lt; nelem(dirtab); i++)
		if(strcmp(name, dirtab[i].name) == 0){
			qid.path = i | fs-&gt;qid;
			qid.vers = 0;
			qid.type = dirtab[i].mode &gt;&gt; 24;
			fid-&gt;qid = qid;
			return 0;
		}
	werrstr(Enotfound);
	return -1;
}

static void
dostat(Usbfs *fs, int path, Dir *d)
{
	Dirtab *t;
	Umsc *lun;

	t = &amp;dirtab[path];
	d-&gt;qid.path = path;
	d-&gt;qid.type = t-&gt;mode &gt;&gt; 24;
	d-&gt;mode = t-&gt;mode;
	d-&gt;name = t-&gt;name;
	lun = fs-&gt;aux;
	if(path == Qdata)
		d-&gt;length = lun-&gt;capacity;
	else
		d-&gt;length = 0;
}

static int
dirgen(Usbfs *fs, Qid, int i, Dir *d, void*)
{
	i++;	/* skip dir */
	if(i &gt;= Qmax)
		return -1;
	else{
		dostat(fs, i, d);
		d-&gt;qid.path |= fs-&gt;qid;
		return 0;
	}
}

static int
dstat(Usbfs *fs, Qid qid, Dir *d)
{
	int path;

	path = qid.path &amp; ~fs-&gt;qid;
	dostat(fs, path, d);
	d-&gt;qid.path |= fs-&gt;qid;
	return 0;
}

static int
dopen(Usbfs *fs, Fid *fid, int)
{
	ulong path;
	Umsc *lun;

	path = fid-&gt;qid.path &amp; ~fs-&gt;qid;
	lun = fs-&gt;aux;
	switch(path){
	case Qraw:
		lun-&gt;phase = Pcmd;
		break;
	}
	return 0;
}

/*
 * check i/o parameters and compute values needed later.
 * we shift &amp; mask manually to avoid run-time calls to _divv and _modv,
 * since we don't need general division nor its cost.
 */
static int
setup(Umsc *lun, char *data, int count, vlong offset)
{
	long nb, lbsize, lbshift, lbmask;
	uvlong bno;

	if(count &lt; 0 || lun-&gt;lbsize &lt;= 0 &amp;&amp; umscapacity(lun) &lt; 0 ||
	    lun-&gt;lbsize == 0)
		return -1;
	lbsize = lun-&gt;lbsize;
	assert(ispow2(lbsize));
	lbshift = log2(lbsize);
	lbmask = lbsize - 1;

	bno = offset &gt;&gt; lbshift;	/* offset / lbsize */
	nb = ((offset + count + lbsize - 1) &gt;&gt; lbshift) - bno;

	if(bno + nb &gt; lun-&gt;blocks)		/* past end of device? */
		nb = lun-&gt;blocks - bno;
	if(nb * lbsize &gt; Maxiosize)
		nb = Maxiosize / lbsize;
	lun-&gt;nb = nb;
	if(bno &gt;= lun-&gt;blocks || nb == 0)
		return 0;

	lun-&gt;offset = bno;
	lun-&gt;off = offset &amp; lbmask;		/* offset % lbsize */
	if(lun-&gt;off == 0 &amp;&amp; (count &amp; lbmask) == 0)
		lun-&gt;bufp = data;
	else
		/* not transferring full, aligned blocks; need intermediary */
		lun-&gt;bufp = lun-&gt;buf;
	return count;
}

/*
 * Upon SRread/SRwrite errors we assume the medium may have changed,
 * and ask again for the capacity of the media.
 * BUG: How to proceed to avoid confusing dossrv??
 *
 * ctl reads must match the format documented in sd(3) exactly
 * to interoperate with the rest of the system.
 */
static long
dread(Usbfs *fs, Fid *fid, void *data, long count, vlong offset)
{
	long n;
	ulong path;
	char buf[1024];
	char *s, *e;
	Umsc *lun;
	Ums *ums;
	Qid q;

	q = fid-&gt;qid;
	path = fid-&gt;qid.path &amp; ~fs-&gt;qid;
	ums = fs-&gt;dev-&gt;aux;
	lun = fs-&gt;aux;

	qlock(ums);
	switch(path){
	case Qdir:
		count = usbdirread(fs, q, data, count, offset, dirgen, nil);
		break;
	case Qctl:
		/*
		 * Some usb disks need an extra opportunity to divulge their
		 * capacity (e.g. M-Systems/SanDisk 1GB flash drive).
		 */
		if(lun-&gt;lbsize &lt;= 0)
			umscapacity(lun);

		s = buf;
		e = buf + sizeof(buf);
		if(lun-&gt;flags &amp; Finqok)
			s = seprint(s, e, "inquiry %s lun %ld: %s\n",
				fs-&gt;dev-&gt;dir, lun - &amp;ums-&gt;lun[0], lun-&gt;inq);
		if(lun-&gt;blocks &gt; 0)
			s = seprint(s, e, "geometry %llud %ld\n",
				lun-&gt;blocks, lun-&gt;lbsize);
		count = usbreadbuf(data, count, offset, buf, s - buf);
		break;
	case Qraw:
		if(lun-&gt;lbsize &lt;= 0 &amp;&amp; umscapacity(lun) &lt; 0){
			count = -1;
			break;
		}
		switch(lun-&gt;phase){
		case Pcmd:
			qunlock(ums);
			werrstr("phase error");
			return -1;
		case Pdata:
			lun-&gt;data.p = data;
			lun-&gt;data.count = count;
			lun-&gt;data.write = 0;
			count = umsrequest(lun,&amp;lun-&gt;cmd,&amp;lun-&gt;data,&amp;lun-&gt;status);
			lun-&gt;phase = Pstatus;
			if(count &lt; 0)
				lun-&gt;lbsize = 0;  /* medium may have changed */
			break;
		case Pstatus:
			n = snprint(buf, sizeof buf, "%11.0ud ", lun-&gt;status);
			count = usbreadbuf(data, count, 0LL, buf, n);
			lun-&gt;phase = Pcmd;
			break;
		}
		break;
	case Qdata:
		count = setup(lun, data, count, offset);
		if (count &lt;= 0)
			break;
		n = SRread(lun, lun-&gt;bufp, lun-&gt;nb * lun-&gt;lbsize);
		if(n &lt; 0){
			lun-&gt;lbsize = 0;	/* medium may have changed */
			count = -1;
		} else if (lun-&gt;bufp == data)
			count = n;
		else{
			/*
			 * if n == lun-&gt;nb*lun-&gt;lbsize (as expected),
			 * just copy count bytes.
			 */
			if(lun-&gt;off + count &gt; n)
				count = n - lun-&gt;off; /* short read */
			if(count &gt; 0)
				memmove(data, lun-&gt;bufp + lun-&gt;off, count);
		}
		break;
	}
	qunlock(ums);
	return count;
}

static long
dwrite(Usbfs *fs, Fid *fid, void *data, long count, vlong offset)
{
	long len, ocount;
	ulong path;
	uvlong bno;
	Ums *ums;
	Umsc *lun;

	ums = fs-&gt;dev-&gt;aux;
	lun = fs-&gt;aux;
	path = fid-&gt;qid.path &amp; ~fs-&gt;qid;

	qlock(ums);
	switch(path){
	default:
		werrstr(Eperm);
		count = -1;
		break;
	case Qctl:
		dprint(2, "usb/disk: ctl ignored\n");
		break;
	case Qraw:
		if(lun-&gt;lbsize &lt;= 0 &amp;&amp; umscapacity(lun) &lt; 0){
			count = -1;
			break;
		}
		switch(lun-&gt;phase){
		case Pcmd:
			if(count != 6 &amp;&amp; count != 10){
				qunlock(ums);
				werrstr("bad command length");
				return -1;
			}
			memmove(lun-&gt;rawcmd, data, count);
			lun-&gt;cmd.p = lun-&gt;rawcmd;
			lun-&gt;cmd.count = count;
			lun-&gt;cmd.write = 1;
			lun-&gt;phase = Pdata;
			break;
		case Pdata:
			lun-&gt;data.p = data;
			lun-&gt;data.count = count;
			lun-&gt;data.write = 1;
			count = umsrequest(lun,&amp;lun-&gt;cmd,&amp;lun-&gt;data,&amp;lun-&gt;status);
			lun-&gt;phase = Pstatus;
			if(count &lt; 0)
				lun-&gt;lbsize = 0;  /* medium may have changed */
			break;
		case Pstatus:
			lun-&gt;phase = Pcmd;
			werrstr("phase error");
			count = -1;
			break;
		}
		break;
	case Qdata:
		len = ocount = count;
		count = setup(lun, data, count, offset);
		if (count &lt;= 0)
			break;
		bno = lun-&gt;offset;
		if (lun-&gt;bufp == lun-&gt;buf) {
			count = SRread(lun, lun-&gt;bufp, lun-&gt;nb * lun-&gt;lbsize);
			if(count &lt; 0) {
				lun-&gt;lbsize = 0;  /* medium may have changed */
				break;
			}
			/*
			 * if count == lun-&gt;nb*lun-&gt;lbsize, as expected, just
			 * copy len (the original count) bytes of user data.
			 */
			if(lun-&gt;off + len &gt; count)
				len = count - lun-&gt;off; /* short read */
			if(len &gt; 0)
				memmove(lun-&gt;bufp + lun-&gt;off, data, len);
		}

		lun-&gt;offset = bno;
		count = SRwrite(lun, lun-&gt;bufp, lun-&gt;nb * lun-&gt;lbsize);
		if(count &lt; 0)
			lun-&gt;lbsize = 0;	/* medium may have changed */
		else{
			if(lun-&gt;off + len &gt; count)
				count -= lun-&gt;off; /* short write */
			/* never report more bytes written than requested */
			if(count &lt; 0)
				count = 0;
			else if(count &gt; ocount)
				count = ocount;
		}
		break;
	}
	qunlock(ums);
	return count;
}

int
findendpoints(Ums *ums)
{
	Ep *ep;
	Usbdev *ud;
	ulong csp, sc;
	int i, epin, epout;

	epin = epout = -1;
	ud = ums-&gt;dev-&gt;usb;
	for(i = 0; i &lt; nelem(ud-&gt;ep); i++){
		if((ep = ud-&gt;ep[i]) == nil)
			continue;
		csp = ep-&gt;iface-&gt;csp;
		sc = Subclass(csp);
		if(!(Class(csp) == Clstorage &amp;&amp; (Proto(csp) == Protobulk)))
			continue;
		if(sc != Subatapi &amp;&amp; sc != Sub8070 &amp;&amp; sc != Subscsi)
			fprint(2, "disk: subclass %#ulx not supported. trying anyway\n", sc);
		if(ep-&gt;type == Ebulk){
			if(ep-&gt;dir == Eboth || ep-&gt;dir == Ein)
				if(epin == -1)
					epin =  ep-&gt;id;
			if(ep-&gt;dir == Eboth || ep-&gt;dir == Eout)
				if(epout == -1)
					epout = ep-&gt;id;
		}
	}
	dprint(2, "disk: ep ids: in %d out %d\n", epin, epout);
	if(epin == -1 || epout == -1)
		return -1;
	ums-&gt;epin = openep(ums-&gt;dev, epin);
	if(ums-&gt;epin == nil){
		fprint(2, "disk: openep %d: %r\n", epin);
		return -1;
	}
	if(epout == epin){
		incref(ums-&gt;epin);
		ums-&gt;epout = ums-&gt;epin;
	}else
		ums-&gt;epout = openep(ums-&gt;dev, epout);
	if(ums-&gt;epout == nil){
		fprint(2, "disk: openep %d: %r\n", epout);
		closedev(ums-&gt;epin);
		return -1;
	}
	if(ums-&gt;epin == ums-&gt;epout)
		opendevdata(ums-&gt;epin, ORDWR);
	else{
		opendevdata(ums-&gt;epin, OREAD);
		opendevdata(ums-&gt;epout, OWRITE);
	}
	if(ums-&gt;epin-&gt;dfd &lt; 0 || ums-&gt;epout-&gt;dfd &lt; 0){
		fprint(2, "disk: open i/o ep data: %r\n");
		closedev(ums-&gt;epin);
		closedev(ums-&gt;epout);
		return -1;
	}
	dprint(2, "disk: ep in %s out %s\n", ums-&gt;epin-&gt;dir, ums-&gt;epout-&gt;dir);

	devctl(ums-&gt;epin, "timeout 2000");
	devctl(ums-&gt;epout, "timeout 2000");

	if(usbdebug &gt; 1 || diskdebug &gt; 2){
		devctl(ums-&gt;epin, "debug 1");
		devctl(ums-&gt;epout, "debug 1");
		devctl(ums-&gt;dev, "debug 1");
	}
	return 0;
}

static int
usage(void)
{
	werrstr("usage: usb/disk [-d] [-N nb]");
	return -1;
}

static void
umsdevfree(void *a)
{
	Ums *ums = a;

	if(ums == nil)
		return;
	closedev(ums-&gt;epin);
	closedev(ums-&gt;epout);
	ums-&gt;epin = ums-&gt;epout = nil;
	free(ums-&gt;lun);
	free(ums);
}

static Usbfs diskfs = {
	.walk = dwalk,
	.open =	 dopen,
	.read =	 dread,
	.write = dwrite,
	.stat =	 dstat,
};

int
diskmain(Dev *dev, int argc, char **argv)
{
	Ums *ums;
	Umsc *lun;
	int i, devid;

	devid = dev-&gt;id;
	ARGBEGIN{
	case 'd':
		scsidebug(diskdebug);
		diskdebug++;
		break;
	case 'N':
		devid = atoi(EARGF(usage()));
		break;
	default:
		return usage();
	}ARGEND
	if(argc != 0) {
		return usage();
	}

//	notify(ding);
	ums = dev-&gt;aux = emallocz(sizeof(Ums), 1);
	ums-&gt;maxlun = -1;
	ums-&gt;dev = dev;
	dev-&gt;free = umsdevfree;
	if(findendpoints(ums) &lt; 0){
		werrstr("disk: endpoints not found");
		return -1;
	}

	/*
	 * SanDISK 512M gets residues wrong.
	 */
	if(dev-&gt;usb-&gt;vid == 0x0781 &amp;&amp; dev-&gt;usb-&gt;did == 0x5150)
		ums-&gt;wrongresidues = 1;

	if(umsinit(ums) &lt; 0){
		dprint(2, "disk: umsinit: %r\n");
		return -1;
	}

	for(i = 0; i &lt;= ums-&gt;maxlun; i++){
		lun = &amp;ums-&gt;lun[i];
		lun-&gt;fs = diskfs;
		snprint(lun-&gt;fs.name, sizeof(lun-&gt;fs.name), "sdU%d.%d", devid, i);
		lun-&gt;fs.dev = dev;
		incref(dev);
		lun-&gt;fs.aux = lun;
		usbfsadd(&amp;lun-&gt;fs);
	}
	return 0;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

