<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/venti/srv/bloom.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/venti/srv/bloom.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * Bloom filter tracking which scores are present in our arenas
 * and (more importantly) which are not.  
 */

#include "stdinc.h"
#include "dat.h"
#include "fns.h"

int ignorebloom;

int
bloominit(Bloom *b, vlong vsize, u8int *data)
{
	ulong size;
	
	size = vsize;
	if(size != vsize){	/* truncation */
		werrstr("bloom data too big");
		return -1;
	}
	
	b-&gt;size = size;
	b-&gt;nhash = 32;	/* will be fixed by caller on initialization */
	if(data != nil)
		if(unpackbloomhead(b, data) &lt; 0)
			return -1;
	
	b-&gt;bitmask = (b-&gt;size&lt;&lt;3) - 1;
	b-&gt;data = data;
	return 0;
}

void
wbbloomhead(Bloom *b)
{
	packbloomhead(b, b-&gt;data);
}

Bloom*
readbloom(Part *p)
{
	uchar buf[512];
	Bloom *b;
	
	b = vtmallocz(sizeof *b);
	if(readpart(p, 0, buf, sizeof buf) &lt; 0)
		return nil;
	/*
	 * pass buf as b-&gt;data so that bloominit
	 * can parse header.  won't be used for
	 * accessing bits (cleared below).
	 */
	if(bloominit(b, 0, buf) &lt; 0){
		vtfree(b);
		return nil;
	}else{
		/*
		 * default block size is system page size.
		 * the bloom filter is usually very big.
		 * bump the block size up to speed i/o.
		 */
		if(p-&gt;blocksize &lt; (1&lt;&lt;20)){
			p-&gt;blocksize = 1&lt;&lt;20;
			if(p-&gt;blocksize &gt; p-&gt;size)
				p-&gt;blocksize = p-&gt;size;
		}
	}
	b-&gt;part = p;
	b-&gt;data = nil;
	return b;
}

int
resetbloom(Bloom *b)
{
	uchar *data;
	
	data = vtmallocz(b-&gt;size);
	b-&gt;data = data;
	if(b-&gt;size == MaxBloomSize)	/* 2^32 overflows ulong */
		addstat(StatBloomBits, b-&gt;size*8-1);
	else
		addstat(StatBloomBits, b-&gt;size*8);
	return 0;
}

int
loadbloom(Bloom *b)
{
	int i, n;
	uint ones;
	uchar *data;
	u32int *a;
	
	data = vtmallocz(b-&gt;size);
	if(readpart(b-&gt;part, 0, data, b-&gt;size) &lt; 0){
		vtfree(b);
		vtfree(data);
		return -1;
	}
	b-&gt;data = data;

	a = (u32int*)b-&gt;data;
	n = b-&gt;size/4;
	ones = 0;
	for(i=0; i&lt;n; i++)
		ones += countbits(a[i]); 
	addstat(StatBloomOnes, ones);

	if(b-&gt;size == MaxBloomSize)	/* 2^32 overflows ulong */
		addstat(StatBloomBits, b-&gt;size*8-1);
	else
		addstat(StatBloomBits, b-&gt;size*8);
		
	return 0;
}

int
writebloom(Bloom *b)
{
	wbbloomhead(b);
	if(writepart(b-&gt;part, 0, b-&gt;data, b-&gt;size) &lt; 0)
		return -1;
	if(flushpart(b-&gt;part) &lt; 0)
		return -1;
	return 0;
}

/*
 * Derive two random 32-bit quantities a, b from the score
 * and then use a+b*i as a sequence of bloom filter indices.
 * Michael Mitzenmacher has a recent (2005) paper saying this is okay.
 * We reserve the bottom bytes (BloomHeadSize*8 bits) for the header.
 */
static void
gethashes(u8int *score, ulong *h)
{
	int i;
	u32int a, b;

	a = 0;
	b = 0;
	for(i=4; i+8&lt;=VtScoreSize; i+=8){
		a ^= *(u32int*)(score+i);
		b ^= *(u32int*)(score+i+4);
	}
	if(i+4 &lt;= VtScoreSize)	/* 20 is not 4-aligned */
		a ^= *(u32int*)(score+i);
	for(i=0; i&lt;BloomMaxHash; i++, a+=b)
		h[i] = a &lt; BloomHeadSize*8 ? BloomHeadSize*8 : a;
}

static void
_markbloomfilter(Bloom *b, u8int *score)
{
	int i, nnew;
	ulong h[BloomMaxHash];
	u32int x, *y, z, *tab;

	trace("markbloomfilter", "markbloomfilter %V", score);
	gethashes(score, h);
	nnew = 0;
	tab = (u32int*)b-&gt;data;
	for(i=0; i&lt;b-&gt;nhash; i++){
		x = h[i];
		y = &amp;tab[(x&amp;b-&gt;bitmask)&gt;&gt;5];
		z = 1&lt;&lt;(x&amp;31);
		if(!(*y&amp;z)){
			nnew++;
			*y |= z;
		}
	}
	if(nnew)
		addstat(StatBloomOnes, nnew);

	trace("markbloomfilter", "markbloomfilter exit");
}

static int
_inbloomfilter(Bloom *b, u8int *score)
{
	int i;
	ulong h[BloomMaxHash], x;
	u32int *tab;

	gethashes(score, h);
	tab = (u32int*)b-&gt;data;
	for(i=0; i&lt;b-&gt;nhash; i++){
		x = h[i];
		if(!(tab[(x&amp;b-&gt;bitmask)&gt;&gt;5] &amp; (1&lt;&lt;(x&amp;31))))
			return 0;
	}
	return 1;
}

int
inbloomfilter(Bloom *b, u8int *score)
{
	int r;

	if(b == nil || b-&gt;data == nil)
		return 1;

	if(ignorebloom)
		return 1;
	
	rlock(&amp;b-&gt;lk);
	r = _inbloomfilter(b, score);
	runlock(&amp;b-&gt;lk);
	addstat(StatBloomLookup, 1);
	if(r)
		addstat(StatBloomMiss, 1);
	else
		addstat(StatBloomHit, 1);
	return r;
}

void
markbloomfilter(Bloom *b, u8int *score)
{
	if(b == nil || b-&gt;data == nil)
		return;

	rlock(&amp;b-&gt;lk);
	qlock(&amp;b-&gt;mod);
	_markbloomfilter(b, score);
	qunlock(&amp;b-&gt;mod);
	runlock(&amp;b-&gt;lk);
}

static void
bloomwriteproc(void *v)
{
	int ret;
	Bloom *b;

	threadsetname("bloomwriteproc");	
	b = v;
	for(;;){
		recv(b-&gt;writechan, 0);
		if((ret=writebloom(b)) &lt; 0)
			fprint(2, "oops! writing bloom: %r\n");
		else
			ret = 0;
		sendul(b-&gt;writedonechan, ret);
	}
}

void
startbloomproc(Bloom *b)
{
	b-&gt;writechan = chancreate(sizeof(void*), 0);
	b-&gt;writedonechan = chancreate(sizeof(ulong), 0);
	vtproc(bloomwriteproc, b);	
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

