<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/venti/srv/dcache.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/venti/srv/dcache.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * Disk cache.
 * 
 * Caches raw disk blocks.  Getdblock() gets a block, putdblock puts it back.
 * Getdblock has a mode parameter that determines i/o and access to a block:
 * if mode is OREAD or ORDWR, it is read from disk if not already in memory.
 * If mode is ORDWR or OWRITE, it is locked for exclusive use before being returned.
 * It is *not* marked dirty -- once changes have been made, they should be noted
 * by using dirtydblock() before putdblock().  
 *
 * There is a global cache lock as well as a lock on each block. 
 * Within a thread, the cache lock can be acquired while holding a block lock,
 * but not vice versa; and a block cannot be locked if you already hold the lock
 * on another block.
 * 
 * The flush proc writes out dirty blocks in batches, one batch per dirty tag.
 * For example, the DirtyArena blocks are all written to disk before any of the
 * DirtyArenaCib blocks.
 *
 * This code used to be in charge of flushing the dirty index blocks out to 
 * disk, but updating the index turned out to benefit from extra care.
 * Now cached index blocks are never marked dirty.  The index.c code takes
 * care of updating them behind our back, and uses _getdblock to update any
 * cached copies of the blocks as it changes them on disk.
 */

#include "stdinc.h"
#include "dat.h"
#include "fns.h"

typedef struct DCache	DCache;

enum
{
	HashLog		= 9,
	HashSize	= 1&lt;&lt;HashLog,
	HashMask	= HashSize - 1,
};

struct DCache
{
	QLock		lock;
	RWLock		dirtylock;		/* must be held to inspect or set b-&gt;dirty */
	Rendez		full;
	Round		round;
	DBlock		*free;			/* list of available lumps */
	u32int		now;			/* ticks for usage timestamps */
	int		size;			/* max. size of any block; allocated to each block */
	DBlock		**heads;		/* hash table for finding address */
	int		nheap;			/* number of available victims */
	DBlock		**heap;			/* heap for locating victims */
	int		nblocks;		/* number of blocks allocated */
	DBlock		*blocks;		/* array of block descriptors */
	DBlock		**write;		/* array of block pointers to be written */
	u8int		*mem;			/* memory for all block descriptors */
	int		ndirty;			/* number of dirty blocks */
	int		maxdirty;		/* max. number of dirty blocks */
};

typedef struct Ra Ra;
struct Ra
{
	Part *part;
	u64int addr;
};

static DCache	dcache;

static int	downheap(int i, DBlock *b);
static int	upheap(int i, DBlock *b);
static DBlock	*bumpdblock(void);
static void	delheap(DBlock *db);
static void	fixheap(int i, DBlock *b);
static void	flushproc(void*);
static void	writeproc(void*);

void
initdcache(u32int mem)
{
	DBlock *b, *last;
	u32int nblocks, blocksize;
	int i;
	u8int *p;

	if(mem &lt; maxblocksize * 2)
		sysfatal("need at least %d bytes for the disk cache", maxblocksize * 2);
	if(maxblocksize == 0)
		sysfatal("no max. block size given for disk cache");
	blocksize = maxblocksize;
	nblocks = mem / blocksize;
	dcache.full.l = &amp;dcache.lock;
	dcache.nblocks = nblocks;
	dcache.maxdirty = (nblocks * 2) / 3;
	trace(TraceProc, "initialize disk cache with %d blocks of %d bytes, maximum %d dirty blocks\n",
			nblocks, blocksize, dcache.maxdirty);
	dcache.size = blocksize;
	dcache.heads = MKNZ(DBlock*, HashSize);
	dcache.heap = MKNZ(DBlock*, nblocks);
	dcache.blocks = MKNZ(DBlock, nblocks);
	dcache.write = MKNZ(DBlock*, nblocks);
	dcache.mem = MKNZ(u8int, (nblocks+1+128) * blocksize);

	last = nil;
	p = (u8int*)(((uintptr)dcache.mem+blocksize-1)&amp;~(uintptr)(blocksize-1));
	for(i = 0; i &lt; nblocks; i++){
		b = &amp;dcache.blocks[i];
		b-&gt;data = &amp;p[i * blocksize];
		b-&gt;heap = TWID32;
		b-&gt;writedonechan = chancreate(sizeof(void*), 1);
		b-&gt;next = last;
		last = b;
	}

	dcache.free = last;
	dcache.nheap = 0;
	setstat(StatDcacheSize, nblocks);
	initround(&amp;dcache.round, "dcache", 120*1000);

	vtproc(flushproc, nil);
	vtproc(delaykickroundproc, &amp;dcache.round);
}

static u32int
pbhash(u64int addr)
{
	u32int h;

#define hashit(c)	((((c) * 0x6b43a9b5) &gt;&gt; (32 - HashLog)) &amp; HashMask)
	h = (addr &gt;&gt; 32) ^ addr;
	return hashit(h);
}

DBlock*
getdblock(Part *part, u64int addr, int mode)
{
	DBlock *b;
	
	b = _getdblock(part, addr, mode, 1);
	if(mode == OREAD || mode == ORDWR)
		addstat(StatDcacheRead, 1);
	if(mode == OWRITE || mode == ORDWR)
		addstat(StatDcacheWrite, 1);
	return b;
}

DBlock*
_getdblock(Part *part, u64int addr, int mode, int load)
{
	DBlock *b;
	u32int h, size, ms;

	ms = 0;
	trace(TraceBlock, "getdblock enter %s 0x%llux", part-&gt;name, addr);
	size = part-&gt;blocksize;
	if(size &gt; dcache.size){
		seterr(EAdmin, "block size %d too big for cache with size %d", size, dcache.size);
		if(load)
			addstat(StatDcacheLookup, 1);
		return nil;
	}
	h = pbhash(addr);

	/*
	 * look for the block in the cache
	 */
	qlock(&amp;dcache.lock);
again:
	for(b = dcache.heads[h]; b != nil; b = b-&gt;next){
		if(b-&gt;part == part &amp;&amp; b-&gt;addr == addr){
			if(load)
				addstat2(StatDcacheHit, 1, StatDcacheLookup, 1);
			goto found;
		}
	}

	/*
	 * missed: locate the block with the oldest second to last use.
	 * remove it from the heap, and fix up the heap.
	 */
	if(!load){
		qunlock(&amp;dcache.lock);
		return nil;
	}

	/*
	 * Only start timer here, on cache miss - calling msec() on plain cache hits
	 * makes cache hits system-call bound.
	 */
	ms = msec();
	addstat2(StatDcacheLookup, 1, StatDcacheMiss, 1);

	b = bumpdblock();
	if(b == nil){
		trace(TraceBlock, "all disk cache blocks in use");
		addstat(StatDcacheStall, 1);
		rsleep(&amp;dcache.full);
		addstat(StatDcacheStall, -1);
		goto again;
	}

	assert(!b-&gt;dirty);

	/*
	 * the new block has no last use, so assume it happens sometime in the middle
ZZZ this is not reasonable
	 */
	b-&gt;used = (b-&gt;used2 + dcache.now) / 2;

	/*
	 * rechain the block on the correct hash chain
	 */
	b-&gt;next = dcache.heads[h];
	dcache.heads[h] = b;
	if(b-&gt;next != nil)
		b-&gt;next-&gt;prev = b;
	b-&gt;prev = nil;

	b-&gt;addr = addr;
	b-&gt;part = part;
	b-&gt;size = 0;

found:
	b-&gt;ref++;
	b-&gt;used2 = b-&gt;used;
	b-&gt;used = dcache.now++;
	if(b-&gt;heap != TWID32)
		fixheap(b-&gt;heap, b);

	if((mode == ORDWR || mode == OWRITE) &amp;&amp; part-&gt;writechan == nil){
		trace(TraceBlock, "getdblock allocwriteproc %s", part-&gt;name);
		part-&gt;writechan = chancreate(sizeof(DBlock*), dcache.nblocks);
		vtproc(writeproc, part);
	}
	qunlock(&amp;dcache.lock);

	trace(TraceBlock, "getdblock lock");
	addstat(StatDblockStall, 1);
	if(mode == OREAD)
		rlock(&amp;b-&gt;lock);
	else
		wlock(&amp;b-&gt;lock);
	addstat(StatDblockStall, -1);
	trace(TraceBlock, "getdblock locked");

	if(b-&gt;size != size){
		if(mode == OREAD){
			addstat(StatDblockStall, 1);
			runlock(&amp;b-&gt;lock);
			wlock(&amp;b-&gt;lock);
			addstat(StatDblockStall, -1);
		}
		if(b-&gt;size &lt; size){
			if(mode == OWRITE)
				memset(&amp;b-&gt;data[b-&gt;size], 0, size - b-&gt;size);
			else{
				trace(TraceBlock, "getdblock readpart %s 0x%llux", part-&gt;name, addr);
				diskaccess(0);
				if(readpart(part, addr + b-&gt;size, &amp;b-&gt;data[b-&gt;size], size - b-&gt;size) &lt; 0){
					b-&gt;mode = ORDWR;	/* so putdblock wunlocks */
					putdblock(b);
					return nil;
				}
				trace(TraceBlock, "getdblock readpartdone");
				addstat(StatApartRead, 1);
				addstat(StatApartReadBytes, size-b-&gt;size);
			}
		}
		b-&gt;size = size;
		if(mode == OREAD){
			addstat(StatDblockStall, 1);
			wunlock(&amp;b-&gt;lock);
			rlock(&amp;b-&gt;lock);
			addstat(StatDblockStall, -1);
		}
	}

	b-&gt;mode = mode;
	trace(TraceBlock, "getdblock exit");
	if(ms)
		addstat(StatDcacheLookupTime, msec() - ms);
	return b;
}

void
putdblock(DBlock *b)
{
	if(b == nil)
		return;

	trace(TraceBlock, "putdblock %s 0x%llux", b-&gt;part-&gt;name, b-&gt;addr);

	if(b-&gt;mode == OREAD)
		runlock(&amp;b-&gt;lock);
	else
		wunlock(&amp;b-&gt;lock);

	qlock(&amp;dcache.lock);
	if(--b-&gt;ref == 0 &amp;&amp; !b-&gt;dirty){
		if(b-&gt;heap == TWID32)
			upheap(dcache.nheap++, b);
		rwakeupall(&amp;dcache.full);
	}
	qunlock(&amp;dcache.lock);
}

void
dirtydblock(DBlock *b, int dirty)
{
	int odirty;

	trace(TraceBlock, "dirtydblock enter %s 0x%llux %d from 0x%lux",
		b-&gt;part-&gt;name, b-&gt;addr, dirty, getcallerpc(&amp;b));
	assert(b-&gt;ref != 0);
	assert(b-&gt;mode==ORDWR || b-&gt;mode==OWRITE);

	odirty = b-&gt;dirty;
	if(b-&gt;dirty)
		assert(b-&gt;dirty == dirty);
	else
		b-&gt;dirty = dirty;

	qlock(&amp;dcache.lock);
	if(!odirty){
		dcache.ndirty++;
		setstat(StatDcacheDirty, dcache.ndirty);
		if(dcache.ndirty &gt;= dcache.maxdirty)
			kickround(&amp;dcache.round, 0);
		else
			delaykickround(&amp;dcache.round);
	}
	qunlock(&amp;dcache.lock);
}

static void
unchain(DBlock *b)
{
	ulong h;
	
	/*
	 * unchain the block
	 */
	if(b-&gt;prev == nil){
		h = pbhash(b-&gt;addr);
		if(dcache.heads[h] != b)
			sysfatal("bad hash chains in disk cache");
		dcache.heads[h] = b-&gt;next;
	}else
		b-&gt;prev-&gt;next = b-&gt;next;
	if(b-&gt;next != nil)
		b-&gt;next-&gt;prev = b-&gt;prev;
}

/*
 * remove some block from use and update the free list and counters
 */
static DBlock*
bumpdblock(void)
{
	DBlock *b;

	trace(TraceBlock, "bumpdblock enter");
	b = dcache.free;
	if(b != nil){
		dcache.free = b-&gt;next;
		return b;
	}

	if(dcache.ndirty &gt;= dcache.maxdirty)
		kickdcache();

	/*
	 * remove blocks until we find one that is unused
	 * referenced blocks are left in the heap even though
	 * they can't be scavenged; this is simple a speed optimization
	 */
	for(;;){
		if(dcache.nheap == 0){
			kickdcache();
			trace(TraceBlock, "bumpdblock gotnothing");
			return nil;
		}
		b = dcache.heap[0];
		delheap(b);
		if(!b-&gt;ref &amp;&amp; !b-&gt;dirty)
			break;
	}

	trace(TraceBlock, "bumpdblock bumping %s 0x%llux", b-&gt;part-&gt;name, b-&gt;addr);

	unchain(b);
	return b;
}

void
emptydcache(void)
{
	DBlock *b;
	
	qlock(&amp;dcache.lock);
	while(dcache.nheap &gt; 0){
		b = dcache.heap[0];
		delheap(b);
		if(!b-&gt;ref &amp;&amp; !b-&gt;dirty){
			unchain(b);
			b-&gt;next = dcache.free;
			dcache.free = b;
		}
	}
	qunlock(&amp;dcache.lock);
}

/*
 * delete an arbitrary block from the heap
 */
static void
delheap(DBlock *db)
{
	if(db-&gt;heap == TWID32)
		return;
	fixheap(db-&gt;heap, dcache.heap[--dcache.nheap]);
	db-&gt;heap = TWID32;
}

/*
 * push an element up or down to it's correct new location
 */
static void
fixheap(int i, DBlock *b)
{
	if(upheap(i, b) == i)
		downheap(i, b);
}

static int
upheap(int i, DBlock *b)
{
	DBlock *bb;
	u32int now;
	int p;

	now = dcache.now;
	for(; i != 0; i = p){
		p = (i - 1) &gt;&gt; 1;
		bb = dcache.heap[p];
		if(b-&gt;used2 - now &gt;= bb-&gt;used2 - now)
			break;
		dcache.heap[i] = bb;
		bb-&gt;heap = i;
	}

	dcache.heap[i] = b;
	b-&gt;heap = i;
	return i;
}

static int
downheap(int i, DBlock *b)
{
	DBlock *bb;
	u32int now;
	int k;

	now = dcache.now;
	for(; ; i = k){
		k = (i &lt;&lt; 1) + 1;
		if(k &gt;= dcache.nheap)
			break;
		if(k + 1 &lt; dcache.nheap &amp;&amp; dcache.heap[k]-&gt;used2 - now &gt; dcache.heap[k + 1]-&gt;used2 - now)
			k++;
		bb = dcache.heap[k];
		if(b-&gt;used2 - now &lt;= bb-&gt;used2 - now)
			break;
		dcache.heap[i] = bb;
		bb-&gt;heap = i;
	}

	dcache.heap[i] = b;
	b-&gt;heap = i;
	return i;
}

static void
findblock(DBlock *bb)
{
	DBlock *b, *last;
	int h;

	last = nil;
	h = pbhash(bb-&gt;addr);
	for(b = dcache.heads[h]; b != nil; b = b-&gt;next){
		if(last != b-&gt;prev)
			sysfatal("bad prev link");
		if(b == bb)
			return;
		last = b;
	}
	sysfatal("block missing from hash table");
}

void
checkdcache(void)
{
	DBlock *b;
	u32int size, now;
	int i, k, refed, nfree;

	qlock(&amp;dcache.lock);
	size = dcache.size;
	now = dcache.now;
	for(i = 0; i &lt; dcache.nheap; i++){
		if(dcache.heap[i]-&gt;heap != i)
			sysfatal("dc: mis-heaped at %d: %d", i, dcache.heap[i]-&gt;heap);
		if(i &gt; 0 &amp;&amp; dcache.heap[(i - 1) &gt;&gt; 1]-&gt;used2 - now &gt; dcache.heap[i]-&gt;used2 - now)
			sysfatal("dc: bad heap ordering");
		k = (i &lt;&lt; 1) + 1;
		if(k &lt; dcache.nheap &amp;&amp; dcache.heap[i]-&gt;used2 - now &gt; dcache.heap[k]-&gt;used2 - now)
			sysfatal("dc: bad heap ordering");
		k++;
		if(k &lt; dcache.nheap &amp;&amp; dcache.heap[i]-&gt;used2 - now &gt; dcache.heap[k]-&gt;used2 - now)
			sysfatal("dc: bad heap ordering");
	}

	refed = 0;
	for(i = 0; i &lt; dcache.nblocks; i++){
		b = &amp;dcache.blocks[i];
		if(b-&gt;data != &amp;dcache.mem[i * size])
			sysfatal("dc: mis-blocked at %d", i);
		if(b-&gt;ref &amp;&amp; b-&gt;heap == TWID32)
			refed++;
		if(b-&gt;addr)
			findblock(b);
		if(b-&gt;heap != TWID32
		&amp;&amp; dcache.heap[b-&gt;heap] != b)
			sysfatal("dc: spurious heap value");
	}

	nfree = 0;
	for(b = dcache.free; b != nil; b = b-&gt;next){
		if(b-&gt;addr != 0 || b-&gt;heap != TWID32)
			sysfatal("dc: bad free list");
		nfree++;
	}

	if(dcache.nheap + nfree + refed != dcache.nblocks)
		sysfatal("dc: missing blocks: %d %d %d", dcache.nheap, refed, dcache.nblocks);
	qunlock(&amp;dcache.lock);
}

void
flushdcache(void)
{
	trace(TraceProc, "flushdcache enter");
	kickround(&amp;dcache.round, 1);
	trace(TraceProc, "flushdcache exit");
}

void
kickdcache(void)
{
	kickround(&amp;dcache.round, 0);
}

static int
parallelwrites(DBlock **b, DBlock **eb, int dirty)
{
	DBlock **p, **q;
	Part *part;

	for(p=b; p&lt;eb &amp;&amp; (*p)-&gt;dirty == dirty; p++){
		assert(b&lt;=p &amp;&amp; p&lt;eb);
		sendp((*p)-&gt;part-&gt;writechan, *p);
	}
	q = p;
	for(p=b; p&lt;q; p++){
		assert(b&lt;=p &amp;&amp; p&lt;eb);
		recvp((*p)-&gt;writedonechan);
	}
	
	/*
	 * Flush the partitions that have been written to.
	 */
	part = nil;
	for(p=b; p&lt;q; p++){
		if(part != (*p)-&gt;part){
			part = (*p)-&gt;part;
			flushpart(part);	/* what if it fails? */
		}
	}

	return p-b;
}

/*
 * Sort first by dirty flag, then by partition, then by address in partition.
 */
static int
writeblockcmp(const void *va, const void *vb)
{
	DBlock *a, *b;

	a = *(DBlock**)va;
	b = *(DBlock**)vb;

	if(a-&gt;dirty != b-&gt;dirty)
		return a-&gt;dirty - b-&gt;dirty;
	if(a-&gt;part != b-&gt;part){
		if(a-&gt;part &lt; b-&gt;part)
			return -1;
		if(a-&gt;part &gt; b-&gt;part)
			return 1;
	}
	if(a-&gt;addr &lt; b-&gt;addr)
		return -1;
	return 1;
}

static void
flushproc(void *v)
{
	int i, j, n;
	ulong t0;
	DBlock *b, **write;

	USED(v);
	threadsetname("flushproc");
	for(;;){
		waitforkick(&amp;dcache.round);

		trace(TraceWork, "start");
		t0 = nsec()/1000;
		trace(TraceProc, "build t=%lud", (ulong)(nsec()/1000)-t0);

		write = dcache.write;
		n = 0;
		for(i=0; i&lt;dcache.nblocks; i++){
			b = &amp;dcache.blocks[i];
			if(b-&gt;dirty)
				write[n++] = b;
		}

		qsort(write, n, sizeof(write[0]), writeblockcmp);

		/* Write each stage of blocks out. */
		trace(TraceProc, "writeblocks t=%lud", (ulong)(nsec()/1000)-t0);
		i = 0;
		for(j=1; j&lt;DirtyMax; j++){
			trace(TraceProc, "writeblocks.%d t=%lud",
				j, (ulong)(nsec()/1000)-t0);
			i += parallelwrites(write+i, write+n, j);
		}
		if(i != n){
			fprint(2, "in flushproc i=%d n=%d\n", i, n);
			for(i=0; i&lt;n; i++)
				fprint(2, "\tblock %d: dirty=%d\n",
					i, write[i]-&gt;dirty);
			abort();
		}

		/*
		 * b-&gt;dirty is protected by b-&gt;lock while ndirty is protected
		 * by dcache.lock, so the --ndirty below is the delayed one
		 * from clearing b-&gt;dirty in the write proc.  It may happen
		 * that some other proc has come along and redirtied b since
		 * the write.  That's okay, it just means that ndirty may be
		 * one too high until we catch up and do the decrement.
		 */
		trace(TraceProc, "undirty.%d t=%lud", j, (ulong)(nsec()/1000)-t0);
		qlock(&amp;dcache.lock);
		for(i=0; i&lt;n; i++){
			b = write[i];
			--dcache.ndirty;
			if(b-&gt;ref == 0 &amp;&amp; b-&gt;heap == TWID32){
				upheap(dcache.nheap++, b);
				rwakeupall(&amp;dcache.full);
			}
		}
		setstat(StatDcacheDirty, dcache.ndirty);
		qunlock(&amp;dcache.lock);
		addstat(StatDcacheFlush, 1);
		trace(TraceWork, "finish");
	}
}

static void
writeproc(void *v)
{
	DBlock *b;
	Part *p;

	p = v;

	threadsetname("writeproc:%s", p-&gt;name);
	for(;;){
		b = recvp(p-&gt;writechan);
		trace(TraceWork, "start");
		assert(b-&gt;part == p);
		trace(TraceProc, "wlock %s 0x%llux", p-&gt;name, b-&gt;addr);
		wlock(&amp;b-&gt;lock);
		trace(TraceProc, "writepart %s 0x%llux", p-&gt;name, b-&gt;addr);
		diskaccess(0);
		if(writepart(p, b-&gt;addr, b-&gt;data, b-&gt;size) &lt; 0)
			fprint(2, "%s: writeproc: part %s addr 0x%llux: write error: %r\n",
				argv0, p-&gt;name, b-&gt;addr);
		addstat(StatApartWrite, 1);
		addstat(StatApartWriteBytes, b-&gt;size);
		b-&gt;dirty = 0;
		wunlock(&amp;b-&gt;lock);
		trace(TraceProc, "finish %s 0x%llux", p-&gt;name, b-&gt;addr);
		trace(TraceWork, "finish");
		sendp(b-&gt;writedonechan, b);
	}
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

