<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/cmd/wikifs/io.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/cmd/wikifs/io.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * I/O for a Wiki document set.
 * 
 * The files are kept in one flat directory.
 * There are three files for each document:
 *	nnn	- current version of the document
 *	nnn.hist	- history (all old versions) of the document
 *		append-only
 *	L.nnn	- write lock file for the document
 * 
 * At the moment, since we don't have read/write locks
 * in the file system, we use the L.nnn file as a read lock too.
 * It's a hack but there aren't supposed to be many readers
 * anyway.
 *
 * The nnn.hist file is in the format read by Brdwhist.
 * The nnn file is in that format too, but only contains the
 * last entry of the nnn.hist file.
 *
 * In addition to this set of files, there is an append-only
 * map file that provides a mapping between numbers and titles.
 * The map file is a sequence of lines of the form
 *	nnn Title Here
 * The lock file L.map must be held to add to the end, to
 * make sure that the numbers are allocated sequentially.
 * 
 * We assume that writes to the map file will fit in one message,
 * so that we don't have to read-lock the file.
 */

#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;bio.h&gt;
#include &lt;String.h&gt;
#include &lt;thread.h&gt;
#include "wiki.h"

enum {
	Nhash = 64,
	Mcache = 128,
};

typedef struct Wcache Wcache;
struct Wcache {
	int n;
	ulong use;
	RWLock;
	ulong tcurrent;
	ulong thist;
	Whist *hist;
	Whist *current;
	Qid qid;
	Qid qidhist;
	Wcache *hash;
};

static RWLock cachelock;
static Wcache *tab[Nhash];
int ncache;

void
closewhist(Whist *wh)
{
	int i;

	if(wh &amp;&amp; decref(wh) == 0){
		free(wh-&gt;title);
		for(i=0; i&lt;wh-&gt;ndoc; i++){
			free(wh-&gt;doc[i].author);
			free(wh-&gt;doc[i].comment);
			freepage(wh-&gt;doc[i].wtxt);
		}
		free(wh-&gt;doc);
		free(wh);
	}
}

void
freepage(Wpage *p)
{
	Wpage *next;

	for(; p; p=next){
		next = p-&gt;next;
		free(p-&gt;text);
		free(p-&gt;url);
		free(p);
	}
}

static Wcache*
findcache(int n)
{
	Wcache *w;

	for(w=tab[n%Nhash]; w; w=w-&gt;hash)
		if(w-&gt;n == n){
			w-&gt;use = time(0);
			return w;
		}
	return nil;
}

static int
getlock(char *lock)
{
	char buf[ERRMAX];
	int i, fd;
	enum { SECS = 200 };

	for(i=0; i&lt;SECS*10; i++){
		fd = wcreate(lock, ORDWR, DMEXCL|0666);
		if(fd &gt;= 0)
			return fd;
		buf[0] = '\0';
		rerrstr(buf, sizeof buf);
		if(strstr(buf, "locked") == nil)
			break;
		sleep(1000/10);
	}
	werrstr("couldn't acquire lock %s: %r", lock);
	return -1;
}

static Whist*
readwhist(char *file, char *lock, Qid *qid)
{
	int lfd;
	Biobuf *b;
	Dir *d;
	Whist *wh;

	if((lfd=getlock(lock)) &lt; 0)	// LOG?
		return nil;

	if(qid){
		if((d = wdirstat(file)) == nil){
			close(lfd);
			return nil;
		}
		*qid = d-&gt;qid;
		free(d);
	}

	if((b = wBopen(file, OREAD)) == nil){	//LOG?
		close(lfd);
		return nil;
	}

	wh = Brdwhist(b);

	Bterm(b);
	close(lfd);
	return wh;
}

static void
gencurrent(Wcache *w, Qid *q, char *file, char *lock, ulong *t, Whist **wp, int n)
{
	Dir *d;
	Whist *wh;

	if(*wp &amp;&amp; *t+Tcache &gt;= time(0))
		return;

	wlock(w);
	if(*wp &amp;&amp; *t+Tcache &gt;= time(0)){
		wunlock(w);
		return;
	}

	if(((d = wdirstat(file)) == nil) || (d-&gt;qid.path==q-&gt;path &amp;&amp; d-&gt;qid.vers==q-&gt;vers)){
		*t = time(0);
		wunlock(w);
		free(d);
		return;
	}

	free(d);
	if(wh = readwhist(file, lock, q)){
		wh-&gt;n = n;
		*t = time(0);
		closewhist(*wp);
		*wp = wh;
	}
else fprint(2, "error file=%s lock=%s %r\n", file, lock);
	wunlock(w);
}

static void
current(Wcache *w)
{
	char tmp[40];
	char tmplock[40];

	sprint(tmplock, "d/L.%d", w-&gt;n);
	sprint(tmp, "d/%d", w-&gt;n);
	gencurrent(w, &amp;w-&gt;qid, tmp, tmplock, &amp;w-&gt;tcurrent, &amp;w-&gt;current, w-&gt;n);
}

static void
currenthist(Wcache *w)
{
	char hist[40], lock[40];

	sprint(hist, "d/%d.hist", w-&gt;n);
	sprint(lock, "d/L.%d", w-&gt;n);

	gencurrent(w, &amp;w-&gt;qidhist, hist, lock, &amp;w-&gt;thist, &amp;w-&gt;hist, w-&gt;n);
}

void
voidcache(int n)
{
	Wcache *c;

	rlock(&amp;cachelock);
	if(c = findcache(n)){
		wlock(c);
		c-&gt;tcurrent = 0;
		c-&gt;thist = 0;
		/* aggressively free memory */
		closewhist(c-&gt;hist);
		c-&gt;hist = nil;
		closewhist(c-&gt;current);
		c-&gt;current = nil;
		wunlock(c);
	}
	runlock(&amp;cachelock);
}

static Whist*
getcache(int n, int hist)
{
	int i, isw;
	ulong t;
	Wcache *c, **cp, **evict;
	Whist *wh;

	isw = 0;
	rlock(&amp;cachelock);
	if(c = findcache(n)){
	Found:
		current(c);
		if(hist)
			currenthist(c);
		rlock(c);
		if(hist)
			wh = c-&gt;hist;
		else
			wh = c-&gt;current;
		if(wh)
			incref(wh);
		runlock(c);
		if(isw)
			wunlock(&amp;cachelock);
		else
			runlock(&amp;cachelock);
		return wh;
	}
	runlock(&amp;cachelock);

	wlock(&amp;cachelock);
	if(c = findcache(n)){
		isw = 1;	/* better to downgrade lock but can't */
		goto Found;
	}

	if(ncache &lt; Mcache){
	Alloc:
		c = emalloc(sizeof *c);
		ncache++;
	}else{
		/* find something to evict. */
		t = ~0;
		evict = nil;
		for(i=0; i&lt;Nhash; i++){
			for(cp=&amp;tab[i], c=*cp; c; cp=&amp;c-&gt;hash, c=*cp){
				if(c-&gt;use &lt; t
				&amp;&amp; (!c-&gt;hist || c-&gt;hist-&gt;ref==1)
				&amp;&amp; (!c-&gt;current || c-&gt;current-&gt;ref==1)){
					evict = cp;
					t = c-&gt;use;
				}
			}
		}

		if(evict == nil){
			fprint(2, "wikifs: nothing to evict\n");
			goto Alloc;
		}

		c = *evict;
		*evict = c-&gt;hash;

		closewhist(c-&gt;current);
		closewhist(c-&gt;hist);
		memset(c, 0, sizeof *c);
	}

	c-&gt;n = n;
	c-&gt;hash = tab[n%Nhash];
	tab[n%Nhash] = c;
	isw = 1;
	goto Found;
}

Whist*
getcurrent(int n)
{
	return getcache(n, 0);
}

Whist*
gethistory(int n)
{
	return getcache(n, 1);
}

RWLock maplock;
Map *map;

static int
mapcmp(const void *va, const void *vb)
{
	Mapel *a, *b;

	a = (Mapel*)va;
	b = (Mapel*)vb;

	return strcmp(a-&gt;s, b-&gt;s);
}

void
closemap(Map *m)
{
	if(decref(m)==0){
		free(m-&gt;buf);
		free(m-&gt;el);
		free(m);
	}
}

void
currentmap(int force)
{
	char *p, *q, *r;
	int lfd, fd, m, n;
	Dir *d;
	Map *nmap;
	char *err = nil;

	lfd = -1;
	fd = -1;
	d = nil;
	nmap = nil;
	if(!force &amp;&amp; map &amp;&amp; map-&gt;t+Tcache &gt;= time(0))
		return;

	wlock(&amp;maplock);
	if(!force &amp;&amp; map &amp;&amp; map-&gt;t+Tcache &gt;= time(0))
		goto Return;

	if((lfd = getlock("d/L.map")) &lt; 0){
		err = "can't lock";
		goto Return;
	}

	if((d = wdirstat("d/map")) == nil)
		goto Return;

	if(map &amp;&amp; d-&gt;qid.path == map-&gt;qid.path &amp;&amp; d-&gt;qid.vers == map-&gt;qid.vers){
		map-&gt;t = time(0);
		goto Return;
	}

	if(d-&gt;length &gt; Maxmap){
		//LOG
		err = "too long";
		goto Return;
	}

	if((fd = wopen("d/map", OREAD)) &lt; 0)
		goto Return;

	nmap = emalloc(sizeof *nmap);
	nmap-&gt;buf = emalloc(d-&gt;length+1);
	n = readn(fd, nmap-&gt;buf, d-&gt;length);
	if(n != d-&gt;length){
		err = "bad length";
		goto Return;
	}
	nmap-&gt;buf[n] = '\0';

	n = 0;
	for(p=nmap-&gt;buf; p; p=strchr(p+1, '\n'))
		n++;
	nmap-&gt;el = emalloc(n*sizeof(nmap-&gt;el[0]));

	m = 0;
	for(p=nmap-&gt;buf; p &amp;&amp; *p &amp;&amp; m &lt; n; p=q){
		if(q = strchr(p+1, '\n'))
			*q++ = '\0';
		nmap-&gt;el[m].n = strtol(p, &amp;r, 10);
		if(*r == ' ')
			r++;
		else
			{}//LOG?
		nmap-&gt;el[m].s = strcondense(r, 1);
		m++;
	}
	//LOG if m != n

	nmap-&gt;qid = d-&gt;qid;
	nmap-&gt;t = time(0);
	nmap-&gt;nel = m;
	qsort(nmap-&gt;el, nmap-&gt;nel, sizeof(nmap-&gt;el[0]), mapcmp);
	if(map)
		closemap(map);
	map = nmap;
	incref(map);
	nmap = nil;
	
Return:
	free(d);
	if(nmap){
		free(nmap-&gt;el);
		free(nmap-&gt;buf);
		free(nmap);
	}
	if(map == nil)
		sysfatal("cannot get map: %s: %r", err);

	if(fd &gt;= 0)
		close(fd);
	if(lfd &gt;= 0)
		close(lfd);
	wunlock(&amp;maplock);
}

int
allocnum(char *title, int mustbenew)
{
	char *p, *q;
	int lfd, fd, n;
	Biobuf b;

	if(strcmp(title, "map")==0 || strcmp(title, "new")==0){
		werrstr("reserved title name");
		return -1;
	}

	if(title[0]=='\0' || strpbrk(title, "/&lt;&gt;:?")){
		werrstr("invalid character in name");
		return -1;
	}
	if((n = nametonum(title)) &gt;= 0){
		if(mustbenew){
			werrstr("duplicate title");
			return -1;
		}
		return n;
	}

	title = estrdup(title);
	strcondense(title, 1);
	strlower(title);
	if(strchr(title, '\n') || strlen(title) &gt; 200){
		werrstr("bad title");
		free(title);
		return -1;
	}

	if((lfd = getlock("d/L.map")) &lt; 0){
		free(title);
		return -1;
	}

	if((fd = wopen("d/map", ORDWR)) &lt; 0){	// LOG?
		close(lfd);
		free(title);
		return -1;
	}

	/*
	 * What we really need to do here is make sure the 
	 * map is up-to-date, then make sure the title isn't
	 * taken, and then add it, all without dropping the locks.
	 *
	 * This turns out to be a mess when you start adding
	 * all the necessary dolock flags, so instead we just
	 * read through the file ourselves, and let our 
	 * map catch up on its own.
	 */
	Binit(&amp;b, fd, OREAD);
	n = 0;
	while(p = Brdline(&amp;b, '\n')){
		p[Blinelen(&amp;b)-1] = '\0';
		n = atoi(p)+1;
		q = strchr(p, ' ');
		if(q == nil)
			continue;
		if(strcmp(q+1, title) == 0){
			free(title);
			close(fd);
			close(lfd);
			if(mustbenew){
				werrstr("duplicate title");
				return -1;
			}else
				return n;
		}
	}

	seek(fd, 0, 2);	/* just in case it's not append only */
	fprint(fd, "%d %s\n", n, title);
	close(fd);
	close(lfd);
	free(title);
	/* kick the map */
	currentmap(1);
	return n;
}

int
nametonum(char *s)
{
	char *p;
	int i, lo, hi, m, rv;

	s = estrdup(s);
	strlower(s);
	for(p=s; *p; p++)
		if(*p=='_')
			*p = ' ';

	currentmap(0);
	rlock(&amp;maplock);
	lo = 0;
	hi = map-&gt;nel;
	while(hi-lo &gt; 1){
		m = (lo+hi)/2;
		i = strcmp(s, map-&gt;el[m].s);
		if(i &lt; 0)
			hi = m;
		else
			lo = m;
	}
	if(hi-lo == 1 &amp;&amp; strcmp(s, map-&gt;el[lo].s)==0)
		rv = map-&gt;el[lo].n;
	else
		rv = -1;
	runlock(&amp;maplock);
	free(s);
	return rv;
}

char*
numtoname(int n)
{
	int i;
	char *s;

	currentmap(0);
	rlock(&amp;maplock);
	for(i=0; i&lt;map-&gt;nel; i++){
		if(map-&gt;el[i].n==n)
			break;
	}
	if(i==map-&gt;nel){
		runlock(&amp;maplock);
		return nil;
	}
	s = estrdup(map-&gt;el[i].s);
	runlock(&amp;maplock);
	return s;
}

Whist*
getcurrentbyname(char *s)
{
	int n;

	if((n = nametonum(s)) &lt; 0)
		return nil;
	return getcache(n, 0);
}

static String*
Brdstring(Biobuf *b)
{
	long len;
	String *s;
	Dir *d;

	d = dirfstat(Bfildes(b));
	if (d == nil)	/* shouldn't happen, we just opened it */
		len = 0;
	else
		len = d-&gt;length;
	free(d);
	s = s_newalloc(len);
	s_read(b, s, len);
	return s;
}

/*
 * Attempt to install a new page.  If t==0 we are creating.
 * Otherwise, we are editing and t must be set to the current
 * version (t is the version we started with) to avoid conflicting
 * writes.
 *
 * If there is a conflicting write, we still write the page to 
 * the history file, but mark it as a failed write.
 */
int
writepage(int num, ulong t, String *s, char *title)
{
	char tmp[40], tmplock[40], err[ERRMAX], hist[40], *p;
	int conflict, lfd, fd;
	Biobuf *b;
	String *os;

	sprint(tmp, "d/%d", num);
	sprint(tmplock, "d/L.%d", num);
	sprint(hist, "d/%d.hist", num);
	if((lfd = getlock(tmplock)) &lt; 0)
		return -1;

	conflict = 0;
	if(b = wBopen(tmp, OREAD)){
		Brdline(b, '\n');	/* title */
		if(p = Brdline(b, '\n'))		/* version */
			p[Blinelen(b)-1] = '\0';
		if(p==nil || p[0] != 'D'){
			snprint(err, sizeof err, "bad format in extant file");
			conflict = 1;
		}else if(strtoul(p+1, 0, 0) != t){
			os = Brdstring(b);	/* why read the whole file? */
			p = strchr(s_to_c(s), '\n');
			if(p!=nil &amp;&amp; strcmp(p+1, s_to_c(os))==0){	/* ignore dup write */
				close(lfd);
				s_free(os);
				Bterm(b);
				return 0;
			}
			s_free(os);
			snprint(err, sizeof err, "update conflict %lud != %s", t, p+1);
			conflict = 1;
		}
		Bterm(b);
	}else{
		if(t != 0){
			close(lfd);
			werrstr("did not expect to create");
			return -1;
		}
	}

	if((fd = wopen(hist, OWRITE)) &lt; 0){
		if((fd = wcreate(hist, OWRITE, 0666)) &lt; 0){
			close(lfd);
			return -1;
		}else
			fprint(fd, "%s\n", title);
	}
	if(seek(fd, 0, 2) &lt; 0
	|| (conflict &amp;&amp; write(fd, "X\n", 2) != 2)
	|| write(fd, s_to_c(s), s_len(s)) != s_len(s)){
		close(fd);
		close(lfd);
		return -1;
	}
	close(fd);

	if(conflict){
		close(lfd);
		voidcache(num);
		werrstr(err);
		return -1;
	}

	if((fd = wcreate(tmp, OWRITE, 0666)) &lt; 0){
		close(lfd);
		voidcache(num);
		return -1;
	}
	if(write(fd, title, strlen(title)) != strlen(title)
	|| write(fd, "\n", 1) != 1
	|| write(fd, s_to_c(s), s_len(s)) != s_len(s)){
		close(fd);
		close(lfd);
		voidcache(num);
		return -1;
	}
	close(fd);
	close(lfd);
	voidcache(num);
	return 0;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

