<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/libgeometry/tstack.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/libgeometry/tstack.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*% cc -gpc %
 * These transformation routines maintain stacks of transformations
 * and their inverses.  
 * t=pushmat(t)		push matrix stack
 * t=popmat(t)		pop matrix stack
 * rot(t, a, axis)	multiply stack top by rotation
 * qrot(t, q)		multiply stack top by rotation, q is unit quaternion
 * scale(t, x, y, z)	multiply stack top by scale
 * move(t, x, y, z)	multiply stack top by translation
 * xform(t, m)		multiply stack top by m
 * ixform(t, m, inv)	multiply stack top by m.  inv is the inverse of m.
 * look(t, e, l, u)	multiply stack top by viewing transformation
 * persp(t, fov, n, f)	multiply stack top by perspective transformation
 * viewport(t, r, aspect)
 *			multiply stack top by window-&gt;viewport transformation.
 */
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;draw.h&gt;
#include &lt;geometry.h&gt;
Space *pushmat(Space *t){
	Space *v;
	v=malloc(sizeof(Space));
	if(t==0){
		ident(v-&gt;t);
		ident(v-&gt;tinv);
	}
	else
		*v=*t;
	v-&gt;next=t;
	return v;
}
Space *popmat(Space *t){
	Space *v;
	if(t==0) return 0;
	v=t-&gt;next;
	free(t);
	return v;
}
void rot(Space *t, double theta, int axis){
	double s=sin(radians(theta)), c=cos(radians(theta));
	Matrix m, inv;
	register i=(axis+1)%3, j=(axis+2)%3;
	ident(m);
	m[i][i] = c;
	m[i][j] = -s;
	m[j][i] = s;
	m[j][j] = c;
	ident(inv);
	inv[i][i] = c;
	inv[i][j] = s;
	inv[j][i] = -s;
	inv[j][j] = c;
	ixform(t, m, inv);
}
void qrot(Space *t, Quaternion q){
	Matrix m, inv;
	int i, j;
	qtom(m, q);
	for(i=0;i!=4;i++) for(j=0;j!=4;j++) inv[i][j]=m[j][i];
	ixform(t, m, inv);
}
void scale(Space *t, double x, double y, double z){
	Matrix m, inv;
	ident(m);
	m[0][0]=x;
	m[1][1]=y;
	m[2][2]=z;
	ident(inv);
	inv[0][0]=1/x;
	inv[1][1]=1/y;
	inv[2][2]=1/z;
	ixform(t, m, inv);
}
void move(Space *t, double x, double y, double z){
	Matrix m, inv;
	ident(m);
	m[0][3]=x;
	m[1][3]=y;
	m[2][3]=z;
	ident(inv);
	inv[0][3]=-x;
	inv[1][3]=-y;
	inv[2][3]=-z;
	ixform(t, m, inv);
}
void xform(Space *t, Matrix m){
	Matrix inv;
	if(invertmat(m, inv)==0) return;
	ixform(t, m, inv);
}
void ixform(Space *t, Matrix m, Matrix inv){
	matmul(t-&gt;t, m);
	matmulr(t-&gt;tinv, inv);
}
/*
 * multiply the top of the matrix stack by a view-pointing transformation
 * with the eyepoint at e, looking at point l, with u at the top of the screen.
 * The coordinate system is deemed to be right-handed.
 * The generated transformation transforms this view into a view from
 * the origin, looking in the positive y direction, with the z axis pointing up,
 * and x to the right.
 */
void look(Space *t, Point3 e, Point3 l, Point3 u){
	Matrix m, inv;
	Point3 r;
	l=unit3(sub3(l, e));
	u=unit3(vrem3(sub3(u, e), l));
	r=cross3(l, u);
	/* make the matrix to transform from (rlu) space to (xyz) space */
	ident(m);
	m[0][0]=r.x; m[0][1]=r.y; m[0][2]=r.z;
	m[1][0]=l.x; m[1][1]=l.y; m[1][2]=l.z;
	m[2][0]=u.x; m[2][1]=u.y; m[2][2]=u.z;
	ident(inv);
	inv[0][0]=r.x; inv[0][1]=l.x; inv[0][2]=u.x;
	inv[1][0]=r.y; inv[1][1]=l.y; inv[1][2]=u.y;
	inv[2][0]=r.z; inv[2][1]=l.z; inv[2][2]=u.z;
	ixform(t, m, inv);
	move(t, -e.x, -e.y, -e.z);
}
/*
 * generate a transformation that maps the frustum with apex at the origin,
 * apex angle=fov and clipping planes y=n and y=f into the double-unit cube.
 * plane y=n maps to y'=-1, y=f maps to y'=1
 */
int persp(Space *t, double fov, double n, double f){
	Matrix m;
	double z;
	if(n&lt;=0 || f&lt;=n || fov&lt;=0 || 180&lt;=fov) /* really need f!=n &amp;&amp; sin(v)!=0 */
		return -1;
	z=1/tan(radians(fov)/2);
	m[0][0]=z; m[0][1]=0;           m[0][2]=0; m[0][3]=0;
	m[1][0]=0; m[1][1]=(f+n)/(f-n); m[1][2]=0; m[1][3]=f*(1-m[1][1]);
	m[2][0]=0; m[2][1]=0;           m[2][2]=z; m[2][3]=0;
	m[3][0]=0; m[3][1]=1;           m[3][2]=0; m[3][3]=0;
	xform(t, m);
	return 0;
}
/*
 * Map the unit-cube window into the given screen viewport.
 * r has min at the top left, max just outside the lower right.  Aspect is the
 * aspect ratio (dx/dy) of the viewport's pixels (not of the whole viewport!)
 * The whole window is transformed to fit centered inside the viewport with equal
 * slop on either top and bottom or left and right, depending on the viewport's
 * aspect ratio.
 * The window is viewed down the y axis, with x to the left and z up.  The viewport
 * has x increasing to the right and y increasing down.  The window's y coordinates
 * are mapped, unchanged, into the viewport's z coordinates.
 */
void viewport(Space *t, Rectangle r, double aspect){
	Matrix m;
	double xc, yc, wid, hgt, scale;
	xc=.5*(r.min.x+r.max.x);
	yc=.5*(r.min.y+r.max.y);
	wid=(r.max.x-r.min.x)*aspect;
	hgt=r.max.y-r.min.y;
	scale=.5*(wid&lt;hgt?wid:hgt);
	ident(m);
	m[0][0]=scale;
	m[0][3]=xc;
	m[1][1]=0;
	m[1][2]=-scale;
	m[1][3]=yc;
	m[2][1]=1;
	m[2][2]=0;
	/* should get inverse by hand */
	xform(t, m);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

