<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/libsec/port/tlshand.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/libsec/port/tlshand.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;bio.h&gt;
#include &lt;auth.h&gt;
#include &lt;mp.h&gt;
#include &lt;libsec.h&gt;

// The main groups of functions are:
//		client/server - main handshake protocol definition
//		message functions - formating handshake messages
//		cipher choices - catalog of digest and encrypt algorithms
//		security functions - PKCS#1, sslHMAC, session keygen
//		general utility functions - malloc, serialization
// The handshake protocol builds on the TLS/SSL3 record layer protocol,
// which is implemented in kernel device #a.  See also /lib/rfc/rfc2246.

enum {
	TLSFinishedLen = 12,
	SSL3FinishedLen = MD5dlen+SHA1dlen,
	MaxKeyData = 136,	// amount of secret we may need
	MaxChunk = 1&lt;&lt;14,
	RandomSize = 32,
	SidSize = 32,
	MasterSecretSize = 48,
	AQueue = 0,
	AFlush = 1,
};

typedef struct TlsSec TlsSec;

typedef struct Bytes{
	int len;
	uchar data[1];  // [len]
} Bytes;

typedef struct Ints{
	int len;
	int data[1];  // [len]
} Ints;

typedef struct Algs{
	char *enc;
	char *digest;
	int nsecret;
	int tlsid;
	int ok;
} Algs;

typedef struct Finished{
	uchar verify[SSL3FinishedLen];
	int n;
} Finished;

typedef struct TlsConnection{
	TlsSec *sec;	// security management goo
	int hand, ctl;	// record layer file descriptors
	int erred;		// set when tlsError called
	int (*trace)(char*fmt, ...); // for debugging
	int version;	// protocol we are speaking
	int verset;		// version has been set
	int ver2hi;		// server got a version 2 hello
	int isClient;	// is this the client or server?
	Bytes *sid;		// SessionID
	Bytes *cert;	// only last - no chain

	Lock statelk;
	int state;		// must be set using setstate

	// input buffer for handshake messages
	uchar buf[MaxChunk+2048];
	uchar *rp, *ep;

	uchar crandom[RandomSize];	// client random
	uchar srandom[RandomSize];	// server random
	int clientVersion;	// version in ClientHello
	char *digest;	// name of digest algorithm to use
	char *enc;		// name of encryption algorithm to use
	int nsecret;	// amount of secret data to init keys

	// for finished messages
	MD5state	hsmd5;	// handshake hash
	SHAstate	hssha1;	// handshake hash
	Finished	finished;
} TlsConnection;

typedef struct Msg{
	int tag;
	union {
		struct {
			int version;
			uchar 	random[RandomSize];
			Bytes*	sid;
			Ints*	ciphers;
			Bytes*	compressors;
		} clientHello;
		struct {
			int version;
			uchar 	random[RandomSize];
			Bytes*	sid;
			int cipher;
			int compressor;
		} serverHello;
		struct {
			int ncert;
			Bytes **certs;
		} certificate;
		struct {
			Bytes *types;
			int nca;
			Bytes **cas;
		} certificateRequest;
		struct {
			Bytes *key;
		} clientKeyExchange;
		Finished finished;
	} u;
} Msg;

typedef struct TlsSec{
	char *server;	// name of remote; nil for server
	int ok;	// &lt;0 killed; == 0 in progress; &gt;0 reusable
	RSApub *rsapub;
	AuthRpc *rpc;	// factotum for rsa private key
	uchar sec[MasterSecretSize];	// master secret
	uchar crandom[RandomSize];	// client random
	uchar srandom[RandomSize];	// server random
	int clientVers;		// version in ClientHello
	int vers;			// final version
	// byte generation and handshake checksum
	void (*prf)(uchar*, int, uchar*, int, char*, uchar*, int, uchar*, int);
	void (*setFinished)(TlsSec*, MD5state, SHAstate, uchar*, int);
	int nfin;
} TlsSec;


enum {
	TLSVersion = 0x0301,
	SSL3Version = 0x0300,
	ProtocolVersion = 0x0301,	// maximum version we speak
	MinProtoVersion = 0x0300,	// limits on version we accept
	MaxProtoVersion	= 0x03ff,
};

// handshake type
enum {
	HHelloRequest,
	HClientHello,
	HServerHello,
	HSSL2ClientHello = 9,  /* local convention;  see devtls.c */
	HCertificate = 11,
	HServerKeyExchange,
	HCertificateRequest,
	HServerHelloDone,
	HCertificateVerify,
	HClientKeyExchange,
	HFinished = 20,
	HMax
};

// alerts
enum {
	ECloseNotify = 0,
	EUnexpectedMessage = 10,
	EBadRecordMac = 20,
	EDecryptionFailed = 21,
	ERecordOverflow = 22,
	EDecompressionFailure = 30,
	EHandshakeFailure = 40,
	ENoCertificate = 41,
	EBadCertificate = 42,
	EUnsupportedCertificate = 43,
	ECertificateRevoked = 44,
	ECertificateExpired = 45,
	ECertificateUnknown = 46,
	EIllegalParameter = 47,
	EUnknownCa = 48,
	EAccessDenied = 49,
	EDecodeError = 50,
	EDecryptError = 51,
	EExportRestriction = 60,
	EProtocolVersion = 70,
	EInsufficientSecurity = 71,
	EInternalError = 80,
	EUserCanceled = 90,
	ENoRenegotiation = 100,
	EMax = 256
};

// cipher suites
enum {
	TLS_NULL_WITH_NULL_NULL	 		= 0x0000,
	TLS_RSA_WITH_NULL_MD5 			= 0x0001,
	TLS_RSA_WITH_NULL_SHA 			= 0x0002,
	TLS_RSA_EXPORT_WITH_RC4_40_MD5 		= 0x0003,
	TLS_RSA_WITH_RC4_128_MD5 		= 0x0004,
	TLS_RSA_WITH_RC4_128_SHA 		= 0x0005,
	TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5	= 0X0006,
	TLS_RSA_WITH_IDEA_CBC_SHA 		= 0X0007,
	TLS_RSA_EXPORT_WITH_DES40_CBC_SHA	= 0X0008,
	TLS_RSA_WITH_DES_CBC_SHA		= 0X0009,
	TLS_RSA_WITH_3DES_EDE_CBC_SHA		= 0X000A,
	TLS_DH_DSS_EXPORT_WITH_DES40_CBC_SHA	= 0X000B,
	TLS_DH_DSS_WITH_DES_CBC_SHA		= 0X000C,
	TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA	= 0X000D,
	TLS_DH_RSA_EXPORT_WITH_DES40_CBC_SHA	= 0X000E,
	TLS_DH_RSA_WITH_DES_CBC_SHA		= 0X000F,
	TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA	= 0X0010,
	TLS_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA	= 0X0011,
	TLS_DHE_DSS_WITH_DES_CBC_SHA		= 0X0012,
	TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA	= 0X0013,	// ZZZ must be implemented for tls1.0 compliance
	TLS_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA	= 0X0014,
	TLS_DHE_RSA_WITH_DES_CBC_SHA		= 0X0015,
	TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA	= 0X0016,
	TLS_DH_anon_EXPORT_WITH_RC4_40_MD5	= 0x0017,
	TLS_DH_anon_WITH_RC4_128_MD5 		= 0x0018,
	TLS_DH_anon_EXPORT_WITH_DES40_CBC_SHA	= 0X0019,
	TLS_DH_anon_WITH_DES_CBC_SHA		= 0X001A,
	TLS_DH_anon_WITH_3DES_EDE_CBC_SHA	= 0X001B,

	TLS_RSA_WITH_AES_128_CBC_SHA		= 0X002f,	// aes, aka rijndael with 128 bit blocks
	TLS_DH_DSS_WITH_AES_128_CBC_SHA		= 0X0030,
	TLS_DH_RSA_WITH_AES_128_CBC_SHA		= 0X0031,
	TLS_DHE_DSS_WITH_AES_128_CBC_SHA	= 0X0032,
	TLS_DHE_RSA_WITH_AES_128_CBC_SHA	= 0X0033,
	TLS_DH_anon_WITH_AES_128_CBC_SHA	= 0X0034,
	TLS_RSA_WITH_AES_256_CBC_SHA		= 0X0035,
	TLS_DH_DSS_WITH_AES_256_CBC_SHA		= 0X0036,
	TLS_DH_RSA_WITH_AES_256_CBC_SHA		= 0X0037,
	TLS_DHE_DSS_WITH_AES_256_CBC_SHA	= 0X0038,
	TLS_DHE_RSA_WITH_AES_256_CBC_SHA	= 0X0039,
	TLS_DH_anon_WITH_AES_256_CBC_SHA	= 0X003A,
	CipherMax
};

// compression methods
enum {
	CompressionNull = 0,
	CompressionMax
};

static Algs cipherAlgs[] = {
	{"rc4_128", "md5", 2*(16+MD5dlen), TLS_RSA_WITH_RC4_128_MD5},
	{"rc4_128", "sha1", 2*(16+SHA1dlen), TLS_RSA_WITH_RC4_128_SHA},
	{"3des_ede_cbc", "sha1", 2*(4*8+SHA1dlen), TLS_RSA_WITH_3DES_EDE_CBC_SHA},
	{"aes_128_cbc", "sha1", 2*(16+16+SHA1dlen), TLS_RSA_WITH_AES_128_CBC_SHA},
	{"aes_256_cbc", "sha1", 2*(32+16+SHA1dlen), TLS_RSA_WITH_AES_256_CBC_SHA}
};

static uchar compressors[] = {
	CompressionNull,
};

static TlsConnection *tlsServer2(int ctl, int hand, uchar *cert, int ncert, int (*trace)(char*fmt, ...), PEMChain *chain);
static TlsConnection *tlsClient2(int ctl, int hand, uchar *csid, int ncsid, int (*trace)(char*fmt, ...));

static void	msgClear(Msg *m);
static char* msgPrint(char *buf, int n, Msg *m);
static int	msgRecv(TlsConnection *c, Msg *m);
static int	msgSend(TlsConnection *c, Msg *m, int act);
static void	tlsError(TlsConnection *c, int err, char *msg, ...);
#pragma	varargck argpos	tlsError 3
static int setVersion(TlsConnection *c, int version);
static int finishedMatch(TlsConnection *c, Finished *f);
static void tlsConnectionFree(TlsConnection *c);

static int setAlgs(TlsConnection *c, int a);
static int okCipher(Ints *cv);
static int okCompression(Bytes *cv);
static int initCiphers(void);
static Ints* makeciphers(void);

static TlsSec* tlsSecInits(int cvers, uchar *csid, int ncsid, uchar *crandom, uchar *ssid, int *nssid, uchar *srandom);
static int	tlsSecSecrets(TlsSec *sec, int vers, uchar *epm, int nepm, uchar *kd, int nkd);
static TlsSec*	tlsSecInitc(int cvers, uchar *crandom);
static int	tlsSecSecretc(TlsSec *sec, uchar *sid, int nsid, uchar *srandom, uchar *cert, int ncert, int vers, uchar **epm, int *nepm, uchar *kd, int nkd);
static int	tlsSecFinished(TlsSec *sec, MD5state md5, SHAstate sha1, uchar *fin, int nfin, int isclient);
static void	tlsSecOk(TlsSec *sec);
static void	tlsSecKill(TlsSec *sec);
static void	tlsSecClose(TlsSec *sec);
static void	setMasterSecret(TlsSec *sec, Bytes *pm);
static void	serverMasterSecret(TlsSec *sec, uchar *epm, int nepm);
static void	setSecrets(TlsSec *sec, uchar *kd, int nkd);
static int	clientMasterSecret(TlsSec *sec, RSApub *pub, uchar **epm, int *nepm);
static Bytes *pkcs1_encrypt(Bytes* data, RSApub* key, int blocktype);
static Bytes *pkcs1_decrypt(TlsSec *sec, uchar *epm, int nepm);
static void	tlsSetFinished(TlsSec *sec, MD5state hsmd5, SHAstate hssha1, uchar *finished, int isClient);
static void	sslSetFinished(TlsSec *sec, MD5state hsmd5, SHAstate hssha1, uchar *finished, int isClient);
static void	sslPRF(uchar *buf, int nbuf, uchar *key, int nkey, char *label,
			uchar *seed0, int nseed0, uchar *seed1, int nseed1);
static int setVers(TlsSec *sec, int version);

static AuthRpc* factotum_rsa_open(uchar *cert, int certlen);
static mpint* factotum_rsa_decrypt(AuthRpc *rpc, mpint *cipher);
static void factotum_rsa_close(AuthRpc*rpc);

static void* emalloc(int);
static void* erealloc(void*, int);
static void put32(uchar *p, u32int);
static void put24(uchar *p, int);
static void put16(uchar *p, int);
static u32int get32(uchar *p);
static int get24(uchar *p);
static int get16(uchar *p);
static Bytes* newbytes(int len);
static Bytes* makebytes(uchar* buf, int len);
static void freebytes(Bytes* b);
static Ints* newints(int len);
static Ints* makeints(int* buf, int len);
static void freeints(Ints* b);

//================= client/server ========================

//	push TLS onto fd, returning new (application) file descriptor
//		or -1 if error.
int
tlsServer(int fd, TLSconn *conn)
{
	char buf[8];
	char dname[64];
	int n, data, ctl, hand;
	TlsConnection *tls;

	if(conn == nil)
		return -1;
	ctl = open("#a/tls/clone", ORDWR);
	if(ctl &lt; 0)
		return -1;
	n = read(ctl, buf, sizeof(buf)-1);
	if(n &lt; 0){
		close(ctl);
		return -1;
	}
	buf[n] = 0;
	sprint(conn-&gt;dir, "#a/tls/%s", buf);
	sprint(dname, "#a/tls/%s/hand", buf);
	hand = open(dname, ORDWR);
	if(hand &lt; 0){
		close(ctl);
		return -1;
	}
	fprint(ctl, "fd %d 0x%x", fd, ProtocolVersion);
	tls = tlsServer2(ctl, hand, conn-&gt;cert, conn-&gt;certlen, conn-&gt;trace, conn-&gt;chain);
	sprint(dname, "#a/tls/%s/data", buf);
	data = open(dname, ORDWR);
	close(fd);
	close(hand);
	close(ctl);
	if(data &lt; 0){
		return -1;
	}
	if(tls == nil){
		close(data);
		return -1;
	}
	if(conn-&gt;cert)
		free(conn-&gt;cert);
	conn-&gt;cert = 0;  // client certificates are not yet implemented
	conn-&gt;certlen = 0;
	conn-&gt;sessionIDlen = tls-&gt;sid-&gt;len;
	conn-&gt;sessionID = emalloc(conn-&gt;sessionIDlen);
	memcpy(conn-&gt;sessionID, tls-&gt;sid-&gt;data, conn-&gt;sessionIDlen);
	if(conn-&gt;sessionKey != nil &amp;&amp; conn-&gt;sessionType != nil &amp;&amp; strcmp(conn-&gt;sessionType, "ttls") == 0)
		tls-&gt;sec-&gt;prf(conn-&gt;sessionKey, conn-&gt;sessionKeylen, tls-&gt;sec-&gt;sec, MasterSecretSize, conn-&gt;sessionConst,  tls-&gt;sec-&gt;crandom, RandomSize, tls-&gt;sec-&gt;srandom, RandomSize);
	tlsConnectionFree(tls);
	return data;
}

//	push TLS onto fd, returning new (application) file descriptor
//		or -1 if error.
int
tlsClient(int fd, TLSconn *conn)
{
	char buf[8];
	char dname[64];
	int n, data, ctl, hand;
	TlsConnection *tls;

	if(!conn)
		return -1;
	ctl = open("#a/tls/clone", ORDWR);
	if(ctl &lt; 0)
		return -1;
	n = read(ctl, buf, sizeof(buf)-1);
	if(n &lt; 0){
		close(ctl);
		return -1;
	}
	buf[n] = 0;
	sprint(conn-&gt;dir, "#a/tls/%s", buf);
	sprint(dname, "#a/tls/%s/hand", buf);
	hand = open(dname, ORDWR);
	if(hand &lt; 0){
		close(ctl);
		return -1;
	}
	sprint(dname, "#a/tls/%s/data", buf);
	data = open(dname, ORDWR);
	if(data &lt; 0)
		return -1;
	fprint(ctl, "fd %d 0x%x", fd, ProtocolVersion);
	tls = tlsClient2(ctl, hand, conn-&gt;sessionID, conn-&gt;sessionIDlen, conn-&gt;trace);
	close(fd);
	close(hand);
	close(ctl);
	if(tls == nil){
		close(data);
		return -1;
	}
	conn-&gt;certlen = tls-&gt;cert-&gt;len;
	conn-&gt;cert = emalloc(conn-&gt;certlen);
	memcpy(conn-&gt;cert, tls-&gt;cert-&gt;data, conn-&gt;certlen);
	conn-&gt;sessionIDlen = tls-&gt;sid-&gt;len;
	conn-&gt;sessionID = emalloc(conn-&gt;sessionIDlen);
	memcpy(conn-&gt;sessionID, tls-&gt;sid-&gt;data, conn-&gt;sessionIDlen);
	if(conn-&gt;sessionKey != nil &amp;&amp; conn-&gt;sessionType != nil &amp;&amp; strcmp(conn-&gt;sessionType, "ttls") == 0)
		tls-&gt;sec-&gt;prf(conn-&gt;sessionKey, conn-&gt;sessionKeylen, tls-&gt;sec-&gt;sec, MasterSecretSize, conn-&gt;sessionConst,  tls-&gt;sec-&gt;crandom, RandomSize, tls-&gt;sec-&gt;srandom, RandomSize);
	tlsConnectionFree(tls);
	return data;
}

static int
countchain(PEMChain *p)
{
	int i = 0;

	while (p) {
		i++;
		p = p-&gt;next;
	}
	return i;
}

static TlsConnection *
tlsServer2(int ctl, int hand, uchar *cert, int ncert, int (*trace)(char*fmt, ...), PEMChain *chp)
{
	TlsConnection *c;
	Msg m;
	Bytes *csid;
	uchar sid[SidSize], kd[MaxKeyData];
	char *secrets;
	int cipher, compressor, nsid, rv, numcerts, i;

	if(trace)
		trace("tlsServer2\n");
	if(!initCiphers())
		return nil;
	c = emalloc(sizeof(TlsConnection));
	c-&gt;ctl = ctl;
	c-&gt;hand = hand;
	c-&gt;trace = trace;
	c-&gt;version = ProtocolVersion;

	memset(&amp;m, 0, sizeof(m));
	if(!msgRecv(c, &amp;m)){
		if(trace)
			trace("initial msgRecv failed\n");
		goto Err;
	}
	if(m.tag != HClientHello) {
		tlsError(c, EUnexpectedMessage, "expected a client hello");
		goto Err;
	}
	c-&gt;clientVersion = m.u.clientHello.version;
	if(trace)
		trace("ClientHello version %x\n", c-&gt;clientVersion);
	if(setVersion(c, m.u.clientHello.version) &lt; 0) {
		tlsError(c, EIllegalParameter, "incompatible version");
		goto Err;
	}

	memmove(c-&gt;crandom, m.u.clientHello.random, RandomSize);
	cipher = okCipher(m.u.clientHello.ciphers);
	if(cipher &lt; 0) {
		// reply with EInsufficientSecurity if we know that's the case
		if(cipher == -2)
			tlsError(c, EInsufficientSecurity, "cipher suites too weak");
		else
			tlsError(c, EHandshakeFailure, "no matching cipher suite");
		goto Err;
	}
	if(!setAlgs(c, cipher)){
		tlsError(c, EHandshakeFailure, "no matching cipher suite");
		goto Err;
	}
	compressor = okCompression(m.u.clientHello.compressors);
	if(compressor &lt; 0) {
		tlsError(c, EHandshakeFailure, "no matching compressor");
		goto Err;
	}

	csid = m.u.clientHello.sid;
	if(trace)
		trace("  cipher %d, compressor %d, csidlen %d\n", cipher, compressor, csid-&gt;len);
	c-&gt;sec = tlsSecInits(c-&gt;clientVersion, csid-&gt;data, csid-&gt;len, c-&gt;crandom, sid, &amp;nsid, c-&gt;srandom);
	if(c-&gt;sec == nil){
		tlsError(c, EHandshakeFailure, "can't initialize security: %r");
		goto Err;
	}
	c-&gt;sec-&gt;rpc = factotum_rsa_open(cert, ncert);
	if(c-&gt;sec-&gt;rpc == nil){
		tlsError(c, EHandshakeFailure, "factotum_rsa_open: %r");
		goto Err;
	}
	c-&gt;sec-&gt;rsapub = X509toRSApub(cert, ncert, nil, 0);
	msgClear(&amp;m);

	m.tag = HServerHello;
	m.u.serverHello.version = c-&gt;version;
	memmove(m.u.serverHello.random, c-&gt;srandom, RandomSize);
	m.u.serverHello.cipher = cipher;
	m.u.serverHello.compressor = compressor;
	c-&gt;sid = makebytes(sid, nsid);
	m.u.serverHello.sid = makebytes(c-&gt;sid-&gt;data, c-&gt;sid-&gt;len);
	if(!msgSend(c, &amp;m, AQueue))
		goto Err;
	msgClear(&amp;m);

	m.tag = HCertificate;
	numcerts = countchain(chp);
	m.u.certificate.ncert = 1 + numcerts;
	m.u.certificate.certs = emalloc(m.u.certificate.ncert * sizeof(Bytes));
	m.u.certificate.certs[0] = makebytes(cert, ncert);
	for (i = 0; i &lt; numcerts &amp;&amp; chp; i++, chp = chp-&gt;next)
		m.u.certificate.certs[i+1] = makebytes(chp-&gt;pem, chp-&gt;pemlen);
	if(!msgSend(c, &amp;m, AQueue))
		goto Err;
	msgClear(&amp;m);

	m.tag = HServerHelloDone;
	if(!msgSend(c, &amp;m, AFlush))
		goto Err;
	msgClear(&amp;m);

	if(!msgRecv(c, &amp;m))
		goto Err;
	if(m.tag != HClientKeyExchange) {
		tlsError(c, EUnexpectedMessage, "expected a client key exchange");
		goto Err;
	}
	if(tlsSecSecrets(c-&gt;sec, c-&gt;version, m.u.clientKeyExchange.key-&gt;data, m.u.clientKeyExchange.key-&gt;len, kd, c-&gt;nsecret) &lt; 0){
		tlsError(c, EHandshakeFailure, "couldn't set secrets: %r");
		goto Err;
	}
	if(trace)
		trace("tls secrets\n");
	secrets = (char*)emalloc(2*c-&gt;nsecret);
	enc64(secrets, 2*c-&gt;nsecret, kd, c-&gt;nsecret);
	rv = fprint(c-&gt;ctl, "secret %s %s 0 %s", c-&gt;digest, c-&gt;enc, secrets);
	memset(secrets, 0, 2*c-&gt;nsecret);
	free(secrets);
	memset(kd, 0, c-&gt;nsecret);
	if(rv &lt; 0){
		tlsError(c, EHandshakeFailure, "can't set keys: %r");
		goto Err;
	}
	msgClear(&amp;m);

	/* no CertificateVerify; skip to Finished */
	if(tlsSecFinished(c-&gt;sec, c-&gt;hsmd5, c-&gt;hssha1, c-&gt;finished.verify, c-&gt;finished.n, 1) &lt; 0){
		tlsError(c, EInternalError, "can't set finished: %r");
		goto Err;
	}
	if(!msgRecv(c, &amp;m))
		goto Err;
	if(m.tag != HFinished) {
		tlsError(c, EUnexpectedMessage, "expected a finished");
		goto Err;
	}
	if(!finishedMatch(c, &amp;m.u.finished)) {
		tlsError(c, EHandshakeFailure, "finished verification failed");
		goto Err;
	}
	msgClear(&amp;m);

	/* change cipher spec */
	if(fprint(c-&gt;ctl, "changecipher") &lt; 0){
		tlsError(c, EInternalError, "can't enable cipher: %r");
		goto Err;
	}

	if(tlsSecFinished(c-&gt;sec, c-&gt;hsmd5, c-&gt;hssha1, c-&gt;finished.verify, c-&gt;finished.n, 0) &lt; 0){
		tlsError(c, EInternalError, "can't set finished: %r");
		goto Err;
	}
	m.tag = HFinished;
	m.u.finished = c-&gt;finished;
	if(!msgSend(c, &amp;m, AFlush))
		goto Err;
	msgClear(&amp;m);
	if(trace)
		trace("tls finished\n");

	if(fprint(c-&gt;ctl, "opened") &lt; 0)
		goto Err;
	tlsSecOk(c-&gt;sec);
	return c;

Err:
	msgClear(&amp;m);
	tlsConnectionFree(c);
	return 0;
}

static TlsConnection *
tlsClient2(int ctl, int hand, uchar *csid, int ncsid, int (*trace)(char*fmt, ...))
{
	TlsConnection *c;
	Msg m;
	uchar kd[MaxKeyData], *epm;
	char *secrets;
	int creq, nepm, rv;

	if(!initCiphers())
		return nil;
	epm = nil;
	c = emalloc(sizeof(TlsConnection));
	c-&gt;version = ProtocolVersion;
	c-&gt;ctl = ctl;
	c-&gt;hand = hand;
	c-&gt;trace = trace;
	c-&gt;isClient = 1;
	c-&gt;clientVersion = c-&gt;version;

	c-&gt;sec = tlsSecInitc(c-&gt;clientVersion, c-&gt;crandom);
	if(c-&gt;sec == nil)
		goto Err;

	/* client hello */
	memset(&amp;m, 0, sizeof(m));
	m.tag = HClientHello;
	m.u.clientHello.version = c-&gt;clientVersion;
	memmove(m.u.clientHello.random, c-&gt;crandom, RandomSize);
	m.u.clientHello.sid = makebytes(csid, ncsid);
	m.u.clientHello.ciphers = makeciphers();
	m.u.clientHello.compressors = makebytes(compressors,sizeof(compressors));
	if(!msgSend(c, &amp;m, AFlush))
		goto Err;
	msgClear(&amp;m);

	/* server hello */
	if(!msgRecv(c, &amp;m))
		goto Err;
	if(m.tag != HServerHello) {
		tlsError(c, EUnexpectedMessage, "expected a server hello");
		goto Err;
	}
	if(setVersion(c, m.u.serverHello.version) &lt; 0) {
		tlsError(c, EIllegalParameter, "incompatible version %r");
		goto Err;
	}
	memmove(c-&gt;srandom, m.u.serverHello.random, RandomSize);
	c-&gt;sid = makebytes(m.u.serverHello.sid-&gt;data, m.u.serverHello.sid-&gt;len);
	if(c-&gt;sid-&gt;len != 0 &amp;&amp; c-&gt;sid-&gt;len != SidSize) {
		tlsError(c, EIllegalParameter, "invalid server session identifier");
		goto Err;
	}
	if(!setAlgs(c, m.u.serverHello.cipher)) {
		tlsError(c, EIllegalParameter, "invalid cipher suite");
		goto Err;
	}
	if(m.u.serverHello.compressor != CompressionNull) {
		tlsError(c, EIllegalParameter, "invalid compression");
		goto Err;
	}
	msgClear(&amp;m);

	/* certificate */
	if(!msgRecv(c, &amp;m) || m.tag != HCertificate) {
		tlsError(c, EUnexpectedMessage, "expected a certificate");
		goto Err;
	}
	if(m.u.certificate.ncert &lt; 1) {
		tlsError(c, EIllegalParameter, "runt certificate");
		goto Err;
	}
	c-&gt;cert = makebytes(m.u.certificate.certs[0]-&gt;data, m.u.certificate.certs[0]-&gt;len);
	msgClear(&amp;m);

	/* server key exchange (optional) */
	if(!msgRecv(c, &amp;m))
		goto Err;
	if(m.tag == HServerKeyExchange) {
		tlsError(c, EUnexpectedMessage, "got an server key exchange");
		goto Err;
		// If implementing this later, watch out for rollback attack
		// described in Wagner Schneier 1996, section 4.4.
	}

	/* certificate request (optional) */
	creq = 0;
	if(m.tag == HCertificateRequest) {
		creq = 1;
		msgClear(&amp;m);
		if(!msgRecv(c, &amp;m))
			goto Err;
	}

	if(m.tag != HServerHelloDone) {
		tlsError(c, EUnexpectedMessage, "expected a server hello done");
		goto Err;
	}
	msgClear(&amp;m);

	if(tlsSecSecretc(c-&gt;sec, c-&gt;sid-&gt;data, c-&gt;sid-&gt;len, c-&gt;srandom,
			c-&gt;cert-&gt;data, c-&gt;cert-&gt;len, c-&gt;version, &amp;epm, &amp;nepm,
			kd, c-&gt;nsecret) &lt; 0){
		tlsError(c, EBadCertificate, "invalid x509/rsa certificate");
		goto Err;
	}
	secrets = (char*)emalloc(2*c-&gt;nsecret);
	enc64(secrets, 2*c-&gt;nsecret, kd, c-&gt;nsecret);
	rv = fprint(c-&gt;ctl, "secret %s %s 1 %s", c-&gt;digest, c-&gt;enc, secrets);
	memset(secrets, 0, 2*c-&gt;nsecret);
	free(secrets);
	memset(kd, 0, c-&gt;nsecret);
	if(rv &lt; 0){
		tlsError(c, EHandshakeFailure, "can't set keys: %r");
		goto Err;
	}

	if(creq) {
		/* send a zero length certificate */
		m.tag = HCertificate;
		if(!msgSend(c, &amp;m, AFlush))
			goto Err;
		msgClear(&amp;m);
	}

	/* client key exchange */
	m.tag = HClientKeyExchange;
	m.u.clientKeyExchange.key = makebytes(epm, nepm);
	free(epm);
	epm = nil;
	if(m.u.clientKeyExchange.key == nil) {
		tlsError(c, EHandshakeFailure, "can't set secret: %r");
		goto Err;
	}
	if(!msgSend(c, &amp;m, AFlush))
		goto Err;
	msgClear(&amp;m);

	/* change cipher spec */
	if(fprint(c-&gt;ctl, "changecipher") &lt; 0){
		tlsError(c, EInternalError, "can't enable cipher: %r");
		goto Err;
	}

	// Cipherchange must occur immediately before Finished to avoid
	// potential hole;  see section 4.3 of Wagner Schneier 1996.
	if(tlsSecFinished(c-&gt;sec, c-&gt;hsmd5, c-&gt;hssha1, c-&gt;finished.verify, c-&gt;finished.n, 1) &lt; 0){
		tlsError(c, EInternalError, "can't set finished 1: %r");
		goto Err;
	}
	m.tag = HFinished;
	m.u.finished = c-&gt;finished;

	if(!msgSend(c, &amp;m, AFlush)) {
		fprint(2, "tlsClient nepm=%d\n", nepm);
		tlsError(c, EInternalError, "can't flush after client Finished: %r");
		goto Err;
	}
	msgClear(&amp;m);

	if(tlsSecFinished(c-&gt;sec, c-&gt;hsmd5, c-&gt;hssha1, c-&gt;finished.verify, c-&gt;finished.n, 0) &lt; 0){
		fprint(2, "tlsClient nepm=%d\n", nepm);
		tlsError(c, EInternalError, "can't set finished 0: %r");
		goto Err;
	}
	if(!msgRecv(c, &amp;m)) {
		fprint(2, "tlsClient nepm=%d\n", nepm);
		tlsError(c, EInternalError, "can't read server Finished: %r");
		goto Err;
	}
	if(m.tag != HFinished) {
		fprint(2, "tlsClient nepm=%d\n", nepm);
		tlsError(c, EUnexpectedMessage, "expected a Finished msg from server");
		goto Err;
	}

	if(!finishedMatch(c, &amp;m.u.finished)) {
		tlsError(c, EHandshakeFailure, "finished verification failed");
		goto Err;
	}
	msgClear(&amp;m);

	if(fprint(c-&gt;ctl, "opened") &lt; 0){
		if(trace)
			trace("unable to do final open: %r\n");
		goto Err;
	}
	tlsSecOk(c-&gt;sec);
	return c;

Err:
	free(epm);
	msgClear(&amp;m);
	tlsConnectionFree(c);
	return 0;
}


//================= message functions ========================

static uchar sendbuf[9000], *sendp;

static int
msgSend(TlsConnection *c, Msg *m, int act)
{
	uchar *p; // sendp = start of new message;  p = write pointer
	int nn, n, i;

	if(sendp == nil)
		sendp = sendbuf;
	p = sendp;
	if(c-&gt;trace)
		c-&gt;trace("send %s", msgPrint((char*)p, (sizeof sendbuf) - (p-sendbuf), m));

	p[0] = m-&gt;tag;	// header - fill in size later
	p += 4;

	switch(m-&gt;tag) {
	default:
		tlsError(c, EInternalError, "can't encode a %d", m-&gt;tag);
		goto Err;
	case HClientHello:
		// version
		put16(p, m-&gt;u.clientHello.version);
		p += 2;

		// random
		memmove(p, m-&gt;u.clientHello.random, RandomSize);
		p += RandomSize;

		// sid
		n = m-&gt;u.clientHello.sid-&gt;len;
		assert(n &lt; 256);
		p[0] = n;
		memmove(p+1, m-&gt;u.clientHello.sid-&gt;data, n);
		p += n+1;

		n = m-&gt;u.clientHello.ciphers-&gt;len;
		assert(n &gt; 0 &amp;&amp; n &lt; 200);
		put16(p, n*2);
		p += 2;
		for(i=0; i&lt;n; i++) {
			put16(p, m-&gt;u.clientHello.ciphers-&gt;data[i]);
			p += 2;
		}

		n = m-&gt;u.clientHello.compressors-&gt;len;
		assert(n &gt; 0);
		p[0] = n;
		memmove(p+1, m-&gt;u.clientHello.compressors-&gt;data, n);
		p += n+1;
		break;
	case HServerHello:
		put16(p, m-&gt;u.serverHello.version);
		p += 2;

		// random
		memmove(p, m-&gt;u.serverHello.random, RandomSize);
		p += RandomSize;

		// sid
		n = m-&gt;u.serverHello.sid-&gt;len;
		assert(n &lt; 256);
		p[0] = n;
		memmove(p+1, m-&gt;u.serverHello.sid-&gt;data, n);
		p += n+1;

		put16(p, m-&gt;u.serverHello.cipher);
		p += 2;
		p[0] = m-&gt;u.serverHello.compressor;
		p += 1;
		break;
	case HServerHelloDone:
		break;
	case HCertificate:
		nn = 0;
		for(i = 0; i &lt; m-&gt;u.certificate.ncert; i++)
			nn += 3 + m-&gt;u.certificate.certs[i]-&gt;len;
		if(p + 3 + nn - sendbuf &gt; sizeof(sendbuf)) {
			tlsError(c, EInternalError, "output buffer too small for certificate");
			goto Err;
		}
		put24(p, nn);
		p += 3;
		for(i = 0; i &lt; m-&gt;u.certificate.ncert; i++){
			put24(p, m-&gt;u.certificate.certs[i]-&gt;len);
			p += 3;
			memmove(p, m-&gt;u.certificate.certs[i]-&gt;data, m-&gt;u.certificate.certs[i]-&gt;len);
			p += m-&gt;u.certificate.certs[i]-&gt;len;
		}
		break;
	case HClientKeyExchange:
		n = m-&gt;u.clientKeyExchange.key-&gt;len;
		if(c-&gt;version != SSL3Version){
			put16(p, n);
			p += 2;
		}
		memmove(p, m-&gt;u.clientKeyExchange.key-&gt;data, n);
		p += n;
		break;
	case HFinished:
		memmove(p, m-&gt;u.finished.verify, m-&gt;u.finished.n);
		p += m-&gt;u.finished.n;
		break;
	}

	// go back and fill in size
	n = p-sendp;
	assert(p &lt;= sendbuf+sizeof(sendbuf));
	put24(sendp+1, n-4);

	// remember hash of Handshake messages
	if(m-&gt;tag != HHelloRequest) {
		md5(sendp, n, 0, &amp;c-&gt;hsmd5);
		sha1(sendp, n, 0, &amp;c-&gt;hssha1);
	}

	sendp = p;
	if(act == AFlush){
		sendp = sendbuf;
		if(write(c-&gt;hand, sendbuf, p-sendbuf) &lt; 0){
			fprint(2, "write error: %r\n");
			goto Err;
		}
	}
	msgClear(m);
	return 1;
Err:
	msgClear(m);
	return 0;
}

static uchar*
tlsReadN(TlsConnection *c, int n)
{
	uchar *p;
	int nn, nr;

	nn = c-&gt;ep - c-&gt;rp;
	if(nn &lt; n){
		if(c-&gt;rp != c-&gt;buf){
			memmove(c-&gt;buf, c-&gt;rp, nn);
			c-&gt;rp = c-&gt;buf;
			c-&gt;ep = &amp;c-&gt;buf[nn];
		}
		for(; nn &lt; n; nn += nr) {
			nr = read(c-&gt;hand, &amp;c-&gt;rp[nn], n - nn);
			if(nr &lt;= 0)
				return nil;
			c-&gt;ep += nr;
		}
	}
	p = c-&gt;rp;
	c-&gt;rp += n;
	return p;
}

static int
msgRecv(TlsConnection *c, Msg *m)
{
	uchar *p;
	int type, n, nn, i, nsid, nrandom, nciph;

	for(;;) {
		p = tlsReadN(c, 4);
		if(p == nil)
			return 0;
		type = p[0];
		n = get24(p+1);

		if(type != HHelloRequest)
			break;
		if(n != 0) {
			tlsError(c, EDecodeError, "invalid hello request during handshake");
			return 0;
		}
	}

	if(n &gt; sizeof(c-&gt;buf)) {
		tlsError(c, EDecodeError, "handshake message too long %d %d", n, sizeof(c-&gt;buf));
		return 0;
	}

	if(type == HSSL2ClientHello){
		/* Cope with an SSL3 ClientHello expressed in SSL2 record format.
			This is sent by some clients that we must interoperate
			with, such as Java's JSSE and Microsoft's Internet Explorer. */
		p = tlsReadN(c, n);
		if(p == nil)
			return 0;
		md5(p, n, 0, &amp;c-&gt;hsmd5);
		sha1(p, n, 0, &amp;c-&gt;hssha1);
		m-&gt;tag = HClientHello;
		if(n &lt; 22)
			goto Short;
		m-&gt;u.clientHello.version = get16(p+1);
		p += 3;
		n -= 3;
		nn = get16(p); /* cipher_spec_len */
		nsid = get16(p + 2);
		nrandom = get16(p + 4);
		p += 6;
		n -= 6;
		if(nsid != 0 	/* no sid's, since shouldn't restart using ssl2 header */
				|| nrandom &lt; 16 || nn % 3)
			goto Err;
		if(c-&gt;trace &amp;&amp; (n - nrandom != nn))
			c-&gt;trace("n-nrandom!=nn: n=%d nrandom=%d nn=%d\n", n, nrandom, nn);
		/* ignore ssl2 ciphers and look for {0x00, ssl3 cipher} */
		nciph = 0;
		for(i = 0; i &lt; nn; i += 3)
			if(p[i] == 0)
				nciph++;
		m-&gt;u.clientHello.ciphers = newints(nciph);
		nciph = 0;
		for(i = 0; i &lt; nn; i += 3)
			if(p[i] == 0)
				m-&gt;u.clientHello.ciphers-&gt;data[nciph++] = get16(&amp;p[i + 1]);
		p += nn;
		m-&gt;u.clientHello.sid = makebytes(nil, 0);
		if(nrandom &gt; RandomSize)
			nrandom = RandomSize;
		memset(m-&gt;u.clientHello.random, 0, RandomSize - nrandom);
		memmove(&amp;m-&gt;u.clientHello.random[RandomSize - nrandom], p, nrandom);
		m-&gt;u.clientHello.compressors = newbytes(1);
		m-&gt;u.clientHello.compressors-&gt;data[0] = CompressionNull;
		goto Ok;
	}

	md5(p, 4, 0, &amp;c-&gt;hsmd5);
	sha1(p, 4, 0, &amp;c-&gt;hssha1);

	p = tlsReadN(c, n);
	if(p == nil)
		return 0;

	md5(p, n, 0, &amp;c-&gt;hsmd5);
	sha1(p, n, 0, &amp;c-&gt;hssha1);

	m-&gt;tag = type;

	switch(type) {
	default:
		tlsError(c, EUnexpectedMessage, "can't decode a %d", type);
		goto Err;
	case HClientHello:
		if(n &lt; 2)
			goto Short;
		m-&gt;u.clientHello.version = get16(p);
		p += 2;
		n -= 2;

		if(n &lt; RandomSize)
			goto Short;
		memmove(m-&gt;u.clientHello.random, p, RandomSize);
		p += RandomSize;
		n -= RandomSize;
		if(n &lt; 1 || n &lt; p[0]+1)
			goto Short;
		m-&gt;u.clientHello.sid = makebytes(p+1, p[0]);
		p += m-&gt;u.clientHello.sid-&gt;len+1;
		n -= m-&gt;u.clientHello.sid-&gt;len+1;

		if(n &lt; 2)
			goto Short;
		nn = get16(p);
		p += 2;
		n -= 2;

		if((nn &amp; 1) || n &lt; nn || nn &lt; 2)
			goto Short;
		m-&gt;u.clientHello.ciphers = newints(nn &gt;&gt; 1);
		for(i = 0; i &lt; nn; i += 2)
			m-&gt;u.clientHello.ciphers-&gt;data[i &gt;&gt; 1] = get16(&amp;p[i]);
		p += nn;
		n -= nn;

		if(n &lt; 1 || n &lt; p[0]+1 || p[0] == 0)
			goto Short;
		nn = p[0];
		m-&gt;u.clientHello.compressors = newbytes(nn);
		memmove(m-&gt;u.clientHello.compressors-&gt;data, p+1, nn);
		n -= nn + 1;
		break;
	case HServerHello:
		if(n &lt; 2)
			goto Short;
		m-&gt;u.serverHello.version = get16(p);
		p += 2;
		n -= 2;

		if(n &lt; RandomSize)
			goto Short;
		memmove(m-&gt;u.serverHello.random, p, RandomSize);
		p += RandomSize;
		n -= RandomSize;

		if(n &lt; 1 || n &lt; p[0]+1)
			goto Short;
		m-&gt;u.serverHello.sid = makebytes(p+1, p[0]);
		p += m-&gt;u.serverHello.sid-&gt;len+1;
		n -= m-&gt;u.serverHello.sid-&gt;len+1;

		if(n &lt; 3)
			goto Short;
		m-&gt;u.serverHello.cipher = get16(p);
		m-&gt;u.serverHello.compressor = p[2];
		n -= 3;
		break;
	case HCertificate:
		if(n &lt; 3)
			goto Short;
		nn = get24(p);
		p += 3;
		n -= 3;
		if(n != nn)
			goto Short;
		/* certs */
		i = 0;
		while(n &gt; 0) {
			if(n &lt; 3)
				goto Short;
			nn = get24(p);
			p += 3;
			n -= 3;
			if(nn &gt; n)
				goto Short;
			m-&gt;u.certificate.ncert = i+1;
			m-&gt;u.certificate.certs = erealloc(m-&gt;u.certificate.certs, (i+1)*sizeof(Bytes));
			m-&gt;u.certificate.certs[i] = makebytes(p, nn);
			p += nn;
			n -= nn;
			i++;
		}
		break;
	case HCertificateRequest:
		if(n &lt; 1)
			goto Short;
		nn = p[0];
		p += 1;
		n -= 1;
		if(nn &lt; 1 || nn &gt; n)
			goto Short;
		m-&gt;u.certificateRequest.types = makebytes(p, nn);
		p += nn;
		n -= nn;
		if(n &lt; 2)
			goto Short;
		nn = get16(p);
		p += 2;
		n -= 2;
		/* nn == 0 can happen; yahoo's servers do it */
		if(nn != n)
			goto Short;
		/* cas */
		i = 0;
		while(n &gt; 0) {
			if(n &lt; 2)
				goto Short;
			nn = get16(p);
			p += 2;
			n -= 2;
			if(nn &lt; 1 || nn &gt; n)
				goto Short;
			m-&gt;u.certificateRequest.nca = i+1;
			m-&gt;u.certificateRequest.cas = erealloc(
				m-&gt;u.certificateRequest.cas, (i+1)*sizeof(Bytes));
			m-&gt;u.certificateRequest.cas[i] = makebytes(p, nn);
			p += nn;
			n -= nn;
			i++;
		}
		break;
	case HServerHelloDone:
		break;
	case HClientKeyExchange:
		/*
		 * this message depends upon the encryption selected
		 * assume rsa.
		 */
		if(c-&gt;version == SSL3Version)
			nn = n;
		else{
			if(n &lt; 2)
				goto Short;
			nn = get16(p);
			p += 2;
			n -= 2;
		}
		if(n &lt; nn)
			goto Short;
		m-&gt;u.clientKeyExchange.key = makebytes(p, nn);
		n -= nn;
		break;
	case HFinished:
		m-&gt;u.finished.n = c-&gt;finished.n;
		if(n &lt; m-&gt;u.finished.n)
			goto Short;
		memmove(m-&gt;u.finished.verify, p, m-&gt;u.finished.n);
		n -= m-&gt;u.finished.n;
		break;
	}

	if(type != HClientHello &amp;&amp; n != 0)
		goto Short;
Ok:
	if(c-&gt;trace){
		char *buf;
		buf = emalloc(8000);
		c-&gt;trace("recv %s", msgPrint(buf, 8000, m));
		free(buf);
	}
	return 1;
Short:
	tlsError(c, EDecodeError, "handshake message has invalid length");
Err:
	msgClear(m);
	return 0;
}

static void
msgClear(Msg *m)
{
	int i;

	switch(m-&gt;tag) {
	default:
		sysfatal("msgClear: unknown message type: %d", m-&gt;tag);
	case HHelloRequest:
		break;
	case HClientHello:
		freebytes(m-&gt;u.clientHello.sid);
		freeints(m-&gt;u.clientHello.ciphers);
		freebytes(m-&gt;u.clientHello.compressors);
		break;
	case HServerHello:
		freebytes(m-&gt;u.clientHello.sid);
		break;
	case HCertificate:
		for(i=0; i&lt;m-&gt;u.certificate.ncert; i++)
			freebytes(m-&gt;u.certificate.certs[i]);
		free(m-&gt;u.certificate.certs);
		break;
	case HCertificateRequest:
		freebytes(m-&gt;u.certificateRequest.types);
		for(i=0; i&lt;m-&gt;u.certificateRequest.nca; i++)
			freebytes(m-&gt;u.certificateRequest.cas[i]);
		free(m-&gt;u.certificateRequest.cas);
		break;
	case HServerHelloDone:
		break;
	case HClientKeyExchange:
		freebytes(m-&gt;u.clientKeyExchange.key);
		break;
	case HFinished:
		break;
	}
	memset(m, 0, sizeof(Msg));
}

static char *
bytesPrint(char *bs, char *be, char *s0, Bytes *b, char *s1)
{
	int i;

	if(s0)
		bs = seprint(bs, be, "%s", s0);
	bs = seprint(bs, be, "[");
	if(b == nil)
		bs = seprint(bs, be, "nil");
	else
		for(i=0; i&lt;b-&gt;len; i++)
			bs = seprint(bs, be, "%.2x ", b-&gt;data[i]);
	bs = seprint(bs, be, "]");
	if(s1)
		bs = seprint(bs, be, "%s", s1);
	return bs;
}

static char *
intsPrint(char *bs, char *be, char *s0, Ints *b, char *s1)
{
	int i;

	if(s0)
		bs = seprint(bs, be, "%s", s0);
	bs = seprint(bs, be, "[");
	if(b == nil)
		bs = seprint(bs, be, "nil");
	else
		for(i=0; i&lt;b-&gt;len; i++)
			bs = seprint(bs, be, "%x ", b-&gt;data[i]);
	bs = seprint(bs, be, "]");
	if(s1)
		bs = seprint(bs, be, "%s", s1);
	return bs;
}

static char*
msgPrint(char *buf, int n, Msg *m)
{
	int i;
	char *bs = buf, *be = buf+n;

	switch(m-&gt;tag) {
	default:
		bs = seprint(bs, be, "unknown %d\n", m-&gt;tag);
		break;
	case HClientHello:
		bs = seprint(bs, be, "ClientHello\n");
		bs = seprint(bs, be, "\tversion: %.4x\n", m-&gt;u.clientHello.version);
		bs = seprint(bs, be, "\trandom: ");
		for(i=0; i&lt;RandomSize; i++)
			bs = seprint(bs, be, "%.2x", m-&gt;u.clientHello.random[i]);
		bs = seprint(bs, be, "\n");
		bs = bytesPrint(bs, be, "\tsid: ", m-&gt;u.clientHello.sid, "\n");
		bs = intsPrint(bs, be, "\tciphers: ", m-&gt;u.clientHello.ciphers, "\n");
		bs = bytesPrint(bs, be, "\tcompressors: ", m-&gt;u.clientHello.compressors, "\n");
		break;
	case HServerHello:
		bs = seprint(bs, be, "ServerHello\n");
		bs = seprint(bs, be, "\tversion: %.4x\n", m-&gt;u.serverHello.version);
		bs = seprint(bs, be, "\trandom: ");
		for(i=0; i&lt;RandomSize; i++)
			bs = seprint(bs, be, "%.2x", m-&gt;u.serverHello.random[i]);
		bs = seprint(bs, be, "\n");
		bs = bytesPrint(bs, be, "\tsid: ", m-&gt;u.serverHello.sid, "\n");
		bs = seprint(bs, be, "\tcipher: %.4x\n", m-&gt;u.serverHello.cipher);
		bs = seprint(bs, be, "\tcompressor: %.2x\n", m-&gt;u.serverHello.compressor);
		break;
	case HCertificate:
		bs = seprint(bs, be, "Certificate\n");
		for(i=0; i&lt;m-&gt;u.certificate.ncert; i++)
			bs = bytesPrint(bs, be, "\t", m-&gt;u.certificate.certs[i], "\n");
		break;
	case HCertificateRequest:
		bs = seprint(bs, be, "CertificateRequest\n");
		bs = bytesPrint(bs, be, "\ttypes: ", m-&gt;u.certificateRequest.types, "\n");
		bs = seprint(bs, be, "\tcertificateauthorities\n");
		for(i=0; i&lt;m-&gt;u.certificateRequest.nca; i++)
			bs = bytesPrint(bs, be, "\t\t", m-&gt;u.certificateRequest.cas[i], "\n");
		break;
	case HServerHelloDone:
		bs = seprint(bs, be, "ServerHelloDone\n");
		break;
	case HClientKeyExchange:
		bs = seprint(bs, be, "HClientKeyExchange\n");
		bs = bytesPrint(bs, be, "\tkey: ", m-&gt;u.clientKeyExchange.key, "\n");
		break;
	case HFinished:
		bs = seprint(bs, be, "HFinished\n");
		for(i=0; i&lt;m-&gt;u.finished.n; i++)
			bs = seprint(bs, be, "%.2x", m-&gt;u.finished.verify[i]);
		bs = seprint(bs, be, "\n");
		break;
	}
	USED(bs);
	return buf;
}

static void
tlsError(TlsConnection *c, int err, char *fmt, ...)
{
	char msg[512];
	va_list arg;

	va_start(arg, fmt);
	vseprint(msg, msg+sizeof(msg), fmt, arg);
	va_end(arg);
	if(c-&gt;trace)
		c-&gt;trace("tlsError: %s\n", msg);
	else if(c-&gt;erred)
		fprint(2, "double error: %r, %s", msg);
	else
		werrstr("tls: local %s", msg);
	c-&gt;erred = 1;
	fprint(c-&gt;ctl, "alert %d", err);
}

// commit to specific version number
static int
setVersion(TlsConnection *c, int version)
{
	if(c-&gt;verset || version &gt; MaxProtoVersion || version &lt; MinProtoVersion)
		return -1;
	if(version &gt; c-&gt;version)
		version = c-&gt;version;
	if(version == SSL3Version) {
		c-&gt;version = version;
		c-&gt;finished.n = SSL3FinishedLen;
	}else if(version == TLSVersion){
		c-&gt;version = version;
		c-&gt;finished.n = TLSFinishedLen;
	}else
		return -1;
	c-&gt;verset = 1;
	return fprint(c-&gt;ctl, "version 0x%x", version);
}

// confirm that received Finished message matches the expected value
static int
finishedMatch(TlsConnection *c, Finished *f)
{
	return memcmp(f-&gt;verify, c-&gt;finished.verify, f-&gt;n) == 0;
}

// free memory associated with TlsConnection struct
//		(but don't close the TLS channel itself)
static void
tlsConnectionFree(TlsConnection *c)
{
	tlsSecClose(c-&gt;sec);
	freebytes(c-&gt;sid);
	freebytes(c-&gt;cert);
	memset(c, 0, sizeof(c));
	free(c);
}


//================= cipher choices ========================

static int weakCipher[CipherMax] =
{
	1,	/* TLS_NULL_WITH_NULL_NULL */
	1,	/* TLS_RSA_WITH_NULL_MD5 */
	1,	/* TLS_RSA_WITH_NULL_SHA */
	1,	/* TLS_RSA_EXPORT_WITH_RC4_40_MD5 */
	0,	/* TLS_RSA_WITH_RC4_128_MD5 */
	0,	/* TLS_RSA_WITH_RC4_128_SHA */
	1,	/* TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5 */
	0,	/* TLS_RSA_WITH_IDEA_CBC_SHA */
	1,	/* TLS_RSA_EXPORT_WITH_DES40_CBC_SHA */
	0,	/* TLS_RSA_WITH_DES_CBC_SHA */
	0,	/* TLS_RSA_WITH_3DES_EDE_CBC_SHA */
	1,	/* TLS_DH_DSS_EXPORT_WITH_DES40_CBC_SHA */
	0,	/* TLS_DH_DSS_WITH_DES_CBC_SHA */
	0,	/* TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA */
	1,	/* TLS_DH_RSA_EXPORT_WITH_DES40_CBC_SHA */
	0,	/* TLS_DH_RSA_WITH_DES_CBC_SHA */
	0,	/* TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA */
	1,	/* TLS_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA */
	0,	/* TLS_DHE_DSS_WITH_DES_CBC_SHA */
	0,	/* TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA */
	1,	/* TLS_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA */
	0,	/* TLS_DHE_RSA_WITH_DES_CBC_SHA */
	0,	/* TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA */
	1,	/* TLS_DH_anon_EXPORT_WITH_RC4_40_MD5 */
	1,	/* TLS_DH_anon_WITH_RC4_128_MD5 */
	1,	/* TLS_DH_anon_EXPORT_WITH_DES40_CBC_SHA */
	1,	/* TLS_DH_anon_WITH_DES_CBC_SHA */
	1,	/* TLS_DH_anon_WITH_3DES_EDE_CBC_SHA */
};

static int
setAlgs(TlsConnection *c, int a)
{
	int i;

	for(i = 0; i &lt; nelem(cipherAlgs); i++){
		if(cipherAlgs[i].tlsid == a){
			c-&gt;enc = cipherAlgs[i].enc;
			c-&gt;digest = cipherAlgs[i].digest;
			c-&gt;nsecret = cipherAlgs[i].nsecret;
			if(c-&gt;nsecret &gt; MaxKeyData)
				return 0;
			return 1;
		}
	}
	return 0;
}

static int
okCipher(Ints *cv)
{
	int weak, i, j, c;

	weak = 1;
	for(i = 0; i &lt; cv-&gt;len; i++) {
		c = cv-&gt;data[i];
		if(c &gt;= CipherMax)
			weak = 0;
		else
			weak &amp;= weakCipher[c];
		for(j = 0; j &lt; nelem(cipherAlgs); j++)
			if(cipherAlgs[j].ok &amp;&amp; cipherAlgs[j].tlsid == c)
				return c;
	}
	if(weak)
		return -2;
	return -1;
}

static int
okCompression(Bytes *cv)
{
	int i, j, c;

	for(i = 0; i &lt; cv-&gt;len; i++) {
		c = cv-&gt;data[i];
		for(j = 0; j &lt; nelem(compressors); j++) {
			if(compressors[j] == c)
				return c;
		}
	}
	return -1;
}

static Lock	ciphLock;
static int	nciphers;

static int
initCiphers(void)
{
	enum {MaxAlgF = 1024, MaxAlgs = 10};
	char s[MaxAlgF], *flds[MaxAlgs];
	int i, j, n, ok;

	lock(&amp;ciphLock);
	if(nciphers){
		unlock(&amp;ciphLock);
		return nciphers;
	}
	j = open("#a/tls/encalgs", OREAD);
	if(j &lt; 0){
		werrstr("can't open #a/tls/encalgs: %r");
		return 0;
	}
	n = read(j, s, MaxAlgF-1);
	close(j);
	if(n &lt;= 0){
		werrstr("nothing in #a/tls/encalgs: %r");
		return 0;
	}
	s[n] = 0;
	n = getfields(s, flds, MaxAlgs, 1, " \t\r\n");
	for(i = 0; i &lt; nelem(cipherAlgs); i++){
		ok = 0;
		for(j = 0; j &lt; n; j++){
			if(strcmp(cipherAlgs[i].enc, flds[j]) == 0){
				ok = 1;
				break;
			}
		}
		cipherAlgs[i].ok = ok;
	}

	j = open("#a/tls/hashalgs", OREAD);
	if(j &lt; 0){
		werrstr("can't open #a/tls/hashalgs: %r");
		return 0;
	}
	n = read(j, s, MaxAlgF-1);
	close(j);
	if(n &lt;= 0){
		werrstr("nothing in #a/tls/hashalgs: %r");
		return 0;
	}
	s[n] = 0;
	n = getfields(s, flds, MaxAlgs, 1, " \t\r\n");
	for(i = 0; i &lt; nelem(cipherAlgs); i++){
		ok = 0;
		for(j = 0; j &lt; n; j++){
			if(strcmp(cipherAlgs[i].digest, flds[j]) == 0){
				ok = 1;
				break;
			}
		}
		cipherAlgs[i].ok &amp;= ok;
		if(cipherAlgs[i].ok)
			nciphers++;
	}
	unlock(&amp;ciphLock);
	return nciphers;
}

static Ints*
makeciphers(void)
{
	Ints *is;
	int i, j;

	is = newints(nciphers);
	j = 0;
	for(i = 0; i &lt; nelem(cipherAlgs); i++){
		if(cipherAlgs[i].ok)
			is-&gt;data[j++] = cipherAlgs[i].tlsid;
	}
	return is;
}



//================= security functions ========================

// given X.509 certificate, set up connection to factotum
//	for using corresponding private key
static AuthRpc*
factotum_rsa_open(uchar *cert, int certlen)
{
	int afd;
	char *s;
	mpint *pub = nil;
	RSApub *rsapub;
	AuthRpc *rpc;

	// start talking to factotum
	if((afd = open("/mnt/factotum/rpc", ORDWR)) &lt; 0)
		return nil;
	if((rpc = auth_allocrpc(afd)) == nil){
		close(afd);
		return nil;
	}
	s = "proto=rsa service=tls role=client";
	if(auth_rpc(rpc, "start", s, strlen(s)) != ARok){
		factotum_rsa_close(rpc);
		return nil;
	}

	// roll factotum keyring around to match certificate
	rsapub = X509toRSApub(cert, certlen, nil, 0);
	while(1){
		if(auth_rpc(rpc, "read", nil, 0) != ARok){
			factotum_rsa_close(rpc);
			rpc = nil;
			goto done;
		}
		pub = strtomp(rpc-&gt;arg, nil, 16, nil);
		assert(pub != nil);
		if(mpcmp(pub,rsapub-&gt;n) == 0)
			break;
	}
done:
	mpfree(pub);
	rsapubfree(rsapub);
	return rpc;
}

static mpint*
factotum_rsa_decrypt(AuthRpc *rpc, mpint *cipher)
{
	char *p;
	int rv;

	if((p = mptoa(cipher, 16, nil, 0)) == nil)
		return nil;
	rv = auth_rpc(rpc, "write", p, strlen(p));
	free(p);
	if(rv != ARok || auth_rpc(rpc, "read", nil, 0) != ARok)
		return nil;
	mpfree(cipher);
	return strtomp(rpc-&gt;arg, nil, 16, nil);
}

static void
factotum_rsa_close(AuthRpc*rpc)
{
	if(!rpc)
		return;
	close(rpc-&gt;afd);
	auth_freerpc(rpc);
}

static void
tlsPmd5(uchar *buf, int nbuf, uchar *key, int nkey, uchar *label, int nlabel, uchar *seed0, int nseed0, uchar *seed1, int nseed1)
{
	uchar ai[MD5dlen], tmp[MD5dlen];
	int i, n;
	MD5state *s;

	// generate a1
	s = hmac_md5(label, nlabel, key, nkey, nil, nil);
	s = hmac_md5(seed0, nseed0, key, nkey, nil, s);
	hmac_md5(seed1, nseed1, key, nkey, ai, s);

	while(nbuf &gt; 0) {
		s = hmac_md5(ai, MD5dlen, key, nkey, nil, nil);
		s = hmac_md5(label, nlabel, key, nkey, nil, s);
		s = hmac_md5(seed0, nseed0, key, nkey, nil, s);
		hmac_md5(seed1, nseed1, key, nkey, tmp, s);
		n = MD5dlen;
		if(n &gt; nbuf)
			n = nbuf;
		for(i = 0; i &lt; n; i++)
			buf[i] ^= tmp[i];
		buf += n;
		nbuf -= n;
		hmac_md5(ai, MD5dlen, key, nkey, tmp, nil);
		memmove(ai, tmp, MD5dlen);
	}
}

static void
tlsPsha1(uchar *buf, int nbuf, uchar *key, int nkey, uchar *label, int nlabel, uchar *seed0, int nseed0, uchar *seed1, int nseed1)
{
	uchar ai[SHA1dlen], tmp[SHA1dlen];
	int i, n;
	SHAstate *s;

	// generate a1
	s = hmac_sha1(label, nlabel, key, nkey, nil, nil);
	s = hmac_sha1(seed0, nseed0, key, nkey, nil, s);
	hmac_sha1(seed1, nseed1, key, nkey, ai, s);

	while(nbuf &gt; 0) {
		s = hmac_sha1(ai, SHA1dlen, key, nkey, nil, nil);
		s = hmac_sha1(label, nlabel, key, nkey, nil, s);
		s = hmac_sha1(seed0, nseed0, key, nkey, nil, s);
		hmac_sha1(seed1, nseed1, key, nkey, tmp, s);
		n = SHA1dlen;
		if(n &gt; nbuf)
			n = nbuf;
		for(i = 0; i &lt; n; i++)
			buf[i] ^= tmp[i];
		buf += n;
		nbuf -= n;
		hmac_sha1(ai, SHA1dlen, key, nkey, tmp, nil);
		memmove(ai, tmp, SHA1dlen);
	}
}

// fill buf with md5(args)^sha1(args)
static void
tlsPRF(uchar *buf, int nbuf, uchar *key, int nkey, char *label, uchar *seed0, int nseed0, uchar *seed1, int nseed1)
{
	int i;
	int nlabel = strlen(label);
	int n = (nkey + 1) &gt;&gt; 1;

	for(i = 0; i &lt; nbuf; i++)
		buf[i] = 0;
	tlsPmd5(buf, nbuf, key, n, (uchar*)label, nlabel, seed0, nseed0, seed1, nseed1);
	tlsPsha1(buf, nbuf, key+nkey-n, n, (uchar*)label, nlabel, seed0, nseed0, seed1, nseed1);
}

/*
 * for setting server session id's
 */
static Lock	sidLock;
static long	maxSid = 1;

/* the keys are verified to have the same public components
 * and to function correctly with pkcs 1 encryption and decryption. */
static TlsSec*
tlsSecInits(int cvers, uchar *csid, int ncsid, uchar *crandom, uchar *ssid, int *nssid, uchar *srandom)
{
	TlsSec *sec = emalloc(sizeof(*sec));

	USED(csid); USED(ncsid);  // ignore csid for now

	memmove(sec-&gt;crandom, crandom, RandomSize);
	sec-&gt;clientVers = cvers;

	put32(sec-&gt;srandom, time(0));
	genrandom(sec-&gt;srandom+4, RandomSize-4);
	memmove(srandom, sec-&gt;srandom, RandomSize);

	/*
	 * make up a unique sid: use our pid, and and incrementing id
	 * can signal no sid by setting nssid to 0.
	 */
	memset(ssid, 0, SidSize);
	put32(ssid, getpid());
	lock(&amp;sidLock);
	put32(ssid+4, maxSid++);
	unlock(&amp;sidLock);
	*nssid = SidSize;
	return sec;
}

static int
tlsSecSecrets(TlsSec *sec, int vers, uchar *epm, int nepm, uchar *kd, int nkd)
{
	if(epm != nil){
		if(setVers(sec, vers) &lt; 0)
			goto Err;
		serverMasterSecret(sec, epm, nepm);
	}else if(sec-&gt;vers != vers){
		werrstr("mismatched session versions");
		goto Err;
	}
	setSecrets(sec, kd, nkd);
	return 0;
Err:
	sec-&gt;ok = -1;
	return -1;
}

static TlsSec*
tlsSecInitc(int cvers, uchar *crandom)
{
	TlsSec *sec = emalloc(sizeof(*sec));
	sec-&gt;clientVers = cvers;
	put32(sec-&gt;crandom, time(0));
	genrandom(sec-&gt;crandom+4, RandomSize-4);
	memmove(crandom, sec-&gt;crandom, RandomSize);
	return sec;
}

static int
tlsSecSecretc(TlsSec *sec, uchar *sid, int nsid, uchar *srandom, uchar *cert, int ncert, int vers, uchar **epm, int *nepm, uchar *kd, int nkd)
{
	RSApub *pub;

	pub = nil;

	USED(sid);
	USED(nsid);
	
	memmove(sec-&gt;srandom, srandom, RandomSize);

	if(setVers(sec, vers) &lt; 0)
		goto Err;

	pub = X509toRSApub(cert, ncert, nil, 0);
	if(pub == nil){
		werrstr("invalid x509/rsa certificate");
		goto Err;
	}
	if(clientMasterSecret(sec, pub, epm, nepm) &lt; 0)
		goto Err;
	rsapubfree(pub);
	setSecrets(sec, kd, nkd);
	return 0;

Err:
	if(pub != nil)
		rsapubfree(pub);
	sec-&gt;ok = -1;
	return -1;
}

static int
tlsSecFinished(TlsSec *sec, MD5state md5, SHAstate sha1, uchar *fin, int nfin, int isclient)
{
	if(sec-&gt;nfin != nfin){
		sec-&gt;ok = -1;
		werrstr("invalid finished exchange");
		return -1;
	}
	md5.malloced = 0;
	sha1.malloced = 0;
	(*sec-&gt;setFinished)(sec, md5, sha1, fin, isclient);
	return 1;
}

static void
tlsSecOk(TlsSec *sec)
{
	if(sec-&gt;ok == 0)
		sec-&gt;ok = 1;
}

static void
tlsSecKill(TlsSec *sec)
{
	if(!sec)
		return;
	factotum_rsa_close(sec-&gt;rpc);
	sec-&gt;ok = -1;
}

static void
tlsSecClose(TlsSec *sec)
{
	if(!sec)
		return;
	factotum_rsa_close(sec-&gt;rpc);
	free(sec-&gt;server);
	free(sec);
}

static int
setVers(TlsSec *sec, int v)
{
	if(v == SSL3Version){
		sec-&gt;setFinished = sslSetFinished;
		sec-&gt;nfin = SSL3FinishedLen;
		sec-&gt;prf = sslPRF;
	}else if(v == TLSVersion){
		sec-&gt;setFinished = tlsSetFinished;
		sec-&gt;nfin = TLSFinishedLen;
		sec-&gt;prf = tlsPRF;
	}else{
		werrstr("invalid version");
		return -1;
	}
	sec-&gt;vers = v;
	return 0;
}

/*
 * generate secret keys from the master secret.
 *
 * different crypto selections will require different amounts
 * of key expansion and use of key expansion data,
 * but it's all generated using the same function.
 */
static void
setSecrets(TlsSec *sec, uchar *kd, int nkd)
{
	(*sec-&gt;prf)(kd, nkd, sec-&gt;sec, MasterSecretSize, "key expansion",
			sec-&gt;srandom, RandomSize, sec-&gt;crandom, RandomSize);
}

/*
 * set the master secret from the pre-master secret.
 */
static void
setMasterSecret(TlsSec *sec, Bytes *pm)
{
	(*sec-&gt;prf)(sec-&gt;sec, MasterSecretSize, pm-&gt;data, MasterSecretSize, "master secret",
			sec-&gt;crandom, RandomSize, sec-&gt;srandom, RandomSize);
}

static void
serverMasterSecret(TlsSec *sec, uchar *epm, int nepm)
{
	Bytes *pm;

	pm = pkcs1_decrypt(sec, epm, nepm);

	// if the client messed up, just continue as if everything is ok,
	// to prevent attacks to check for correctly formatted messages.
	// Hence the fprint(2,) can't be replaced by tlsError(), which sends an Alert msg to the client.
	if(sec-&gt;ok &lt; 0 || pm == nil || get16(pm-&gt;data) != sec-&gt;clientVers){
		fprint(2, "serverMasterSecret failed ok=%d pm=%p pmvers=%x cvers=%x nepm=%d\n",
			sec-&gt;ok, pm, pm ? get16(pm-&gt;data) : -1, sec-&gt;clientVers, nepm);
		sec-&gt;ok = -1;
		if(pm != nil)
			freebytes(pm);
		pm = newbytes(MasterSecretSize);
		genrandom(pm-&gt;data, MasterSecretSize);
	}
	setMasterSecret(sec, pm);
	memset(pm-&gt;data, 0, pm-&gt;len);	
	freebytes(pm);
}

static int
clientMasterSecret(TlsSec *sec, RSApub *pub, uchar **epm, int *nepm)
{
	Bytes *pm, *key;

	pm = newbytes(MasterSecretSize);
	put16(pm-&gt;data, sec-&gt;clientVers);
	genrandom(pm-&gt;data+2, MasterSecretSize - 2);

	setMasterSecret(sec, pm);

	key = pkcs1_encrypt(pm, pub, 2);
	memset(pm-&gt;data, 0, pm-&gt;len);
	freebytes(pm);
	if(key == nil){
		werrstr("tls pkcs1_encrypt failed");
		return -1;
	}

	*nepm = key-&gt;len;
	*epm = malloc(*nepm);
	if(*epm == nil){
		freebytes(key);
		werrstr("out of memory");
		return -1;
	}
	memmove(*epm, key-&gt;data, *nepm);

	freebytes(key);

	return 1;
}

static void
sslSetFinished(TlsSec *sec, MD5state hsmd5, SHAstate hssha1, uchar *finished, int isClient)
{
	DigestState *s;
	uchar h0[MD5dlen], h1[SHA1dlen], pad[48];
	char *label;

	if(isClient)
		label = "CLNT";
	else
		label = "SRVR";

	md5((uchar*)label, 4, nil, &amp;hsmd5);
	md5(sec-&gt;sec, MasterSecretSize, nil, &amp;hsmd5);
	memset(pad, 0x36, 48);
	md5(pad, 48, nil, &amp;hsmd5);
	md5(nil, 0, h0, &amp;hsmd5);
	memset(pad, 0x5C, 48);
	s = md5(sec-&gt;sec, MasterSecretSize, nil, nil);
	s = md5(pad, 48, nil, s);
	md5(h0, MD5dlen, finished, s);

	sha1((uchar*)label, 4, nil, &amp;hssha1);
	sha1(sec-&gt;sec, MasterSecretSize, nil, &amp;hssha1);
	memset(pad, 0x36, 40);
	sha1(pad, 40, nil, &amp;hssha1);
	sha1(nil, 0, h1, &amp;hssha1);
	memset(pad, 0x5C, 40);
	s = sha1(sec-&gt;sec, MasterSecretSize, nil, nil);
	s = sha1(pad, 40, nil, s);
	sha1(h1, SHA1dlen, finished + MD5dlen, s);
}

// fill "finished" arg with md5(args)^sha1(args)
static void
tlsSetFinished(TlsSec *sec, MD5state hsmd5, SHAstate hssha1, uchar *finished, int isClient)
{
	uchar h0[MD5dlen], h1[SHA1dlen];
	char *label;

	// get current hash value, but allow further messages to be hashed in
	md5(nil, 0, h0, &amp;hsmd5);
	sha1(nil, 0, h1, &amp;hssha1);

	if(isClient)
		label = "client finished";
	else
		label = "server finished";
	tlsPRF(finished, TLSFinishedLen, sec-&gt;sec, MasterSecretSize, label, h0, MD5dlen, h1, SHA1dlen);
}

static void
sslPRF(uchar *buf, int nbuf, uchar *key, int nkey, char *label, uchar *seed0, int nseed0, uchar *seed1, int nseed1)
{
	DigestState *s;
	uchar sha1dig[SHA1dlen], md5dig[MD5dlen], tmp[26];
	int i, n, len;

	USED(label);
	len = 1;
	while(nbuf &gt; 0){
		if(len &gt; 26)
			return;
		for(i = 0; i &lt; len; i++)
			tmp[i] = 'A' - 1 + len;
		s = sha1(tmp, len, nil, nil);
		s = sha1(key, nkey, nil, s);
		s = sha1(seed0, nseed0, nil, s);
		sha1(seed1, nseed1, sha1dig, s);
		s = md5(key, nkey, nil, nil);
		md5(sha1dig, SHA1dlen, md5dig, s);
		n = MD5dlen;
		if(n &gt; nbuf)
			n = nbuf;
		memmove(buf, md5dig, n);
		buf += n;
		nbuf -= n;
		len++;
	}
}

static mpint*
bytestomp(Bytes* bytes)
{
	mpint* ans;

	ans = betomp(bytes-&gt;data, bytes-&gt;len, nil);
	return ans;
}

/*
 * Convert mpint* to Bytes, putting high order byte first.
 */
static Bytes*
mptobytes(mpint* big)
{
	int n, m;
	uchar *a;
	Bytes* ans;

	a = nil;
	n = (mpsignif(big)+7)/8;
	m = mptobe(big, nil, n, &amp;a);
	ans = makebytes(a, m);
	if(a != nil)
		free(a);
	return ans;
}

// Do RSA computation on block according to key, and pad
// result on left with zeros to make it modlen long.
static Bytes*
rsacomp(Bytes* block, RSApub* key, int modlen)
{
	mpint *x, *y;
	Bytes *a, *ybytes;
	int ylen;

	x = bytestomp(block);
	y = rsaencrypt(key, x, nil);
	mpfree(x);
	ybytes = mptobytes(y);
	ylen = ybytes-&gt;len;

	if(ylen &lt; modlen) {
		a = newbytes(modlen);
		memset(a-&gt;data, 0, modlen-ylen);
		memmove(a-&gt;data+modlen-ylen, ybytes-&gt;data, ylen);
		freebytes(ybytes);
		ybytes = a;
	}
	else if(ylen &gt; modlen) {
		// assume it has leading zeros (mod should make it so)
		a = newbytes(modlen);
		memmove(a-&gt;data, ybytes-&gt;data, modlen);
		freebytes(ybytes);
		ybytes = a;
	}
	mpfree(y);
	return ybytes;
}

// encrypt data according to PKCS#1, /lib/rfc/rfc2437 9.1.2.1
static Bytes*
pkcs1_encrypt(Bytes* data, RSApub* key, int blocktype)
{
	Bytes *pad, *eb, *ans;
	int i, dlen, padlen, modlen;

	modlen = (mpsignif(key-&gt;n)+7)/8;
	dlen = data-&gt;len;
	if(modlen &lt; 12 || dlen &gt; modlen - 11)
		return nil;
	padlen = modlen - 3 - dlen;
	pad = newbytes(padlen);
	genrandom(pad-&gt;data, padlen);
	for(i = 0; i &lt; padlen; i++) {
		if(blocktype == 0)
			pad-&gt;data[i] = 0;
		else if(blocktype == 1)
			pad-&gt;data[i] = 255;
		else if(pad-&gt;data[i] == 0)
			pad-&gt;data[i] = 1;
	}
	eb = newbytes(modlen);
	eb-&gt;data[0] = 0;
	eb-&gt;data[1] = blocktype;
	memmove(eb-&gt;data+2, pad-&gt;data, padlen);
	eb-&gt;data[padlen+2] = 0;
	memmove(eb-&gt;data+padlen+3, data-&gt;data, dlen);
	ans = rsacomp(eb, key, modlen);
	freebytes(eb);
	freebytes(pad);
	return ans;
}

// decrypt data according to PKCS#1, with given key.
// expect a block type of 2.
static Bytes*
pkcs1_decrypt(TlsSec *sec, uchar *epm, int nepm)
{
	Bytes *eb, *ans = nil;
	int i, modlen;
	mpint *x, *y;

	modlen = (mpsignif(sec-&gt;rsapub-&gt;n)+7)/8;
	if(nepm != modlen)
		return nil;
	x = betomp(epm, nepm, nil);
	y = factotum_rsa_decrypt(sec-&gt;rpc, x);
	if(y == nil)
		return nil;
	eb = mptobytes(y);
	if(eb-&gt;len &lt; modlen){ // pad on left with zeros
		ans = newbytes(modlen);
		memset(ans-&gt;data, 0, modlen-eb-&gt;len);
		memmove(ans-&gt;data+modlen-eb-&gt;len, eb-&gt;data, eb-&gt;len);
		freebytes(eb);
		eb = ans;
	}
	if(eb-&gt;data[0] == 0 &amp;&amp; eb-&gt;data[1] == 2) {
		for(i = 2; i &lt; modlen; i++)
			if(eb-&gt;data[i] == 0)
				break;
		if(i &lt; modlen - 1)
			ans = makebytes(eb-&gt;data+i+1, modlen-(i+1));
	}
	freebytes(eb);
	return ans;
}


//================= general utility functions ========================

static void *
emalloc(int n)
{
	void *p;
	if(n==0)
		n=1;
	p = malloc(n);
	if(p == nil){
		exits("out of memory");
	}
	memset(p, 0, n);
	return p;
}

static void *
erealloc(void *ReallocP, int ReallocN)
{
	if(ReallocN == 0)
		ReallocN = 1;
	if(!ReallocP)
		ReallocP = emalloc(ReallocN);
	else if(!(ReallocP = realloc(ReallocP, ReallocN))){
		exits("out of memory");
	}
	return(ReallocP);
}

static void
put32(uchar *p, u32int x)
{
	p[0] = x&gt;&gt;24;
	p[1] = x&gt;&gt;16;
	p[2] = x&gt;&gt;8;
	p[3] = x;
}

static void
put24(uchar *p, int x)
{
	p[0] = x&gt;&gt;16;
	p[1] = x&gt;&gt;8;
	p[2] = x;
}

static void
put16(uchar *p, int x)
{
	p[0] = x&gt;&gt;8;
	p[1] = x;
}

static u32int
get32(uchar *p)
{
	return (p[0]&lt;&lt;24)|(p[1]&lt;&lt;16)|(p[2]&lt;&lt;8)|p[3];
}

static int
get24(uchar *p)
{
	return (p[0]&lt;&lt;16)|(p[1]&lt;&lt;8)|p[2];
}

static int
get16(uchar *p)
{
	return (p[0]&lt;&lt;8)|p[1];
}

#define OFFSET(x, s) offsetof(s, x)

/*
 * malloc and return a new Bytes structure capable of
 * holding len bytes. (len &gt;= 0)
 * Used to use crypt_malloc, which aborts if malloc fails.
 */
static Bytes*
newbytes(int len)
{
	Bytes* ans;

	ans = (Bytes*)malloc(OFFSET(data[0], Bytes) + len);
	ans-&gt;len = len;
	return ans;
}

/*
 * newbytes(len), with data initialized from buf
 */
static Bytes*
makebytes(uchar* buf, int len)
{
	Bytes* ans;

	ans = newbytes(len);
	memmove(ans-&gt;data, buf, len);
	return ans;
}

static void
freebytes(Bytes* b)
{
	if(b != nil)
		free(b);
}

/* len is number of ints */
static Ints*
newints(int len)
{
	Ints* ans;

	ans = (Ints*)malloc(OFFSET(data[0], Ints) + len*sizeof(int));
	ans-&gt;len = len;
	return ans;
}

static Ints*
makeints(int* buf, int len)
{
	Ints* ans;

	ans = newints(len);
	if(len &gt; 0)
		memmove(ans-&gt;data, buf, len*sizeof(int));
	return ans;
}

static void
freeints(Ints* b)
{
	if(b != nil)
		free(b);
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

