<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<meta http-equiv=Content-Type content="text/html; charset=utf8">
<title>/n/sources/plan9/sys/src/libventi/file.c - Plan 9 from Bell Labs</title>
<!-- THIS FILE IS AUTOMATICALLY GENERATED. -->
<!-- EDIT sources.tr INSTEAD. -->
</meta>
</head>
<body>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"><a href="http://plan9.bell-labs.com/plan9/">Plan 9 from Bell Labs</a>&rsquo;s /n/sources/plan9/sys/src/libventi/file.c</span></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
Copyright © 2009 Alcatel-Lucent.<br />
Distributed under the
<a href="http://plan9.bell-labs.com/plan9/license.html">Lucent Public License version 1.02</a>.
<br />
<a href="http://plan9.bell-labs.com/plan9/download.html">Download the Plan 9 distribution.</a>
</font>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<table width="100%" cellspacing=0 border=0><tr><td align="center">
<table cellspacing=0 cellpadding=5 bgcolor="#eeeeff"><tr><td align="left">
<pre>
<!-- END HEADER -->
/*
 * Manage tree of VtFiles stored in the block cache.
 *
 * The single point of truth for the info about the VtFiles themselves
 * is the block data.  Because of this, there is no explicit locking of
 * VtFile structures, and indeed there may be more than one VtFile
 * structure for a given Venti file.  They synchronize through the 
 * block cache.
 *
 * This is a bit simpler than fossil because there are no epochs
 * or tags or anything else.  Just mutable local blocks and immutable
 * Venti blocks.
 */

#include &lt;u.h&gt;
#include &lt;libc.h&gt;
#include &lt;venti.h&gt;

#define MaxBlock (1UL&lt;&lt;31)

static char ENotDir[] = "walk in non-directory";
static char ETooBig[] = "file too big";
/* static char EBadAddr[] = "bad address"; */
static char ELabelMismatch[] = "label mismatch";

static int	sizetodepth(uvlong s, int psize, int dsize);
static VtBlock 	*fileload(VtFile *r, VtEntry *e);
static int	shrinkdepth(VtFile*, VtBlock*, VtEntry*, int);
static int	shrinksize(VtFile*, VtEntry*, uvlong);
static int	growdepth(VtFile*, VtBlock*, VtEntry*, int);

#define ISLOCKED(r)	((r)-&gt;b != nil)
#define DEPTH(t)	((t)&amp;VtTypeDepthMask)

static VtFile *
vtfilealloc(VtCache *c, VtBlock *b, VtFile *p, u32int offset, int mode)
{
	int epb;
	u32int size;
	VtEntry e;
	VtFile *r;

	assert(p==nil || ISLOCKED(p));

	if(p == nil){
		assert(offset == 0);
		epb = 1;
	}else
		epb = p-&gt;dsize / VtEntrySize;

	if(b-&gt;type != VtDirType){
		werrstr("bad block type %#uo", b-&gt;type);
		return nil;
	}

	/*
	 * a non-active entry is the only thing that
	 * can legitimately happen here. all the others
	 * get prints.
	 */
	if(vtentryunpack(&amp;e, b-&gt;data, offset % epb) &lt; 0){
		fprint(2, "vtentryunpack failed: %r (%.*H)\n", VtEntrySize, b-&gt;data+VtEntrySize*(offset%epb));
		return nil;
	}
	if(!(e.flags &amp; VtEntryActive)){
		werrstr("entry not active");
		return nil;
	}

	if(DEPTH(e.type) &lt; sizetodepth(e.size, e.psize, e.dsize)){
		fprint(2, "depth %ud size %llud psize %ud dsize %ud\n",
			DEPTH(e.type), e.size, e.psize, e.dsize);
		werrstr("bad depth");
		return nil;
	}

	size = vtcacheblocksize(c);
	if(e.dsize &gt; size || e.psize &gt; size){
		werrstr("block sizes %ud, %ud bigger than cache block size %ud",
			e.psize, e.dsize, size);
		return nil;
	}

	r = vtmallocz(sizeof(VtFile));
	r-&gt;c = c;
	r-&gt;mode = mode;
	r-&gt;dsize = e.dsize;
	r-&gt;psize = e.psize;
	r-&gt;gen = e.gen;
	r-&gt;dir = (e.type &amp; VtTypeBaseMask) == VtDirType;
	r-&gt;ref = 1;
	r-&gt;parent = p;
	if(p){
		qlock(&amp;p-&gt;lk);
		assert(mode == VtOREAD || p-&gt;mode == VtORDWR);
		p-&gt;ref++;
		qunlock(&amp;p-&gt;lk);
	}else{
		assert(b-&gt;addr != NilBlock);
		r-&gt;local = 1;
	}
	memmove(r-&gt;score, b-&gt;score, VtScoreSize);
	r-&gt;offset = offset;
	r-&gt;epb = epb;

	return r;
}

VtFile *
vtfileroot(VtCache *c, u32int addr, int mode)
{
	VtFile *r;
	VtBlock *b;

	b = vtcachelocal(c, addr, VtDirType);
	if(b == nil)
		return nil;
	r = vtfilealloc(c, b, nil, 0, mode);
	vtblockput(b);
	return r;
}

VtFile*
vtfileopenroot(VtCache *c, VtEntry *e)
{
	VtBlock *b;
	VtFile *f;

	b = vtcacheallocblock(c, VtDirType);
	if(b == nil)
		return nil;

	vtentrypack(e, b-&gt;data, 0);
	f = vtfilealloc(c, b, nil, 0, VtORDWR);
	vtblockput(b);
	return f;
}

VtFile *
vtfilecreateroot(VtCache *c, int psize, int dsize, int type)
{
	VtEntry e;

	memset(&amp;e, 0, sizeof e);
	e.flags = VtEntryActive;
	e.psize = psize;
	e.dsize = dsize;
	e.type = type;
	memmove(e.score, vtzeroscore, VtScoreSize);

	return vtfileopenroot(c, &amp;e);
}

VtFile *
vtfileopen(VtFile *r, u32int offset, int mode)
{
	ulong bn;
	VtBlock *b;

	assert(ISLOCKED(r));
	if(!r-&gt;dir){
		werrstr(ENotDir);
		return nil;
	}

	bn = offset/(r-&gt;dsize/VtEntrySize);

	b = vtfileblock(r, bn, mode);
	if(b == nil)
		return nil;
	r = vtfilealloc(r-&gt;c, b, r, offset, mode);
	vtblockput(b);
	return r;
}

VtFile*
vtfilecreate(VtFile *r, int psize, int dsize, int type)
{
	return _vtfilecreate(r, -1, psize, dsize, type);
}

VtFile*
_vtfilecreate(VtFile *r, int o, int psize, int dsize, int type)
{
	int i;
	VtBlock *b;
	u32int bn, size;
	VtEntry e;
	int epb;
	VtFile *rr;
	u32int offset;
	
	assert(ISLOCKED(r));
	assert(psize &lt;= VtMaxLumpSize);
	assert(dsize &lt;= VtMaxLumpSize);
	assert(type == VtDirType || type == VtDataType);

	if(!r-&gt;dir){
		werrstr(ENotDir);
		return nil;
	}

	epb = r-&gt;dsize/VtEntrySize;

	size = vtfilegetdirsize(r);
	/*
	 * look at a random block to see if we can find an empty entry
	 */
	if(o == -1){
		offset = lnrand(size+1);
		offset -= offset % epb;
	}else
		offset = o;

	/* try the given block and then try the last block */
	for(;;){
		bn = offset/epb;
		b = vtfileblock(r, bn, VtORDWR);
		if(b == nil)
			return nil;
		for(i=offset%r-&gt;epb; i&lt;epb; i++){
			if(vtentryunpack(&amp;e, b-&gt;data, i) &lt; 0)
				continue;
			if((e.flags&amp;VtEntryActive) == 0 &amp;&amp; e.gen != ~0)
				goto Found;
		}
		vtblockput(b);
		if(offset == size){
			fprint(2, "vtfilecreate: cannot happen\n");
			werrstr("vtfilecreate: cannot happen");
			return nil;
		}
		offset = size;
	}

Found:
	/* found an entry - gen already set */
	e.psize = psize;
	e.dsize = dsize;
	e.flags = VtEntryActive;
	e.type = type;
	e.size = 0;
	memmove(e.score, vtzeroscore, VtScoreSize);
	vtentrypack(&amp;e, b-&gt;data, i);

	offset = bn*epb + i;
	if(offset+1 &gt; size){
		if(vtfilesetdirsize(r, offset+1) &lt; 0){
			vtblockput(b);
			return nil;
		}
	}

	rr = vtfilealloc(r-&gt;c, b, r, offset, VtORDWR);
	vtblockput(b);
	return rr;
}

static int
vtfilekill(VtFile *r, int doremove)
{
	VtEntry e;
	VtBlock *b;

	assert(ISLOCKED(r));
	b = fileload(r, &amp;e);
	if(b == nil)
		return -1;

	if(doremove==0 &amp;&amp; e.size == 0){
		/* already truncated */
		vtblockput(b);
		return 0;
	}

	if(doremove){
		if(e.gen != ~0)
			e.gen++;
		e.dsize = 0;
		e.psize = 0;
		e.flags = 0;
	}else
		e.flags &amp;= ~VtEntryLocal;
	e.type = 0;
	e.size = 0;
	memmove(e.score, vtzeroscore, VtScoreSize);
	vtentrypack(&amp;e, b-&gt;data, r-&gt;offset % r-&gt;epb);
	vtblockput(b);

	if(doremove){
		vtfileunlock(r);
		vtfileclose(r);
	}

	return 0;
}

int
vtfileremove(VtFile *r)
{
	return vtfilekill(r, 1);
}

int
vtfiletruncate(VtFile *r)
{
	return vtfilekill(r, 0);
}

uvlong
vtfilegetsize(VtFile *r)
{
	VtEntry e;
	VtBlock *b;

	assert(ISLOCKED(r));
	b = fileload(r, &amp;e);
	if(b == nil)
		return ~(uvlong)0;
	vtblockput(b);

	return e.size;
}

static int
shrinksize(VtFile *r, VtEntry *e, uvlong size)
{
	int i, depth, type, isdir, ppb;
	uvlong ptrsz;
	uchar score[VtScoreSize];
	VtBlock *b;

	b = vtcacheglobal(r-&gt;c, e-&gt;score, e-&gt;type);
	if(b == nil)
		return -1;

	ptrsz = e-&gt;dsize;
	ppb = e-&gt;psize/VtScoreSize;
	type = e-&gt;type;
	depth = DEPTH(type);
	for(i=0; i+1&lt;depth; i++)
		ptrsz *= ppb;

	isdir = r-&gt;dir;
	while(depth &gt; 0){
		if(b-&gt;addr == NilBlock){
			/* not worth copying the block just so we can zero some of it */
			vtblockput(b);
			return -1;
		}

		/*
		 * invariant: each pointer in the tree rooted at b accounts for ptrsz bytes
		 */

		/* zero the pointers to unnecessary blocks */
		i = (size+ptrsz-1)/ptrsz;
		for(; i&lt;ppb; i++)
			memmove(b-&gt;data+i*VtScoreSize, vtzeroscore, VtScoreSize);

		/* recurse (go around again) on the partially necessary block */
		i = size/ptrsz;
		size = size%ptrsz;
		if(size == 0){
			vtblockput(b);
			return 0;
		}
		ptrsz /= ppb;
		type--;
		memmove(score, b-&gt;data+i*VtScoreSize, VtScoreSize);
		vtblockput(b);
		b = vtcacheglobal(r-&gt;c, score, type);
		if(b == nil)
			return -1;
	}

	if(b-&gt;addr == NilBlock){
		vtblockput(b);
		return -1;
	}

	/*
	 * No one ever truncates BtDir blocks.
	 */
	if(depth==0 &amp;&amp; !isdir &amp;&amp; e-&gt;dsize &gt; size)
		memset(b-&gt;data+size, 0, e-&gt;dsize-size);
	vtblockput(b);
	return 0;
}

int
vtfilesetsize(VtFile *r, u64int size)
{
	int depth, edepth;
	VtEntry e;
	VtBlock *b;

	assert(ISLOCKED(r));
	if(size == 0)
		return vtfiletruncate(r);

	if(size &gt; VtMaxFileSize || size &gt; ((uvlong)MaxBlock)*r-&gt;dsize){
		werrstr(ETooBig);
		return -1;
	}

	b = fileload(r, &amp;e);
	if(b == nil)
		return -1;

	/* quick out */
	if(e.size == size){
		vtblockput(b);
		return 0;
	}

	depth = sizetodepth(size, e.psize, e.dsize);
	edepth = DEPTH(e.type);
	if(depth &lt; edepth){
		if(shrinkdepth(r, b, &amp;e, depth) &lt; 0){
			vtblockput(b);
			return -1;
		}
	}else if(depth &gt; edepth){
		if(growdepth(r, b, &amp;e, depth) &lt; 0){
			vtblockput(b);
			return -1;
		}
	}

	if(size &lt; e.size)
		shrinksize(r, &amp;e, size);

	e.size = size;
	vtentrypack(&amp;e, b-&gt;data, r-&gt;offset % r-&gt;epb);
	vtblockput(b);

	return 0;
}

int
vtfilesetdirsize(VtFile *r, u32int ds)
{
	uvlong size;
	int epb;

	assert(ISLOCKED(r));
	epb = r-&gt;dsize/VtEntrySize;

	size = (uvlong)r-&gt;dsize*(ds/epb);
	size += VtEntrySize*(ds%epb);
	return vtfilesetsize(r, size);
}

u32int
vtfilegetdirsize(VtFile *r)
{
	ulong ds;
	uvlong size;
	int epb;

	assert(ISLOCKED(r));
	epb = r-&gt;dsize/VtEntrySize;

	size = vtfilegetsize(r);
	ds = epb*(size/r-&gt;dsize);
	ds += (size%r-&gt;dsize)/VtEntrySize;
	return ds;
}

int
vtfilegetentry(VtFile *r, VtEntry *e)
{
	VtBlock *b;

	assert(ISLOCKED(r));
	b = fileload(r, e);
	if(b == nil)
		return -1;
	vtblockput(b);

	return 0;
}

int
vtfilesetentry(VtFile *r, VtEntry *e)
{
	VtBlock *b;
	VtEntry ee;

	assert(ISLOCKED(r));
	b = fileload(r, &amp;ee);
	if(b == nil)
		return -1;
	vtentrypack(e, b-&gt;data, r-&gt;offset % r-&gt;epb);
	vtblockput(b);
	return 0;
}

static VtBlock *
blockwalk(VtBlock *p, int index, VtCache *c, int mode, VtEntry *e)
{
	VtBlock *b;
	int type;
	uchar *score;
	VtEntry oe;

	switch(p-&gt;type){
	case VtDataType:
		assert(0);
	case VtDirType:
		type = e-&gt;type;
		score = e-&gt;score;
		break;
	default:
		type = p-&gt;type - 1;
		score = p-&gt;data+index*VtScoreSize;
		break;
	}
/*print("walk from %V/%d ty %d to %V ty %d\n", p-&gt;score, index, p-&gt;type, score, type); */

	if(mode == VtOWRITE &amp;&amp; vtglobaltolocal(score) == NilBlock){
		b = vtcacheallocblock(c, type);
		if(b)
			goto HaveCopy;
	}else
		b = vtcacheglobal(c, score, type);

	if(b == nil || mode == VtOREAD)
		return b;

	if(vtglobaltolocal(b-&gt;score) != NilBlock)
		return b;

	oe = *e;

	/*
	 * Copy on write.
	 */
	e-&gt;flags |= VtEntryLocal;

	b = vtblockcopy(b/*, e-&gt;tag, fs-&gt;ehi, fs-&gt;elo*/);
	if(b == nil)
		return nil;

HaveCopy:
	if(p-&gt;type == VtDirType){
		memmove(e-&gt;score, b-&gt;score, VtScoreSize);
		vtentrypack(e, p-&gt;data, index);
	}else{
		memmove(p-&gt;data+index*VtScoreSize, b-&gt;score, VtScoreSize);
	}
	return b;
}

/*
 * Change the depth of the VtFile r.
 * The entry e for r is contained in block p.
 */
static int
growdepth(VtFile *r, VtBlock *p, VtEntry *e, int depth)
{
	VtBlock *b, *bb;
	VtEntry oe;

	assert(ISLOCKED(r));
	assert(depth &lt;= VtPointerDepth);

	b = vtcacheglobal(r-&gt;c, e-&gt;score, e-&gt;type);
	if(b == nil)
		return -1;

	oe = *e;

	/*
	 * Keep adding layers until we get to the right depth
	 * or an error occurs.
	 */
	while(DEPTH(e-&gt;type) &lt; depth){
		bb = vtcacheallocblock(r-&gt;c, e-&gt;type+1);
		if(bb == nil)
			break;
		memmove(bb-&gt;data, b-&gt;score, VtScoreSize);
		memmove(e-&gt;score, bb-&gt;score, VtScoreSize);	
		e-&gt;type++;
		e-&gt;flags |= VtEntryLocal;
		vtblockput(b);
		b = bb;
	}

	vtentrypack(e, p-&gt;data, r-&gt;offset % r-&gt;epb);
	vtblockput(b);

	if(DEPTH(e-&gt;type) == depth)
		return 0;
	return -1;
}

static int
shrinkdepth(VtFile *r, VtBlock *p, VtEntry *e, int depth)
{
	VtBlock *b, *nb, *ob, *rb;
	VtEntry oe;

	assert(ISLOCKED(r));
	assert(depth &lt;= VtPointerDepth);

	rb = vtcacheglobal(r-&gt;c, e-&gt;score, e-&gt;type);
	if(rb == nil)
		return -1;

	/*
	 * Walk down to the new root block.
	 * We may stop early, but something is better than nothing.
	 */
	oe = *e;

	ob = nil;
	b = rb;
	for(; DEPTH(e-&gt;type) &gt; depth; e-&gt;type--){
		nb = vtcacheglobal(r-&gt;c, b-&gt;data, e-&gt;type-1);
		if(nb == nil)
			break;
		if(ob!=nil &amp;&amp; ob!=rb)
			vtblockput(ob);
		ob = b;
		b = nb;
	}

	if(b == rb){
		vtblockput(rb);
		return 0;
	}

	/*
	 * Right now, e points at the root block rb, b is the new root block,
	 * and ob points at b.  To update:
	 *
	 *	(i) change e to point at b
	 *	(ii) zero the pointer ob -&gt; b
	 *	(iii) free the root block
	 *
	 * p (the block containing e) must be written before
	 * anything else.
 	 */

	/* (i) */
	memmove(e-&gt;score, b-&gt;score, VtScoreSize);
	vtentrypack(e, p-&gt;data, r-&gt;offset % r-&gt;epb);

	/* (ii) */
	memmove(ob-&gt;data, vtzeroscore, VtScoreSize);

	/* (iii) */
	vtblockput(rb);
	if(ob!=nil &amp;&amp; ob!=rb)
		vtblockput(ob);
	vtblockput(b);

	if(DEPTH(e-&gt;type) == depth)
		return 0;
	return -1;
}

static int
mkindices(VtEntry *e, u32int bn, int *index)
{
	int i, np;

	memset(index, 0, (VtPointerDepth+1)*sizeof(int));

	np = e-&gt;psize/VtScoreSize;
	for(i=0; bn &gt; 0; i++){
		if(i &gt;= VtPointerDepth){
			werrstr("bad address 0x%lux", (ulong)bn);
			return -1;
		}
		index[i] = bn % np;
		bn /= np;
	}
	return i;
}
	
VtBlock *
vtfileblock(VtFile *r, u32int bn, int mode)
{
	VtBlock *b, *bb;
	int index[VtPointerDepth+1];
	VtEntry e;
	int i;
	int m;

	assert(ISLOCKED(r));
	assert(bn != NilBlock);

	b = fileload(r, &amp;e);
	if(b == nil)
		return nil;

	i = mkindices(&amp;e, bn, index);
	if(i &lt; 0)
		goto Err;
	if(i &gt; DEPTH(e.type)){
		if(mode == VtOREAD){
			werrstr("bad address 0x%lux", (ulong)bn);
			goto Err;
		}
		index[i] = 0;
		if(growdepth(r, b, &amp;e, i) &lt; 0)
			goto Err;
	}

assert(b-&gt;type == VtDirType);

	index[DEPTH(e.type)] = r-&gt;offset % r-&gt;epb;

	/* mode for intermediate block */
	m = mode;
	if(m == VtOWRITE)
		m = VtORDWR;

	for(i=DEPTH(e.type); i&gt;=0; i--){
		bb = blockwalk(b, index[i], r-&gt;c, i==0 ? mode : m, &amp;e);
		if(bb == nil)
			goto Err;
		vtblockput(b);
		b = bb;
	}
	b-&gt;pc = getcallerpc(&amp;r);
	return b;
Err:
	vtblockput(b);
	return nil;
}

int
vtfileblockscore(VtFile *r, u32int bn, uchar score[VtScoreSize])
{
	VtBlock *b, *bb;
	int index[VtPointerDepth+1];
	VtEntry e;
	int i;

	assert(ISLOCKED(r));
	assert(bn != NilBlock);

	b = fileload(r, &amp;e);
	if(b == nil)
		return -1;

	if(DEPTH(e.type) == 0){
		memmove(score, e.score, VtScoreSize);
		vtblockput(b);
		return 0;
	}

	i = mkindices(&amp;e, bn, index);
	if(i &lt; 0){
		vtblockput(b);
		return -1;
	}
	if(i &gt; DEPTH(e.type)){
		memmove(score, vtzeroscore, VtScoreSize);
		vtblockput(b);
		return 0;
	}

	index[DEPTH(e.type)] = r-&gt;offset % r-&gt;epb;

	for(i=DEPTH(e.type); i&gt;=1; i--){
		bb = blockwalk(b, index[i], r-&gt;c, VtOREAD, &amp;e);
		if(bb == nil)
			goto Err;
		vtblockput(b);
		b = bb;
		if(memcmp(b-&gt;score, vtzeroscore, VtScoreSize) == 0)
			break;
	}

	memmove(score, b-&gt;data+index[0]*VtScoreSize, VtScoreSize);
	vtblockput(b);
	return 0;

Err:
	vtblockput(b);
	return -1;
}

void
vtfileincref(VtFile *r)
{
	qlock(&amp;r-&gt;lk);
	r-&gt;ref++;
	qunlock(&amp;r-&gt;lk);
}

void
vtfileclose(VtFile *r)
{
	if(r == nil)
		return;
	qlock(&amp;r-&gt;lk);
	r-&gt;ref--;
	if(r-&gt;ref){
		qunlock(&amp;r-&gt;lk);
		return;
	}
	assert(r-&gt;ref == 0);
	qunlock(&amp;r-&gt;lk);
	if(r-&gt;parent)
		vtfileclose(r-&gt;parent);
	memset(r, ~0, sizeof(*r));
	vtfree(r);
}

/*
 * Retrieve the block containing the entry for r.
 * If a snapshot has happened, we might need
 * to get a new copy of the block.  We avoid this
 * in the common case by caching the score for
 * the block and the last epoch in which it was valid.
 *
 * We use r-&gt;mode to tell the difference between active
 * file system VtFiles (VtORDWR) and VtFiles for the
 * snapshot file system (VtOREAD).
 */
static VtBlock*
fileloadblock(VtFile *r, int mode)
{
	char e[ERRMAX];
	u32int addr;
	VtBlock *b;

	switch(r-&gt;mode){
	default:
		assert(0);
	case VtORDWR:
		assert(r-&gt;mode == VtORDWR);
		if(r-&gt;local == 1){
			b = vtcacheglobal(r-&gt;c, r-&gt;score, VtDirType);
			if(b == nil)
				return nil;
			b-&gt;pc = getcallerpc(&amp;r);
			return b;
		}
		assert(r-&gt;parent != nil);
		if(vtfilelock(r-&gt;parent, VtORDWR) &lt; 0)
			return nil;
		b = vtfileblock(r-&gt;parent, r-&gt;offset/r-&gt;epb, VtORDWR);
		vtfileunlock(r-&gt;parent);
		if(b == nil)
			return nil;
		memmove(r-&gt;score, b-&gt;score, VtScoreSize);
		r-&gt;local = 1;
		return b;

	case VtOREAD:
		if(mode == VtORDWR){
			werrstr("read/write lock of read-only file");
			return nil;
		}
		addr = vtglobaltolocal(r-&gt;score);
		if(addr == NilBlock)
			return vtcacheglobal(r-&gt;c, r-&gt;score, VtDirType);

		b = vtcachelocal(r-&gt;c, addr, VtDirType);
		if(b)
			return b;

		/*
		 * If it failed because the epochs don't match, the block has been
		 * archived and reclaimed.  Rewalk from the parent and get the
		 * new pointer.  This can't happen in the VtORDWR case
		 * above because blocks in the current epoch don't get
		 * reclaimed.  The fact that we're VtOREAD means we're
		 * a snapshot.  (Or else the file system is read-only, but then
		 * the archiver isn't going around deleting blocks.)
		 */
		rerrstr(e, sizeof e);
		if(strcmp(e, ELabelMismatch) == 0){
			if(vtfilelock(r-&gt;parent, VtOREAD) &lt; 0)
				return nil;
			b = vtfileblock(r-&gt;parent, r-&gt;offset/r-&gt;epb, VtOREAD);
			vtfileunlock(r-&gt;parent);
			if(b){
				fprint(2, "vtfilealloc: lost %V found %V\n",
					r-&gt;score, b-&gt;score);
				memmove(r-&gt;score, b-&gt;score, VtScoreSize);
				return b;
			}
		}
		return nil;
	}
}

int
vtfilelock(VtFile *r, int mode)
{
	VtBlock *b;

	if(mode == -1)
		mode = r-&gt;mode;

	b = fileloadblock(r, mode);
	if(b == nil)
		return -1;
	/*
	 * The fact that we are holding b serves as the
	 * lock entitling us to write to r-&gt;b.
	 */
	assert(r-&gt;b == nil);
	r-&gt;b = b;
	b-&gt;pc = getcallerpc(&amp;r);
	return 0;
}

/*
 * Lock two (usually sibling) VtFiles.  This needs special care
 * because the Entries for both vtFiles might be in the same block.
 * We also try to lock blocks in left-to-right order within the tree.
 */
int
vtfilelock2(VtFile *r, VtFile *rr, int mode)
{
	VtBlock *b, *bb;

	if(rr == nil)
		return vtfilelock(r, mode);

	if(mode == -1)
		mode = r-&gt;mode;

	if(r-&gt;parent==rr-&gt;parent &amp;&amp; r-&gt;offset/r-&gt;epb == rr-&gt;offset/rr-&gt;epb){
		b = fileloadblock(r, mode);
		if(b == nil)
			return -1;
		vtblockduplock(b);
		bb = b;
	}else if(r-&gt;parent==rr-&gt;parent || r-&gt;offset &gt; rr-&gt;offset){
		bb = fileloadblock(rr, mode);
		b = fileloadblock(r, mode);
	}else{
		b = fileloadblock(r, mode);
		bb = fileloadblock(rr, mode);
	}
	if(b == nil || bb == nil){
		if(b)
			vtblockput(b);
		if(bb)
			vtblockput(bb);
		return -1;
	}

	/*
	 * The fact that we are holding b and bb serves
	 * as the lock entitling us to write to r-&gt;b and rr-&gt;b.
	 */
	r-&gt;b = b;
	rr-&gt;b = bb;
	b-&gt;pc = getcallerpc(&amp;r);
	bb-&gt;pc = getcallerpc(&amp;r);
	return 0;
}

void
vtfileunlock(VtFile *r)
{
	VtBlock *b;

	if(r-&gt;b == nil){
		fprint(2, "vtfileunlock: already unlocked\n");
		abort();
	}
	b = r-&gt;b;
	r-&gt;b = nil;
	vtblockput(b);
}

static VtBlock*
fileload(VtFile *r, VtEntry *e)
{
	VtBlock *b;

	assert(ISLOCKED(r));
	b = r-&gt;b;
	if(vtentryunpack(e, b-&gt;data, r-&gt;offset % r-&gt;epb) &lt; 0)
		return nil;
	vtblockduplock(b);
	return b;
}

static int
sizetodepth(uvlong s, int psize, int dsize)
{
	int np;
	int d;
	
	/* determine pointer depth */
	np = psize/VtScoreSize;
	s = (s + dsize - 1)/dsize;
	for(d = 0; s &gt; 1; d++)
		s = (s + np - 1)/np;
	return d;
}

long
vtfileread(VtFile *f, void *data, long count, vlong offset)
{
	int frag;
	VtBlock *b;
	VtEntry e;

	assert(ISLOCKED(f));

	vtfilegetentry(f, &amp;e);
	if(count == 0)
		return 0;
	if(count &lt; 0 || offset &lt; 0){
		werrstr("vtfileread: bad offset or count");
		return -1;
	}
	if(offset &gt;= e.size)
		return 0;

	if(offset+count &gt; e.size)
		count = e.size - offset;

	frag = offset % e.dsize;
	if(frag+count &gt; e.dsize)
		count = e.dsize - frag;

	b = vtfileblock(f, offset/e.dsize, VtOREAD);
	if(b == nil)
		return -1;

	memmove(data, b-&gt;data+frag, count);
	vtblockput(b);
	return count;
}

static long
filewrite1(VtFile *f, void *data, long count, vlong offset)
{
	int frag, m;
	VtBlock *b;
	VtEntry e;

	vtfilegetentry(f, &amp;e);
	if(count &lt; 0 || offset &lt; 0){
		werrstr("vtfilewrite: bad offset or count");
		return -1;
	}

	frag = offset % e.dsize;
	if(frag+count &gt; e.dsize)
		count = e.dsize - frag;

	m = VtORDWR;
	if(frag == 0 &amp;&amp; count == e.dsize)
		m = VtOWRITE;

	b = vtfileblock(f, offset/e.dsize, m);
	if(b == nil)
		return -1;

	memmove(b-&gt;data+frag, data, count);
	if(m == VtOWRITE &amp;&amp; frag+count &lt; e.dsize)
		memset(b-&gt;data+frag+count, 0, e.dsize-frag-count);

	if(offset+count &gt; e.size){
		vtfilegetentry(f, &amp;e);
		e.size = offset+count;
		vtfilesetentry(f, &amp;e);
	}

	vtblockput(b);
	return count;
}

long
vtfilewrite(VtFile *f, void *data, long count, vlong offset)
{
	long tot, m;

	assert(ISLOCKED(f));

	tot = 0;
	m = 0;
	while(tot &lt; count){
		m = filewrite1(f, (char*)data+tot, count-tot, offset+tot);
		if(m &lt;= 0)
			break;
		tot += m;
	}
	if(tot==0)
		return m;
	return tot;
}

static int
flushblock(VtCache *c, VtBlock *bb, uchar score[VtScoreSize], int ppb, int epb,
	int type)
{
	u32int addr;
	VtBlock *b;
	VtEntry e;
	int i;

	addr = vtglobaltolocal(score);
	if(addr == NilBlock)
		return 0;

	if(bb){
		b = bb;
		if(memcmp(b-&gt;score, score, VtScoreSize) != 0)
			abort();
	}else
		if((b = vtcachelocal(c, addr, type)) == nil)
			return -1;

	switch(type){
	case VtDataType:
		break;

	case VtDirType:
		for(i=0; i&lt;epb; i++){
			if(vtentryunpack(&amp;e, b-&gt;data, i) &lt; 0)
				goto Err;
			if(!(e.flags&amp;VtEntryActive))
				continue;
			if(flushblock(c, nil, e.score, e.psize/VtScoreSize, e.dsize/VtEntrySize,
				e.type) &lt; 0)
				goto Err;
			vtentrypack(&amp;e, b-&gt;data, i);
		}
		break;
	
	default:	/* VtPointerTypeX */
		for(i=0; i&lt;ppb; i++){
			if(flushblock(c, nil, b-&gt;data+VtScoreSize*i, ppb, epb, type-1) &lt; 0)
				goto Err;
		}
		break;
	}

	if(vtblockwrite(b) &lt; 0)
		goto Err;
	memmove(score, b-&gt;score, VtScoreSize);
	if(b != bb)
		vtblockput(b);
	return 0;

Err:
	if(b != bb)
		vtblockput(b);
	return -1;
}

int
vtfileflush(VtFile *f)
{
	int ret;
	VtBlock *b;
	VtEntry e;

	assert(ISLOCKED(f));
	b = fileload(f, &amp;e);
	if(!(e.flags&amp;VtEntryLocal)){
		vtblockput(b);
		return 0;
	}

	ret = flushblock(f-&gt;c, nil, e.score, e.psize/VtScoreSize, e.dsize/VtEntrySize,
		e.type);
	if(ret &lt; 0){
		vtblockput(b);
		return -1;
	}

	vtentrypack(&amp;e, b-&gt;data, f-&gt;offset % f-&gt;epb);
	vtblockput(b);
	return 0;
}

int
vtfileflushbefore(VtFile *r, u64int offset)
{
	VtBlock *b, *bb;
	VtEntry e;
	int i, base, depth, ppb, epb, doflush;
	int index[VtPointerDepth+1], j, ret;
	VtBlock *bi[VtPointerDepth+2];
	uchar *score;

	assert(ISLOCKED(r));
	if(offset == 0)
		return 0;

	b = fileload(r, &amp;e);
	if(b == nil)
		return -1;

	/*
	 * compute path through tree for the last written byte and the next one.
	 */
	ret = -1;
	memset(bi, 0, sizeof bi);
	depth = DEPTH(e.type);
	bi[depth+1] = b;
	i = mkindices(&amp;e, (offset-1)/e.dsize, index);
	if(i &lt; 0)
		goto Err;
	if(i &gt; depth)
		goto Err;
	ppb = e.psize / VtScoreSize;
	epb = e.dsize / VtEntrySize;

	/*
	 * load the blocks along the last written byte
	 */
	index[depth] = r-&gt;offset % r-&gt;epb;
	for(i=depth; i&gt;=0; i--){
		bb = blockwalk(b, index[i], r-&gt;c, VtORDWR, &amp;e);
		if(bb == nil)
			goto Err;
		bi[i] = bb;
		b = bb;
	}
	ret = 0;

	/*
	 * walk up the path from leaf to root, flushing anything that
	 * has been finished.
	 */
	base = e.type&amp;~VtTypeDepthMask;
	for(i=0; i&lt;=depth; i++){
		doflush = 0;
		if(i == 0){
			/* leaf: data or dir block */
			if(offset%e.dsize == 0)
				doflush = 1;
		}else{
			/*
			 * interior node: pointer blocks.
			 * specifically, b = bi[i] is a block whose index[i-1]'th entry 
			 * points at bi[i-1].  
			 */
			b = bi[i];

			/*
			 * the index entries up to but not including index[i-1] point at 
			 * finished blocks, so flush them for sure.
			 */
			for(j=0; j&lt;index[i-1]; j++)
				if(flushblock(r-&gt;c, nil, b-&gt;data+j*VtScoreSize, ppb, epb, base+i-1) &lt; 0)
					goto Err;

			/*
			 * if index[i-1] is the last entry in the block and is global
			 * (i.e. the kid is flushed), then we can flush this block.
			 */
			if(j==ppb-1 &amp;&amp; vtglobaltolocal(b-&gt;data+j*VtScoreSize)==NilBlock)
				doflush = 1;
		}
		if(doflush){
			if(i == depth)
				score = e.score;
			else
				score = bi[i+1]-&gt;data+index[i]*VtScoreSize;
			if(flushblock(r-&gt;c, bi[i], score, ppb, epb, base+i) &lt; 0)
				goto Err;
		}
	}

Err:
	/* top: entry.  do this always so that the score is up-to-date */
	vtentrypack(&amp;e, bi[depth+1]-&gt;data, index[depth]);
	for(i=0; i&lt;nelem(bi); i++)
		if(bi[i])
			vtblockput(bi[i]);
	return ret;
}
<!-- BEGIN TAIL -->
</pre>
</td></tr></table>
</td></tr></table>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<p style="line-height: 1.2em; margin-left: 1.00in; text-indent: 0.00in; margin-right: 1.00in; margin-top: 0; margin-bottom: 0; text-align: center;">
<span style="font-size: 10pt"></span></p>
<p style="margin-top: 0; margin-bottom: 0.50in"></p>
<p style="margin-top: 0; margin-bottom: 0.33in"></p>
<center><table border="0"><tr>
<td valign="middle"><a href="http://www.alcatel-lucent.com/"><img border="0" src="/plan9/img/logo_ft.gif" alt="Bell Labs" />
</a></td>
<td valign="middle"><a href="http://www.opensource.org"><img border="0" alt="OSI certified" src="/plan9/img/osi-certified-60x50.gif" />
</a></td>
<td><img style="padding-right: 45px;" alt="Powered by Plan 9" src="/plan9/img/power36.gif" />
</td>
</tr></table></center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center>
<span style="font-size: 10pt">(<a href="/plan9/">Return to Plan 9 Home Page</a>)</span>
</center>
<p style="margin-top: 0; margin-bottom: 0.17in"></p>
<center><font size=-1>
<span style="font-size: 10pt"><a href="http://www.lucent.com/copyright.html">Copyright</a></span>
<span style="font-size: 10pt">© 2009 Alcatel-Lucent.</span>
<span style="font-size: 10pt">All Rights Reserved.</span>
<br />
<span style="font-size: 10pt">Comments to</span>
<span style="font-size: 10pt"><a href="mailto:webmaster@plan9.bell-labs.com">webmaster@plan9.bell-labs.com</a>.</span>
</font></center>
</body>
</html>

